#!/usr/bin/env python3
"""Migrate contact handles.

This is a data migration after 'ab05205d07e0'.
"""
import asyncio

from regal.exceptions import ContactDoesNotExist
from setapp.constants import Verbosity
from sqlalchemy import select

from verification.commands.utils import echo, verification_command
from verification.database import atomic, get_engine
from verification.models import Verification
from verification.settings import APPLICATION
from verification.utils import Result

_OUT_MSG = ('Migrate contact handles: {total} total, {result.successes} migrated, '
            '{result.errors} errors.')


async def migrate_handles() -> Result:
    """Actually migrate contact handles."""
    result = Result()
    async with get_engine().begin() as conn:
        id_query = select(Verification.id).filter_by(contact_handle=None)
        ids = (await conn.execute(id_query)).scalars()

    for id in ids:
        try:
            async with atomic() as session:
                query = select(Verification).with_for_update().filter_by(id=id)
                verification = (await session.execute(query)).scalars().one()
                try:
                    contact = await verification.contact
                except ContactDoesNotExist:
                    # Contact is already deleted, dig through history.
                    history = await APPLICATION.contact_client.get_contact_history(verification.contact_id)
                    contact = await APPLICATION.contact_client.get_contact_info(
                        verification.contact_id, history['timeline'][-1]['contact_history_id'])
                verification.contact_handle = contact.contact_handle
                result.mark_success()
        except Exception as error:
            echo(f"Error in migration of {verification}: {error}", Verbosity.SILENT, err=True)
            result.mark_error()
    return result


@verification_command()
def main() -> None:
    """Migrate contact handles."""
    result = asyncio.run(migrate_handles())
    echo(_OUT_MSG.format(total=result.total(), result=result), Verbosity.INFO)


if __name__ == '__main__':
    main()
