#!/usr/bin/env python3
"""Migrate verification domains.

This is a data migration after 'ec2101e7d847'.
"""
import asyncio

from regal.exceptions import ContactDoesNotExist
from setapp.constants import Verbosity
from sqlalchemy import select

from verification.commands.utils import echo, verification_command
from verification.database import atomic, get_engine
from verification.models import Domain, Verification
from verification.utils import Result

_OUT_MSG = ('Migrate domains: {total} total, {result.successes} migrated, {result.skips} skipped (no domains),'
            '{result.errors} errors.')


async def migrate_domains() -> Result:
    """Actually migrate domains."""
    result = Result()
    # Migrate verifications
    async with get_engine().begin() as conn:
        id_query = select(Verification.id).where(~Verification.domains.any())
        ids = (await conn.execute(id_query)).scalars()

    for id in ids:
        try:
            async with atomic() as session:
                query = select(Verification).with_for_update().filter_by(id=id)
                verification = (await session.execute(query)).scalars().one()
                if await verification.awaitable_attrs.domains:
                    # Already migrated.
                    continue
                try:
                    domains = await verification.affected_domains
                except ContactDoesNotExist:
                    # Contact is already deleted and it had no domains.
                    domains = []
                if domains:
                    for domain in domains:
                        ver_domain = Domain(verification_id=verification.id, domain_id=domain.id, fqdn=domain.fqdn)
                        session.add(ver_domain)
                    result.mark_success()
                else:
                    result.mark_skip()
        except Exception as error:
            echo(f"Error in migration of {verification}: {error}", Verbosity.SILENT, err=True)
            result.mark_error()

    return result


@verification_command()
def main() -> None:
    """Migrate domains."""
    result = asyncio.run(migrate_domains())
    echo(_OUT_MSG.format(total=result.total(), result=result), Verbosity.INFO)


if __name__ == '__main__':
    main()
