"""Checks for contact addresses."""
import difflib
import logging
import re
from functools import wraps
from itertools import filterfalse
from typing import Any, Callable, Dict, Iterable, Iterator, Sequence, Set, Tuple, TypeVar, Union, cast

import requests
from regal import Address, Contact

from verification.exceptions import SkipEvaluation

from .base import BaseCheck

_LOGGER = logging.getLogger(__name__)

X = TypeVar('X')
F = TypeVar('F', bound=Callable[..., Iterable])


def _deduplicate(func: F) -> F:
    """Filter generator results to return only the first copy of each object."""
    @wraps(func)
    def wrapped(*args: Any, **kwargs: Any) -> Iterator[X]:
        # Copied from itertools unique_everseen recipe.
        # See https://docs.python.org/3/library/itertools.html#itertools-recipes
        seen: Set[X] = set()
        # Python 3.9 coverage bug
        for item in filterfalse(seen.__contains__, func(*args, **kwargs)):
            seen.add(item)
            yield item
    return cast(F, wrapped)


class RuianAddressCheck(BaseCheck):
    """Check registry address against RUAIN.

    RUAIN - Registry of territorial identification, addresses and real estate, https://www.cuzk.cz/ruian.

    Attributes:
        query_threshold: Minimal value of a result to not try another query to RUIAN.
    """

    RUIAN_URL = ('https://ags.cuzk.cz/arcgis/rest/services/RUIAN/'
                 'Vyhledavaci_sluzba_nad_daty_RUIAN/MapServer/exts/GeocodeSOE/tables/1/suggest')

    def __init__(self, *, name: str, url: str = RUIAN_URL, timeout: Union[float, Tuple[float, float]] = 3.05,
                 query_threshold: float = 0.5):
        super().__init__(name=name)
        self.url = url
        self.timeout = timeout
        self.query_threshold = query_threshold

    @_deduplicate
    def _get_streets(self, street: str) -> Iterable[str]:
        yield street
        yield from (i.strip() for i in street.split(','))

    @_deduplicate
    def _get_queries(self, address: Address) -> Iterable[str]:
        """Generate queries for RUIAN."""
        # Simple street-city query.
        if address.street:
            base_street = address.street[0]
        else:
            base_street = ''

        # Drop all whitespaces from postal code
        postal_code = ''.join((cast(str, address.postal_code)).split())

        house_number_queries = []
        for street in self._get_streets(base_street):
            yield '{street}, {address.city}'.format(street=street, address=address)
            # A lot of addresses have incorrect city, try postal code instead.
            yield '{street}, {postal_code}'.format(street=street, postal_code=postal_code)

            # Some addresses are formatted incorrectly and RUIAN doesn't resolve those.
            # Replace \ with /
            fixed_street = street.replace('\\', '/')
            # Ensure space after a dot.
            fixed_street = re.sub(r'\.([^ ])', r'. \1', fixed_street)
            # Ensure space after a number in the beginning of the street name.
            fixed_street = re.sub(r'^([0-9.]+)(\s*)(\w)', r'\1 \3', fixed_street)
            # Ensure space before the number in the street name
            fixed_street = re.sub(r'([^\s\d/])(\d\w*)$', r'\1 \2', fixed_street)
            yield '{street}, {address.city}'.format(street=fixed_street, address=address)
            yield '{street}, {postal_code}'.format(street=fixed_street, postal_code=postal_code)

            hn_match = re.search(r'([\d\s/-]+)$', fixed_street)
            if hn_match:
                house_number_queries.append('{number}, {postal_code}'.format(number=hn_match.groups()[0].strip(),
                                                                             postal_code=postal_code))

            # Drop letters from the street number (most likely a house number)
            fixed_street = re.sub(r'([\w]\s*[0-9])\s*[-/]?\s*[a-zA-Z]+$', r'\1', fixed_street)
            yield '{street}, {address.city}'.format(street=fixed_street, address=address)
            yield '{street}, {postal_code}'.format(street=fixed_street, postal_code=postal_code)

            hn_match = re.search(r'([\d\s/-]+)$', fixed_street)
            if hn_match:
                house_number_queries.append('{number}, {postal_code}'.format(number=hn_match.groups()[0].strip(),
                                                                             postal_code=postal_code))

            # Drop any chars after the house number, e.g. '1/2' '1/A' '1/2A' '1 - A'
            fixed_street = re.sub(r'([0-9])\s*[/-]\s*.{1,6}$', r'\1', fixed_street)
            yield '{street}, {address.city}'.format(street=fixed_street, address=address)
            yield '{street}, {postal_code}'.format(street=fixed_street, postal_code=postal_code)

            hn_match = re.search(r'([\d\s/-]+)$', fixed_street)
            if hn_match:
                house_number_queries.append('{number}, {postal_code}'.format(number=hn_match.groups()[0].strip(),
                                                                             postal_code=postal_code))

        # Last resort - house number - postal code combinations
        yield from house_number_queries

    def _do_ruian(self, query: str) -> Sequence[Dict[str, Any]]:
        """Perform a query to RUIAN, return suggestions."""
        try:
            response = requests.post(self.url, {"f": "json", "text": query}, timeout=self.timeout)
            response.raise_for_status()
            return cast(Sequence[Dict[str, Any]], response.json()['suggestions'])
        except (requests.RequestException, KeyError) as error:
            _LOGGER.info("RUIAN request [query=%r] failed with %r", query, error)
            raise

    async def _get_results(self, contact: Contact) -> Tuple[Dict[str, Any], float]:
        assert contact.place is not None  # noqa: S101
        if contact.place.country_code != 'CZ':
            raise SkipEvaluation

        # Flatten the registry address
        street = ', '.join([s for s in contact.place.street if s])
        registry_address = "{street}, {place.postal_code} {place.city}".format(street=street, place=contact.place)
        if contact.place.state_or_province:
            registry_address += ', ' + contact.place.state_or_province

        # Query RUIAN
        value = 0.
        found = ''
        for query in self._get_queries(contact.place):
            suggestions = self._do_ruian(query)
            for suggestion in suggestions:
                new_value = difflib.SequenceMatcher(None, registry_address, suggestion['text']).ratio()
                if new_value > value:
                    value = new_value
                    found = suggestion['text']
            if value >= self.query_threshold:
                # Threshold was reached, address was resolved.
                break

        return {'registry': registry_address, 'source': found, 'query': query}, value
