#!/usr/bin/python3
"""Load contact IDs from file to score queue."""
import asyncio
import itertools
import logging
from typing import Iterable, TextIO, Tuple, TypeVar, cast

import click
from setapp.constants import Verbosity
from sqlalchemy import insert, select

from verification.commands.utils import echo, verification_command
from verification.database import atomic
from verification.models import ScoreQueue
from verification.utils import Result

_LOGGER = logging.getLogger(__name__)
_OUT_MSG = ('Load score queue: {total} total, {result.successes} loaded, {result.skips} skipped.')


X = TypeVar('X')


# XXX: Use itertools.batched when 3.12 is minimal supported python.
def _batched(iterable: Iterable[X], n: int) -> Iterable[Tuple[X, ...]]:
    if n < 1:  # pragma: no cover
        raise ValueError('n must be at least one')
    it = iter(iterable)
    while batch := tuple(itertools.islice(it, n)):
        yield batch


async def _load_score_queue(input: TextIO, *, batch_size: int = 100) -> Result:
    """Load the contacts IDs from input to score queue.

    Arguments:
        input: Buffer with the contact IDs.
        batch_size: Number of IDs processed at once.
    """
    result = Result()
    for batch in _batched(input, batch_size):
        async with atomic() as session:
            ids = tuple(i.strip() for i in batch if i.strip() != '')
            id_query = select(ScoreQueue.id).filter(ScoreQueue.id.in_(ids))
            found_ids = tuple((await session.execute(id_query)).scalars())
            queue_data = [{"id": i} for i in ids if i not in found_ids]
            if queue_data:
                await session.execute(insert(ScoreQueue), queue_data)
            result.update(success=len(ids) - len(found_ids), skip=len(found_ids))
    return result


@verification_command()
@click.option('--file', type=click.File('r'), default='-', show_default=True,
              help="A file to read the contact IDs from.")
def main(file: click.File) -> None:
    """Load contact IDs from file to score queue."""
    result = asyncio.run(_load_score_queue(cast(TextIO, file)))
    echo(_OUT_MSG.format(total=result.total(), result=result), Verbosity.INFO, log=_LOGGER.info)


if __name__ == '__main__':
    main()
