#!/usr/bin/python3
"""Run verification gRPC server."""
import asyncio
import multiprocessing
import signal
import types
from functools import partial

# Recommended for annotations instead of multiprocessing.Event directly by mypy's typeshed.
from multiprocessing.synchronize import Event
from typing import Iterable, Optional

import click
import grpc
from fred_api.verification import service_verification_grpc_pb2
from fred_api.verification.service_verification_grpc_pb2_grpc import add_VerificationServicer_to_server

from verification.commands.utils import verification_command
from verification.service import VerificationServicer
from verification.settings import APPLICATION

try:
    from grpc_reflection.v1alpha import reflection
except ImportError:
    reflection = None


async def _serve(stop_event: Event) -> None:
    server = grpc.aio.server(options=APPLICATION.settings.grpc_options,
                             maximum_concurrent_rpcs=APPLICATION.settings.grpc_maximum_concurrent_rpcs)
    add_VerificationServicer_to_server(VerificationServicer(), server)
    if reflection is not None:
        service_name = service_verification_grpc_pb2.DESCRIPTOR.services_by_name['Verification'].full_name
        reflection.enable_server_reflection([service_name, reflection.SERVICE_NAME], server)

    server.add_insecure_port('[::]:{port}'.format(port=APPLICATION.settings.grpc_port))
    await server.start()
    # Sadly this is the recommended way to keep the server running.
    try:
        # XXX: Coverage ignored, tested in a subprocess and not recorded.
        while not stop_event.is_set():  # pragma: no branch
            await asyncio.sleep(1)
    except asyncio.CancelledError:
        pass
    finally:
        await server.stop(0)


# XXX: Coverage ignored, tested in a subprocess and not recorded.
def _sigterm_handler(signum: int, frame: Optional[types.FrameType], *, stop_event: Event) -> None:  # pragma: no cover
    """Capture SIGTERM signal and set stop event."""
    stop_event.set()


@verification_command()
@click.option('-p', '--port', type=int, help="Set custom port.")
@click.option('-j', '--jobs', type=int, help="Set custom number of jobs to start. If 0, use all processors.")
@click.option('--grpc-options', multiple=True,
              help="Set custom options to the gRPC server, can be used multiple times.")
@click.option('--max-concurrency', type=int, help="Set custom maximum number of concurrent RPCs.")
def main(port: int, jobs: Optional[int], grpc_options: Iterable[str], max_concurrency: int) -> None:
    """Start the server."""
    if port:
        APPLICATION.settings.grpc_port = port
    if jobs is not None:
        APPLICATION.settings.grpc_jobs = jobs
    if grpc_options:
        options = [p.split('=', 1) for p in grpc_options]
        APPLICATION.settings.grpc_options = options
    if max_concurrency:
        APPLICATION.settings.grpc_maximum_concurrent_rpcs = max_concurrency

    # Credits https://the-fonz.gitlab.io/posts/python-multiprocessing/
    stop_event = multiprocessing.Event()
    signal.signal(signal.SIGTERM, partial(_sigterm_handler, stop_event=stop_event))
    workers = []
    for _ in range(APPLICATION.settings.grpc_jobs or multiprocessing.cpu_count()):
        worker = multiprocessing.Process(target=asyncio.run, args=(_serve(stop_event), ))
        worker.start()
        workers.append(worker)
    for worker in workers:
        worker.join()


if __name__ == '__main__':
    main()
