"""Command utilities."""
from functools import wraps
from typing import Any, Callable, Optional, TypeVar, cast

import click
from click import echo as _echo
from setapp.constants import Verbosity

import verification
from verification.settings import APPLICATION


def echo(message: str, verbosity: Verbosity, *, nl: bool = True, err: bool = False,
         log: Optional[Callable[[str], None]] = None) -> None:
    """Print message with respect to verbosity level.

    Arguments:
        message: The message to print.
        verbosity: Verbosity of the message.
        nl: Whether to print newline after message.
        err: Whether to print the message to error output instead.
        log: Callable to log the message as well.
    """
    if verbosity <= APPLICATION.verbosity:
        _echo(message, nl=nl, err=err)
    if log is not None:
        log(message)


F = TypeVar('F', bound=Callable[..., Any])


def verification_command(command: Callable = click.command) -> Callable:
    """Turn function into a verification command.

    Arguments:
        command: A click `command` callable to attach command to a group.
    """
    def decorator(func: F) -> F:
        # This decorator turns the function into a click command. To keep a reasonable order of options, that is
        # end usage with --connection, --config, --version and --help in this order,
        # we need to keep this order of click decorators.
        @click.version_option(version=verification.__version__, message="fred-verification %(version)s")
        @click.option('-v', '--verbosity', type=int, default=Verbosity.INFO, show_default=True,
                      help="Set verbosity level.")
        @click.option('--config', type=click.Path(exists=True, dir_okay=False), help="Set custom config file.")
        @click.option('--connection', help="Set custom database connection string.")
        @command()
        @wraps(func)
        def wrapped(*, config: Optional[str], connection: str, verbosity: int, **kwargs: Any) -> Any:
            APPLICATION.load(config)
            APPLICATION.verbosity = cast(Verbosity, verbosity)
            if connection:
                APPLICATION.settings.db_connection = connection
            return func(**kwargs)

        return cast(F, wrapped)
    return decorator
