"""Constants for verification."""
from enum import Enum, unique
from typing import Any


@unique
class VerificationStatus(str, Enum):
    """Enum of verification statuses.

    Members:
        CREATED: Verification is new.
        INITIALIZED: Initial verification steps have been finished.
        ACCESSED: Contact has performed at least one interaction with the verification process.
        PENDING: Contact has chosen verification method and verification is waiting for contact input or action.
        CONSTRAINED: Constraints have been imposed since insufficient progress has been made.
        FUTILE: Verification was unsuccessful, but irreversible effects have not been imposed yet.
        PAUSED: Verification is paused, automatic processes are stopped.
        SUCCEEDED: Verification has been completed successfully.
        FAILED: Verification has been completed unsuccessfully.
        CANCELED: Verification has been canceled.
        MANUAL: Verification has been completed successfully, manually outside of the verification system.

    To ease comparison of statuses, members are partially ordered.
    Open statuses have ascending order up to closed statuses.
    Closed statuses return ``False`` when compared to each other.
    """

    CREATED = 'created'
    INITIALIZED = 'initialized'
    ACCESSED = 'accessed'
    PENDING = 'pending'
    CONSTRAINED = 'constrained'
    FUTILE = 'futile'
    PAUSED = 'paused'
    SUCCEEDED = 'succeeded'
    FAILED = 'failed'
    CANCELED = 'canceled'
    MANUAL = 'manual'

    def __ge__(self, other: Any) -> bool:
        if not isinstance(other, VerificationStatus):
            return NotImplemented
        return self == other or self > other

    def __gt__(self, other: Any) -> bool:
        if not isinstance(other, VerificationStatus):
            return NotImplemented
        # Explicit check for closed statuses.
        if self not in OPEN_STATUSES and other not in OPEN_STATUSES:
            return False
        # Compare open (or open and closed) statuses by their order in the enum.
        items = list(VerificationStatus)
        return items.index(self) > items.index(other)

    def __le__(self, other: Any) -> bool:
        if not isinstance(other, VerificationStatus):
            return NotImplemented
        return self == other or self < other

    def __lt__(self, other: Any) -> bool:
        if not isinstance(other, VerificationStatus):
            return NotImplemented
        # A < B <==> B > A
        return other > self


OPEN_STATUSES = {VerificationStatus.CREATED, VerificationStatus.INITIALIZED, VerificationStatus.ACCESSED,
                 VerificationStatus.PENDING, VerificationStatus.CONSTRAINED, VerificationStatus.FUTILE,
                 VerificationStatus.PAUSED}


@unique
class EventType(str, Enum):
    """Enum of event types."""

    CANCEL = 'cancel'
    CHECK_DOMAINS = 'check_domains'
    COMPLETE = 'complete'
    CREATE = 'create'
    DELAY_STATUS = 'delay_status'
    DOMAINS_CHANGED = 'domains_changed'
    FAILED = 'failed'
    FUTILE = 'futile'
    MANUAL = 'manual'
    OUTZONE = 'outzone'
    REQUEST_MANUAL_CHECK = 'request_manual_check'
    RESOLVE_MANUAL_CHECK = 'resolve_manual_check'
    RESUME = 'resume'
    SEND_MESSAGE = 'send_message'
    SET_CONTACT_REPRESENTATIVE = 'set_contact_representative'
    USE_METHOD = 'use_method'
    VERIFY_CONTACT_DATA = 'verify_contact_data'
    VERIFY_TOKEN = 'verify_token'  # noqa: S105


@unique
class TransportMethod(str, Enum):
    """Enum of transport methods."""

    EMAIL = 'email'
    LETTER = 'letter'
    SMS = 'sms'


@unique
class MessageType(str, Enum):
    """Enum of message types."""

    CANCELED = 'verification_canceled'
    CONFIRMATION_LETTER = 'verification_confirmation_letter'
    FAILED = 'verification_failed'
    FUTILE_REMINDER = 'verification_futile_reminder'
    INITIAL_EMAIL = 'verification_initial'
    MANUAL_CHECK_RESOLUTION = 'verification_manual_check_resolution'
    OUTZONE = 'verification_outzone'
    OUTZONE_REMINDER = 'verification_outzone_reminder'
    PAUSED_NOTIFICATION = 'verification_paused_notification'
    REPRESENTATIVE_LETTER = 'verification_representative_letter'
    SUCCEEDED = 'verification_succeeded'


@unique
class Template(str, Enum):
    """Enum of templates."""

    CANCELED_SUBJECT = 'verification-canceled-subject.txt'
    CANCELED_BODY = 'verification-canceled-body.txt'
    CANCELED_ADMIN_BODY = 'verification-canceled-admin-body.txt'
    CONFIRMATION_LETTER = 'verification-confirmation-letter.html'
    FAILED_SUBJECT = 'verification-failed-subject.txt'
    FAILED_BODY = 'verification-failed-body.txt'
    FUTILE_REMINDER_SUBJECT = 'verification-futile-reminder-subject.txt'
    FUTILE_REMINDER_BODY = 'verification-futile-reminder-body.txt'
    FUTILE_REMINDER_ADMIN_BODY = 'verification-futile-reminder-admin-body.txt'
    INITIAL_EMAIL_SUBJECT = 'verification-initial-subject.txt'
    INITIAL_EMAIL_BODY = 'verification-initial-body.txt'
    INITIAL_EMAIL_ADMIN_BODY = 'verification-initial-admin-body.txt'
    MANUAL_CHECK_RESOLUTION_SUBJECT = 'verification-manual-check-resolution-subject.txt'
    MANUAL_CHECK_RESOLUTION_BODY = 'verification-manual-check-resolution-body.txt'
    OUTZONE_SUBJECT = 'verification-outzone-subject.txt'
    OUTZONE_BODY = 'verification-outzone-body.txt'
    OUTZONE_REMINDER_SUBJECT = 'verification-outzone-reminder-subject.txt'
    OUTZONE_REMINDER_BODY = 'verification-outzone-reminder-body.txt'
    OUTZONE_REMINDER_ADMIN_BODY = 'verification-outzone-reminder-admin-body.txt'
    PAUSED_NOTIFICATION_SUBJECT = 'verification-paused-subject.txt'
    PAUSED_NOTIFICATION_BODY = 'verification-paused-body.txt'
    REPRESENTATIVE_LETTER = 'verification-representative-letter.html'
    SUCCEEDED_SUBJECT = 'verification-succeeded-subject.txt'
    SUCCEEDED_BODY = 'verification-succeeded-body.txt'
    SUCCEEDED_ADMIN_BODY = 'verification-succeeded-admin-body.txt'


class ObjectState(str, Enum):
    """Enum of registry object state flags."""

    SERVER_BLOCKED = 'serverBlocked'
    SERVER_OUTZONE_MANUAL = 'serverOutzoneManual'
    SERVER_LINK_PROHIBITED = 'serverLinkProhibited'
    CONTACT_IN_MANUAL_VERIFICATION = 'contactInManualVerification'
    CONTACT_PASSED_MANUAL_VERIFICATION = 'contactPassedManualVerification'
    CONTACT_FAILED_MANUAL_VERIFICATION = 'contactFailedManualVerification'


@unique
class ReferenceType(str, Enum):
    """Enum of referenced object types."""

    CONTACT = 'contact'
    VERIFICATION = 'verification'
