"""Database utilities."""
from functools import lru_cache
from typing import AsyncGenerator

from asyncstdlib.contextlib import contextmanager
from sqlalchemy.ext.asyncio import (
    AsyncEngine,
    AsyncSession,
    async_object_session,
    async_sessionmaker,
    create_async_engine,
)

from .settings import APPLICATION
from .utils import import_path


@lru_cache
def get_engine() -> AsyncEngine:
    """Return engine as defined by settings."""
    if APPLICATION.settings.db_poolclass:
        poolclass = import_path(APPLICATION.settings.db_poolclass)
    else:
        poolclass = None
    engine = create_async_engine(APPLICATION.settings.db_connection, echo=APPLICATION.settings.db_echo,
                                 poolclass=poolclass)
    # Redirect all queries to schema defined in settings.
    # See https://docs.sqlalchemy.org/en/20/core/connections.html#translation-of-schema-names
    engine.update_execution_options(schema_translate_map={None: APPLICATION.settings.db_schema})
    return engine


@lru_cache
def get_session_cls() -> async_sessionmaker[AsyncSession]:
    """Return session class for defined engine."""
    return async_sessionmaker(bind=get_engine())


@contextmanager
async def atomic() -> AsyncGenerator[AsyncSession, None]:
    """Context manager which wraps content into a transaction."""
    async with get_session_cls()() as session:
        try:
            yield session
            await session.commit()
        except BaseException:
            await session.rollback()
            raise


def object_session(model: object) -> AsyncSession:
    """Return the object session or raise an error if not bound to one.

    Raises:
        ValueError: Object is not bound to a session.

    Modified version of SQLAlchemy's object_session which raises error rather than return None if session is not found.
    """
    session = async_object_session(model)
    if session is None:
        raise ValueError("{} is not bound to a session.".format(model))
    return session
