"""Evaluator - the top level handler of contact scoring."""
import logging
import math
import signal
import statistics
import types
from asyncio import sleep
from datetime import datetime, timedelta
from functools import cached_property
from typing import Dict, Iterable, Optional, Sequence, cast

from fred_types import ContactId
from pytz import UTC
from regal.exceptions import ContactDoesNotExist
from sqlalchemy import or_, select
from sqlalchemy.sql import expression, func
from sqlalchemy.sql.elements import SQLCoreOperations

from .checks import BaseCheck
from .database import atomic, get_engine, object_session
from .exceptions import SkipEvaluation
from .models import Score, ScoreQueue
from .settings import APPLICATION
from .utils import Result, load_instance

_LOGGER = logging.getLogger(__name__)
_LAST_SCORE_QUERY = select(
    expression.true().label("has_score"),
    Score.contact_id,
    Score.create_datetime,
    func.lag(Score.id, 1).over(partition_by=Score.contact_id,
                               order_by=Score.create_datetime.desc()).label("next_id"))


class Evaluator:
    """Evaluates contact scores."""

    FAIL_LIMIT: int = 10

    def __init__(self) -> None:
        # Whether to stop `process_all`.
        self.stop_process: bool = False
        signal.signal(signal.SIGTERM, self._sigterm_handler)
        # Counter of contact evaluation failures.
        self._failures: Dict[ContactId, int] = {}

    def _sigterm_handler(self, signum: int, frame: Optional[types.FrameType]) -> None:
        """Capture SIGTERM signal and setup graceful stop."""
        _LOGGER.info("Captured %s, stopping evaluator...", signal)
        self.stop_process = True

    @cached_property
    def checks(self) -> Sequence[BaseCheck]:
        """Return check instances."""
        checks = []
        for name, settings in APPLICATION.settings.checks.items():
            settings = settings.copy()
            cls_name = settings.pop('class', None)
            if cls_name is None:
                raise RuntimeError("Check {} has no class defined.".format(name))
            try:
                checks.append(load_instance(cls_name, name=name, **settings))
            except Exception as error:
                raise RuntimeError("Error in loading check: {}".format(error)) from error
        return tuple(checks)

    def rms(self, values: Iterable[float]) -> float:
        """Return root mean square of the values."""
        return math.sqrt(statistics.mean(v ** 2 for v in values))

    async def evaluate(self, queue: ScoreQueue) -> bool:
        """Completely evaluate the contact.

        Performs the evaluation of the contact and stores the score into database.

        Returns: Whether the contact was processed.
        """
        if not self.checks:
            return False

        session = object_session(queue)
        try:
            contact = await APPLICATION.contact_client.get_contact_info(queue.id)
        except ContactDoesNotExist:
            _LOGGER.info('Contact %s deleted from queue, because it does not exist.', queue.id)
            await session.delete(queue)
            return True

        try:
            results = [await check.evaluate(contact) for check in self.checks]
        except SkipEvaluation:
            queue.skipped_at = cast(SQLCoreOperations[Optional[datetime]], func.now())
            return False
        except Exception:
            failures = self._failures.get(contact.contact_id, 0) + 1
            if failures >= self.FAIL_LIMIT:
                # Too many failures, store empty score and move on.
                score = Score(contact_id=contact.contact_id, contact_history_id=contact.contact_history_id, value=0)
                session.add(score)
                return True
            self._failures[contact.contact_id] = failures
            raise
        score = Score(contact_id=contact.contact_id, contact_history_id=contact.contact_history_id,
                      value=self.rms(r.value for r in results))
        session.add(score)
        # Flush model to database
        await session.flush()

        for result in results:
            result.score_id = score.id
            session.add(result)
        return True

    def _get_cutoff(self, delay: Optional[timedelta] = None) -> datetime:
        """Return cutoff based on delay."""
        return datetime.now(UTC) - (delay or timedelta(0))

    async def evaluate_all(self, renew_delay: Optional[timedelta] = None, throttle: int = 0, batch_size: int = 10_000,
                           *, skip_delay: timedelta = timedelta(days=1)) -> Result:
        """Evaluate all contacts in the queue.

        Arguments:
            renew_delay: A minimal delay between two evaluations of the same contact.
            throttle: A delay in seconds between any two evaluations.
            batch_size: A maximal number of contacts evaluated in a single batch.
            skip_delay: A minimal delay before new evaluation of a contact which was skipped.
        """
        self.stop_process = False
        result = Result()

        # Fetch a list of IDs to process.
        last_score_subquery = _LAST_SCORE_QUERY.subquery().alias("last_scores")
        async with get_engine().connect() as connection:
            id_query = select(ScoreQueue.id).outerjoin(last_score_subquery,
                                                       ScoreQueue.id == last_score_subquery.columns.contact_id)
            id_query = id_query.filter(
                or_(ScoreQueue.skipped_at.is_(None),
                    ScoreQueue.skipped_at < self._get_cutoff(skip_delay)),
                last_score_subquery.columns.next_id.is_(None),
                or_(last_score_subquery.columns.create_datetime.is_(None),
                    last_score_subquery.columns.create_datetime < self._get_cutoff(renew_delay)))
            # Sort the query to evaluate new contacts first and then from the oldest.
            id_query = id_query.order_by(last_score_subquery.columns.has_score.desc(),
                                         last_score_subquery.columns.create_datetime.asc())
            ids = (await connection.execute(id_query.limit(batch_size))).scalars()

        # Process individual IDs.
        for id in ids:
            if self.stop_process:
                # Stop the processing gracefully.
                return result
            _LOGGER.debug('Processing [%s]', id)
            try:
                async with atomic() as session:
                    # XXX: Window functions are not allowed in FOR UPDATE statements in postgresql.
                    # We need to check the last score separately.
                    query = select(ScoreQueue).filter_by(id=id).with_for_update(skip_locked=True)
                    query = query.filter(
                        or_(ScoreQueue.skipped_at.is_(None),
                            ScoreQueue.skipped_at < self._get_cutoff(skip_delay)))
                    queue = (await session.execute(query)).scalar()
                    if queue is None:
                        continue
                    # Check the last score is still old enough
                    last_score_query = select(func.max(Score.create_datetime)).filter_by(contact_id=queue.id)
                    last_score_datetime = (await session.execute(last_score_query)).scalar()
                    if last_score_datetime is not None and last_score_datetime > self._get_cutoff(renew_delay):
                        continue
                    _LOGGER.debug('Evaluating [%s]', id)
                    if await self.evaluate(queue):
                        result.mark_success()
                    else:
                        result.mark_skip()
                    await sleep(throttle)
            except Exception as error:
                _LOGGER.exception('An error in evaluation %r: %r', id, error)
                result.mark_error()
        return result
