"""Verification handler - the top level handler of verification processing."""
import logging
import signal
import types
from datetime import datetime, timedelta
from functools import cached_property, partial, wraps
from itertools import takewhile
from operator import attrgetter
from typing import (
    Any,
    AsyncIterable,
    AsyncIterator,
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    TypeVar,
    cast,
)

from asyncstdlib import filterfalse, list as alist
from fred_types import ContactId
from pytz import UTC
from regal import DomainContactRole
from regal.exceptions import ContactDoesNotExist, DomainDoesNotExist
from sqlalchemy import and_, or_, select
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.sql import func
from sqlalchemy.sql.elements import SQLCoreOperations

from .constants import OPEN_STATUSES, EventType, VerificationStatus
from .database import atomic, get_engine, object_session
from .exceptions import VerificationAlreadyExists
from .models import Domain, Event, Verification
from .processors.base import BaseProcessor, ScheduledEvent
from .settings import APPLICATION
from .utils import Result, load_instance

_LOGGER = logging.getLogger(__name__)


X = TypeVar('X')
F = TypeVar('F', bound=Callable[..., AsyncIterator])


def _deduplicate(func: F) -> F:
    """Filter async generator results to return only the first copy of each object."""
    @wraps(func)
    async def wrapped(*args: Any, **kwargs: Any) -> AsyncIterator[X]:
        # Copied from itertools unique_everseen recipe. Used list to support unhashable objects.
        # See https://docs.python.org/3/library/itertools.html#itertools-recipes
        seen: List[X] = []
        # Python 3.9 coverage bug
        async for item in cast(AsyncIterator[X],
                               filterfalse(seen.__contains__, func(*args, **kwargs))):  # pragma: no cover
            seen.append(item)
            yield item
    return cast(F, wrapped)


class VerificationHandler:
    """Verification handler - the top level handler of verification processing.

    Attributes:
        loop_limit: Maximal amount of events expected to be run at once. If exceeded, an error is logged.
        stop_process: Whether the handler processing should be stopped.
    """

    loop_limit = 10

    def __init__(self) -> None:
        # Whether to stop `process_all`.
        self.stop_process: bool = False
        signal.signal(signal.SIGTERM, self._sigterm_handler)

    def _sigterm_handler(self, signum: int, frame: Optional[types.FrameType]) -> None:
        """Capture SIGTERM signal and setup graceful stop."""
        logging.info("Captured %s, stopping handler...", signal)
        self.stop_process = True

    @cached_property
    def processors(self) -> Dict[str, BaseProcessor]:
        """Return processor instances."""
        processors = {}
        for method, settings in APPLICATION.settings.processors.items():
            settings = settings.copy()
            cls_name = settings.pop('class', None)
            if cls_name is None:
                raise RuntimeError("Processor {} has no class defined.".format(method))
            try:
                processors[method] = load_instance(cls_name, method=method, **settings)
            except Exception as error:
                raise RuntimeError("Error in loading processor: {}".format(error)) from error
        return processors

    async def _get_active_processors(self, verification: Verification) -> Sequence[BaseProcessor]:
        """Return an iterable of active processors in forward sorting."""
        active_processors = [p for n, p in self.processors.items() if n in await verification.active_methods]
        if active_processors:
            return active_processors

        # Fallback to the first applicable processor as a default.
        for processor in self.processors.values():
            if await processor.is_applicable(verification):
                return (processor, )

        # Nothing found.
        _LOGGER.warning("No processor found to process %s", verification)
        return ()

    async def get_active_processors(self, verification: Verification, reverse: bool = False
                                    ) -> AsyncIterable[BaseProcessor]:
        """Return an iterable of active processors for the verification.

        Arguments:
            verification: Verification to act upon.
            reverse: Whether to return active processors in reversed order.
        """
        active_processors = await self._get_active_processors(verification)
        if reverse:
            for processor in reversed(active_processors):
                yield processor
        else:
            for processor in active_processors:
                yield processor

    @_deduplicate
    async def get_next_events(self, verification: Verification) -> AsyncIterator[ScheduledEvent]:
        """Collect all scheduled events for the verification."""
        async for processor in self.get_active_processors(verification):
            # Python 3.9 coverage bug
            async for scheduled in processor.get_next_events(verification):  # pragma: no cover
                yield scheduled

    @_deduplicate
    async def get_cancel_events(self, verification: Verification, status: Optional[VerificationStatus] = None,
                                ) -> AsyncIterator[ScheduledEvent]:
        """Collect all cancel events for the verification.

        Arguments:
            verification: Verification to act upon.
            status: If provided, return events to cancel that status if applicable.
                    If None, return events to cancel the whole verification.
        """
        # For cancel, processors are iterated in reversed order.
        async for processor in self.get_active_processors(verification, reverse=True):
            # Python 3.9 coverage bug
            async for scheduled in processor.get_cancel_events(verification, status):  # pragma: no cover
                yield scheduled

    async def _process(self, verification: Verification,
                       event_fetcher: Callable[[Verification], AsyncIterator[ScheduledEvent]]) -> bool:
        """Get and process events on verification.

        Internal method shared by all processing methods. Events are generated by calling the `event_fetcher` provided.

        Return: Whether the verification was actually processed.
        """
        # First, check whether the contact still exists, close the verification if not.
        try:
            await verification.contact
        except ContactDoesNotExist:
            if verification.status in OPEN_STATUSES:
                _LOGGER.info('%s canceled because the contact does not exist.', verification)
                # The contact was deleted. Mark verification as canceled since there's nothing else to be done.
                verification.close_datetime = cast(SQLCoreOperations[Optional[datetime]], func.now())
                verification.status = VerificationStatus.CANCELED
                # Verification was closed as canceled, hence processed.
                return True
            else:
                # Contact doesn't exist and verification is closed, there is nothing else to be done.
                return False

        # Check affected domains on open verification before event processing
        if verification.status in OPEN_STATUSES:
            result = await self.check_domains(verification)
        else:
            result = False

        session = object_session(verification)
        for _ in range(self.loop_limit):
            savepoint = await session.begin_nested()
            async for scheduled in event_fetcher(verification):
                if scheduled.due_datetime is None or scheduled.due_datetime <= datetime.now(UTC):
                    await scheduled.process()
                    session.add(await scheduled.to_event())
                    _LOGGER.debug('An event %s processed for %s.', scheduled.type.value, verification)
                    result = True
                    break
            else:
                return result
            await savepoint.commit()
            await session.refresh(verification)
        else:
            _LOGGER.error('Processing of %s exceeded the loop limit.', verification)
            return result

    async def process(self, verification: Verification) -> bool:
        """Get and process events on verification.

        Verification is processed in these steps:
            1. Check if verification is complete, close the verification if so.
               Do not stop to process any post-complete scheduled events.
            2. A list of active processors is set up as:
               * Processors for all activated methods for the verification.
               * If there are no activated methods for the verification, first applicable processor is used
                 as a default.
            3. A list of events by all active processors is collected.
            4. First due event found is processed.
            5. Cycle is repeated until due events run out.

        Returns: Whether any event was processed.
        """
        # Use `any` to prevent any short-circuiting.
        return any((await self.try_complete(verification),
                    await self._process(verification, self.get_next_events)))

    async def process_all(self, *, closed_limit: int = 0) -> Result:
        """Run through open and recently closed verifications and process scheduled events.

        Arguments:
            closed_limit: A maximum interval (in seconds) passed since a verification is closed to be processed.
        """
        self.stop_process = False
        result = Result()
        async with get_engine().connect() as connection:
            if connection.dialect.name != 'sqlite':
                connection = await connection.execution_options(isolation_level='REPEATABLE READ')
            cutoff = datetime.now(UTC) - timedelta(seconds=closed_limit)
            id_query = select(Verification.id).filter(
                or_(Verification.status.in_(OPEN_STATUSES), Verification.close_datetime >= cutoff))
            for id in (await connection.execute(id_query)).scalars():
                if self.stop_process:
                    # Stop the processing gracefully.
                    return result
                _LOGGER.debug('Processing verification[%s]', id)
                try:
                    async with atomic() as session:
                        query = select(Verification).filter_by(id=id).with_for_update(skip_locked=True)
                        verification = (await session.execute(query)).scalar()
                        if verification is None:
                            continue
                        if await self.process(verification):
                            result.mark_success()
                        else:
                            result.mark_skip()
                except Exception as error:
                    _LOGGER.exception('An error in processing verification[%s]: %r', id, error)
                    result.mark_error()
        return result

    async def cancel(self, verification: Verification, status: Optional[VerificationStatus] = None) -> None:
        """Get and process cancel events on verification.

        Verification cancel is processed in these steps:
            1. A list of active processors is set up as:
               * Processors for all activated methods for the verification in the reversed order.
               * If there are no activated methods for the verification, first applicable processor is used
                 as a default.
            2. A list of cancel events by all active processors is collected.
            3. First event found is processed.
            4. Cycle is repeated until cancel events run out.

        Arguments:
            verification: Verification to act upon.
            status: If provided, return events to cancel that status if applicable.
                    If None, return events to cancel the whole verification.
        """
        event_fetcher = partial(self.get_cancel_events, status=status)
        await self._process(verification, event_fetcher)

    async def complete(self, verification: Verification, processor: BaseProcessor) -> None:
        """Get and process complete events on verification.

        Verification cancel is processed in these steps:
            1. A list of complete events by the processor is collected.
            3. First event found is processed.
            4. Cycle is repeated until complete events run out.

        Arguments:
            verification: Verification to act upon.
            processor: A processor which completes the verification.
        """
        await self._process(verification, processor.get_complete_events)

    async def try_complete(self, verification: Verification) -> bool:
        """Check if the verification is complete, close the verification if so.

        Returns: Whether verification was completed.
        """
        # Python 3.9 coverage bug
        async for processor in self.get_active_processors(verification):  # pragma: no cover
            if await processor.check_complete(verification):
                await self.complete(verification, processor)
                return True
        return False

    async def _should_disperse(self, session: AsyncSession, contact_id: ContactId) -> bool:
        """Return whether verification should disperse to the contact or not."""
        if APPLICATION.settings.dispersion is None:
            # Dispersion is disabled
            return False

        # Prepare cut off for recent verifications.
        if APPLICATION.settings.dispersion.max_age:
            cutoff = datetime.now(UTC) - timedelta(days=APPLICATION.settings.dispersion.max_age)
        else:
            cutoff = datetime.fromtimestamp(0, UTC)

        where = or_(
            # Open verification
            Verification.status.in_(OPEN_STATUSES),
            # Recently successful verification
            and_(Verification.status.in_((VerificationStatus.SUCCEEDED, VerificationStatus.MANUAL)),
                 Verification.close_datetime > cutoff))
        query = select(Verification).filter_by(contact_id=contact_id).filter(where)
        if (await session.execute(query)).scalar():
            # There is open of recently successful verification
            return False

        status = await APPLICATION.contact_client.get_contact_state(contact_id)
        if any(status.get(f, False) for f in APPLICATION.settings.dispersion.verified_statuses):
            # Contact has one of verified status flags
            return False

        for flag in APPLICATION.settings.dispersion.transient_statuses:
            if status.get(flag, False):
                # Temporary flag found, find how old is it.
                history = await APPLICATION.contact_client.get_contact_state_history(contact_id)

                def _filter_rows(row: Dict[str, Any]) -> bool:
                    return cast(bool, row['flags'].get(flag, False))  # noqa: B023

                valid_from = list(takewhile(_filter_rows, history['timeline'][::-1]))[-1]['valid_from']
                if valid_from > cutoff:
                    return False

        return True

    async def check_domains(self, verification: Verification) -> bool:
        """Check domains affected by the verification.

        Returns: Whether affected domains changed.
        """
        session = object_session(verification)
        result = await APPLICATION.domain_client.get_domains_by_contact(verification.contact_id,
                                                                        {DomainContactRole.HOLDER})
        # ID to FQDN map
        registry_domains = {d['domain'].id: d['domain'].fqdn for d in result['domains']}
        # ID to Domain map
        domains = {d.domain_id: d for d in await verification.awaitable_attrs.domains}

        added = []
        for domain_id in (d for d in registry_domains if d not in domains):
            added.append(domain_id)
            session.add(Domain(verification_id=verification.id, domain_id=domain_id, fqdn=registry_domains[domain_id]))

        removed = []
        for domain_id in (d for d in domains if d not in registry_domains):
            removed.append(domain_id)
            await session.delete(domains[domain_id])

            try:
                removed_domain = await APPLICATION.domain_client.get_domain_info(domain_id)
            except DomainDoesNotExist:
                # Domain have been deleted, nothing to do.
                continue

            # Cancel outzone requests on the transferred domain.
            outzone_events = sorted(await verification.get_events(EventType.OUTZONE),
                                    key=attrgetter('event_datetime'), reverse=True)
            if outzone_events and not outzone_events[0].details.get('canceled'):
                request_ids = outzone_events[0].details.get('request_ids', {})
            else:
                request_ids = {}

            # Collect request IDs to cancel
            if isinstance(request_ids, list):
                # Legacy structure, do nothing
                close_requests = []
            else:
                # Decode request IDs for the transferred domain.
                close_requests = [r for d, s in request_ids.items() if d == domain_id for r in s]

            # Cancel the outzone.
            await alist(APPLICATION.domain_admin_client.manage_domain_state_flags(close_requests=close_requests))

            # Handle dispersion
            new_registrant = removed_domain.registrant
            assert new_registrant is not None  # noqa: S101
            if await self._should_disperse(session, new_registrant.id):
                await self.create(session, new_registrant.id)

        if added or removed:
            details = {}
            if added:
                details['added'] = added
            if removed:
                details['removed'] = removed
            event = Event(verification_id=verification.id, type=EventType.DOMAINS_CHANGED, details=details)
            session.add(event)

        return bool(added or removed)

    async def create(self, session: AsyncSession, contact_id: ContactId, *, reason: str = '', reason_url: str = '',
                     ) -> Verification:
        """Create and return new verification.

        Raises:
            ContactDoesNotExist: If contact_id does not exist.
            VerificationAlreadyExists: If contact already has an open verification.
        """
        # Let ContactDoesNotExist be raised.
        contact = await APPLICATION.contact_client.get_contact_info(contact_id)

        # Check if verification already exist.
        query = select(Verification).filter_by(contact_id=contact.contact_id).filter(
            Verification.status.in_(OPEN_STATUSES))
        if (await session.execute(query)).scalar():
            raise VerificationAlreadyExists

        verification = Verification(contact_id=contact.contact_id, contact_handle=contact.contact_handle,
                                    status=VerificationStatus.CREATED, reason=reason, reason_url=reason_url)
        session.add(verification)
        # Flush the model to database to get its id.
        await session.flush()

        # Fetch and store contact domains
        domain_ids = []
        result = await APPLICATION.domain_client.get_domains_by_contact(contact.contact_id,
                                                                        {DomainContactRole.HOLDER})
        for domain in result['domains']:
            ver_domain = Domain(verification_id=verification.id, domain_id=domain['domain'].id,
                                fqdn=domain['domain'].fqdn)
            session.add(ver_domain)
            domain_ids.append(domain['domain'].id)

        event = Event(verification_id=verification.id, type=EventType.CREATE, details={'domains': domain_ids})
        session.add(event)

        return verification
