"""Add Verification.

Revision ID: 02642a30fecc
Revises:
Create Date: 2022-06-16 08:02:00.691972+00:00
"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.sql import expression, func
from sqlalchemy_utc import UtcDateTime
from sqlalchemy_utils import JSONType, ScalarListType, UUIDType

# revision identifiers, used by Alembic.
revision = "02642a30fecc"
down_revision = None
branch_labels = None
depends_on = None


_ADDRESS_CHECK_SQL = (
    "(address_street IS NULL AND coalesce(address_city, address_state, address_postal_code, address_country_code) "
    "IS NULL) = (verified_address IS NULL)"
)


def upgrade() -> None:
    op.create_table(
        "verification_verification",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("uuid", UUIDType(binary=False), nullable=False),
        sa.Column("contact_id", sa.Text(), nullable=False),
        sa.Column("reason", sa.Text(), server_default='', nullable=False),
        sa.Column("access_token", sa.Text(), nullable=False),
        sa.Column(
            "create_datetime",
            UtcDateTime(timezone=True),
            server_default=func.now(),
            nullable=False,
        ),
        sa.Column("close_datetime", UtcDateTime(timezone=True), nullable=True),
        sa.Column("last_access_datetime", UtcDateTime(timezone=True), nullable=True),
        sa.Column(
            "status",
            sa.Enum(
                "CREATED",
                "INITIALIZED",
                "ACCESSED",
                "PENDING",
                "CONSTRAINED",
                "FUTILE",
                "SUCCEEDED",
                "FAILED",
                "CANCELED",
                name="verificationstatus",
                create_constraint=True,
            ),
            nullable=False,
        ),
        sa.Column(
            "manual_check_request",
            sa.Boolean(name="manual_check_request", create_constraint=True),
            nullable=False,
            server_default=expression.false(),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("verification_verification_pkey")),
        sa.UniqueConstraint("uuid", name=op.f("verification_verification_uuid_key")),
        sa.Index(op.f("verification_verification_status_key"), "status", unique=False),
        sa.Index(
            op.f("verification_verification_contact_id_active_key"),
            "contact_id",
            unique=True,
            postgresql_where=sa.text(
                "status IN ('CREATED', 'INITIALIZED', 'ACCESSED', 'PENDING', 'CONSTRAINED', 'FUTILE')"
            ),
            sqlite_where=sa.text(
                "status IN ('CREATED', 'INITIALIZED', 'ACCESSED', 'PENDING', 'CONSTRAINED', 'FUTILE')"
            ),
        )
    )

    op.create_table(
        "verification_token",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("uuid", UUIDType(binary=False), nullable=False),
        sa.Column(
            "create_datetime",
            UtcDateTime(timezone=True),
            server_default=func.now(),
            nullable=False,
        ),
        sa.Column("verification_id", sa.Integer(), nullable=False),
        sa.Column("token", sa.Text(), nullable=False),
        sa.Column("method", sa.Text(), nullable=False),
        sa.ForeignKeyConstraint(
            ["verification_id"], ["verification_verification.id"], name=op.f("verification_token_verification_id_fk")
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("verification_token_pkey")),
        sa.UniqueConstraint("uuid", name=op.f("verification_token_uuid_key")),
        sa.UniqueConstraint("verification_id", name=op.f("verification_token_verification_id_key")),
    )
    op.create_table(
        "verification_event",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column(
            "event_datetime",
            UtcDateTime(timezone=True),
            server_default=func.now(),
            nullable=False,
        ),
        sa.Column("verification_id", sa.Integer(), nullable=False),
        sa.Column("method", sa.Text(), nullable=True),
        sa.Column("type", sa.Text(), nullable=False),
        sa.Column("details", JSONType(), nullable=False),
        sa.ForeignKeyConstraint(
            ["verification_id"], ["verification_verification.id"], name=op.f("verification_event_verification_id_fk")
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("verification_event_pkey")),
    )

    op.create_table(
        "verification_contactdata",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("verification_id", sa.Integer(), nullable=False),
        sa.Column(
            "verify_datetime",
            UtcDateTime(timezone=True),
            server_default=func.now(),
            nullable=False,
        ),
        sa.Column("method", sa.Text(), nullable=True),
        sa.Column("contact_history_id", sa.Text(), nullable=False),
        sa.Column("name", sa.Text(), nullable=True),
        sa.Column(
            "verified_name",
            sa.Boolean(name="verified_name", create_constraint=True),
            sa.CheckConstraint("(name IS NULL) = (verified_name IS NULL)", name="name_null"),
            nullable=True,
        ),
        sa.Column("organization", sa.Text(), nullable=True),
        sa.Column(
            "verified_organization",
            sa.Boolean(name="verified_organization", create_constraint=True),
            sa.CheckConstraint("(organization IS NULL) = (verified_organization IS NULL)", name="organization_null"),
            nullable=True,
        ),
        sa.Column("birthdate", sa.Date(), nullable=True),
        sa.Column(
            "verified_birthdate",
            sa.Boolean(name="verified_birthdate", create_constraint=True),
            sa.CheckConstraint("(birthdate IS NULL) = (verified_birthdate IS NULL)", name="birthdate_null"),
            nullable=True,
        ),
        sa.Column("company_registration_number", sa.Text(), nullable=True),
        sa.Column(
            "verified_company_registration_number",
            sa.Boolean(name="verified_company_registration_number", create_constraint=True),
            sa.CheckConstraint(
                "(company_registration_number IS NULL) = (verified_company_registration_number IS NULL)",
                name="company_registration_number_null",
            ),
            nullable=True,
        ),
        sa.Column(
            "emails",
            sa.ARRAY(sa.Text()).with_variant(ScalarListType(), "sqlite"),
            nullable=True,
        ),
        sa.Column(
            "verified_emails",
            sa.Boolean(name="verified_emails", create_constraint=True),
            sa.CheckConstraint("(emails IS NULL) = (verified_emails IS NULL)", name="emails_null"),
            nullable=True,
        ),
        sa.Column(
            "address_street",
            sa.ARRAY(sa.Text()).with_variant(ScalarListType(), "sqlite"),
            nullable=True,
        ),
        sa.Column("address_city", sa.Text(), nullable=True),
        sa.Column("address_state", sa.Text(), nullable=True),
        sa.Column("address_postal_code", sa.Text(), nullable=True),
        sa.Column("address_country_code", sa.Text(), nullable=True),
        sa.Column(
            "verified_address",
            sa.Boolean(name="verified_address", create_constraint=True),
            sa.CheckConstraint(_ADDRESS_CHECK_SQL, name="address_null"),
            nullable=True,
        ),
        sa.ForeignKeyConstraint(
            ["verification_id"],
            ["verification_verification.id"],
            name=op.f("verification_contactdata_verification_id_fk"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("verification_contactdata_pkey")),
    )


def downgrade() -> None:
    op.drop_table("verification_contactdata")
    op.drop_table("verification_event")
    op.drop_table("verification_token")
    op.drop_table("verification_verification")
