"""Base classes for models."""
from datetime import datetime
from uuid import UUID, uuid4

from fred_types import ContactHistoryId, ContactId, DomainId
from sqlalchemy import Boolean, Column, Integer, MetaData, Text
from sqlalchemy.ext.asyncio import AsyncAttrs
from sqlalchemy.orm import DeclarativeBase, declared_attr, mapped_column
from sqlalchemy.sql.schema import _NamingSchemaTD
from sqlalchemy_utc import UtcDateTime
from sqlalchemy_utils import UUIDType as _UUIDType, generic_repr
from typing_extensions import Annotated

NAMING_CONVENTION: _NamingSchemaTD = {
    'pk': '%(table_name)s_pkey',
    'ix': '%(table_name)s_%(column_0_name)s_key',
    'fk': '%(table_name)s_%(column_0_name)s_fk',
    'ck': '%(table_name)s_%(constraint_name)s',
    'uq': '%(table_name)s_%(column_0_name)s_key',
}
METADATA = MetaData(naming_convention=NAMING_CONVENTION)


APPLICATION_PREFIX = 'verification_'

# Definitions of complete columns.
# See https://docs.sqlalchemy.org/en/20/orm/declarative_tables.html#mapping-whole-column-declarations-to-python-types
UuidType = Annotated[UUID, mapped_column(_UUIDType(binary=False), unique=True, default=uuid4)]


@generic_repr
class BaseModel(AsyncAttrs, DeclarativeBase):
    """Base class for models."""

    metadata = METADATA

    id = Column(Integer, primary_key=True)

    type_annotation_map = {
        bool: Boolean(),
        str: Text(),
        ContactId: Text(),
        ContactHistoryId: Text(),
        DomainId: Text(),
        datetime: UtcDateTime(timezone=True),
    }

    @declared_attr.directive
    def __tablename__(cls) -> str:
        return APPLICATION_PREFIX + cls.__name__.lower()
