"""Models for verifications."""
import asyncio
import operator
import re
import secrets
import unicodedata
from datetime import date, datetime
from itertools import zip_longest
from typing import Any, Collection, Dict, Iterable, List, Optional, Sequence, cast

from asyncstdlib import cached_property
from fred_types import ContactHistoryId, ContactId, DomainId, DomainRef
from regal import Contact, ContactIdentifierType, DomainContactRole
from regal.exceptions import DomainDoesNotExist
from sqlalchemy import ARRAY, CheckConstraint, Enum, ForeignKey, Index, Text, UniqueConstraint
from sqlalchemy.orm import Mapped, mapped_column, relationship
from sqlalchemy.sql import expression, func
from sqlalchemy_utils import JSONType, ScalarListType

from ..constants import OPEN_STATUSES, EventType, MessageType, VerificationStatus
from ..settings import APPLICATION
from .base import BaseModel, UuidType

LEGIBLE_CHARACTERS = 'ACDEFGHJKLMNPQRTUVWXYZ2345679'


def _random_token() -> str:
    return ''.join(secrets.choice(LEGIBLE_CHARACTERS) for i in range(10))


class Verification(BaseModel):
    """Represents a verification."""

    uuid: Mapped[UuidType]
    contact_id: Mapped[ContactId] = mapped_column(index=True)
    contact_handle: Mapped[Optional[str]]
    reason: Mapped[str] = mapped_column(server_default='')
    reason_url: Mapped[str] = mapped_column(server_default='')
    access_token: Mapped[str] = mapped_column(default=_random_token)
    create_datetime: Mapped[datetime] = mapped_column(server_default=func.now())
    close_datetime: Mapped[Optional[datetime]]
    last_access_datetime: Mapped[Optional[datetime]]
    status: Mapped[VerificationStatus] = mapped_column(Enum(VerificationStatus, create_constraint=True), index=True)
    manual_check_request: Mapped[bool] = mapped_column(server_default=expression.false())

    domains: Mapped[List["Domain"]] = relationship('Domain')
    events: Mapped[List["Event"]] = relationship('Event')
    contact_data: Mapped[List["ContactData"]] = relationship('ContactData', back_populates='verification')

    __table_args__ = (
        Index('verification_verification_contact_id_active_key', 'contact_id', unique=True,
              postgresql_where=status.in_(OPEN_STATUSES)).ddl_if(dialect='postgresql'),
        Index('verification_verification_contact_id_active_key', 'contact_id', unique=True,
              sqlite_where=status.in_(OPEN_STATUSES)).ddl_if(dialect='sqlite'),
    )

    @cached_property
    async def contact(self) -> Contact:
        """Return the contact instance."""
        return await APPLICATION.contact_client.get_contact_info(self.contact_id)

    async def get_events(self, type: Optional[EventType] = None) -> Sequence['Event']:
        """Return verification events with given type."""
        events = await self.awaitable_attrs.events
        if type is not None:
            events = (e for e in events if e.type == type)
        return sorted(events, key=operator.attrgetter('id'), reverse=True)

    @property
    async def active_methods(self) -> Collection[str]:
        """Return an iterable of active methods for the verification."""
        return {cast(str, e.method) for e in await self.get_events(EventType.USE_METHOD)}

    async def get_message_events(self, message_type: MessageType) -> Sequence['Event']:
        """Return verification message events with given message type."""
        return [e for e in await self.get_events(EventType.SEND_MESSAGE)
                if e.details.get('message_type') == message_type]

    async def get_status_delay(self, status: VerificationStatus) -> int:
        """Get delay for the verification status transition in seconds.

        Result may be negative if status change should occur sooner.
        """
        delay = 0
        for event in await self.get_events(EventType.DELAY_STATUS):
            if event.details.get('status') == status:
                delay += event.details.get('delay', 0)
        return delay

    async def _get_domain_ref(self, domain_id: DomainId, semaphore: asyncio.Semaphore) -> Optional[DomainRef]:
        """Utility to get domain ref from id."""
        # Limit the number of concurrent requests by semaphore.
        async with semaphore:
            try:
                domain = await APPLICATION.domain_client.get_domain_info(domain_id)
            except DomainDoesNotExist:
                return None
            return DomainRef(id=domain.domain_id, fqdn=domain.fqdn)

    @property
    async def affected_domains(self) -> Collection[DomainRef]:
        """Return a collection of domain identifiers affected by the verification."""
        domains = tuple(d.ref for d in await self.awaitable_attrs.domains)
        if domains:
            return domains

        domains_change_events = await self.get_events(EventType.DOMAINS_CHANGED)
        if domains_change_events:
            # It's a migrated verification. Domains have been recorded, there just aren't any.
            return ()

        create_events = await self.get_events(EventType.CREATE)
        if create_events and 'domains' in create_events[0].details:
            # It's a newly created verification, domains are monitored.
            return ()

        outzone_events = await self.get_events(EventType.OUTZONE)
        if outzone_events and not outzone_events[0].details.get('canceled'):
            # Found outzone event which is not canceled, return outzoned domains.
            domain_ids = outzone_events[0].details.get('domains', ())
            semaphore = asyncio.Semaphore(10)
            domain_refs = await asyncio.gather(*(self._get_domain_ref(d, semaphore) for d in domain_ids))
            return [r for r in domain_refs if r is not None]

        # No clue, fetch domains directly from registry
        result = await APPLICATION.domain_client.get_domains_by_contact(self.contact_id,
                                                                        {DomainContactRole.HOLDER})
        return tuple(DomainRef(id=d['domain'].id, fqdn=d['domain'].fqdn) for d in result['domains'])


class Domain(BaseModel):
    """Represents a domain affected by the verification."""

    verification_id: Mapped[int] = mapped_column(ForeignKey(Verification.id))
    domain_id: Mapped[DomainId]
    fqdn: Mapped[str]

    __table_args__ = (
        UniqueConstraint('verification_id', 'domain_id', name='verification_domain_id_key'),
    )

    @property
    def ref(self) -> DomainRef:
        """Return respective DomainRef."""
        return DomainRef(id=self.domain_id, fqdn=self.fqdn)


class Token(BaseModel):
    """Represents a verification token."""

    uuid: Mapped[UuidType]
    create_datetime: Mapped[datetime] = mapped_column(server_default=func.now())
    verification_id: Mapped[int] = mapped_column(ForeignKey(Verification.id), unique=True)
    token: Mapped[str] = mapped_column(default=_random_token, unique=True)
    method: Mapped[str]

    verification: Mapped["Verification"] = relationship('Verification')


class Event(BaseModel):
    """Represents an event in verification."""

    event_datetime: Mapped[datetime] = mapped_column(server_default=func.now())
    verification_id: Mapped[int] = mapped_column(ForeignKey(Verification.id), index=True)
    method: Mapped[Optional[str]]
    type: Mapped[str]
    details: Mapped[Dict[str, Any]] = mapped_column(JSONType(), default=dict)


_WHITESPACE_PATTERN = re.compile(r'\s+')


def _normalize_string(value: str) -> str:
    """Normalize string for comparisons."""
    value = _WHITESPACE_PATTERN.sub(' ', value).strip().lower()
    return unicodedata.normalize('NFD', value).encode('ASCII', 'ignore').decode('utf-8')


def compare_strings(first: str, second: str) -> bool:
    """Return whether provided strings are equivalent."""
    return _normalize_string(first) == _normalize_string(second)


class ContactData(BaseModel):
    """Represents a verified contact data.

    All data fields are nullable. When the value is NULL, then the data field was not provided.
    Similar logic applies to the verified flags.
    """

    verification_id: Mapped[int] = mapped_column(ForeignKey(Verification.id), index=True)
    verify_datetime: Mapped[datetime] = mapped_column(server_default=func.now())
    method: Mapped[Optional[str]]
    contact_history_id: Mapped[ContactHistoryId]

    name: Mapped[Optional[str]]
    verified_name: Mapped[Optional[bool]] = mapped_column(
        CheckConstraint('(name IS NULL) = (verified_name IS NULL)', name='name_null'))
    organization: Mapped[Optional[str]]
    verified_organization: Mapped[Optional[bool]] = mapped_column(
        CheckConstraint('(organization IS NULL) = (verified_organization IS NULL)', name='organization_null'))
    birthdate: Mapped[Optional[date]]
    verified_birthdate: Mapped[Optional[bool]] = mapped_column(
        CheckConstraint('(birthdate IS NULL) = (verified_birthdate IS NULL)', name='birthdate_null'))
    company_registration_number: Mapped[Optional[str]]
    verified_company_registration_number: Mapped[Optional[bool]] = mapped_column(
        CheckConstraint('(company_registration_number IS NULL) = (verified_company_registration_number IS NULL)',
                        name='company_registration_number_null'))
    emails: Mapped[Optional[List[str]]] = mapped_column(ARRAY(Text).with_variant(ScalarListType(), 'sqlite'))
    verified_emails: Mapped[Optional[bool]] = mapped_column(
        CheckConstraint('(emails IS NULL) = (verified_emails IS NULL)', name='emails_null'))

    _ADDRESS_CHECK_SQL = (
        '(address_street IS NULL AND coalesce(address_city, address_state, address_postal_code, address_country_code) '
        'IS NULL) = (verified_address IS NULL)')
    address_street: Mapped[Optional[List[str]]] = mapped_column(ARRAY(Text).with_variant(ScalarListType(), 'sqlite'))
    address_city: Mapped[Optional[str]]
    address_state: Mapped[Optional[str]]
    address_postal_code: Mapped[Optional[str]]
    address_country_code: Mapped[Optional[str]]
    verified_address: Mapped[Optional[bool]] = mapped_column(
        CheckConstraint(_ADDRESS_CHECK_SQL, name='address_null'))

    # 'load_on_pending' allows loading related objects from unflushed instances.
    verification: Mapped["Verification"] = relationship('Verification', back_populates='contact_data',
                                                        load_on_pending=True)

    def _compare_name(self, reference: str) -> bool:
        """Return whether reference is equivalent to the name."""
        name_chunks = [_normalize_string(c) for c in cast(str, self.name).split()]
        reference_chunks = [_normalize_string(c) for c in reference.split()]
        return sorted(name_chunks) == sorted(reference_chunks)

    def _compare_organization(self, reference: str) -> bool:
        """Return whether reference is equivalent to the organization."""
        return compare_strings(cast(str, self.organization), reference)

    def _compare_company_registration_number(self, reference: str) -> bool:
        """Return whether reference is equivalent to the company registration number."""
        return compare_strings(cast(str, self.company_registration_number), reference)

    def _compare_emails(self, references: Iterable[str]) -> bool:
        """Return whether each of the emails matches at least one of the reference emails."""
        return all(any(e.strip() == r.strip() for r in references) for e in cast(List[str], self.emails))

    def _compare_address(self, street: Optional[Iterable[str]] = None, city: Optional[str] = None,
                         state: Optional[str] = None, postal_code: Optional[str] = None,
                         country_code: Optional[str] = None) -> bool:
        """Return whether reference is equivalent to the address."""
        for first, second in zip_longest(self.address_street or [], street or [], fillvalue=''):
            if not compare_strings(first, second):
                return False
        if not compare_strings(self.address_city or '', city or ''):
            return False
        if not compare_strings(self.address_state or '', state or ''):
            return False
        if not compare_strings(_WHITESPACE_PATTERN.sub('', self.address_postal_code or ''),
                               _WHITESPACE_PATTERN.sub('', postal_code or '')):
            return False
        if not compare_strings(self.address_country_code or '', country_code or ''):
            return False
        return True

    @staticmethod
    def _get_birthdate(contact: Contact) -> Optional[date]:
        """Extract birthdate from contact data."""
        if (contact.additional_identifier is None
                or contact.additional_identifier.type != ContactIdentifierType.BIRTHDATE):
            # Identifier missing or has different type.
            return None
        else:
            try:
                return date.fromisoformat(contact.additional_identifier.value)
            except ValueError:
                # Reference date can't be decoded.
                return None

    async def verify(self) -> None:
        """Perform verification of the contact data and set results."""
        contact = await self.verification.contact
        assert contact.contact_history_id is not None  # noqa: S101
        self.contact_history_id = contact.contact_history_id
        if self.name is not None:
            self.verified_name = self._compare_name(contact.name)
        if self.organization is not None:
            self.verified_organization = self._compare_organization(contact.organization)
        if self.birthdate is not None:
            self.verified_birthdate = (self.birthdate == self._get_birthdate(contact))
        if self.company_registration_number is not None:
            if (contact.additional_identifier is None
                    or contact.additional_identifier.type != ContactIdentifierType.COMPANY_REGISTRATION_NUMBER):
                # Identifier missing or has different type.
                self.verified_company_registration_number = False
            else:
                self.verified_company_registration_number = self._compare_company_registration_number(
                    contact.additional_identifier.value)
        if self.emails is not None:
            self.verified_emails = self._compare_emails(contact.emails)
        if (self.address_street is not None or self.address_city is not None or self.address_state is not None
                or self.address_postal_code is not None or self.address_country_code is not None):
            if contact.place is None:
                # No address in reference data.
                self.verified_address = False
            else:
                self.verified_address = self._compare_address(
                    contact.place.street, contact.place.city, contact.place.state_or_province,
                    contact.place.postal_code, contact.place.country_code)
