"""Processors for verification methods."""
import asyncio
from abc import ABC, abstractmethod
from collections import defaultdict
from datetime import datetime, timedelta
from operator import attrgetter
from typing import Any, AsyncIterator, Dict, Iterable, List, Optional, Union, cast

import requests
from asyncstdlib import list as alist
from fred_types import ContactHistoryId, ContactId, DomainId
from hermes import Email, EmailEnvelope, EmailMessengerClient, Message, MessageEnvelope
from hermes.client import MessengerClientBase
from hermes.message import Reference
from pytz import UTC
from regal import Domain
from regal.domain import DomainContactRole
from regal.exceptions import ContactDoesNotExist
from sqlalchemy import null, or_, select
from sqlalchemy.sql import func
from sqlalchemy.sql.elements import SQLCoreOperations

from ..constants import (
    OPEN_STATUSES,
    EventType,
    MessageType,
    ObjectState,
    ReferenceType,
    Template,
    TransportMethod,
    VerificationStatus,
)
from ..database import object_session
from ..models import ContactData, Event, Token, Verification
from ..settings import APPLICATION, CheckDomainsSettings


class ScheduledEvent(ABC):
    """Base class representing an event in verification process.

    Attributes:
        verification: The verification affected by this event.
        method: The method this event belongs to. `None` if the event is not specific.
        details: The details of the event. May be modified by event processing.
        due_datetime: When is this event due. If None, then it is due now.
    """

    @property
    @abstractmethod
    def type(self) -> EventType:
        """Return event type."""

    def __init__(self, verification: Verification, *, method: Optional[str] = None,
                 details: Optional[Dict[str, Any]] = None, due_datetime: Optional[datetime] = None):
        self.verification = verification
        self.method = method
        self._details = details or {}
        self.due_datetime = due_datetime

    def __repr__(self) -> str:
        return '<{} {} method={}, due_datetime={}, details={}>'.format(
            type(self).__name__, self.verification, self.method, self.due_datetime, self._details)

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, ScheduledEvent):
            return type(self) is type(other) and self.__dict__ == other.__dict__
        else:
            # Mainly for utilities - falls back to default comparisons.
            return NotImplemented

    async def get_details(self) -> Dict[str, Any]:
        """Return event details."""
        return self._details.copy()

    async def to_event(self) -> Event:
        """Return a corresponding event model."""
        return Event(verification_id=self.verification.id, method=self.method, type=self.type,
                     details=await self.get_details())

    @abstractmethod
    async def process(self) -> None:
        """Perform this event."""


class _BaseSendMessageEvent(ScheduledEvent):
    """Base class for send message events.

    Attributes:
        context: Additional context for the message.
        resend: Whether the event is sending again an already sent message.
    """

    type = EventType.SEND_MESSAGE
    status: Optional[VerificationStatus] = None

    def __init__(self, *args: Any, context: Optional[Dict[str, Any]] = None, resend: bool = False, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.context = context or {}
        self.resend = resend

    @property
    @abstractmethod
    def transport_method(self) -> TransportMethod:
        """Return transport method."""

    @property
    @abstractmethod
    def message_type(self) -> MessageType:
        """Return message type."""

    @property
    def _client(self) -> MessengerClientBase:
        """Return client based on transport method."""
        return cast(MessengerClientBase, getattr(APPLICATION, self.transport_method + '_client'))

    async def get_context(self) -> Dict[str, Any]:
        """Return message context."""
        return self.context.copy()

    async def get_details(self) -> Dict[str, Any]:
        """Return event details."""
        details = await super().get_details()
        details['method'] = self.transport_method
        details['message_type'] = self.message_type
        return details


class SendMessageEvent(_BaseSendMessageEvent):
    """Send a verification related message."""

    @abstractmethod
    async def _get_message(self) -> Message:
        """Return the message to be sent."""

    async def process(self) -> None:
        """Send a message."""
        self._details['message_id'] = await self._client.send(
            await self._get_message(),
            references=[Reference(type=ReferenceType.CONTACT, value=self.verification.contact_id),
                        Reference(type=ReferenceType.VERIFICATION, value=str(self.verification.uuid))])
        if self.status is not None and not self.resend:
            self.verification.status = self.status


class SendMessageBatchEvent(_BaseSendMessageEvent):
    """Send a batch of verification related messages."""

    # XXX: batch_send only works on email for now.
    transport_method = TransportMethod.EMAIL

    @abstractmethod
    async def _get_envelopes(self) -> Iterable[MessageEnvelope]:
        """Return envelopes to be send."""

    async def process(self) -> None:
        """Send a message batch."""
        # XXX: batch_send currently available only on EmailMessengerClient.
        self._details['message_ids'] = await alist(cast(EmailMessengerClient, self._client).batch_send(
            cast(Iterable[EmailEnvelope], await self._get_envelopes())))
        if self.status is not None and not self.resend:
            self.verification.status = self.status


class NotifyContactsEvent(SendMessageBatchEvent):
    """Base class to send notification to all verification related contacts."""

    _concurrency_limit = 10

    @property
    @abstractmethod
    def subject_template(self) -> Template:
        """Return subject template."""

    @property
    @abstractmethod
    def body_template(self) -> Template:
        """Return body template."""

    @property
    @abstractmethod
    def admin_body_template(self) -> Template:
        """Return body template for messages to an admin contact."""

    async def get_context(self) -> Dict[str, Any]:
        """Return message context."""
        contact = await self.verification.contact
        context = await super().get_context()
        context['uuid'] = str(self.verification.uuid)
        context['handle'] = contact.contact_handle
        return context

    async def get_details(self) -> Dict[str, Any]:
        """Return event details."""
        if 'recipient' not in self._details:
            contact = await self.verification.contact
            self._details['recipient'] = contact.emails
        return await super().get_details()

    async def _make_envelope(self, emails: List[str], contact_id: ContactId, template: Template,
                             context: Dict[str, Any]) -> MessageEnvelope:
        """Make and return a single envelope."""
        context = dict(await self.get_context(), **context)
        email = Email(recipient=', '.join(emails), subject_template=self.subject_template,
                      body_template=template, type=self.message_type, context=context)
        references = [Reference(type=ReferenceType.CONTACT, value=contact_id),
                      Reference(type=ReferenceType.VERIFICATION, value=str(self.verification.uuid))]
        return EmailEnvelope(message_data=email, references=references)

    async def _make_admin_envelope(self, admin_id: ContactId, domains: List[Domain]) -> MessageEnvelope:
        """Make envelope for an admin contact."""
        admin_contact = await APPLICATION.contact_client.get_contact_info(admin_id)
        context = {'domains': [d.fqdn for d in domains]}
        return await self._make_envelope(admin_contact.emails, admin_id, self.admin_body_template, context)

    async def _get_domain(self, domain_id: DomainId, semaphore: asyncio.Semaphore) -> Domain:
        """Get domain detail."""
        # Limit the number of concurrent requests by semaphore.
        async with semaphore:
            return await APPLICATION.domain_client.get_domain_info(domain_id)

    async def _get_envelopes(self) -> Iterable[MessageEnvelope]:
        """Return envelopes to be send.

        The first message is sent to the contact that is being verified.
        On resend, only the first message is send again.
        """
        envelopes = []
        domains = await self.verification.affected_domains
        recipient = (await self.get_details())['recipient']
        # Add access_token only to the message to the contact itself.
        context = {'access_token': self.verification.access_token,
                   'domains': [d.fqdn for d in domains]}
        envelopes.append(await self._make_envelope(recipient, self.verification.contact_id,
                                                   self.body_template, context))

        # Resend only the first message.
        if self.resend:
            return envelopes

        # Generate messages to domain admin contacts
        semaphore = asyncio.Semaphore(self._concurrency_limit)
        domain_infos = cast(List[Domain], await asyncio.gather(*(self._get_domain(d.id, semaphore) for d in domains)))
        admin_domains: Dict[ContactId, List[Domain]] = defaultdict(list)
        for domain in domain_infos:
            for admin in domain.administrative_contacts:
                admin_domains[admin.id].append(domain)
        envelopes.extend(await asyncio.gather(*(self._make_admin_envelope(a, d) for a, d in admin_domains.items())))
        return envelopes


class InitialEmailEvent(NotifyContactsEvent):
    """Send an initial verification email."""

    message_type = MessageType.INITIAL_EMAIL
    status = VerificationStatus.INITIALIZED
    subject_template = Template.INITIAL_EMAIL_SUBJECT
    body_template = Template.INITIAL_EMAIL_BODY
    admin_body_template = Template.INITIAL_EMAIL_ADMIN_BODY

    async def process(self) -> None:
        """Update contact states and send a message."""
        if not self.resend:
            contact = await self.verification.contact
            flags: Dict[str, bool] = {ObjectState.CONTACT_IN_MANUAL_VERIFICATION: True,
                                      ObjectState.CONTACT_PASSED_MANUAL_VERIFICATION: False,
                                      ObjectState.CONTACT_FAILED_MANUAL_VERIFICATION: False}
            await APPLICATION.contact_client.update_contact_state(contact.contact_id,
                                                                  cast(ContactHistoryId, contact.contact_history_id),
                                                                  flags)

        return await super().process()


class CheckDomainsEvent(ScheduledEvent):
    """Check whether contact has any high ranked domains."""

    type = EventType.CHECK_DOMAINS

    async def process(self) -> None:
        """Check high ranked domains."""
        pause = False
        contact = await self.verification.contact
        settings = cast(CheckDomainsSettings, APPLICATION.settings.check_domains)

        domains = [d.fqdn for d in await self.verification.affected_domains]
        self._details['count'] = len(domains)
        if settings.count_limit and len(domains) >= settings.count_limit:
            pause = True

        # Check high ranked domains explicitely.
        last_month = (datetime.now(UTC).date().replace(day=1) - timedelta(days=1)).replace(day=1)
        query = settings.rank_query.format(handle=contact.contact_handle, last_month=last_month)

        response = requests.get(query, timeout=settings.rank_timeout, headers=settings.rank_headers)
        response.raise_for_status()

        # Filter ranked domains to only those currently owned by the contact
        ranked_domains = [i['domain'] for i in response.json() if i['domain'] in domains]
        self._details['ranked_domains'] = ranked_domains
        if ranked_domains:
            pause = True

        if pause:
            # Pause the verification.
            self.verification.status = VerificationStatus.PAUSED
            if settings.notification_email:
                context = {
                    'uuid': str(self.verification.uuid), 'handle': contact.contact_handle, 'name': contact.name,
                    'organization': contact.organization, 'domains': domains,
                    'ranked_domains': ranked_domains}
                email = Email(
                    recipient=settings.notification_email, subject_template=Template.PAUSED_NOTIFICATION_SUBJECT,
                    body_template=Template.PAUSED_NOTIFICATION_BODY, type=MessageType.PAUSED_NOTIFICATION,
                    context=context)
                await APPLICATION.email_client.send(
                    email,
                    references=[Reference(type=ReferenceType.VERIFICATION, value=str(self.verification.uuid))])


class OutzoneReminderEvent(NotifyContactsEvent):
    """Send an email reminder about impending outzone event."""

    message_type = MessageType.OUTZONE_REMINDER
    subject_template = Template.OUTZONE_REMINDER_SUBJECT
    body_template = Template.OUTZONE_REMINDER_BODY
    admin_body_template = Template.OUTZONE_REMINDER_ADMIN_BODY


class FutileReminderEvent(NotifyContactsEvent):
    """Send an email reminder about impending futile event."""

    message_type = MessageType.FUTILE_REMINDER
    subject_template = Template.FUTILE_REMINDER_SUBJECT
    body_template = Template.FUTILE_REMINDER_BODY
    admin_body_template = Template.FUTILE_REMINDER_ADMIN_BODY


class OutzoneEvent(ScheduledEvent):
    """Move all domains out of the zone."""

    type = EventType.OUTZONE

    async def process(self) -> None:
        """Move all domains out of the zone."""
        contact = await self.verification.contact
        domains = await self.verification.affected_domains
        self._details['domains'] = [d.id for d in domains]
        state_result = APPLICATION.domain_admin_client.manage_domain_state_flags(
            domains=[Domain(domain_id=d.id, fqdn=d.fqdn) for d in domains],
            state_flags={ObjectState.SERVER_BLOCKED: None, ObjectState.SERVER_OUTZONE_MANUAL: None})
        request_ids: Dict[DomainId, List[str]] = {}
        # Python 3.9 coverage bug
        async for result in state_result:  # pragma: no cover
            request_ids.setdefault(result['domain_id'], []).append(result['request_id'])
        self._details['request_ids'] = request_ids
        await APPLICATION.contact_client.update_contact_state(contact.contact_id,
                                                              cast(ContactHistoryId, contact.contact_history_id),
                                                              {ObjectState.SERVER_LINK_PROHIBITED: True})
        self.verification.status = VerificationStatus.CONSTRAINED

        # XXX: Not a nice solution. Check how it's going to be used and refactor appropriately.
        futile_datetime = ''
        # Python 3.9 coverage bug
        async for processor in APPLICATION.handler.get_active_processors(self.verification):  # pragma: no cover
            futile_datetime = (await processor._get_futile_due(self.verification)).isoformat()
            break

        context = {'uuid': str(self.verification.uuid),
                   'access_token': self.verification.access_token,
                   'initial_email_datetime': self.verification.create_datetime.isoformat(),
                   'futile_datetime': futile_datetime,
                   'handle': contact.contact_handle,
                   'domains': [d.fqdn for d in domains]}
        email = Email(recipient=', '.join(contact.emails), subject_template=Template.OUTZONE_SUBJECT,
                      body_template=Template.OUTZONE_BODY, type=MessageType.OUTZONE, context=context)
        await APPLICATION.email_client.send(
            email,
            references=[Reference(type=ReferenceType.CONTACT, value=self.verification.contact_id),
                        Reference(type=ReferenceType.VERIFICATION, value=str(self.verification.uuid))])


class DelayOutzoneEvent(ScheduledEvent):
    """Placeholder event to delay outzone and futile events if manual check request is pending."""

    type = EventType.OUTZONE

    async def process(self) -> None:
        """Delay futile event."""
        session = object_session(self.verification)
        details = {'status': VerificationStatus.FUTILE, 'delay': self._details['delay']}
        event = Event(verification_id=self.verification.id, type=EventType.DELAY_STATUS, details=details)
        session.add(event)

    async def to_event(self) -> Event:
        """Delay outzone event."""
        details = {'status': VerificationStatus.CONSTRAINED, 'delay': self._details['delay']}
        return Event(verification_id=self.verification.id, type=EventType.DELAY_STATUS, details=details)


class FutileEvent(ScheduledEvent):
    """Mark verification as futile."""

    type = EventType.FUTILE

    async def process(self) -> None:
        """Mark verification as futile."""
        self.verification.status = VerificationStatus.FUTILE


class DelayFutileEvent(ScheduledEvent):
    """Placeholder event to delay futile event if manual check request is pending."""

    type = EventType.FUTILE

    async def process(self) -> None:
        """Do nothing."""

    async def to_event(self) -> Event:
        """Delay futile event."""
        details = {'status': VerificationStatus.FUTILE, 'delay': self._details['delay']}
        return Event(verification_id=self.verification.id, type=EventType.DELAY_STATUS, details=details)


class CompleteEvent(ScheduledEvent):
    """Mark verification as complete."""

    type = EventType.COMPLETE

    async def process(self) -> None:
        """Mark verification as completed."""
        contact = await self.verification.contact
        self.verification.close_datetime = cast(SQLCoreOperations[Optional[datetime]], func.now())
        self.verification.status = VerificationStatus.SUCCEEDED

        flags: Dict[str, bool] = {ObjectState.CONTACT_IN_MANUAL_VERIFICATION: False,
                                  ObjectState.CONTACT_PASSED_MANUAL_VERIFICATION: True,
                                  ObjectState.CONTACT_FAILED_MANUAL_VERIFICATION: False}
        await APPLICATION.contact_client.update_contact_state(contact.contact_id,
                                                              cast(ContactHistoryId, contact.contact_history_id), flags)


class NotifyCompleteEvent(NotifyContactsEvent):
    """Send an email notification about successful verification."""

    message_type = MessageType.SUCCEEDED
    subject_template = Template.SUCCEEDED_SUBJECT
    body_template = Template.SUCCEEDED_BODY
    admin_body_template = Template.SUCCEEDED_ADMIN_BODY


class FailedEvent(ScheduledEvent):
    """Mark verification as failed."""

    type = EventType.FAILED

    async def process(self) -> None:
        """Mark verification as failed."""
        contact = await self.verification.contact
        self.verification.close_datetime = cast(SQLCoreOperations[Optional[datetime]], func.now())
        self.verification.status = VerificationStatus.FAILED

        flags: Dict[str, bool] = {ObjectState.CONTACT_IN_MANUAL_VERIFICATION: False,
                                  ObjectState.CONTACT_PASSED_MANUAL_VERIFICATION: False,
                                  ObjectState.CONTACT_FAILED_MANUAL_VERIFICATION: True}
        await APPLICATION.contact_client.update_contact_state(contact.contact_id,
                                                              cast(ContactHistoryId, contact.contact_history_id), flags)


class NotifyFailedEvent(SendMessageEvent):
    """Send an email notification about failed verification."""

    transport_method = TransportMethod.EMAIL
    message_type = MessageType.FAILED
    subject_template = Template.FAILED_SUBJECT
    body_template = Template.FAILED_BODY

    async def _get_message(self) -> Email:
        """Return the message to be send."""
        contact = await self.verification.contact
        return Email(recipient=', '.join(contact.emails), subject_template=Template.FAILED_SUBJECT,
                     body_template=Template.FAILED_BODY, type=MessageType.FAILED, context=await self.get_context())

    async def get_context(self) -> Dict[str, Any]:
        """Return message context."""
        contact = await self.verification.contact
        context = await super().get_context()
        context['handle'] = contact.contact_handle

        # Gather FQDNs of deleted domains.
        domains = []
        failed_events = sorted(await self.verification.get_events(EventType.FAILED), key=attrgetter('event_datetime'),
                               reverse=True)
        if failed_events:
            domain_ids = failed_events[0].details.get('domains', [])
            if domain_ids:
                # Use get_domains_by_contact to retrieve mapping of IDs to FQDNs.
                all_domains = await APPLICATION.domain_client.get_domains_by_contact(
                    contact.contact_id, {DomainContactRole.HOLDER}, include_deleted=True)
                fqdn_map = {d['domain'].id: d['domain'].fqdn for d in all_domains['domains']}
                domains = [fqdn_map[i] for i in domain_ids]
        context['domains'] = domains
        return context

    async def get_details(self) -> Dict[str, Any]:
        """Return event details."""
        contact = await self.verification.contact
        details = await super().get_details()
        details['recipient'] = contact.emails
        return details


class CancelVerificationEvent(ScheduledEvent):
    """Mark verification as canceled."""

    type = EventType.CANCEL

    async def process(self) -> None:
        """Mark verification as canceled."""
        contact = await self.verification.contact
        self.verification.close_datetime = cast(SQLCoreOperations[Optional[datetime]], func.now())
        self.verification.status = VerificationStatus.CANCELED

        flags: Dict[str, bool] = {ObjectState.CONTACT_IN_MANUAL_VERIFICATION: False}
        await APPLICATION.contact_client.update_contact_state(contact.contact_id,
                                                              cast(ContactHistoryId, contact.contact_history_id), flags)


class NotifyCancelEvent(NotifyContactsEvent):
    """Send an email notification about canceled verification."""

    message_type = MessageType.CANCELED
    subject_template = Template.CANCELED_SUBJECT
    body_template = Template.CANCELED_BODY
    admin_body_template = Template.CANCELED_ADMIN_BODY


class CancelOutzoneEvent(ScheduledEvent):
    """Cancel moving domains out of the zone."""

    type = EventType.OUTZONE

    async def get_details(self) -> Dict[str, Any]:
        """Return event details."""
        details = await super().get_details()
        details['canceled'] = True
        return details

    async def process(self) -> None:
        """Cancel moving domains out of the zone."""
        contact = await self.verification.contact

        # Get outzone requests to cancel.
        outzone_events = sorted(await self.verification.get_events(EventType.OUTZONE), key=attrgetter('event_datetime'),
                                reverse=True)
        if outzone_events and not outzone_events[0].details.get('canceled'):
            request_ids = outzone_events[0].details.get('request_ids', {})
        else:
            request_ids = {}

        # Collect request IDs to cancel
        close_requests = []
        if isinstance(request_ids, list):
            # Legacy structure
            close_requests = request_ids
        elif request_ids:
            # Filter request_ids to affected domains only
            domain_ids = [d.id for d in await self.verification.affected_domains]
            request_ids = {d: s for d, s in request_ids.items() if d in domain_ids}
            close_requests = [r for s in request_ids.values() for r in s]

        self._details['request_ids'] = request_ids
        # Cancel the outzone.
        await alist(APPLICATION.domain_admin_client.manage_domain_state_flags(close_requests=close_requests))

        flags: Dict[str, bool] = {ObjectState.SERVER_LINK_PROHIBITED: False}
        await APPLICATION.contact_client.update_contact_state(contact.contact_id,
                                                              cast(ContactHistoryId, contact.contact_history_id), flags)
        if self.verification.last_access_datetime:
            self.verification.status = VerificationStatus.ACCESSED
        else:
            self.verification.status = VerificationStatus.INITIALIZED


class CancelFutileEvent(ScheduledEvent):
    """Cancel marking verification as futile."""

    type = EventType.FUTILE

    async def get_details(self) -> Dict[str, Any]:
        """Return event details."""
        details = await super().get_details()
        details['canceled'] = True
        return details

    async def process(self) -> None:
        """Cancel marking verification as futile."""
        self.verification.status = VerificationStatus.CONSTRAINED


class BaseProcessor:
    """Base class for verification processor.

    Attributes:
        method: The method managed by this processor.
        outzone_limit: The number of days before outzone is scheduled. Counted from the verification start.
        outzone_reminder_limit: The number of days to send reminder before scheduled outzone event.
        futile_limit: The number of days before futile is scheduled. Counted from the verification start.
        futile_reminder_limit: The number of days to send reminder before scheduled futile event.
        manual_check_delay: The number of seconds to delay futile and outzone events in case of pending manual check
                            request.
    """

    def __init__(self, *, method: str, outzone_limit: int = 30, outzone_reminder_limit: int = 7, futile_limit: int = 44,
                 futile_reminder_limit: int = 7, manual_check_delay: int = 86400):
        self.method = method
        self.outzone_limit = outzone_limit
        self.outzone_reminder_limit = outzone_reminder_limit
        self.futile_limit = futile_limit
        self.futile_reminder_limit = futile_reminder_limit
        self.manual_check_delay = manual_check_delay

    def __repr__(self) -> str:
        return '<{} of method {}>'.format(type(self).__name__, self.method)

    async def is_applicable(self, verification: Verification) -> bool:
        """Return whether this method is applicable for the verification."""
        return True

    async def _get_next_events(self, verification: Verification) -> AsyncIterator[ScheduledEvent]:
        """Return scheduled events for the verification.

        Internal processor specific method.
        This method is part of the stable API and incombatible changes will only occur between major versions.
        """
        return
        yield  # type: ignore[unreachable]

    async def _get_initial_email_event(self, verification: Verification, *, resend: bool = False) -> InitialEmailEvent:
        """Return initial email event."""
        contact = await verification.contact
        details = {'contact_history_id': contact.contact_history_id}
        context = {'outzone_datetime': (await self._get_outzone_due(verification)).isoformat(),
                   'futile_datetime': (await self._get_futile_due(verification)).isoformat()}
        return InitialEmailEvent(verification, details=details, context=context, resend=resend)

    async def _get_outzone_due(self, verification: Verification) -> datetime:
        """Return due datetime for outzone event.

        Limited to at most futile due to ensure outzone event occurs before the futile.
        """
        outzone_due = (verification.create_datetime
                       + timedelta(days=self.outzone_limit)
                       + timedelta(seconds=await verification.get_status_delay(VerificationStatus.CONSTRAINED)))
        return min(outzone_due, await self._get_futile_due(verification))

    async def _get_futile_due(self, verification: Verification) -> datetime:
        """Return due datetime for futile event."""
        return (verification.create_datetime
                + timedelta(days=self.futile_limit)
                + timedelta(seconds=await verification.get_status_delay(VerificationStatus.FUTILE)))

    async def get_next_events(self, verification: Verification) -> AsyncIterator[ScheduledEvent]:
        """Return scheduled events for the verification with descending priority."""
        # Check if contact exists, otherwise do not send any notifications.
        send_notification = True
        try:
            await verification.contact
        except ContactDoesNotExist:
            send_notification = False
        if (
                send_notification
                and verification.status == VerificationStatus.SUCCEEDED
                and not await verification.get_message_events(MessageType.SUCCEEDED)):
            yield NotifyCompleteEvent(verification)
        if (
                send_notification
                and verification.status == VerificationStatus.FAILED
                and not await verification.get_message_events(MessageType.FAILED)):
            yield NotifyFailedEvent(verification)
        if (
                send_notification
                and verification.status == VerificationStatus.CANCELED
                and not await verification.get_message_events(MessageType.CANCELED)):
            yield NotifyCancelEvent(verification)
        if verification.status not in OPEN_STATUSES or verification.status == VerificationStatus.PAUSED:
            return

        if verification.status == VerificationStatus.CREATED:
            # A completely new verification which needs to be initialized.
            yield await self._get_initial_email_event(verification)

        # Python 3.9 coverage bug
        async for scheduled in self._get_next_events(verification):  # pragma: no cover
            yield scheduled

        # Python 3.9 coverage bug
        async for scheduled in self._get_restrictive_events(verification):  # pragma: no cover
            yield scheduled

    async def _get_restrictive_events(self, verification: Verification) -> AsyncIterator[ScheduledEvent]:
        """Return next events related to outzone and futile events."""
        if verification.status in (VerificationStatus.CREATED, VerificationStatus.INITIALIZED,
                                   VerificationStatus.ACCESSED, VerificationStatus.PENDING):
            outzone_due = await self._get_outzone_due(verification)
            if not await verification.get_message_events(MessageType.OUTZONE_REMINDER):
                context = {'initial_email_datetime': verification.create_datetime.isoformat(),
                           'outzone_datetime': outzone_due.isoformat(),
                           'futile_datetime': (await self._get_futile_due(verification)).isoformat()}
                yield OutzoneReminderEvent(verification,
                                           due_datetime=outzone_due - timedelta(days=self.outzone_reminder_limit),
                                           context=context)
            if verification.manual_check_request:
                yield DelayOutzoneEvent(verification, due_datetime=outzone_due,
                                        details={'delay': self.manual_check_delay})
            else:
                if APPLICATION.settings.check_domains and not await verification.get_events(EventType.CHECK_DOMAINS):
                    yield CheckDomainsEvent(verification, due_datetime=outzone_due)
                yield OutzoneEvent(verification, due_datetime=outzone_due)

        futile_due = await self._get_futile_due(verification)
        if verification.status != VerificationStatus.FUTILE:
            # Plan futile, unless it already occured.
            if not await verification.get_message_events(MessageType.FUTILE_REMINDER):
                context = {'initial_email_datetime': verification.create_datetime.isoformat(),
                           'futile_datetime': (await self._get_futile_due(verification)).isoformat()}
                yield FutileReminderEvent(verification,
                                          due_datetime=futile_due - timedelta(days=self.futile_reminder_limit),
                                          context=context)
            if verification.manual_check_request:
                yield DelayFutileEvent(verification, due_datetime=futile_due,
                                       details={'delay': self.manual_check_delay})
            else:
                yield FutileEvent(verification, due_datetime=futile_due)

        if not await verification.affected_domains:
            # User has no domains, verification can be failed automatically.
            yield FailedEvent(verification, due_datetime=futile_due)

    async def get_resend_event(self, verification: Verification, message_type: str
                               ) -> Union[SendMessageEvent, SendMessageBatchEvent]:
        """Create and process a new message event.

        Arguments:
            verification: Verification to act upon.
            message_type: Message type of the event.

        Raises:
            ValueError: Message type not found.
        """
        if message_type == MessageType.INITIAL_EMAIL:
            return await self._get_initial_email_event(verification, resend=True)
        raise ValueError("Message type '{}' not found".format(message_type))

    async def _get_cancel_events(self, verification: Verification, status: Optional[VerificationStatus],
                                 ) -> AsyncIterator[ScheduledEvent]:
        """Return scheduled events for cancelations.

        Internal processor specific method.
        This method is part of the stable API and incombatible changes will only occur between major versions.
        """
        return
        yield  # type: ignore[unreachable]

    async def get_cancel_events(self, verification: Verification, status: Optional[VerificationStatus] = None,
                                ) -> AsyncIterator[ScheduledEvent]:
        """Return scheduled events for cancelations.

        Arguments:
            verification: Verification to act upon.
            status: If provided, return events to cancel that status if applicable.
                    If None, return events to cancel the whole verification.
        """
        if status is not None and status not in OPEN_STATUSES:
            raise ValueError("Can't cancel closed status.")
        if verification.status not in OPEN_STATUSES:
            return

        now = datetime.now(UTC)
        # Cancel futile if verification
        # * is FUTILE or PAUSED and should be canceled completely
        # * or is FUTILE and FUTILE is being rolled back and its not due yet.
        if (
            (verification.status >= VerificationStatus.FUTILE and status is None)
            or (verification.status == status == VerificationStatus.FUTILE
                and (await self._get_futile_due(verification)) > now)
        ):
            yield CancelFutileEvent(verification)

        # Cancel outzone if verification
        # * is CONSTRAINED, FUTILE or PAUSED and should be canceled completely
        # * or is CONSTRAINED and CONSTRAINED is being rolled back and its not due yet.
        if (
            (status is None and verification.status >= VerificationStatus.CONSTRAINED)
            or (verification.status == status == VerificationStatus.CONSTRAINED
                and (await self._get_outzone_due(verification)) > now)
        ):
            yield CancelOutzoneEvent(verification)

        # Python 3.9 coverage bug
        async for scheduled in self._get_cancel_events(verification, status):  # pragma: no cover
            yield scheduled

        # Return a verification cancel event if complete cancel is requested.
        if status is None:
            yield CancelVerificationEvent(verification)

    async def _get_complete_events(self, verification: Verification) -> AsyncIterator[ScheduledEvent]:
        """Return scheduled events for verification completion.

        Internal processor specific method.
        This method is part of the stable API and incombatible changes will only occur between major versions.
        """
        return
        yield  # type: ignore[unreachable]

    async def get_complete_events(self, verification: Verification) -> AsyncIterator[ScheduledEvent]:
        """Return scheduled events for the verification completion."""
        if verification.status not in OPEN_STATUSES:
            return
        if verification.status >= VerificationStatus.FUTILE:
            yield CancelFutileEvent(verification)
        if verification.status >= VerificationStatus.CONSTRAINED:
            yield CancelOutzoneEvent(verification)
        # Python 3.9 coverage bug
        async for scheduled in self._get_complete_events(verification):  # pragma: no cover
            yield scheduled
        yield CompleteEvent(verification, method=self.method)

    async def check_complete(self, verification: Verification) -> bool:
        """Return whether all verification steps were already completed."""
        session = object_session(verification)

        query = select(ContactData).filter(
            ContactData.verification_id == verification.id,
            or_(ContactData.method == null(), ContactData.method == self.method))
        query = query.order_by(ContactData.verify_datetime.desc())
        records = (await session.execute(query)).scalars().all()

        # Check whether the result is a company based on the most recent record.
        is_company = None
        for record in records:
            if record.verified_organization:
                is_company = record.organization != ''
                break
        else:
            # It's not even sure whether it's a company or a person.
            return False

        if not is_company and not any(r.verified_name for r in records):
            # It's a person, but its name was never verified.
            return False

        return any(r.verified_address for r in records)

    async def handle_token_verified(self, token: Token) -> None:
        """Perform processors specific action after token is verified.

        The default implementation is noop.
        """
