"""Processor for letter method."""
from typing import Any, AsyncIterator, Collection, Dict, Optional, Union, cast
from uuid import UUID

from hermes import Letter
from hermes.letter import Address
from regal import Address as ContactAddress
from regal.exceptions import ContactDoesNotExist
from sqlalchemy import delete, select

from ..constants import EventType, MessageType, Template, TransportMethod, VerificationStatus
from ..database import object_session
from ..models import ContactData, Event, Token, Verification
from ..settings import APPLICATION
from .base import BaseProcessor, ScheduledEvent, SendMessageBatchEvent, SendMessageEvent


class ConfirmationLetterEvent(SendMessageEvent):
    """Send a letter with a verification token.

    Args:
        letter_limit_bonus: Bonus (in days) to letter verification (see ``LetterProcessor``).
    """

    transport_method = TransportMethod.LETTER
    message_type = MessageType.CONFIRMATION_LETTER
    template = Template.CONFIRMATION_LETTER

    def __init__(self, *args: Any, letter_limit_bonus: int = 0, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self._token: Optional[Token] = None
        self.letter_limit_bonus = letter_limit_bonus

    async def _get_message(self) -> Letter:
        """Return the message to be send."""
        with APPLICATION.file_storage.create(name='verification-letter.pdf', mimetype='application/pdf') as file:
            file.write(
                APPLICATION.secretary_client.render_pdf(self.template, context=await self.get_context()))

        return Letter(recipient=Address(**self._details['recipient']), file=str(file.uuid), type=self.message_type)

    async def get_context(self) -> Dict[str, Any]:
        """Return message context."""
        contact = await self.verification.contact
        context = await super().get_context()
        context['address'] = self._details['recipient']
        context['token'] = cast(Token, self._token).token
        context['handle'] = contact.contact_handle
        return context

    async def process(self) -> None:
        """Generate a token, process a bonus limit and send a message."""
        session = object_session(self.verification)
        await session.execute(delete(Token).where(Token.verification_id == self.verification.id))
        self._token = Token(verification_id=self.verification.id, method=self.method)
        session.add(self._token)
        if self.letter_limit_bonus > 0 and not self.resend:
            session.add_all([  # delay for outzone and futile events (time interval converted from days to seconds)
                Event(verification_id=self.verification.id, type=EventType.DELAY_STATUS,
                      details={'status': i, 'delay': self.letter_limit_bonus * 86400})
                for i in (VerificationStatus.FUTILE, VerificationStatus.CONSTRAINED)
            ])
        # Flush session to get the token itself.
        await session.flush()
        self._details['token'] = str(self._token.uuid)
        await super().process()


class LetterProcessor(BaseProcessor):
    """Processor for letter method.

    Args:
        allowed_countries: List of allowed countries for letter communication (2-letter country codes).
                           If empty (default) all countries are allowed.
        letter_limit_bonus: Additional bonus (in days) to ``futile_limit`` and ``outzone_limit`` in case of
                            letter verification. Defaults to 0 if not set.
        kwargs: For other keyword parameters see parent class.

    Example:
        To limit letter verification procedure to the EU countries use YAML configuration settings for the processor::

         - allowed_countries: [ # EU - 27 countries (as of April 2023)
               AT, BE, BG, CY, CZ, DE, DK, EE, EL, ES,
               FI, FR, HR, HU, IE, IT, LT, LU, LV, MT,
               NL, PL, PT, RO, SI, SK, SW,
           ]
    """

    def __init__(self, *,
                 allowed_countries: Optional[Collection[str]] = None, letter_limit_bonus: int = 0, **kwargs: Any):
        super().__init__(**kwargs)
        self.allowed_countries = allowed_countries
        self.letter_limit_bonus = letter_limit_bonus

    async def is_applicable(self, verification: Verification) -> bool:
        """Return whether this method is applicable for the verification.

        Letter verification is only applicable to contacts based in allowed countries if set.
        """
        try:
            contact = await verification.contact
        except ContactDoesNotExist:
            return False
        # address required, country_code in allowed countries
        return contact.place is not None and \
            (not self.allowed_countries or contact.place.country_code in self.allowed_countries)

    async def _get_confirmation_letter_event(self, verification: Verification, *, resend: bool = False,
                                             ) -> ConfirmationLetterEvent:
        """Return confirmation letter event."""
        contact = await verification.contact
        place = cast(ContactAddress, contact.place)
        address = Address(
            name=contact.name, organization=contact.organization, street='\n'.join(place.street),
            city=place.city, state_or_province=place.state_or_province, postal_code=place.postal_code or "",
            country=place.country_code or "")
        address_data = address.dict()
        context = {'initial_email_datetime': verification.create_datetime.isoformat()}
        return ConfirmationLetterEvent(
            verification, method=self.method, resend=resend, letter_limit_bonus=self.letter_limit_bonus,
            details={'recipient': address_data, 'contact_history_id': contact.contact_history_id}, context=context)

    async def _get_next_events(self, verification: Verification) -> AsyncIterator[ScheduledEvent]:
        """Return a letter event if required."""
        if await verification.get_message_events(MessageType.CONFIRMATION_LETTER):
            # Confirmation letter was already sent. Don't send another.
            return

        yield await self._get_confirmation_letter_event(verification)

    async def get_resend_event(self, verification: Verification, message_type: str
                               ) -> Union[SendMessageEvent, SendMessageBatchEvent]:
        """Return processor specific resend event."""
        if message_type == MessageType.CONFIRMATION_LETTER:
            return await self._get_confirmation_letter_event(verification, resend=True)
        return await super().get_resend_event(verification, message_type)

    async def handle_token_verified(self, token: Token) -> None:
        """Mark address where the token was sent as verified."""
        session = object_session(token)
        query = select(Event).filter_by(verification_id=token.verification_id, type=EventType.SEND_MESSAGE,
                                        method=self.method)
        for event in (await session.execute(query)).scalars():
            try:
                event_token_uuid = UUID(event.details.get('token', ''))
            except ValueError:
                # Ignore errors.
                continue

            if event_token_uuid == token.uuid:
                contact_data = ContactData(
                    verification_id=token.verification_id,
                    method=self.method,
                    contact_history_id=event.details['contact_history_id'],
                    name=event.details['recipient'].get('name'),
                    verified_name=True,
                    organization=event.details['recipient'].get('organization'),
                    verified_organization=True,
                    address_street=event.details['recipient'].get('street'),
                    address_city=event.details['recipient'].get('city'),
                    address_state=event.details['recipient'].get('state_or_province'),
                    address_postal_code=event.details['recipient'].get('postal_code'),
                    address_country_code=event.details['recipient'].get('country'),
                    verified_address=True,
                )
                session.add(contact_data)
                return
