"""Processor for representative method."""
from typing import Any, AsyncIterator, Collection, Optional, Union, cast
from uuid import UUID

from fred_types import ContactId, RepresentativeId
from hermes import Address
from regal import Address as ContactAddress, Representative
from regal.exceptions import ContactDoesNotExist, RepresentativeDoesNotExist
from sqlalchemy import select

from ..constants import EventType, MessageType, Template
from ..database import object_session
from ..models import Event, Token, Verification
from ..settings import APPLICATION
from .base import BaseProcessor, ScheduledEvent, SendMessageBatchEvent, SendMessageEvent
from .letter import ConfirmationLetterEvent


class RepresentativeLetterEvent(ConfirmationLetterEvent):
    """Send a letter with a verification token to a contact representative."""

    message_type = MessageType.REPRESENTATIVE_LETTER
    template = Template.REPRESENTATIVE_LETTER


class RepresentativeProcessor(BaseProcessor):
    """Processor for representative method.

    Args:
        disallowed_countries: List of disallowed countries for representative verification (2-letter country codes).
                              If empty (default) all countries are allowed.
        letter_limit_bonus: Additional bonus (in days) to ``futile_limit`` and ``outzone_limit`` in case of
                            letter verification. Defaults to 0 if not set.
    """

    def __init__(self, *,
                 disallowed_countries: Optional[Collection[str]] = None, letter_limit_bonus: int = 0, **kwargs: Any):
        super().__init__(**kwargs)
        self.disallowed_countries = disallowed_countries
        self.letter_limit_bonus = letter_limit_bonus

    async def is_applicable(self, verification: Verification) -> bool:
        """Return whether this method is applicable for the verification.

        Representative verification is only applicable to contacts not based in disallowed countries if set.
        """
        try:
            contact = await verification.contact
        except ContactDoesNotExist:
            return False
        # address required, country_code in allowed countries
        return (
            contact.place is not None
            and (not self.disallowed_countries or contact.place.country_code not in self.disallowed_countries)
        )

    async def _get_representative_letter_event(self, verification: Verification, *, resend: bool = False,
                                               ) -> RepresentativeLetterEvent:
        """Return representative letter event."""
        representative_events = await verification.get_events(EventType.SET_CONTACT_REPRESENTATIVE)
        # Take the newest representative
        representative = representative_events[0].details
        address = Address(name=representative['name'], organization=representative['organization'],
                          street='\n'.join(representative['address'].get("street", [])),
                          city=representative['address'].get("city", ""),
                          state_or_province=representative['address'].get("state_or_province"),
                          postal_code=representative['address'].get("postal_code", ""),
                          country=representative['address'].get("country_code", ""))
        address_data = address.dict()
        context = {'initial_email_datetime': verification.create_datetime.isoformat()}
        # Store all representative data, so it can be processed from this single event in
        # the `handle_token_verified` below.
        details = {'recipient': address_data, 'telephone': representative['telephone'],
                   'email': representative['email']}
        return RepresentativeLetterEvent(
            verification, method=self.method, resend=resend, letter_limit_bonus=self.letter_limit_bonus,
            details=details, context=context)

    async def _get_next_events(self, verification: Verification) -> AsyncIterator[ScheduledEvent]:
        """Return a representative letter event if required."""
        if await verification.get_message_events(MessageType.REPRESENTATIVE_LETTER):
            # Representative letter was already sent. Don't send another.
            return

        if await verification.get_events(EventType.SET_CONTACT_REPRESENTATIVE):
            yield await self._get_representative_letter_event(verification)

    async def get_resend_event(self, verification: Verification, message_type: str
                               ) -> Union[SendMessageEvent, SendMessageBatchEvent]:
        """Return processor specific resend event."""
        if message_type == MessageType.REPRESENTATIVE_LETTER:
            if await verification.get_events(EventType.SET_CONTACT_REPRESENTATIVE):
                return await self._get_representative_letter_event(verification, resend=True)
            else:
                raise ValueError("Message type '{}' can't be resent".format(message_type))

        return await super().get_resend_event(verification, message_type)

    async def check_complete(self, verification: Verification) -> bool:
        """Return whether all verification steps were already completed."""
        session = object_session(verification)
        query = select(Event).filter_by(verification_id=verification.id, type=EventType.VERIFY_TOKEN,
                                        method=self.method)
        return bool((await session.execute(query)).scalar())

    async def handle_token_verified(self, token: Token) -> None:
        """Mark representative to which the token was sent as verified."""
        session = object_session(token)
        verification = await token.awaitable_attrs.verification
        query = select(Event).filter_by(verification_id=verification.id, type=EventType.SEND_MESSAGE,
                                        method=self.method)
        for event in (await session.execute(query)).scalars():
            try:
                event_token_uuid = UUID(event.details.get('token', ''))
            except ValueError:
                # Ignore errors.
                continue

            if event_token_uuid == token.uuid:
                try:
                    representative = await APPLICATION.representative_client.get(ContactId(verification.contact_id))
                except RepresentativeDoesNotExist:
                    representative = Representative(id=RepresentativeId('0'), contact_id=verification.contact_id)
                    callback = APPLICATION.representative_client.create
                else:
                    callback = APPLICATION.representative_client.update

                representative.name = event.details['recipient'].get('name')
                representative.organization = event.details['recipient'].get('organization')
                representative.telephone = cast(str, event.details.get('telephone'))
                representative.email = cast(str, event.details.get('email'))
                representative.place = ContactAddress(
                    street=[event.details['recipient'].get('street')],
                    city=event.details['recipient'].get('city'),
                    state_or_province=event.details['recipient'].get('state_or_province'),
                    postal_code=event.details['recipient'].get('postal_code'),
                    country_code=event.details['recipient'].get('country'),
                )
                await callback(representative)
                return
