"""Verification gRPC services."""
import asyncio
import logging
import math
from datetime import date, datetime
from functools import wraps
from secrets import compare_digest
from typing import Any, AsyncIterator, Callable, Container, Dict, List, Optional, Tuple, Type, TypeVar, cast
from uuid import UUID

from asyncstdlib import list as alist
from fred_api.verification.common_types_pb2 import (
    ContactData as ProtoContactData,
    ContactDoesNotExist as ProtoContactDoesNotExist,
    IncorrectStatus,
    InvalidData,
    MethodNotApplicable,
    VerificationAlreadyExists as ProtoVerificationAlreadyExists,
    VerificationDoesNotExist,
    VerificationInfo,
    VerificationResult,
    VerificationStatus as ProtoVerificationStatus,
)
from fred_api.verification.service_verification_grpc_pb2 import (
    ApproveFutileReply,
    ApproveFutileRequest,
    ApproveManualReply,
    ApproveManualRequest,
    CancelReply,
    CancelRequest,
    CreateReply,
    CreateRequest,
    DelayStatusReply,
    DelayStatusRequest,
    GetMethodsReply,
    GetMethodsRequest,
    GetReply,
    GetRequest,
    ListReply,
    ListRequest,
    ListVerificationResultsReply,
    ListVerificationResultsRequest,
    RequestManualCheckReply,
    RequestManualCheckRequest,
    ResendMessageReply,
    ResendMessageRequest,
    ResolveManualCheckReply,
    ResolveManualCheckRequest,
    ResumeReply,
    ResumeRequest,
    SetContactRepresentativeReply,
    SetContactRepresentativeRequest,
    UseMethodReply,
    UseMethodRequest,
    VerifyContactDataReply,
    VerifyContactDataRequest,
    VerifyTokenReply,
    VerifyTokenRequest,
)
from fred_api.verification.service_verification_grpc_pb2_grpc import VerificationServicer as ProtoVerificationServicer
from fred_types import ContactHistoryId
from frgal.exceptions import ServiceError
from frgal.service import awrap_stream, awrap_unary
from google.protobuf.message import Message
from grpc import ServicerContext, StatusCode
from hermes import Email
from hermes.message import Reference
from pytz import UTC
from regal import Domain
from regal.domain import DomainContactRole
from regal.exceptions import ContactDoesNotExist
from sqlalchemy import Column, Row, Select, UnaryExpression, literal_column, select
from sqlalchemy.engine import Engine
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import selectinload
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.sql import func, tuple_
from sqlalchemy.sql.elements import SQLCoreOperations
from sqlalchemy.sql.operators import asc_op, desc_op

from .constants import OPEN_STATUSES, EventType, MessageType, ObjectState, ReferenceType, Template, VerificationStatus
from .database import atomic, object_session
from .exceptions import VerificationAlreadyExists
from .models import ContactData, Domain as VerDomain, Event, Token, Verification
from .processors import CancelFutileEvent, CancelOutzoneEvent, FailedEvent
from .settings import APPLICATION

_LOGGER = logging.getLogger(__name__)

F = TypeVar('F', bound=Callable[..., Any])


def _handle_sqlalchemy_error(func: F) -> F:
    """Turn the SQLAlchemyError to FAILED_PRECONDITION."""
    @wraps(func)
    async def wrapped(*args: Any, **kwargs: Any) -> Any:
        try:
            return await func(*args, **kwargs)
        except SQLAlchemyError as error:
            _LOGGER.error("Error in SQL query in %s: %s", func.__name__, error)
            raise ServiceError(StatusCode.FAILED_PRECONDITION, "Database error: {}".format(error)) from error
    return cast(F, wrapped)


class _ServiceException(Exception):
    """Base class for service specific exceptions."""

    attr: str
    error_cls: Type[Message]

    def __init__(self, reply: Message, /, **kwargs: Any):
        super().__init__(reply, kwargs)
        self.reply = reply
        self.kwargs = kwargs


class _ServiceContactDoesNotExist(_ServiceException):
    """Service variant of ContactDoesNotExist."""

    attr = 'contact_does_not_exist'
    error_cls = ProtoContactDoesNotExist


class _ServiceInvalidData(_ServiceException):
    """Service variant of InvalidData."""

    attr = 'invalid_data'
    error_cls = InvalidData


class _ServiceIncorrectStatus(_ServiceException):
    """Service variant of IncorrectStatus."""

    attr = 'incorrect_status'
    error_cls = IncorrectStatus


class _ServiceMethodNotApplicable(_ServiceException):
    """Service variant of MethodNotApplicable."""

    attr = 'method_not_applicable'
    error_cls = MethodNotApplicable


class _ServiceVerificationAlreadyExists(_ServiceException):
    """Service variant of VerificationAlreadyExists."""

    attr = 'verification_already_exists'
    error_cls = ProtoVerificationAlreadyExists


class _ServiceVerificationDoesNotExist(_ServiceException):
    """Service variant of VerificationDoesNotExist."""

    attr = 'verification_does_not_exist'
    error_cls = VerificationDoesNotExist


def _handle_service_exceptions(func: F) -> F:
    """Capture service exception and return the reply with an appropriate exception."""
    @wraps(func)
    async def wrapped(*args: Any, **kwargs: Any) -> Any:
        try:
            return await func(*args, **kwargs)
        except _ServiceException as error:
            reply = error.reply
            msg = reply.exception  # type: ignore[attr-defined]
            getattr(msg, error.attr).CopyFrom(error.error_cls(**error.kwargs))
            return reply
    return cast(F, wrapped)


def _reverse_order_by(query: Select) -> Select:
    """Return query with ORDER BY reversed."""
    order_by = []
    for clause in query._order_by_clauses:
        if isinstance(clause, Column):
            order_by.append(clause.desc())
        else:
            assert isinstance(clause, UnaryExpression)  # noqa: S101
            if clause.modifier == asc_op:
                order_by.append(cast(Column, clause.element).desc())
            else:
                assert clause.modifier == desc_op  # noqa: S101
                order_by.append(cast(Column, clause.element).asc())
    return query.order_by(None).order_by(*order_by)


class VerificationServicer(ProtoVerificationServicer):
    """Verification service."""

    async def _get_verification_info(self, verification: Verification, refresh: bool = True) -> VerificationInfo:
        """Return interface verification info for verification model."""
        if refresh:
            await object_session(verification).refresh(verification)
        info = VerificationInfo()
        info.verification_id.value = str(verification.uuid)
        info.contact_id.value = verification.contact_id
        info.create_datetime.FromDatetime(verification.create_datetime)
        if verification.close_datetime:
            info.close_datetime.FromDatetime(verification.close_datetime)
        if verification.last_access_datetime:
            info.last_access_datetime.FromDatetime(verification.last_access_datetime)
        info.status = ProtoVerificationStatus.Value(verification.status)
        info.manual_check_request = verification.manual_check_request
        info.reason = verification.reason
        info.reason_url = verification.reason_url
        for event in await verification.get_events():
            proto_event = info.events.add()
            proto_event.event_datetime.FromDatetime(event.event_datetime)
            if event.method:
                proto_event.method = event.method
            proto_event.type = event.type
            proto_event.details.update(event.details or {})
        # Python 3.9 coverage bug
        async for next_event in APPLICATION.handler.get_next_events(verification):  # pragma: no cover
            proto_event = info.next_events.add()
            if next_event.due_datetime:
                proto_event.event_datetime.FromDatetime(next_event.due_datetime)
            if next_event.method:
                proto_event.method = next_event.method
            proto_event.type = next_event.type
            proto_event.details.update(await next_event.get_details())
        return info

    # XXX: Ignore complexity limit to keep the ContactData -> VerificationResult serialization in a single method.
    async def _get_verification_result(self, contact_data: ContactData, refresh: bool = True,  # noqa: C901
                                       ) -> VerificationResult:
        """Return interface verification result for contact data model."""
        if refresh:
            await object_session(contact_data).refresh(contact_data)

        result = VerificationResult(method=contact_data.method)
        result.verify_datetime.FromDatetime(contact_data.verify_datetime)
        result.contact_history_id.value = contact_data.contact_history_id
        for field in ('name', 'organization', 'company_registration_number'):
            if getattr(contact_data, 'verified_' + field) is not None:
                setattr(result.data, field, getattr(contact_data, field))
                setattr(result, field, getattr(contact_data, 'verified_' + field))
        if contact_data.verified_birthdate is not None:
            result.data.birthdate = cast(date, contact_data.birthdate).isoformat()
            result.birthdate = contact_data.verified_birthdate
        if contact_data.verified_address is not None:
            if contact_data.address_street is not None:
                result.data.address.street[:] = contact_data.address_street
            if contact_data.address_city is not None:
                result.data.address.city = contact_data.address_city
            if contact_data.address_state is not None:
                result.data.address.state_or_province = contact_data.address_state
            if contact_data.address_postal_code is not None:
                result.data.address.postal_code = contact_data.address_postal_code
            if contact_data.address_country_code is not None:
                result.data.address.country_code = contact_data.address_country_code
            result.address = contact_data.verified_address
        if contact_data.verified_emails is not None:
            result.data.emails.extend(contact_data.emails)
            result.emails = contact_data.verified_emails
        return result

    async def _get_verification(self, session: AsyncSession, raw_uuid: str, reply: Message,
                                *, allowed_statuses: Optional[Container[VerificationStatus]] = None) -> Verification:
        """Return a verification model identified by a raw UUID."""
        try:
            uuid = UUID(raw_uuid)
        except ValueError as error:
            raise _ServiceVerificationDoesNotExist(reply) from error
        query = select(Verification).filter_by(uuid=uuid).with_for_update()
        verification = (await session.execute(query)).scalar()
        if verification is None:
            raise _ServiceVerificationDoesNotExist(reply)
        if allowed_statuses and verification.status not in allowed_statuses:
            raise _ServiceIncorrectStatus(reply)
        return verification

    def _decode_status(self, status: int) -> VerificationStatus:
        """Decode status from protobuf enum value to VerificationStatus."""
        try:
            return VerificationStatus(ProtoVerificationStatus.Name(status))
        except ValueError as error:
            raise ServiceError(StatusCode.INVALID_ARGUMENT,
                               "Unknown status '{}'".format(ProtoVerificationStatus.Name(status))) from error

    @awrap_unary
    @_handle_sqlalchemy_error
    @_handle_service_exceptions
    async def create(self, request: CreateRequest, context: ServicerContext) -> CreateReply:
        """Create new verification."""
        async with atomic() as session:
            try:
                verification = await APPLICATION.handler.create(session, request.contact_id.value,
                                                                reason=request.reason, reason_url=request.reason_url)
            except ContactDoesNotExist as error:
                raise _ServiceContactDoesNotExist(CreateReply()) from error
            except VerificationAlreadyExists as error:
                raise _ServiceVerificationAlreadyExists(CreateReply()) from error

            # Flush the model to database to get its data.
            await session.flush()

            reply = CreateReply()
            reply.data.verification.CopyFrom(await self._get_verification_info(verification))
            return reply

    @awrap_unary
    @_handle_sqlalchemy_error
    @_handle_service_exceptions
    async def get(self, request: GetRequest, context: ServicerContext) -> GetReply:
        """Return verification data."""
        async with atomic() as session:
            verification = await self._get_verification(session, request.verification_id.value, GetReply())
            if request.access_token and not compare_digest(request.access_token, verification.access_token):
                raise _ServiceVerificationDoesNotExist(GetReply())

            if request.access_token:
                # Update last_access_datetime if token is provided.
                verification.last_access_datetime = cast(SQLCoreOperations[Optional[datetime]], func.now())
                # Update status if at most INITIALIZED.
                if verification.status <= VerificationStatus.INITIALIZED:
                    verification.status = VerificationStatus.ACCESSED

                # Mark emails as verified if not verified yet and there's only single email event.
                email_query = select(ContactData).filter_by(verification_id=verification.id, verified_emails=True)
                results = tuple((await session.execute(email_query)).scalars())
                email_events = await verification.get_message_events(MessageType.INITIAL_EMAIL)
                if not results and len(email_events) == 1:
                    event = email_events[0]
                    contact_data = ContactData(
                        verification_id=verification.id, contact_history_id=event.details['contact_history_id'],
                        emails=event.details['recipient'], verified_emails=True)
                    session.add(contact_data)

                # Flush the model to database and update its data.
                await session.flush()

            reply = GetReply()
            reply.data.verification.CopyFrom(await self._get_verification_info(verification))
            return reply

    @awrap_unary
    @_handle_sqlalchemy_error
    @_handle_service_exceptions
    async def cancel(self, request: CancelRequest, context: ServicerContext) -> CancelReply:
        """Cancel the verification."""
        async with atomic() as session:
            verification = await self._get_verification(session, request.verification_id.value, CancelReply(),
                                                        allowed_statuses=OPEN_STATUSES)
            await APPLICATION.handler.cancel(verification)
            return CancelReply()

    @awrap_unary
    @_handle_sqlalchemy_error
    @_handle_service_exceptions
    async def resume(self, request: ResumeRequest, context: ServicerContext) -> ResumeReply:
        """Resume the verification."""
        async with atomic() as session:
            verification = await self._get_verification(session, request.verification_id.value, ResumeReply(),
                                                        allowed_statuses=[VerificationStatus.PAUSED])
            verification.status = VerificationStatus.PENDING
            event = Event(verification_id=verification.id, type=EventType.RESUME)
            session.add(event)
            # Flush the model to database to get its data.
            await session.flush()

            # Delay FUTILE by the period since the CHECK_DOMAINS event. Do nothing if the event is not found.
            check_domains = await verification.get_events(EventType.CHECK_DOMAINS)
            if check_domains:
                delay = math.ceil((datetime.now(UTC) - check_domains[0].event_datetime).total_seconds())
                details = {'status': VerificationStatus.FUTILE, 'delay': delay}
                event = Event(verification_id=verification.id, type=EventType.DELAY_STATUS, details=details)
                session.add(event)

            return ResumeReply()

    async def _return_results(self, request: ListRequest, results: List[Verification], count: int,
                              ) -> AsyncIterator[ListReply]:
        """Return results - utility function for `list` method to generate a gRPC stream of messages."""
        if not results:
            # Nothing to return.
            return

        limit = abs(request.pagination.page_size) or count
        returned, remaining = results[:limit], results[limit:]

        # Handle all, but the last verifications.
        coros = [self._get_verification_info(v, refresh=False) for v in returned[:-1]]
        for info in await asyncio.gather(*coros):
            reply = ListReply()
            reply.data.verifications.append(info)
            yield reply

        # Handle the last verification
        verification = returned[-1]
        reply = ListReply()
        reply.data.verifications.append(await self._get_verification_info(verification, refresh=False))
        # Handle pagination
        if request.pagination.page_size:
            if remaining and request.pagination.page_size > 0:
                # In the forward direction, return the last item UUID as the next page token.
                reply.data.pagination.next_page_token = str(verification.uuid)
            elif remaining:
                # In the backward direction, return the next item UUID as the next page token.
                reply.data.pagination.next_page_token = str(remaining[0].uuid)
            if remaining:
                reply.data.pagination.items_left = count - limit
            else:
                reply.data.pagination.items_left = 0
        yield reply

    async def _get_token_value(self, session: AsyncSession, page_token: str) -> Row[Tuple[datetime, int]]:
        """Dereference page_token to a token value for pagination."""
        # The page token is the UUID of the last item from the previous page.
        query = select(Verification.create_datetime, Verification.id)
        if cast(Engine, session.bind).dialect.name == 'sqlite':
            # SQLite has no datetime type and natively formats datetime in a different manner than sqlalchemy does.
            # For filtering we need to use the exact value that's in the database.
            query = select(literal_column(Verification.create_datetime.name), Verification.id)

        try:
            token_uuid = UUID(page_token)
            query = query.filter(Verification.uuid == token_uuid)
            return (await session.execute(query)).one()
        except (ValueError, NoResultFound) as error:
            raise ServiceError(StatusCode.INVALID_ARGUMENT, "Invalid page token") from error

    def _get_list_query(self, session: AsyncSession, request: ListRequest) -> Select:
        """Set up filters on list query."""
        query = select(Verification)
        if request.contact_id.value:
            query = query.filter_by(contact_id=request.contact_id.value)
        if request.domain_id.value:
            query = query.join(VerDomain).filter(VerDomain.domain_id == request.domain_id.value)
        if request.statuses:
            query = query.filter(Verification.status.in_(self._decode_status(s) for s in request.statuses))
        if request.HasField('manual_check_request'):
            query = query.filter_by(manual_check_request=request.manual_check_request)
        if request.order_by:
            if request.order_by[0] == '-':
                direction = 'desc'
                column = request.order_by[1:]
            elif request.order_by[0] == '+':
                direction = 'asc'
                column = request.order_by[1:]
            else:
                direction = 'asc'
                column = request.order_by
            if column not in ('create_datetime', 'close_datetime', 'status'):
                raise ServiceError(StatusCode.INVALID_ARGUMENT, "Invalid order_by value: {!r}".format(column))
            order_by = getattr(getattr(Verification, column), direction)()
        else:
            # Default ORDER BY
            order_by = Verification.create_datetime.desc()
        query = query.order_by(order_by, Verification.id.desc())
        # Load events right away
        query = query.options(selectinload(Verification.events))
        return query

    # _handle_sqlalchemy_error doesn't work on stream method.
    @awrap_stream
    async def list(self, request: ListRequest, context: ServicerContext) -> AsyncIterator[ListReply]:
        """Return list of verifications."""
        try:
            async with atomic() as session:
                query = self._get_list_query(session, request)
                if request.pagination.page_token:
                    token_value = tuple_(*await self._get_token_value(session, request.pagination.page_token))
                    token_query = tuple_(Verification.create_datetime, Verification.id)
                    if request.pagination.page_size < 0:
                        # Drop default ordering, then set new ordering.
                        query = _reverse_order_by(query)
                        query = query.filter(token_query >= token_value)
                    else:
                        query = query.filter(token_query < token_value)

                # It seems to be faster to just count the number of results than use a window function.
                count = (await session.execute(select(func.count()).select_from(query.subquery()))).scalars().one()
                # Select one additional row for pagination data.
                limit = (abs(request.pagination.page_size) or count) + 1
                results = (await session.execute(query.limit(limit))).scalars()
                # Python 3.9 coverage bug
                async for reply in self._return_results(request, list(results), count):  # pragma: no cover
                    yield reply
        except SQLAlchemyError as error:
            _LOGGER.error("Error in SQL query in list: %s", error)
            raise ServiceError(StatusCode.FAILED_PRECONDITION, "Database error: {}".format(error)) from error

    @awrap_unary
    @_handle_sqlalchemy_error
    @_handle_service_exceptions
    async def get_methods(self, request: GetMethodsRequest, context: ServicerContext) -> GetMethodsReply:
        """Get applicable methods for the verification."""
        async with atomic() as session:
            verification = await self._get_verification(session, request.verification_id.value, GetMethodsReply())

            reply = GetMethodsReply()
            for method, processor in APPLICATION.handler.processors.items():
                if await processor.is_applicable(verification):
                    reply.data.methods.append(method)
            return reply

    @awrap_unary
    @_handle_sqlalchemy_error
    @_handle_service_exceptions
    async def use_method(self, request: UseMethodRequest, context: ServicerContext) -> UseMethodReply:
        """Set the selected method to be used for the verification."""
        async with atomic() as session:
            verification = await self._get_verification(session, request.verification_id.value, UseMethodReply(),
                                                        allowed_statuses=OPEN_STATUSES)

            if (request.method not in APPLICATION.handler.processors
                    or not await APPLICATION.handler.processors[request.method].is_applicable(verification)):
                raise _ServiceMethodNotApplicable(UseMethodReply())

            # Record the event only if the method is not already active.
            if request.method not in await verification.active_methods:
                verification.status = max(verification.status, VerificationStatus.PENDING)

                event = Event(verification_id=verification.id, type=EventType.USE_METHOD, method=request.method)
                session.add(event)
                # Flush the model to database to get its data.
                await session.flush()

            reply = UseMethodReply()
            reply.data.verification.CopyFrom(await self._get_verification_info(verification))
            return reply

    def _decode_contact_data(self, proto_data: ProtoContactData, reply: Message) -> Tuple[ContactData, Dict[str, Any]]:
        """Decode contact data from request into contact data and event details."""
        contact_data = ContactData()
        event_details = {}
        if proto_data.HasField('name'):
            contact_data.name = event_details['name'] = proto_data.name
        if proto_data.HasField('organization'):
            contact_data.organization = event_details['organization'] = proto_data.organization
        if proto_data.HasField('birthdate'):
            try:
                contact_data.birthdate = date.fromisoformat(proto_data.birthdate)
            except ValueError as error:
                raise _ServiceInvalidData(reply, fields=['birthdate']) from error
            event_details['birthdate'] = proto_data.birthdate
        if proto_data.HasField('company_registration_number'):
            contact_data.company_registration_number = proto_data.company_registration_number
            event_details['company_registration_number'] = proto_data.company_registration_number
        if proto_data.HasField('address'):
            address = {}
            contact_data.address_street = address['street'] = proto_data.address.street[:]
            contact_data.address_city = address['city'] = proto_data.address.city
            contact_data.address_state = address['state_or_province'] = proto_data.address.state_or_province
            contact_data.address_postal_code = address['postal_code'] = proto_data.address.postal_code
            contact_data.address_country_code = address['country_code'] = proto_data.address.country_code
            event_details['address'] = address
        return contact_data, event_details

    async def _try_complete(self, verification: Verification) -> None:
        """Try to complete the verification in a nested session."""
        try:
            async with object_session(verification).begin_nested():
                await APPLICATION.handler.try_complete(verification)
        except Exception:  # noqa: S110
            pass

    @awrap_unary
    @_handle_sqlalchemy_error
    @_handle_service_exceptions
    async def verify_contact_data(self, request: VerifyContactDataRequest, context: ServicerContext,
                                  ) -> VerifyContactDataReply:
        """Verify contact data supplied for the verification."""
        async with atomic() as session:
            verification = await self._get_verification(session, request.verification_id.value,
                                                        VerifyContactDataReply(), allowed_statuses=OPEN_STATUSES)

            if (request.method not in APPLICATION.handler.processors
                    or not await APPLICATION.handler.processors[request.method].is_applicable(verification)):
                raise _ServiceMethodNotApplicable(VerifyContactDataReply())

            contact_data, event_details = self._decode_contact_data(request.data, VerifyContactDataReply())
            contact_data.verification_id = verification.id
            contact_data.method = request.method
            session.add(contact_data)
            await contact_data.verify()

            event = Event(verification_id=verification.id, type=EventType.VERIFY_CONTACT_DATA, method=request.method,
                          details=event_details)
            session.add(event)

            await self._try_complete(verification)

            reply = VerifyContactDataReply()
            reply.data.verification.CopyFrom(await self._get_verification_info(verification))
            reply.data.result.CopyFrom(await self._get_verification_result(contact_data))
            return reply

    @awrap_unary
    @_handle_sqlalchemy_error
    @_handle_service_exceptions
    async def request_manual_check(self, request: RequestManualCheckRequest, context: ServicerContext,
                                   ) -> RequestManualCheckReply:
        """Request manual check on verification."""
        async with atomic() as session:
            verification = await self._get_verification(session, request.verification_id.value,
                                                        RequestManualCheckReply(), allowed_statuses=OPEN_STATUSES)
            verification.manual_check_request = True

            event = Event(verification_id=verification.id, type=EventType.REQUEST_MANUAL_CHECK)
            session.add(event)
            return RequestManualCheckReply()

    @awrap_unary
    @_handle_sqlalchemy_error
    @_handle_service_exceptions
    async def resolve_manual_check(self, request: ResolveManualCheckRequest, context: ServicerContext,
                                   ) -> ResolveManualCheckReply:
        """Resolve manual check of verification."""
        async with atomic() as session:
            verification = await self._get_verification(session, request.verification_id.value,
                                                        ResolveManualCheckReply(), allowed_statuses=OPEN_STATUSES)

            # Drop the manual check flag.
            verification.manual_check_request = False

            contact_data, event_details = self._decode_contact_data(request.data, ResolveManualCheckReply())
            contact_data.verification_id = verification.id
            contact_data.contact_history_id = request.contact_history_id.value
            if request.data.HasField('name'):
                contact_data.verified_name = request.verified
            if request.data.HasField('organization'):
                contact_data.verified_organization = request.verified
            if request.data.HasField('birthdate'):
                contact_data.verified_birthdate = request.verified
            if request.data.HasField('company_registration_number'):
                contact_data.verified_company_registration_number = request.verified
            if request.data.HasField('address'):
                contact_data.verified_address = request.verified
            session.add(contact_data)

            event_details['verified'] = request.verified
            event = Event(verification_id=verification.id, type=EventType.RESOLVE_MANUAL_CHECK, details=event_details)
            session.add(event)

            # get recepients from registry and send email notification
            contact = await verification.contact
            if contact.emails:
                # XXX: Not a nice solution. Check how it's going to be used and refactor appropriately.
                futile_datetime = ''
                # Python 3.9 coverage bug
                async for processor in APPLICATION.handler.get_active_processors(verification):  # pragma: no cover
                    futile_datetime = (await processor._get_futile_due(verification)).isoformat()
                    break

                # XXX: Variables `contact_data.verified` and `verified` are duplicates.
                context = {
                    'uuid': str(verification.uuid), 'access_token': verification.access_token,
                    'contact_data': event_details, 'verified': request.verified, 'handle': contact.contact_handle,
                    'registrar_handle': contact.sponsoring_registrar, 'futile_datetime': futile_datetime}
                await APPLICATION.email_client.send(
                    Email(
                        recipient=', '.join(contact.emails),
                        subject_template=Template.MANUAL_CHECK_RESOLUTION_SUBJECT,
                        body_template=Template.MANUAL_CHECK_RESOLUTION_BODY,
                        type=MessageType.MANUAL_CHECK_RESOLUTION,
                        context=context,
                    ),
                    references=[Reference(type=ReferenceType.CONTACT, value=verification.contact_id),
                                Reference(type=ReferenceType.VERIFICATION, value=str(verification.uuid))]
                )

            await self._try_complete(verification)

            return ResolveManualCheckReply()

    @awrap_unary
    @_handle_sqlalchemy_error
    @_handle_service_exceptions
    async def list_verification_results(self, request: ListVerificationResultsRequest, context: ServicerContext,
                                        ) -> ListVerificationResultsReply:
        """Return verification results."""
        async with atomic() as session:
            verification = await self._get_verification(session, request.verification_id.value,
                                                        ListVerificationResultsReply())

            reply = ListVerificationResultsReply()
            for contact_data in await verification.awaitable_attrs.contact_data:
                reply.data.results.append(await self._get_verification_result(contact_data, refresh=False))
            return reply

    @awrap_unary
    @_handle_sqlalchemy_error
    @_handle_service_exceptions
    async def verify_token(self, request: VerifyTokenRequest, context: ServicerContext) -> VerifyTokenReply:
        """Verify the verification's token."""
        async with atomic() as session:
            verification = await self._get_verification(session, request.verification_id.value, VerifyTokenReply(),
                                                        allowed_statuses=OPEN_STATUSES)

            query = select(Token).filter_by(verification_id=verification.id).filter(
                func.upper(Token.token) == request.token.upper())
            token = (await session.execute(query)).scalar()
            if token and not request.check:
                event = Event(verification_id=verification.id, type=EventType.VERIFY_TOKEN, method=token.method)
                session.add(event)

                # Call the callback.
                if token.method not in APPLICATION.handler.processors:
                    _LOGGER.warning("Token's %s method %s not found in processors.", token, token.method)
                else:
                    await APPLICATION.handler.processors[token.method].handle_token_verified(token)

                # And delete the token.
                await session.delete(token)

                await self._try_complete(verification)
            else:
                pass

            reply = VerifyTokenReply()
            reply.data.token_matches = (token is not None)
            return reply

    @awrap_unary
    @_handle_sqlalchemy_error
    @_handle_service_exceptions
    async def delay_status(self, request: DelayStatusRequest, context: ServicerContext) -> DelayStatusReply:
        """Delay verification status change."""
        async with atomic() as session:
            verification = await self._get_verification(session, request.verification_id.value, DelayStatusReply(),
                                                        allowed_statuses=OPEN_STATUSES)
            # Store the delay
            status = self._decode_status(request.status)
            details = {'status': status, 'delay': request.delay}
            event = Event(verification_id=verification.id, type=EventType.DELAY_STATUS, details=details)
            session.add(event)

            # Flush models to database to update their data.
            await session.flush()

            # Try partial cancel.
            await APPLICATION.handler.cancel(verification, status)

            reply = DelayStatusReply()
            reply.data.verification.CopyFrom(await self._get_verification_info(verification))
            return reply

    @awrap_unary
    @_handle_sqlalchemy_error
    @_handle_service_exceptions
    async def resend_message(self, request: ResendMessageRequest, context: ServicerContext) -> ResendMessageReply:
        """Resend a message related to the verification."""
        async with atomic() as session:
            verification = await self._get_verification(session, request.verification_id.value, ResendMessageReply(),
                                                        allowed_statuses=OPEN_STATUSES - {VerificationStatus.CREATED})
            scheduled = None
            # Python 3.9 coverage bug
            async for processor in APPLICATION.handler.get_active_processors(verification):  # pragma: no cover
                try:
                    scheduled = await processor.get_resend_event(verification, request.message_type)
                except ValueError:
                    continue
                else:
                    await scheduled.process()
                    session.add(await scheduled.to_event())
                    break
            if not scheduled:
                raise ServiceError(StatusCode.FAILED_PRECONDITION, "Message type not found.")
            return ResendMessageReply()

    @awrap_unary
    @_handle_sqlalchemy_error
    @_handle_service_exceptions
    async def approve_futile(self, request: ApproveFutileRequest, context: ServicerContext) -> ApproveFutileReply:
        """Approve verification is futile."""
        async with atomic() as session:
            verification = await self._get_verification(session, request.verification_id.value, ApproveFutileReply(),
                                                        allowed_statuses=[VerificationStatus.FUTILE])
            if verification.manual_check_request:
                raise _ServiceIncorrectStatus(ApproveFutileReply())
            failed_event = FailedEvent(verification)
            await failed_event.process()

            result = await APPLICATION.domain_client.get_domains_by_contact(verification.contact_id,
                                                                            {DomainContactRole.HOLDER})
            domains = [
                Domain(domain_id=d['domain'].id, domain_history_id=d['domain_history_id'], fqdn=d['domain'].fqdn)
                for d in result['domains']]

            undeleted = await alist(APPLICATION.domain_admin_client.batch_delete_domains(domains))
            undeleted_data = {d['domain_id']: d['reason'].name.lower() for d in undeleted}
            details = {'domains': [d.domain_id for d in domains], 'undeleted': undeleted_data}
            event = await failed_event.to_event()
            event.details.update(details)
            session.add(event)
            # Flush models to database
            await session.flush()

            reply = ApproveFutileReply()
            reply.data.verification.CopyFrom(await self._get_verification_info(verification))
            return reply

    @awrap_unary
    @_handle_sqlalchemy_error
    @_handle_service_exceptions
    async def approve_manual(self, request: ApproveManualRequest, context: ServicerContext) -> ApproveManualReply:
        """Close verification as manual."""
        async with atomic() as session:
            verification = await self._get_verification(session, request.verification_id.value, ApproveManualReply(),
                                                        allowed_statuses=OPEN_STATUSES)
            if verification.status == VerificationStatus.FUTILE:
                cancel_futile = CancelFutileEvent(verification)
                await cancel_futile.process()
                session.add(await cancel_futile.to_event())
            if verification.status == VerificationStatus.CONSTRAINED:
                cancel_outzone = CancelOutzoneEvent(verification)
                await cancel_outzone.process()
                session.add(await cancel_outzone.to_event())

            verification.status = VerificationStatus.MANUAL
            verification.close_datetime = cast(SQLCoreOperations[Optional[datetime]], func.now())
            event = Event(verification_id=verification.id, type=EventType.MANUAL, details={})
            session.add(event)
            # Flush models to database
            await session.flush()

            contact = await verification.contact
            flags: Dict[str, bool] = {ObjectState.CONTACT_IN_MANUAL_VERIFICATION: False,
                                      ObjectState.CONTACT_PASSED_MANUAL_VERIFICATION: True,
                                      ObjectState.CONTACT_FAILED_MANUAL_VERIFICATION: False}
            await APPLICATION.contact_client.update_contact_state(
                contact.contact_id, cast(ContactHistoryId, contact.contact_history_id), flags)

            reply = ApproveManualReply()
            reply.data.verification.CopyFrom(await self._get_verification_info(verification))
            return reply

    @awrap_unary
    @_handle_sqlalchemy_error
    @_handle_service_exceptions
    async def set_contact_representative(self, request: SetContactRepresentativeRequest, context: ServicerContext
                                         ) -> SetContactRepresentativeReply:
        """Set contact representative."""
        async with atomic() as session:
            verification = await self._get_verification(session, request.verification_id.value,
                                                        SetContactRepresentativeReply(),
                                                        allowed_statuses=OPEN_STATUSES)

            address = {"street": list(request.address.street), "city": request.address.city,
                       "state_or_province": request.address.state_or_province,
                       "postal_code": request.address.postal_code, "country_code": request.address.country_code}
            details = {"name": request.name, "organization": request.organization, "address": address,
                       "telephone": request.telephone, "email": request.email}
            event = Event(verification_id=verification.id, type=EventType.SET_CONTACT_REPRESENTATIVE, details=details)
            session.add(event)

            return SetContactRepresentativeReply()
