"""Verification settings management."""
from functools import cached_property
from typing import TYPE_CHECKING, Any, Dict, Optional, Sequence, Tuple, Union

from filed import FileClient, Storage
from frgal import make_credentials
from hermes import EmailMessengerClient, LetterMessengerClient, SmsMessengerClient
from pydantic import BaseModel, Json
from regal import ContactClient, DomainAdminClient, DomainClient, RepresentativeClient
from setapp import BaseApplication, BaseSettings, SettingsConfigDict
from typist import HTTPTokenAuth, SecretaryClient

if TYPE_CHECKING:
    from .handler import VerificationHandler

ENVIRON_NAME = 'VERIFICATION_CONFIG'
CONFIG_FILES = ('~/.fred/verification.conf', '/etc/fred/verification.conf')


class CheckDomainsSettings(BaseModel):
    """Settings for high rank domains check."""

    count_limit: int = 0
    notification_email: Optional[str] = None
    rank_query: str
    rank_headers: Union[Dict[str, str], Json[Dict[str, str]]] = {}
    rank_timeout: Union[float, Tuple[float, float]] = 3.05


class DispersionSettings(BaseModel):
    """Settings for dispersion."""

    max_age: int = 0
    transient_statuses: Tuple[str, ...] = ()
    verified_statuses: Tuple[str, ...] = ()


class Settings(BaseSettings):
    """Verification settings."""

    model_config = SettingsConfigDict(env_prefix='VERIFICATION_',
                                      config_files=CONFIG_FILES,
                                      config_environ=ENVIRON_NAME)

    check_domains: Optional[CheckDomainsSettings] = None
    checks: Dict[str, Dict[str, Any]] = {}
    db_connection: str = 'sqlite+aiosqlite:///:memory:'
    db_echo: bool = False
    db_poolclass: Optional[str] = None
    db_schema: Optional[str] = None
    dispersion: Optional[DispersionSettings] = None
    fileman_netloc: str = 'localhost:50051'
    fileman_ssl_cert: Optional[str] = None
    grpc_jobs: int = 0
    grpc_maximum_concurrent_rpcs: Optional[int] = None
    grpc_options: Optional[Sequence] = None
    grpc_port: int = 50051
    messenger_netloc: str = 'localhost:50051'
    messenger_ssl_cert: Optional[str] = None
    processors: Dict[str, Dict[str, Any]] = {}
    registry_netloc: str = 'localhost:50050'
    registry_ssl_cert: Optional[str] = None
    secretary_timeout: Union[float, Tuple[float, float]] = 3.05
    secretary_token: Optional[str] = None
    secretary_url: str = 'http://localhost/'
    test_db_connection: str = 'sqlite+aiosqlite:///:memory:'


class Application(BaseApplication[Settings]):
    """Represent application settings and basic setup."""

    @cached_property
    def handler(self) -> 'VerificationHandler':
        """Return a verification handler."""
        from .handler import VerificationHandler
        return VerificationHandler()

    @cached_property
    def contact_client(self) -> ContactClient:
        """Return a contact client."""
        return ContactClient(self.settings.registry_netloc, make_credentials(self.settings.registry_ssl_cert))

    @cached_property
    def domain_admin_client(self) -> DomainAdminClient:
        """Return a domain_admin client."""
        return DomainAdminClient(self.settings.registry_netloc, make_credentials(self.settings.registry_ssl_cert))

    @cached_property
    def domain_client(self) -> DomainClient:
        """Return a domain client."""
        return DomainClient(self.settings.registry_netloc, make_credentials(self.settings.registry_ssl_cert))

    @cached_property
    def email_client(self) -> EmailMessengerClient:
        """Return a email client."""
        return EmailMessengerClient(self.settings.messenger_netloc, make_credentials(self.settings.messenger_ssl_cert))

    @cached_property
    def file_storage(self) -> Storage:
        """Return a file storage."""
        return Storage(FileClient(self.settings.fileman_netloc, make_credentials(self.settings.fileman_ssl_cert)))

    @cached_property
    def letter_client(self) -> LetterMessengerClient:
        """Return a letter client."""
        return LetterMessengerClient(self.settings.messenger_netloc, make_credentials(self.settings.messenger_ssl_cert))

    @cached_property
    def representative_client(self) -> RepresentativeClient:
        """Return a representative client."""
        return RepresentativeClient(self.settings.registry_netloc, make_credentials(self.settings.registry_ssl_cert))

    @cached_property
    def secretary_client(self) -> SecretaryClient:
        """Return a secretary client."""
        auth = None
        if self.settings.secretary_token:
            auth = HTTPTokenAuth(self.settings.secretary_token)
        return SecretaryClient(self.settings.secretary_url, timeout=self.settings.secretary_timeout, auth=auth)

    @cached_property
    def sms_client(self) -> SmsMessengerClient:
        """Return a sms client."""
        return SmsMessengerClient(self.settings.messenger_netloc, make_credentials(self.settings.messenger_ssl_cert))

    def clear_cache(self) -> None:
        """Clear all caches."""
        self._del('handler')
        self._del('contact_client')
        self._del('domain_admin_client')
        self._del('domain_client')
        self._del('email_client')
        self._del('file_storage')
        self._del('letter_client')
        self._del('representative_client')
        self._del('secretary_client')
        self._del('sms_client')


APPLICATION = Application(Settings)
