from typing import Any, Dict, List, Optional, Type
from unittest import IsolatedAsyncioTestCase
from unittest.mock import ANY, sentinel

import responses
from fred_types import ContactHistoryId, ContactId
from regal import Address, Contact
from requests import RequestException
from responses.matchers import request_kwargs_matcher, urlencoded_params_matcher

from verification.checks import RuianAddressCheck
from verification.exceptions import SkipEvaluation


class RuianAddressCheckTest(IsolatedAsyncioTestCase):
    contact_id = ContactId('KRYTEN')
    handle = '2X4B'
    contact_history_id = ContactHistoryId('523P')

    def test_get_streets(self):
        data = (
            # street, streets
            ('', ('', )),
            ('Deck 16', ('Deck 16', )),
            ('Deck 16, Upper bank', ('Deck 16, Upper bank', 'Deck 16', 'Upper bank')),
        )
        check = RuianAddressCheck(name=sentinel.name)
        for street, streets in data:
            with self.subTest(street=street):
                self.assertEqual(tuple(check._get_streets(street)), streets)

    def test_get_queries(self):
        data = (
            # address, queries
            (Address(street=[], city='', postal_code=''), (', ', )),
            (Address(street=['Deck 16'], city='Red Dwarf', postal_code='JMC'),
             ('Deck 16, Red Dwarf', 'Deck 16, JMC', '16, JMC')),
            (Address(street=['Deck 16/42'], city='Red Dwarf', postal_code='JMC'),
             ('Deck 16/42, Red Dwarf', 'Deck 16/42, JMC', 'Deck 16, Red Dwarf', 'Deck 16, JMC',
              '16/42, JMC', '16, JMC')),
            # Multiple streets
            (Address(street=['Upper bank', 'Deck 16'], city='Red Dwarf', postal_code='JMC'),
             ('Upper bank, Red Dwarf', 'Upper bank, JMC')),
            # Multiple streets as a single string
            (Address(street=['Deck 16, Upper bank'], city='Red Dwarf', postal_code='JMC'),
             ('Deck 16, Upper bank, Red Dwarf', 'Deck 16, Upper bank, JMC',
              'Deck 16, Red Dwarf', 'Deck 16, JMC', 'Upper bank, Red Dwarf', 'Upper bank, JMC', '16, JMC')),
            # Whitespaces in postal code
            (Address(street=['Deck 16'], city='Red Dwarf', postal_code='\tJ M\nC  '),
             ('Deck 16, Red Dwarf', 'Deck 16, JMC', '16, JMC')),
            # Replace \ with /
            (Address(street=['Deck 16\\42'], city='Red Dwarf', postal_code='JMC'),
             ('Deck 16\\42, Red Dwarf', 'Deck 16\\42, JMC', 'Deck 16/42, Red Dwarf', 'Deck 16/42, JMC',
              'Deck 16, Red Dwarf', 'Deck 16, JMC', '16/42, JMC', '16, JMC')),
            # Add a space after a dot or a number if not present
            (Address(street=['16. deck'], city='Red Dwarf', postal_code='JMC'),
             ('16. deck, Red Dwarf', '16. deck, JMC')),
            (Address(street=['16.deck'], city='Red Dwarf', postal_code='JMC'),
             ('16.deck, Red Dwarf', '16.deck, JMC', '16. deck, Red Dwarf', '16. deck, JMC')),
            (Address(street=['16. deck 42'], city='Red Dwarf', postal_code='JMC'),
             ('16. deck 42, Red Dwarf', '16. deck 42, JMC', '42, JMC')),
            (Address(street=['16.deck 42'], city='Red Dwarf', postal_code='JMC'),
             ('16.deck 42, Red Dwarf', '16.deck 42, JMC', '16. deck 42, Red Dwarf', '16. deck 42, JMC', '42, JMC')),
            (Address(street=['16 deck'], city='Red Dwarf', postal_code='JMC'),
             ('16 deck, Red Dwarf', '16 deck, JMC', ANY, ANY)),
            (Address(street=['16deck'], city='Red Dwarf', postal_code='JMC'),
             ('16deck, Red Dwarf', '16deck, JMC', '16 deck, Red Dwarf', '16 deck, JMC', ANY, ANY)),
            (Address(street=['16 deck 42'], city='Red Dwarf', postal_code='JMC'),
             ('16 deck 42, Red Dwarf', '16 deck 42, JMC', '42, JMC')),
            (Address(street=['16deck 42'], city='Red Dwarf', postal_code='JMC'),
             ('16deck 42, Red Dwarf', '16deck 42, JMC', '16 deck 42, Red Dwarf', '16 deck 42, JMC', '42, JMC')),
            # Add a space before the number in the street name
            (Address(street=['Deck16'], city='Red Dwarf', postal_code='JMC'),
             ('Deck16, Red Dwarf', 'Deck16, JMC', 'Deck 16, Red Dwarf', 'Deck 16, JMC', '16, JMC')),
            # Drop letters from the street number
            (Address(street=['Deck 16I'], city='Red Dwarf', postal_code='JMC'),
             ('Deck 16I, Red Dwarf', 'Deck 16I, JMC', 'Deck 16, Red Dwarf', 'Deck 16, JMC', '16, JMC')),
            (Address(street=['Deck 16 I'], city='Red Dwarf', postal_code='JMC'),
             ('Deck 16 I, Red Dwarf', 'Deck 16 I, JMC', 'Deck 16, Red Dwarf', 'Deck 16, JMC', '16, JMC')),
            (Address(street=['Deck 16III'], city='Red Dwarf', postal_code='JMC'),
             ('Deck 16III, Red Dwarf', 'Deck 16III, JMC', 'Deck 16, Red Dwarf', 'Deck 16, JMC', '16, JMC')),
            (Address(street=['Deck 16/I'], city='Red Dwarf', postal_code='JMC'),
             ('Deck 16/I, Red Dwarf', 'Deck 16/I, JMC', 'Deck 16, Red Dwarf', 'Deck 16, JMC', '16, JMC')),
            (Address(street=['Deck 16 / I'], city='Red Dwarf', postal_code='JMC'),
             ('Deck 16 / I, Red Dwarf', 'Deck 16 / I, JMC', 'Deck 16, Red Dwarf', 'Deck 16, JMC', '16, JMC')),
            (Address(street=['Deck 16/III'], city='Red Dwarf', postal_code='JMC'),
             ('Deck 16/III, Red Dwarf', 'Deck 16/III, JMC', 'Deck 16, Red Dwarf', 'Deck 16, JMC', '16, JMC')),
            (Address(street=['Deck 16-I'], city='Red Dwarf', postal_code='JMC'),
             ('Deck 16-I, Red Dwarf', 'Deck 16-I, JMC', 'Deck 16, Red Dwarf', 'Deck 16, JMC', '16, JMC')),
            (Address(street=['Deck 16 - I'], city='Red Dwarf', postal_code='JMC'),
             ('Deck 16 - I, Red Dwarf', 'Deck 16 - I, JMC', 'Deck 16, Red Dwarf', 'Deck 16, JMC', '16, JMC')),
            (Address(street=['Deck 16-III'], city='Red Dwarf', postal_code='JMC'),
             ('Deck 16-III, Red Dwarf', 'Deck 16-III, JMC', 'Deck 16, Red Dwarf', 'Deck 16, JMC', '16, JMC')),
        )
        check = RuianAddressCheck(name=sentinel.name)
        for address, queries in data:
            with self.subTest(address=address):
                self.assertEqual(tuple(check._get_queries(address)), queries)

    async def test_evaluate_not_czech(self):
        check = RuianAddressCheck(name=sentinel.name)

        contact = Contact(contact_id=self.contact_id, contact_handle=self.handle,
                          contact_history_id=self.contact_history_id,
                          place=Address(street=[], city='', country_code='JU'))
        with self.assertRaises(SkipEvaluation):
            await check.evaluate(contact)

    async def _test_evaluate(self, place: Address, value: float, details: Dict[str, Any],
                             *, suggestions: Optional[List[Dict[str, Any]]] = None) -> None:
        check = RuianAddressCheck(name=sentinel.name)

        contact = Contact(contact_id=self.contact_id, contact_handle=self.handle,
                          contact_history_id=self.contact_history_id, place=place)
        with responses.RequestsMock() as rsps:
            if suggestions is None:
                suggestions = [{'text': 'Deck 16, JMC Red Dwarf'}]
            rsps.add(responses.POST, RuianAddressCheck.RUIAN_URL, json={'suggestions': suggestions})
            result = await check.evaluate(contact)

        self.assertEqual(result.name, sentinel.name)
        self.assertAlmostEqual(result.value, value)
        self.assertEqual(result.details, details)

    async def test_evaluate_empty(self):
        place = Address(street=[], city='', postal_code='', country_code='CZ')
        details = {'registry': ',  ', 'source': 'Deck 16, JMC Red Dwarf', 'query': ', '}
        await self._test_evaluate(place, 0.24, details)

    async def test_evaluate(self):
        place = Address(street=['Deck 16'], city='Red Dwarf', postal_code='JMC', country_code='CZ')
        details = {'registry': 'Deck 16, JMC Red Dwarf', 'source': 'Deck 16, JMC Red Dwarf',
                   'query': 'Deck 16, Red Dwarf'}
        await self._test_evaluate(place, 1.0, details)

    async def test_evaluate_state_or_province(self):
        place = Address(street=['Deck 16'], city='Red Dwarf', state_or_province='Deep Space', postal_code='JMC',
                        country_code='CZ')
        details = {'registry': 'Deck 16, JMC Red Dwarf, Deep Space', 'source': 'Deck 16, JMC Red Dwarf',
                   'query': 'Deck 16, Red Dwarf'}
        await self._test_evaluate(place, 0.7857143, details)

    async def test_evaluate_no_suggestion(self):
        # Test no addresses were returned.
        place = Address(street=[], city='', postal_code='', country_code='CZ')
        details = {'registry': ',  ', 'source': '', 'query': ', '}
        await self._test_evaluate(place, 0.0, details, suggestions=[])

    async def test_evaluate_multiple_suggestions(self):
        # Test multiple addresses were returned.
        place = Address(street=['Deck 16'], city='Red Dwarf', postal_code='JMC', country_code='CZ')
        details = {'registry': 'Deck 16, JMC Red Dwarf', 'source': 'Deck 16, JMC Red Dwarf',
                   'query': 'Deck 16, Red Dwarf'}
        suggestions = [{'text': 'Cockpit, JMC Nova 5'},
                       {'text': 'Deck 16, JMC Red Dwarf'},
                       {'text': 'Starbug, JMC Red Dwarf'}]
        await self._test_evaluate(place, 1.0, details, suggestions=suggestions)

    async def test_evaluate_first_match(self):
        # Test evaluate with first query over threshold.
        check = RuianAddressCheck(name=sentinel.name)

        place = Address(street=['Deck 16'], city='Red Dwarf', postal_code='JMC', country_code='CZ')
        contact = Contact(contact_id=self.contact_id, contact_handle=self.handle,
                          contact_history_id=self.contact_history_id, place=place)
        suggestions = [{'text': 'Deck 16, JMC Red Dwarf'}]
        with responses.RequestsMock() as rsps:
            rsps.add(responses.POST, RuianAddressCheck.RUIAN_URL,
                     json={'suggestions': suggestions},
                     match=[urlencoded_params_matcher({"f": "json", "text": 'Deck 16, Red Dwarf'})])
            result = await check.evaluate(contact)

        self.assertEqual(result.name, sentinel.name)
        self.assertAlmostEqual(result.value, 1.0)
        details = {'registry': 'Deck 16, JMC Red Dwarf', 'source': 'Deck 16, JMC Red Dwarf',
                   'query': 'Deck 16, Red Dwarf'}
        self.assertEqual(result.details, details)

    async def test_evaluate_second_match(self):
        # Test evaluate with first and second query under and over threshold, respectively.
        check = RuianAddressCheck(name=sentinel.name)

        place = Address(street=['Deck 16'], city='Red Dwarf', postal_code='JMC', country_code='CZ')
        contact = Contact(contact_id=self.contact_id, contact_handle=self.handle,
                          contact_history_id=self.contact_history_id, place=place)
        suggestions = [{'text': 'Deck 16, JMC Red Dwarf'}]
        with responses.RequestsMock() as rsps:
            # First query -> no suggestions
            rsps.add(responses.POST, RuianAddressCheck.RUIAN_URL,
                     json={'suggestions': []},
                     match=[urlencoded_params_matcher({"f": "json", "text": 'Deck 16, Red Dwarf'})])
            # Second query -> suggestions
            rsps.add(responses.POST, RuianAddressCheck.RUIAN_URL,
                     json={'suggestions': suggestions},
                     match=[urlencoded_params_matcher({"f": "json", "text": 'Deck 16, JMC'})])
            result = await check.evaluate(contact)

        self.assertEqual(result.name, sentinel.name)
        self.assertAlmostEqual(result.value, 1.0)
        details = {'registry': 'Deck 16, JMC Red Dwarf', 'source': 'Deck 16, JMC Red Dwarf',
                   'query': 'Deck 16, JMC'}
        self.assertEqual(result.details, details)

    async def test_evaluate_cache_query_result(self):
        # Test evaluate doesn't requests RUIAN the same query twice.
        check = RuianAddressCheck(name=sentinel.name)

        place = Address(street=[], city='', postal_code='', country_code='CZ')
        contact = Contact(contact_id=self.contact_id, contact_handle=self.handle,
                          contact_history_id=self.contact_history_id, place=place)
        with responses.RequestsMock() as rsps:
            rsps.add(responses.POST, RuianAddressCheck.RUIAN_URL, json={'suggestions': []})
            result = await check.evaluate(contact)

            self.assertEqual(len(rsps.calls), 1)

        self.assertEqual(result.name, sentinel.name)
        self.assertAlmostEqual(result.value, 0.0)
        self.assertEqual(result.details, {'registry': ',  ', 'source': '', 'query': ', '})

    async def _test_evaluate_ruian_error(self, exception: Type[Exception]) -> None:
        check = RuianAddressCheck(name=sentinel.name)

        contact = Contact(contact_id=self.contact_id, contact_handle=self.handle,
                          contact_history_id=self.contact_history_id,
                          place=Address(street=[], city='', postal_code='', country_code='CZ'))
        with self.assertRaises(exception):
            await check.evaluate(contact)

    async def test_evaluate_ruian_error_status(self):
        with responses.RequestsMock() as rsps:
            rsps.add(responses.POST, RuianAddressCheck.RUIAN_URL, status=500)
            await self._test_evaluate_ruian_error(RequestException)

    async def test_evaluate_ruian_error_json(self):
        with responses.RequestsMock() as rsps:
            rsps.add(responses.POST, RuianAddressCheck.RUIAN_URL, json={"error": {"code": 400}})
            await self._test_evaluate_ruian_error(KeyError)

    async def test_evaluate_ruian_invalid_json(self):
        with responses.RequestsMock() as rsps:
            rsps.add(responses.POST, RuianAddressCheck.RUIAN_URL, body='{INVALID')
            await self._test_evaluate_ruian_error(RequestException)

    async def test_evaluate_ruian_timeout(self):
        with responses.RequestsMock() as rsps:
            rsps.add(responses.POST, RuianAddressCheck.RUIAN_URL, body=RequestException())
            await self._test_evaluate_ruian_error(RequestException)

    async def test_evaluate_url(self):
        check = RuianAddressCheck(name=sentinel.name, url='http://example.org/')

        contact = Contact(contact_id=self.contact_id, contact_handle=self.handle,
                          contact_history_id=self.contact_history_id,
                          place=Address(street=[], city='', postal_code='', country_code='CZ'))
        with responses.RequestsMock() as rsps:
            rsps.add(responses.POST, "http://example.org/", json={'suggestions': [{'text': 'Deck 16, JMC Red Dwarf'}]})
            result = await check.evaluate(contact)

        self.assertEqual(result.name, sentinel.name)

    async def test_evaluate_timeout(self):
        check = RuianAddressCheck(name=sentinel.name, timeout=42)

        contact = Contact(contact_id=self.contact_id, contact_handle=self.handle,
                          contact_history_id=self.contact_history_id,
                          place=Address(street=[], city='', postal_code='', country_code='CZ'))
        with responses.RequestsMock() as rsps:
            rsps.add(responses.POST, RuianAddressCheck.RUIAN_URL,
                     json={'suggestions': [{'text': 'Deck 16, JMC Red Dwarf'}]},
                     match=[request_kwargs_matcher({'timeout': 42})])
            result = await check.evaluate(contact)

        self.assertEqual(result.name, sentinel.name)
