import signal
from datetime import timedelta
from typing import Optional, Sequence
from unittest import TestCase
from unittest.mock import ANY, AsyncMock, _Call, call, patch

from verification.commands.evaluate import main
from verification.evaluator import Evaluator
from verification.utils import Result

from .utils import CommandMixin, mock_awaits


class TestEvaluator(Evaluator):
    """Simple evaluator to test the `evaluate` command.

    The tests for the `evaluate` command can't async, because the command itself contains `asyncio.run`.
    Thus the evaluator is overridden to mock the `evaluate_all` parts.
    """

    def __init__(self, result: Result):
        super().__init__()
        self._result = result

    async def evaluate_all(self, renew_delay: Optional[timedelta] = None, throttle: int = 0, batch_size: int = 10_000,
                           *, skip_delay: timedelta = timedelta(days=1)) -> Result:
        return self._result


class MainTest(CommandMixin, TestCase):
    command = main
    contact_id = 'KRYTEN'

    def setUp(self):
        self.sigterm_handler = signal.getsignal(signal.SIGTERM)

    def tearDown(self):
        signal.signal(signal.SIGTERM, self.sigterm_handler)

    def test_main_results(self):
        data = (
            # result, stdout
            (Result(), "Evaluate: 0 total, 0 evaluated, 0 skipped, 0 errors.\n"),
            (Result(success=1), "Evaluate: 1 total, 1 evaluated, 0 skipped, 0 errors.\n"),
            (Result(skip=1), "Evaluate: 1 total, 0 evaluated, 1 skipped, 0 errors.\n"),
            (Result(error=1), "Evaluate: 1 total, 0 evaluated, 0 skipped, 1 errors.\n"),
        )
        for result, stdout in data:
            with self.subTest(result=result):
                with patch('verification.commands.evaluate.Evaluator', return_value=TestEvaluator(result)):
                    self.assertCommandSuccess(stdout=stdout)

    def test_main_renew_delay(self):
        evaluator_mock = AsyncMock(Evaluator)
        evaluator_mock.evaluate_all.return_value = Result(success=1)
        evaluator_mock.stop_process = False

        with patch('verification.commands.evaluate.Evaluator', return_value=evaluator_mock):
            self.assertCommandSuccess(['--renew-delay', '42'],
                                      stdout="Evaluate: 1 total, 1 evaluated, 0 skipped, 0 errors.\n")

        calls = [call.evaluate_all(renew_delay=timedelta(seconds=42), throttle=10, skip_delay=timedelta(days=1))]
        self.assertEqual(mock_awaits(evaluator_mock), calls)

    def test_main_throttle(self):
        evaluator_mock = AsyncMock(Evaluator)
        evaluator_mock.evaluate_all.return_value = Result(success=1)
        evaluator_mock.stop_process = False

        with patch('verification.commands.evaluate.Evaluator', return_value=evaluator_mock):
            self.assertCommandSuccess(['--throttle', '42'],
                                      stdout="Evaluate: 1 total, 1 evaluated, 0 skipped, 0 errors.\n")

        calls = [call.evaluate_all(renew_delay=timedelta(days=30), throttle=42, skip_delay=timedelta(days=1))]
        self.assertEqual(mock_awaits(evaluator_mock), calls)

    def test_main_skip_delay(self):
        evaluator_mock = AsyncMock(Evaluator)
        evaluator_mock.evaluate_all.return_value = Result(success=1)
        evaluator_mock.stop_process = False

        with patch('verification.commands.evaluate.Evaluator', return_value=evaluator_mock):
            self.assertCommandSuccess(['--skip-delay', '42'],
                                      stdout="Evaluate: 1 total, 1 evaluated, 0 skipped, 0 errors.\n")

        calls = [call.evaluate_all(renew_delay=timedelta(days=30), throttle=10, skip_delay=timedelta(seconds=42))]
        self.assertEqual(mock_awaits(evaluator_mock), calls)

    def _test_main_loop(self, args: Sequence[str], sleep_calls: Sequence[_Call]) -> None:
        # Hack the time.sleep to trigger keyboard interrupt to stop the cycle.
        with patch('verification.commands.evaluate.Evaluator', return_value=TestEvaluator(Result())):
            with patch('verification.commands.evaluate.sleep') as sleep_mock:
                sleep_mock.side_effect = [None, KeyboardInterrupt]

                # Interrupting the command will cause non-zero exit code.
                self.assertCommandFail(args, stdout=ANY)

        # Check sleep was called
        self.assertEqual(sleep_mock.mock_calls, sleep_calls)

    def test_main_loop(self):
        # Test the loop mode.
        self._test_main_loop(['--loop'], [call(3600), call(3600)])

    def test_main_loop_delay(self):
        self._test_main_loop(['--loop', '--loop-delay', '1024'], [call(1024), call(1024)])

    def test_main_loop_stopped(self):
        evaluator = AsyncMock(Evaluator, stop_process=False)

        counter = []

        def test_evaluate_all(*args, **kwargs):
            counter.append(1)
            if len(counter) == 2:
                evaluator.stop_process = True
            elif len(counter) >= 3:
                raise self.fail("Shall not be called!")  # pragma: no cover
            return Result(success=1)

        evaluator.evaluate_all.side_effect = test_evaluate_all

        with patch('verification.commands.evaluate.Evaluator', return_value=evaluator):
            self.assertCommandSuccess(['--loop', '--loop-delay', '0'],
                                      stdout='Evaluate: 1 total, 1 evaluated, 0 skipped, 0 errors.\n' * 2)

        calls = [call.evaluate_all(renew_delay=timedelta(days=30), throttle=10, skip_delay=timedelta(days=1))] * 2
        self.assertEqual(mock_awaits(evaluator), calls)

    def test_output(self):
        data = (
            # verbosity, stdout
            ('0', ''),
            ('1', 'Evaluate: 1 total, 1 evaluated, 0 skipped, 0 errors.\n'),
            ('2', 'Evaluate: 1 total, 1 evaluated, 0 skipped, 0 errors.\n'),
        )
        for verbosity, stdout in data:
            with self.subTest(verbosity=verbosity):
                with patch('verification.commands.evaluate.Evaluator', return_value=TestEvaluator(Result(success=1))):
                    self.assertCommandSuccess(['--verbosity', verbosity], stdout=stdout)
