import tempfile
from functools import partial
from io import StringIO
from typing import TextIO
from unittest import IsolatedAsyncioTestCase, TestCase
from unittest.mock import patch

from sqlalchemy import select

from verification.commands.load_score_queue import _load_score_queue, main
from verification.database import atomic
from verification.models import ScoreQueue
from verification.utils import Result

from .utils import CommandMixin, make_score_queue, override_database, test_session_cls


@override_database()
class LoadScoreQueueTest(IsolatedAsyncioTestCase):
    async def test_load_score_queue_empty(self):
        data = (
            # input, result
            ('', Result()),
            # XXX: In python<=3.9 counters don't ignore 0 items.
            ('\n\n', Result(success=0, skip=0)),
            ('     \n', Result(success=0, skip=0)),
        )
        for input, result in data:
            with self.subTest(input=input):
                buff = StringIO(input)
                self.assertEqual(await _load_score_queue(buff), result)

                # Check no contact loaded
                async with test_session_cls()() as session:
                    self.assertIsNone((await session.execute(select(ScoreQueue))).scalar())

    async def test_load_score_queue(self):
        buff = StringIO('KRYTEN\n')
        # XXX: In python<=3.9 counters don't ignore 0 items.
        self.assertEqual(await _load_score_queue(buff), Result(success=1, skip=0))

        # Check contact loaded
        async with test_session_cls()() as session:
            query = select(ScoreQueue)
            queue = (await session.execute(query)).scalars().one()
            self.assertEqual(queue.id, 'KRYTEN')

    async def test_load_score_queue_existing(self):
        async with atomic() as session:
            await make_score_queue(session, id='KRYTEN')

        buff = StringIO('KRYTEN\n')
        # XXX: In python<=3.9 counters don't ignore 0 items.
        self.assertEqual(await _load_score_queue(buff), Result(success=0, skip=1))

        # Check no new contact loaded
        async with test_session_cls()() as session:
            query = select(ScoreQueue)
            queue = (await session.execute(query)).scalars().one()
            self.assertEqual(queue.id, 'KRYTEN')

    async def test_load_score_queue_batched(self):
        buff = StringIO(''.join(f'KRYTEN{i}\n' for i in range(10)))
        # XXX: In python<=3.9 counters don't ignore 0 items.
        self.assertEqual(await _load_score_queue(buff, batch_size=1), Result(success=10, skip=0))

        # Check contacts loaded
        async with test_session_cls()() as session:
            query = select(ScoreQueue)
            queue = tuple((await session.execute(query)).scalars())
            self.assertEqual(len(queue), 10)


class MainTest(CommandMixin, TestCase):
    command = main

    def test_main(self):
        content = 'RIMMER\nKRYTEN\n'

        async def test_load(input: TextIO) -> Result:
            self.assertEqual(input.read(), content)
            return Result(success=1)

        with tempfile.NamedTemporaryFile(mode='w') as tmpfile:
            tmpfile.write(content)
            tmpfile.flush()

            with patch('verification.commands.load_score_queue._load_score_queue', side_effect=test_load):
                self.assertCommandSuccess(['--file', tmpfile.name])

    def test_main_results(self):
        data = (
            # result, stdout
            (Result(), "Load score queue: 0 total, 0 loaded, 0 skipped.\n"),
            (Result(success=1), "Load score queue: 1 total, 1 loaded, 0 skipped.\n"),
            (Result(skip=1), "Load score queue: 1 total, 0 loaded, 1 skipped.\n"),
        )
        for result, stdout in data:
            with self.subTest(result=result):
                async def test_load(input: TextIO, _result: Result) -> Result:
                    return _result

                with patch('verification.commands.load_score_queue._load_score_queue',
                           side_effect=partial(test_load, _result=result)):
                    self.assertCommandSuccess(stdout=stdout)
