import signal
from typing import Sequence
from unittest import TestCase
from unittest.mock import ANY, AsyncMock, _Call, call, patch

from verification.commands.process import main
from verification.handler import VerificationHandler
from verification.settings import APPLICATION
from verification.utils import Result

from .utils import CommandMixin, mock_awaits


class TestHandler(VerificationHandler):
    """Simple handler to test the `process` command.

    The tests for the `process` command can't async, because the command itself contains `asyncio.run`.
    Thus the handler is overridden to mock the `process_all` parts.
    """

    def __init__(self, result: Result):
        self.result = result
        super().__init__()

    async def process_all(self, closed_limit: int = 0) -> Result:
        return self.result


class MainTest(CommandMixin, TestCase):
    command = main
    contact_id = 'KRYTEN'

    def setUp(self):
        self.sigterm_handler = signal.getsignal(signal.SIGTERM)

    def tearDown(self):
        signal.signal(signal.SIGTERM, self.sigterm_handler)
        APPLICATION.clear_cache()

    def test_main_results(self):
        data = (
            # result, stdout
            (Result(), "Verifications: 0 total, 0 processed, 0 skipped, 0 errors.\n"),
            (Result(success=1), "Verifications: 1 total, 1 processed, 0 skipped, 0 errors.\n"),
            (Result(skip=1), "Verifications: 1 total, 0 processed, 1 skipped, 0 errors.\n"),
            (Result(error=1), "Verifications: 1 total, 0 processed, 0 skipped, 1 errors.\n"),
        )
        for result, stdout in data:
            with self.subTest(result=result):
                APPLICATION.handler = TestHandler(result)
                self.assertCommandSuccess(stdout=stdout)

    def test_main_closed_limit(self):
        handler_mock = AsyncMock(VerificationHandler)
        handler_mock.process_all.return_value = Result(success=1)
        handler_mock.stop_process = False
        APPLICATION.handler = handler_mock

        self.assertCommandSuccess(['--closed-limit', '42'],
                                  stdout="Verifications: 1 total, 1 processed, 0 skipped, 0 errors.\n")

        self.assertEqual(mock_awaits(handler_mock), [call.process_all(closed_limit=42)])

    def _test_main_loop(self, args: Sequence[str], sleep_calls: Sequence[_Call]) -> None:
        # Hack the time.sleep to trigger keyboard interrupt to stop the cycle.
        APPLICATION.handler = TestHandler(Result())
        with patch('verification.commands.process.sleep') as sleep_mock:
            sleep_mock.side_effect = [None, KeyboardInterrupt]

            # Interrupting the command will cause non-zero exit code.
            self.assertCommandFail(args, stdout=ANY)

        # Check sleep was called
        self.assertEqual(sleep_mock.mock_calls, sleep_calls)

    def test_main_loop(self):
        # Test the loop mode.
        self._test_main_loop(['--loop'], [call(300), call(300)])

    def test_main_loop_delay(self):
        self._test_main_loop(['--loop', '--loop-delay', '1024'], [call(1024), call(1024)])

    def test_main_loop_stopped(self):
        handler = AsyncMock(VerificationHandler, stop_process=False)

        counter = []

        def test_process_all(*args, **kwargs):
            counter.append(1)
            if len(counter) == 2:
                handler.stop_process = True
            elif len(counter) >= 3:
                raise self.fail("Shall not be called!")  # pragma: no cover
            return Result(success=1)

        handler.process_all.side_effect = test_process_all

        APPLICATION.handler = handler
        self.assertCommandSuccess(['--loop', '--loop-delay', '0'],
                                  stdout='Verifications: 1 total, 1 processed, 0 skipped, 0 errors.\n' * 2)

        self.assertEqual(mock_awaits(handler), [call.process_all(closed_limit=86400)] * 2)

    def test_output(self):
        data = (
            # verbosity, stdout
            ('0', ''),
            ('1', 'Verifications: 1 total, 1 processed, 0 skipped, 0 errors.\n'),
            ('2', 'Verifications: 1 total, 1 processed, 0 skipped, 0 errors.\n'),
        )
        for verbosity, stdout in data:
            with self.subTest(verbosity=verbosity):
                APPLICATION.handler = TestHandler(Result(success=1))
                self.assertCommandSuccess(['--verbosity', verbosity], stdout=stdout)
