import asyncio
import multiprocessing
import os
import signal
import time
from typing import Callable, Dict, List, Optional, Sequence, cast
from unittest import TestCase
from unittest.mock import ANY, AsyncMock, call, patch, sentinel

from grpc._utilities import DictionaryGenericHandler
from testfixtures import Comparison

from verification.commands.server import main

from .utils import CommandMixin, mock_awaits


class TestProcess:
    """Fake multiprocess.Process class to test multiprocessing tests."""

    def __init__(self, *, target: Callable, args: Sequence):
        self.target = target
        self.args = args

    def start(self) -> None:
        self.target(*self.args)

    def join(self) -> None:
        pass


class ServerMainTest(CommandMixin, TestCase):
    command = main

    def setUp(self):
        # Wrap 'multiprocessing.Process' to run the target inline and count calls.
        process_patcher = patch('multiprocessing.Process', wraps=TestProcess)
        self.addCleanup(process_patcher.stop)
        self.process_mock = process_patcher.start()

        server_mock = AsyncMock()
        grpc_patcher = patch('grpc.aio.server', return_value=server_mock)
        self.addCleanup(grpc_patcher.stop)
        self.grpc_mock = grpc_patcher.start()
        self.grpc_mock.attach_mock(server_mock, '()')
        # Disable reflection
        ref_patcher = patch('verification.commands.server.reflection', new=None)
        self.addCleanup(ref_patcher.stop)
        ref_patcher.start()

        self.sigterm_handler = signal.getsignal(signal.SIGTERM)

    def tearDown(self):
        signal.signal(signal.SIGTERM, self.sigterm_handler)

    def _test_main(self, args: List[str], *, port: int = 50051, options: Optional[Sequence] = None,
                   max_concurrency: Optional[int] = None, jobs: int = 1, settings: Optional[Dict] = None) -> None:
        with patch('multiprocessing.cpu_count', return_value=1):
            with patch('asyncio.sleep', side_effect=asyncio.CancelledError):
                self.assertCommandSuccess(args, settings=settings)

        calls = [call(options=options, maximum_concurrent_rpcs=max_concurrency),
                 call().add_generic_rpc_handlers((Comparison(DictionaryGenericHandler), )),
                 call().add_insecure_port('[::]:{}'.format(port)),
                 call().start(),
                 call().stop(0)] * jobs
        self.assertEqual(self.grpc_mock.mock_calls, calls)
        self.assertEqual(mock_awaits(self.grpc_mock), [call().start(), call().stop(0)] * jobs)
        self.assertEqual(self.process_mock.mock_calls, [call(target=asyncio.run, args=ANY)] * jobs)

    def test_main(self):
        self._test_main([])

    def test_port(self):
        self._test_main(['--port', '12345'], port=12345)

    def test_grpc_options(self):
        self._test_main(['--grpc-options', 'name=Kryten'], options=[['name', 'Kryten']])

    def test_jobs_config(self):
        self._test_main([], settings={'grpc_jobs': 2}, jobs=2)

    def test_jobs_option(self):
        self._test_main(['--jobs', '2'], jobs=2)

    def test_jobs_option_override(self):
        # Test --jobs=0 actually overrides the value from settings.
        self._test_main(['--jobs', '0'], settings={'grpc_jobs': 2}, jobs=1)

    def test_max_concurrency_config(self):
        self._test_main([], settings={'grpc_maximum_concurrent_rpcs': 42}, max_concurrency=42)

    def test_max_concurrency_option(self):
        self._test_main(['--max-concurrency', '42'], max_concurrency=42)

    def test_reflection(self):
        with patch('verification.commands.server.reflection') as reflection_mock:
            reflection_mock.SERVICE_NAME = sentinel.reflection
            self._test_main([])

        ref_call = call.enable_server_reflection(['Fred.Verification.Api.Verification', sentinel.reflection],
                                                 self.grpc_mock.return_value)
        self.assertEqual(reflection_mock.mock_calls, [ref_call])


class ServerMainMultiprocessingTest(CommandMixin, TestCase):
    command = main

    def setUp(self):
        self.sigterm_handler = signal.getsignal(signal.SIGTERM)

        server_mock = AsyncMock()
        grpc_patcher = patch('grpc.aio.server', return_value=server_mock)
        self.addCleanup(grpc_patcher.stop)
        self.grpc_mock = grpc_patcher.start()
        self.grpc_mock.attach_mock(server_mock, '()')
        # Disable reflection
        ref_patcher = patch('verification.commands.server.reflection', new=None)
        self.addCleanup(ref_patcher.stop)
        ref_patcher.start()

    def tearDown(self):
        signal.signal(signal.SIGTERM, self.sigterm_handler)

    def test_main_sigterm(self):
        # Test SIGTERM handling.
        # Start the server in a subprocess
        # XXX: Ignore warnings in stderr, probably happens only in tests.
        server = multiprocessing.Process(target=self.assertCommandSuccess, args=(['--jobs', '3'], ),
                                         kwargs={'stderr': ANY})
        server.start()

        # Wait a bit for a server to start before sending the SIGTERM signal.
        time.sleep(0.1)
        os.kill(cast(int, server.pid), signal.SIGTERM)

        # Wait for the server to finish.
        server.join(2)
        # Check server finished.
        self.assertFalse(server.is_alive())
