import tempfile
from typing import cast
from unittest import TestCase
from unittest.mock import Mock, call

import click
import yaml
from click.testing import CliRunner
from setapp.constants import Verbosity
from setapp.utils import override_environ
from testfixtures import OutputCapture

import verification
from verification.commands.utils import echo, verification_command
from verification.settings import APPLICATION, ENVIRON_NAME


class EchoTest(TestCase):
    def test_output(self):
        message = 'Smeghead!'
        output = 'Smeghead!\n'
        data = (
            # settings, verbosity, result
            (Verbosity.SILENT, Verbosity.SILENT, output),
            (Verbosity.SILENT, Verbosity.INFO, ''),
            (Verbosity.SILENT, Verbosity.DETAIL, ''),
            (Verbosity.SILENT, Verbosity.DEBUG, ''),
            (Verbosity.INFO, Verbosity.SILENT, output),
            (Verbosity.INFO, Verbosity.INFO, output),
            (Verbosity.INFO, Verbosity.DETAIL, ''),
            (Verbosity.INFO, Verbosity.DEBUG, ''),
            (Verbosity.DETAIL, Verbosity.SILENT, output),
            (Verbosity.DETAIL, Verbosity.INFO, output),
            (Verbosity.DETAIL, Verbosity.DETAIL, output),
            (Verbosity.DETAIL, Verbosity.DEBUG, ''),
            (Verbosity.DEBUG, Verbosity.SILENT, output),
            (Verbosity.DEBUG, Verbosity.INFO, output),
            (Verbosity.DEBUG, Verbosity.DETAIL, output),
            (Verbosity.DEBUG, Verbosity.DEBUG, output),
            (1, Verbosity.SILENT, output),
            (1, Verbosity.INFO, output),
            (1, Verbosity.DETAIL, ''),
            (1, Verbosity.DEBUG, ''),
        )

        for settings, verbosity, result in data:
            with self.subTest(settings=settings, verbosity=verbosity):
                APPLICATION.verbosity = cast(Verbosity, settings)

                with OutputCapture(separate=True) as capture:
                    echo(message, verbosity)

                self.assertEqual(capture.stdout.getvalue(), result)
                self.assertEqual(capture.stderr.getvalue(), '')

    def test_error(self):
        APPLICATION.verbosity = Verbosity.DEBUG

        with OutputCapture(separate=True) as capture:
            echo('Smeghead!', Verbosity.DEBUG, err=True)

        self.assertEqual(capture.stdout.getvalue(), '')
        self.assertEqual(capture.stderr.getvalue(), 'Smeghead!\n')

    def test_no_newline(self):
        APPLICATION.verbosity = Verbosity.DEBUG

        with OutputCapture(separate=True) as capture:
            echo('Smeghead!', Verbosity.DEBUG, nl=False)

        self.assertEqual(capture.stdout.getvalue(), 'Smeghead!')
        self.assertEqual(capture.stderr.getvalue(), '')

    def test_log(self):
        # Test message is logged, regardless of the verbosity setting.
        message = 'Smeghead!'

        for settings in (Verbosity.DEBUG, Verbosity.SILENT):
            log_mock = Mock()
            with OutputCapture(separate=True):
                APPLICATION.verbosity = settings
                echo(message, Verbosity.DEBUG, log=log_mock)

            self.assertEqual(log_mock.mock_calls, [call(message)])


@verification_command()
def test_command():
    pass


class VerificationCommandTest(TestCase):
    def setUp(self):
        self.runner = CliRunner(env={ENVIRON_NAME: None})

        # Delete settings for tests to check loading.
        self.old_settings = APPLICATION.settings
        APPLICATION._settings = None

    def tearDown(self):
        APPLICATION.settings = self.old_settings

    def test_help(self):
        result = self.runner.invoke(test_command, ['--help'])

        self.assertEqual(result.exit_code, 0)
        self.assertIn('--help', result.stdout)

    def test_version(self):
        result = self.runner.invoke(test_command, ['--version'])

        self.assertEqual(result.exit_code, 0)
        self.assertIn('fred-verification ' + verification.__version__, result.stdout)

    def test_config(self):
        db_connection = 'file://example/connection'
        config_file = tempfile.NamedTemporaryFile(mode='w')
        yaml.dump({'db_connection': db_connection}, config_file)

        with override_environ():
            result = self.runner.invoke(test_command, ['--config', config_file.name])

        self.assertEqual(APPLICATION.settings.db_connection, db_connection)
        self.assertEqual(result.exit_code, 0)

    def test_connection(self):
        db_connection = 'args://example/connection'

        result = self.runner.invoke(test_command, ['--connection', db_connection])

        self.assertEqual(APPLICATION.settings.db_connection, db_connection)
        self.assertEqual(result.exit_code, 0)

    def test_verbosity_long(self):
        result = self.runner.invoke(test_command, ['--verbosity', '3'])

        self.assertEqual(APPLICATION.verbosity, Verbosity.DEBUG)
        self.assertEqual(result.exit_code, 0)

    def test_verbosity_short(self):
        result = self.runner.invoke(test_command, ['-v', '3'])

        self.assertEqual(APPLICATION.verbosity, Verbosity.DEBUG)
        self.assertEqual(result.exit_code, 0)

    def test_group(self):
        # Test command can be set as a member of a click group.
        @click.group()
        def group():
            pass

        @verification_command(command=group.command)
        def test_command():
            click.echo('Subgroup!')

        result = self.runner.invoke(group, ['test-command'])

        self.assertEqual(result.exit_code, 0)
        # In py36-postgres the output contains a spurious ResourceWarning.
        # Thus, we use assertIn instead of assertEqual to avoid that.
        self.assertIn('Subgroup!\n', result.stdout)
