from tempfile import NamedTemporaryFile
from unittest import IsolatedAsyncioTestCase
from unittest.mock import call, patch

from sqlalchemy import select
from sqlalchemy.pool import NullPool

from verification.constants import VerificationStatus
from verification.database import atomic, get_engine, get_session_cls, object_session
from verification.models import Verification

from .utils import override_database, override_settings, test_session_cls


class GetEngineTest(IsolatedAsyncioTestCase):
    def setUp(self):
        # File is deleted by the override_settings cleanup.
        self.tmp_db = NamedTemporaryFile(delete=False)

    async def test_result(self):
        async with override_database(db_connection='sqlite+aiosqlite:///{}'.format(self.tmp_db.name)):
            engine = get_engine()

        self.assertEqual(str(engine.url), 'sqlite+aiosqlite:///{}'.format(self.tmp_db.name))

    async def test_cached(self):
        async with override_database(db_connection='sqlite+aiosqlite:///{}'.format(self.tmp_db.name)):
            engine1 = get_engine()
            engine2 = get_engine()

        self.assertIs(engine1, engine2)

    async def test_poolclass(self):
        with override_settings():
            with patch('verification.database.create_async_engine') as create_engine_mock:
                with patch('verification.database.APPLICATION') as application_mock:
                    application_mock.settings.db_poolclass = 'sqlalchemy.pool.NullPool'
                    get_engine()

        engine_calls = [
            call(application_mock.settings.db_connection, echo=application_mock.settings.db_echo, poolclass=NullPool),
            call().update_execution_options(schema_translate_map={None: application_mock.settings.db_schema})]
        self.assertEqual(create_engine_mock.mock_calls, engine_calls)
        self.assertEqual(application_mock.mock_calls, [])

    async def test_schema(self):
        with override_settings():
            with patch('verification.database.create_async_engine') as create_engine_mock:
                with patch('verification.database.APPLICATION') as application_mock:
                    application_mock.settings.db_poolclass = None
                    get_engine()

        engine_calls = [
            call(application_mock.settings.db_connection, echo=application_mock.settings.db_echo, poolclass=None),
            call().update_execution_options(schema_translate_map={None: application_mock.settings.db_schema})]
        self.assertEqual(create_engine_mock.mock_calls, engine_calls)
        self.assertEqual(application_mock.mock_calls, [])


class GetSessionClsTest(IsolatedAsyncioTestCase):
    def setUp(self):
        # File is deleted by the override_settings cleanup.
        self.tmp_db = NamedTemporaryFile(delete=False)

    async def test_result(self):
        async with override_database(db_connection='sqlite+aiosqlite:///{}'.format(self.tmp_db.name)):
            session_cls = get_session_cls()

        self.assertEqual(str(session_cls.kw['bind'].url), 'sqlite+aiosqlite:///{}'.format(self.tmp_db.name))

    async def test_cached(self):
        async with override_database(db_connection='sqlite+aiosqlite:///{}'.format(self.tmp_db.name)):
            cls1 = get_session_cls()
            cls2 = get_session_cls()

        self.assertIs(cls1, cls2)


@override_database()
class AtomicTest(IsolatedAsyncioTestCase):
    async def test_commit(self):
        async with atomic() as session:
            obj = Verification(contact_id='KRYTEN', status=VerificationStatus.CREATED)
            session.add(obj)

        async with test_session_cls()() as session:
            verification = (await session.execute(select(Verification))).scalars().one()
            self.assertEqual(verification.contact_id, 'KRYTEN')

    async def test_rollback(self):
        with self.assertRaisesRegex(Exception, 'Gazpacho!'):
            async with atomic() as session:
                obj = Verification(contact_id='KRYTEN', status=VerificationStatus.CREATED)
                session.add(obj)
                raise Exception('Gazpacho!')

        async with test_session_cls()() as session:
            verification = (await session.execute(select(Verification))).scalar()
            self.assertIsNone(verification)


@override_database()
class ObjectSessionTest(IsolatedAsyncioTestCase):
    async def test_session(self):
        async with test_session_cls()() as session:
            obj = Verification(contact_id='KRYTEN', status=VerificationStatus.CREATED)
            session.add(obj)

            session_get = object_session(obj)

            self.assertEqual(session_get, session)

    def test_no_session(self):
        obj = Verification()

        with self.assertRaisesRegex(ValueError, "is not bound to a session"):
            object_session(obj)
