import asyncio
import signal
from datetime import datetime, timedelta
from typing import Any, Dict, Sequence, Tuple, cast
from unittest import IsolatedAsyncioTestCase, skipIf
from unittest.mock import AsyncMock, call, patch, sentinel

from fred_types import ContactHistoryId, ContactId
from freezegun import freeze_time
from pytz import UTC
from regal import Contact, ContactClient
from regal.exceptions import ContactDoesNotExist
from sqlalchemy import select
from testfixtures import Comparison, LogCapture, StringComparison

from verification.checks import BaseCheck
from verification.database import atomic
from verification.evaluator import Evaluator
from verification.exceptions import SkipEvaluation
from verification.models import CheckResult, Score, ScoreQueue
from verification.settings import APPLICATION
from verification.utils import Result

from .utils import (
    TransactionTestCase,
    make_score,
    make_score_queue,
    mock_awaits,
    override_database,
    override_settings,
    test_session_cls,
)


class TestCheck(BaseCheck):
    """Simple check for testing."""

    def __init__(self, *, name: str, value: float = 0.07):
        super().__init__(name=name)
        self.value = value

    async def _get_results(self, contact: Contact) -> Tuple[Dict[str, Any], float]:
        return {'value': self.value}, self.value


class TestEvaluator(Evaluator):
    """Configurable evaluator for tests."""

    def __init__(self, checks: Sequence[BaseCheck]):
        super().__init__()
        self._checks = checks

    @property
    def checks(self) -> Sequence[BaseCheck]:
        return self._checks


class EvaluatorTest(TransactionTestCase):
    contact_id = ContactId('2X4B')
    handle = 'KRYTEN'
    contact_history_id = ContactHistoryId('523P')

    async def asyncSetUp(self):
        APPLICATION.contact_client = AsyncMock(ContactClient)
        APPLICATION.contact_client.get_contact_info.return_value = Contact(
            contact_id=self.contact_id, contact_handle=self.handle, contact_history_id=self.contact_history_id)
        self.log_handler = LogCapture('verification', propagate=False)

    async def asyncTearDown(self):
        self.log_handler.uninstall()

    async def test_checks_empty(self):
        evaluator = Evaluator()

        with override_settings(checks={}):
            self.assertEqual(evaluator.checks, ())

    async def test_checks_no_class(self):
        evaluator = Evaluator()

        with override_settings(checks={'psirens': {}}):
            with self.assertRaisesRegex(RuntimeError, 'Check psirens has no class defined.'):
                evaluator.checks

    async def test_checks_invalid(self):
        evaluator = Evaluator()

        settings = {'psirens': {'class': 'verification.tests.test_evaluator.InvalidCheck'}}
        with override_settings(checks=settings):
            with self.assertRaisesRegex(RuntimeError, "Error in loading check:"):
                evaluator.checks

    async def test_checks_args(self):
        evaluator = Evaluator()

        settings = {'psirens': {'class': 'verification.tests.test_evaluator.TestCheck', 'value': 0.42}}
        checks = (Comparison(TestCheck, name='psirens', value=0.42), )
        with override_settings(checks=settings):
            self.assertEqual(evaluator.checks, checks)

    def test_rms(self):
        data = (
            # values, rms
            ((0.0, ), 0.0),
            ((0.5, ), 0.5),
            ((1.0, ), 1.0),
            ((0.0, 0.5), 0.35355),
            ((1.0, 0.5), 0.79057),
        )
        evaluator = TestEvaluator(())
        for values, rms in data:
            self.assertAlmostEqual(evaluator.rms(values), rms, places=5)

    async def test_evaluate_no_checks(self):
        queue = await make_score_queue(self.session, id=self.contact_id)
        evaluator = TestEvaluator(())

        self.assertFalse(await evaluator.evaluate(queue))

        # Check no score was stored.
        self.assertIsNone((await self.session.execute(select(Score))).scalar())
        self.assertIsNone((await self.session.execute(select(CheckResult))).scalar())

    async def test_evaluate(self):
        queue = await make_score_queue(self.session, id=self.contact_id)
        evaluator = TestEvaluator((TestCheck(name='psirens', value=0.42), ))

        self.assertTrue(await evaluator.evaluate(queue))

        # Check score was stored.
        score = (await self.session.execute(select(Score))).scalars().one()
        self.assertEqual(score.contact_id, self.contact_id)
        self.assertEqual(score.contact_history_id, self.contact_history_id)
        self.assertEqual(score.value, 0.42)
        result = (await self.session.execute(select(CheckResult))).scalars().one()
        self.assertEqual(result.name, 'psirens')
        self.assertEqual(result.value, 0.42)
        self.assertEqual(result.details, {'value': 0.42})

    async def test_evaluate_skip(self):
        queue = await make_score_queue(self.session, id=self.contact_id)
        check = AsyncMock(BaseCheck)
        check.evaluate.side_effect = SkipEvaluation
        evaluator = TestEvaluator([check])

        self.assertFalse(await evaluator.evaluate(queue))

        # Check queue skipped_at was updated.
        queue = (await self.session.execute(select(ScoreQueue))).scalars().one()
        self.assertEqual(queue.id, self.contact_id)
        self.assertIsNotNone(queue.skipped_at)
        # Check no score was stored.
        self.assertIsNone((await self.session.execute(select(Score))).scalar())
        self.assertIsNone((await self.session.execute(select(CheckResult))).scalar())

    async def test_evaluate_contact_does_not_exist(self):
        # Test contact is deleted from queue if it no longer exist.
        cast(AsyncMock, APPLICATION.contact_client).get_contact_info.side_effect = ContactDoesNotExist
        queue = await make_score_queue(self.session, id=self.contact_id)
        evaluator = TestEvaluator((TestCheck(name='psirens', value=0.42), ))

        self.assertTrue(await evaluator.evaluate(queue))

        # Check queue was deleted.
        self.assertIsNone((await self.session.execute(select(ScoreQueue))).scalar())
        # Check no score was stored.
        self.assertIsNone((await self.session.execute(select(Score))).scalar())
        self.assertIsNone((await self.session.execute(select(CheckResult))).scalar())

    async def test_evaluate_error(self):
        queue = await make_score_queue(self.session, id=self.contact_id)
        check = AsyncMock(BaseCheck)
        check.evaluate.side_effect = Exception('Gazpacho!')
        evaluator = TestEvaluator([check])

        with self.assertRaisesRegex(Exception, 'Gazpacho!'):
            await evaluator.evaluate(queue)

        # Check failure gets noted.
        self.assertEqual(evaluator._failures, {self.contact_id: 1})
        # Check queue is unchanged.
        queue = (await self.session.execute(select(ScoreQueue))).scalars().one()
        self.assertEqual(queue.id, self.contact_id)
        self.assertIsNone(queue.skipped_at)
        # Check no score was stored.
        self.assertIsNone((await self.session.execute(select(Score))).scalar())
        self.assertIsNone((await self.session.execute(select(CheckResult))).scalar())

    async def test_evaluate_error_limit_reached(self):
        queue = await make_score_queue(self.session, id=self.contact_id)
        check = AsyncMock(BaseCheck)
        check.evaluate.side_effect = Exception('Gazpacho!')
        evaluator = TestEvaluator([check])
        evaluator._failures[self.contact_id] = 9

        self.assertTrue(await evaluator.evaluate(queue))

        # Check queue is unchanged.
        queue = (await self.session.execute(select(ScoreQueue))).scalars().one()
        self.assertEqual(queue.id, self.contact_id)
        self.assertIsNone(queue.skipped_at)
        # Check score with 0 value have been stored.
        score = (await self.session.execute(select(Score))).scalars().one()
        self.assertEqual(score.contact_id, self.contact_id)
        self.assertEqual(score.contact_history_id, self.contact_history_id)
        self.assertEqual(score.value, 0)
        # No check result have been stored though.
        self.assertIsNone((await self.session.execute(select(CheckResult))).scalar())


@override_database()
class EvaluatorEvaluateAllTest(IsolatedAsyncioTestCase):
    contact_id = ContactId('2X4B')
    handle = 'KRYTEN'
    contact_history_id = ContactHistoryId('523P')

    async def asyncSetUp(self):
        APPLICATION.contact_client = AsyncMock(ContactClient)
        APPLICATION.contact_client.get_contact_info.return_value = Contact(
            contact_id=self.contact_id, contact_handle=self.handle, contact_history_id=self.contact_history_id)
        self.log_handler = LogCapture('verification', propagate=False)
        self.sigterm_handler = signal.getsignal(signal.SIGTERM)

    async def asyncTearDown(self):
        signal.signal(signal.SIGTERM, self.sigterm_handler)
        self.log_handler.uninstall()

    async def test_evaluate_all_no_queue(self):
        evaluator = TestEvaluator([])

        self.assertEqual(await evaluator.evaluate_all(), Result())

        # Check score wasn't evaluated.
        async with test_session_cls()() as session:
            self.assertIsNone((await session.execute(select(ScoreQueue))).scalar())
            self.assertIsNone((await session.execute(select(Score))).scalar())
            self.assertIsNone((await session.execute(select(CheckResult))).scalar())

    async def test_evaluate_all_no_checks(self):
        async with atomic() as session:
            await make_score_queue(session, id=self.contact_id)
        evaluator = TestEvaluator([])

        self.assertEqual(await evaluator.evaluate_all(), Result(skip=1))

        # Check score wasn't evaluated.
        async with test_session_cls()() as session:
            queue = (await session.execute(select(ScoreQueue))).scalars().one()
            self.assertEqual(queue.id, self.contact_id)
            self.assertIsNone(queue.skipped_at)
            self.assertIsNone((await session.execute(select(Score))).scalar())
            self.assertIsNone((await session.execute(select(CheckResult))).scalar())

    async def test_evaluate_all(self):
        async with atomic() as session:
            await make_score_queue(session, id=self.contact_id)
        evaluator = TestEvaluator([TestCheck(name='psirens', value=0.42)])

        self.assertEqual(await evaluator.evaluate_all(), Result(success=1))

        # Check contact was evaluated.
        async with test_session_cls()() as session:
            queue = (await session.execute(select(ScoreQueue))).scalars().one()
            self.assertEqual(queue.id, self.contact_id)
            self.assertIsNone(queue.skipped_at)
            score = (await session.execute(select(Score))).scalars().one()
            self.assertEqual(score.contact_id, self.contact_id)
            self.assertEqual(score.value, 0.42)
            result = (await session.execute(select(CheckResult))).scalars().one()
            self.assertEqual(result.name, 'psirens')

    async def test_evaluate_all_recent(self):
        # Test recently evaluated contact is skipped.
        async with atomic() as session:
            await make_score_queue(session, id=self.contact_id)
            await make_score(session, contact_id=self.contact_id, contact_history_id=self.contact_history_id,
                             create_datetime=datetime(1988, 2, 15, 21, 0, tzinfo=UTC), value=0.85)
        evaluator = TestEvaluator([TestCheck(name='psirens', value=0.42)])

        with freeze_time('1988-02-15T21:05:00'):
            self.assertEqual(await evaluator.evaluate_all(renew_delay=timedelta(hours=1)), Result())

        # Check score wasn't evaluated.
        async with test_session_cls()() as session:
            queue = (await session.execute(select(ScoreQueue))).scalars().one()
            self.assertEqual(queue.id, self.contact_id)
            self.assertIsNone(queue.skipped_at)
            self.assertEqual(len(tuple((await session.execute(select(Score))).scalars())), 1)
            self.assertIsNone((await session.execute(select(CheckResult))).scalar())

    async def test_evaluate_all_old(self):
        # Test not recently evaluated contact is evaluated again.
        async with atomic() as session:
            await make_score_queue(session, id=self.contact_id)
            await make_score(session, contact_id=self.contact_id, contact_history_id=self.contact_history_id,
                             create_datetime=datetime(1988, 2, 15, 21, 0, tzinfo=UTC), value=0.85)
        evaluator = TestEvaluator([TestCheck(name='psirens', value=0.42)])

        with freeze_time('1988-02-15T22:15:00'):
            self.assertEqual(await evaluator.evaluate_all(renew_delay=timedelta(hours=1)), Result(success=1))

        # Check score was evaluated.
        async with test_session_cls()() as session:
            queue = (await session.execute(select(ScoreQueue))).scalars().one()
            self.assertEqual(queue.id, self.contact_id)
            self.assertIsNone(queue.skipped_at)
            self.assertEqual(len(tuple((await session.execute(select(Score))).scalars())), 2)
            score = cast(Score, (await session.execute(select(Score).order_by(Score.id.desc()))).scalars().first())
            self.assertEqual(score.contact_id, self.contact_id)
            self.assertEqual(score.value, 0.42)
            result = (await session.execute(select(CheckResult))).scalars().one()
            self.assertEqual(result.name, 'psirens')

    async def test_evaluate_all_old_ensure(self):
        # Test last score datetime is checked again within evaluation cycle.
        async with atomic() as session:
            await make_score_queue(session, id=self.contact_id)
            await make_score(session, contact_id=self.contact_id, contact_history_id=self.contact_history_id,
                             create_datetime=datetime(1988, 2, 15, 21, 0, tzinfo=UTC), value=0.85)
        evaluator = TestEvaluator([TestCheck(name='psirens', value=0.42)])

        with patch('verification.evaluator.datetime') as datetime_mock:
            # XXX: It's uneasy create a new score because of async, instead move the clock back.
            datetime_mock.now.side_effect = [datetime(1988, 2, 15, 23, 0, tzinfo=UTC),  # skip
                                             datetime(1988, 2, 15, 23, 0, tzinfo=UTC),  # renew
                                             datetime(1988, 2, 15, 23, 0, tzinfo=UTC),  # skip
                                             datetime(1988, 2, 15, 21, 0, tzinfo=UTC)]  # renew
            self.assertEqual(await evaluator.evaluate_all(renew_delay=timedelta(hours=1)), Result())

        # Check score wasn't evaluated.
        async with test_session_cls()() as session:
            queue = (await session.execute(select(ScoreQueue))).scalars().one()
            self.assertEqual(queue.id, self.contact_id)
            self.assertIsNone(queue.skipped_at)
            self.assertEqual(len(tuple((await session.execute(select(Score))).scalars())), 1)
            self.assertIsNone((await session.execute(select(CheckResult))).scalar())

    async def test_evaluate_all_skip_delay(self):
        # Test recently skipped contact is skipped.
        async with atomic() as session:
            await make_score_queue(session, id=self.contact_id, skipped_at=datetime(1988, 2, 15, 21, 0, tzinfo=UTC))
        evaluator = TestEvaluator([TestCheck(name='psirens', value=0.42)])

        with freeze_time('1988-02-15T21:05:00'):
            self.assertEqual(await evaluator.evaluate_all(skip_delay=timedelta(hours=1)), Result())

        # Check score wasn't evaluated.
        async with test_session_cls()() as session:
            queue = (await session.execute(select(ScoreQueue))).scalars().one()
            self.assertEqual(queue.id, self.contact_id)
            self.assertEqual(queue.skipped_at, datetime(1988, 2, 15, 21, 0, tzinfo=UTC))
            self.assertIsNone((await session.execute(select(Score))).scalar())
            self.assertIsNone((await session.execute(select(CheckResult))).scalar())

    async def test_evaluate_all_skip_delay_ensure(self):
        # Test recently skipped contact is skipped.
        async with atomic() as session:
            await make_score_queue(session, id=self.contact_id, skipped_at=datetime(1988, 2, 15, 21, 0, tzinfo=UTC))
        evaluator = TestEvaluator([TestCheck(name='psirens', value=0.42)])

        with patch('verification.evaluator.datetime') as datetime_mock:
            # XXX: It's uneasy create a new score because of async, instead move the clock back.
            datetime_mock.now.side_effect = [datetime(1988, 2, 15, 23, 0, tzinfo=UTC),  # skip
                                             datetime(1988, 2, 15, 23, 0, tzinfo=UTC),  # renew
                                             datetime(1988, 2, 15, 21, 0, tzinfo=UTC),  # skip
                                             datetime(1988, 2, 15, 23, 0, tzinfo=UTC)]  # renew
            self.assertEqual(await evaluator.evaluate_all(skip_delay=timedelta(hours=1)), Result())

        # Check score wasn't evaluated.
        async with test_session_cls()() as session:
            queue = (await session.execute(select(ScoreQueue))).scalars().one()
            self.assertEqual(queue.id, self.contact_id)
            self.assertEqual(queue.skipped_at, datetime(1988, 2, 15, 21, 0, tzinfo=UTC))
            self.assertIsNone((await session.execute(select(Score))).scalar())
            self.assertIsNone((await session.execute(select(CheckResult))).scalar())

    async def test_evaluate_all_error(self):
        # Test an error occurs during evaluation.
        async with atomic() as session:
            await make_score_queue(session, id=self.contact_id)
        check = AsyncMock(BaseCheck)
        check.evaluate.side_effect = Exception('Gazpacho!')
        evaluator = TestEvaluator([check])

        self.assertEqual(await evaluator.evaluate_all(), Result(error=1))

        # Check score wasn't evaluated.
        async with test_session_cls()() as session:
            queue = (await session.execute(select(ScoreQueue))).scalars().one()
            self.assertIsNone(queue.skipped_at)
            self.assertEqual(queue.id, self.contact_id)
            self.assertIsNone((await session.execute(select(Score))).scalar())
            self.assertIsNone((await session.execute(select(CheckResult))).scalar())
        # Check logs
        msg = StringComparison(r"An error in evaluation '.*': .*Exception.*Gazpacho!.*")
        self.log_handler.check_present(('verification.evaluator', 'ERROR', msg))

    @skipIf(APPLICATION.settings.test_db_connection.startswith('sqlite'), 'Does not work on SQLite database.')
    async def test_evaluate_all_order(self):
        # Test ordering of evaluations.
        async with atomic() as session:
            # Evaluated recently
            await make_score_queue(session, id=self.contact_id)
            await make_score(session, contact_id=self.contact_id, contact_history_id=self.contact_history_id,
                             create_datetime=datetime(1988, 2, 15, 21, 0, tzinfo=UTC), value=0.85)
            # Not evaluated recently
            await make_score_queue(session, id='523P')
            await make_score(session, contact_id='523P', contact_history_id=self.contact_history_id,
                             create_datetime=datetime(1988, 2, 14, 21, 0, tzinfo=UTC), value=0.85)
            # One more old score
            await make_score_queue(session, id='3X4B')
            await make_score(session, contact_id='3X4B', contact_history_id=self.contact_history_id,
                             create_datetime=datetime(1988, 2, 13, 21, 0, tzinfo=UTC), value=0.85)
            # Never evaluated.
            await make_score_queue(session, id='4X4B')

        class TestOrderEvaluator(Evaluator):
            def __init__(self):
                super().__init__()
                self.ids = []

            async def evaluate(self, queue: ScoreQueue) -> bool:
                # Do nothing, just record the id.
                self.ids.append(queue.id)
                return True

        evaluator = TestOrderEvaluator()

        self.assertEqual(await evaluator.evaluate_all(), Result(success=4))

        # Check order of evaluation
        self.assertEqual(evaluator.ids, ['4X4B', '3X4B', '523P', self.contact_id])

    @skipIf(APPLICATION.settings.test_db_connection.startswith('sqlite'), 'Does not work on SQLite database.')
    async def test_evaluate_all_skip_locked(self):
        # Test locked rows are skipped.
        async with atomic() as session:
            await make_score_queue(session, id=self.contact_id)
        evaluator = TestEvaluator([TestCheck(name='psirens', value=0.42)])

        # Lock the verification in test transaction.
        session = test_session_cls()()
        query = select(ScoreQueue).with_for_update()
        await session.execute(query)

        self.assertEqual(await evaluator.evaluate_all(), Result())

        # Release the lock.
        await session.rollback()

        # Check score wasn't evaluated.
        async with test_session_cls()() as session:
            self.assertIsNone((await session.execute(select(Score))).scalar())
            self.assertIsNone((await session.execute(select(CheckResult))).scalar())

    async def test_evaluate_all_graceful_stop(self):
        # Test graceful stop.
        # Steps:
        # 1. Start evaluation of the contact.
        step_1 = asyncio.Event()
        # 2. Send SIGTERM signal.
        step_2 = asyncio.Event()
        # 3. Continue evaluation of the contact.
        # 4. End evaluation cycle.
        async with atomic() as session:
            await make_score_queue(session, id=self.contact_id)
            await make_score_queue(session, id='523P')

        # Custom evaluate on the first event to trigger events in correct order.
        async def evaluate(contact: Contact) -> CheckResult:
            step_1.set()
            await asyncio.wait_for(step_2.wait(), timeout=3)
            return CheckResult(name='psirens', details={'value': 0.42}, value=0.42)

        check = AsyncMock(BaseCheck)
        check.evaluate.side_effect = evaluate
        evaluator = TestEvaluator([check])

        async def send_signal():
            await asyncio.wait_for(step_1.wait(), timeout=3)
            step_2.set()
            signal.raise_signal(signal.SIGTERM)

        results = await asyncio.gather(evaluator.evaluate_all(), send_signal())
        self.assertEqual(results[0], Result(success=1))

        # Check only the first score is evaluated.
        async with test_session_cls()() as session:
            score = (await session.execute(select(Score))).scalars().one()
            self.assertEqual(score.contact_id, self.contact_id)
            self.assertEqual(score.value, 0.42)
            result = (await session.execute(select(CheckResult))).scalars().one()
            self.assertEqual(result.name, 'psirens')

    async def test_evaluate_throttle(self):
        async with atomic() as session:
            await make_score_queue(session, id=self.contact_id)
            await make_score_queue(session, id='523P')
        evaluator = TestEvaluator([TestCheck(name='psirens', value=0.42)])

        with patch('verification.evaluator.sleep') as sleep_mock:
            self.assertEqual(await evaluator.evaluate_all(throttle=sentinel.throttle), Result(success=2))

        self.assertEqual(mock_awaits(sleep_mock), [call(sentinel.throttle), call(sentinel.throttle)])

        # Check contacts were evaluated.
        async with test_session_cls()() as session:
            self.assertEqual(len(tuple((await session.execute(select(Score))).scalars())), 2)
