import asyncio
import signal
from datetime import datetime
from functools import partial
from typing import Any, AsyncIterator, Callable, Dict, Iterable, List, Optional, Sequence, cast
from unittest import IsolatedAsyncioTestCase, skipIf
from unittest.mock import ANY, AsyncMock, Mock, _Call, call, sentinel

from asyncstdlib import any_iter, iter as aiter, tuple as atuple
from asyncstdlib._typing import AnyIterable
from fred_types import ContactId, ContactRef, DomainId, DomainRef
from freezegun import freeze_time
from pytz import UTC
from regal import Contact, ContactClient, Domain as _Domain, DomainAdminClient, DomainClient, DomainContactRole
from regal.exceptions import ContactDoesNotExist, DomainDoesNotExist
from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncSession
from testfixtures import Comparison, LogCapture, StringComparison

from verification.constants import EventType, VerificationStatus
from verification.database import atomic
from verification.exceptions import VerificationAlreadyExists
from verification.handler import VerificationHandler
from verification.models import Domain, Event, Verification
from verification.processors.base import BaseProcessor, ScheduledEvent
from verification.settings import APPLICATION
from verification.utils import Result

from .utils import (
    TransactionTestCase,
    make_domain,
    make_event,
    make_verification,
    mock_awaits,
    override_database,
    override_settings,
    test_session_cls,
)


class TestEvent(ScheduledEvent):
    """Simple event which switches the verification to canceled status."""

    type = EventType.CANCEL

    def __init__(self, verification: Verification, *, status: VerificationStatus = VerificationStatus.CANCELED,
                 **kwargs: Any):
        super().__init__(verification, **kwargs)
        self.status = status

    async def process(self) -> None:
        self.verification.status = self.status


class TestErrorEvent(ScheduledEvent):
    """Simple test event which raises error on processing.

    Designed to test events that shouldn't be processed.
    """

    type = 'error'  # type: ignore[assignment]

    def __init__(self, verification: Verification, **kwargs: Any):
        super().__init__(verification, **kwargs)  # pragma: no cover

    async def process(self) -> None:
        raise RuntimeError("YOU SHALL NOT PROCESS!")  # pragma: no cover


class TestImportProcessor(BaseProcessor):
    """Simple processor for testing VerificationHandler.processors."""

    def __init__(self, *, task: str = 'Play the guitar', **kwargs: Any):
        super().__init__(**kwargs)
        self.task = task


def noop_callback(verification: Verification) -> Iterable[ScheduledEvent]:
    return ()  # pragma: no cover


class TestProcessor(BaseProcessor):
    """Simple processor for testing VerificationHandler."""

    def __init__(self, next_callback: Callable[[Verification], AnyIterable[ScheduledEvent]] = noop_callback,
                 cancel_callback: Callable[[Verification], AnyIterable[ScheduledEvent]] = noop_callback,
                 complete_callback: Callable[[Verification], AnyIterable[ScheduledEvent]] = noop_callback,
                 complete: bool = False, **kwargs: Any):
        super().__init__(**kwargs)
        self.next_callback = next_callback
        self.cancel_callback = cancel_callback
        self.complete_callback = complete_callback
        self.complete = complete

    async def get_next_events(self, verification: Verification) -> AsyncIterator[ScheduledEvent]:
        async for scheduled in any_iter(self.next_callback(verification)):
            yield scheduled

    async def get_cancel_events(self, verification: Verification, status: Optional[VerificationStatus] = None,
                                ) -> AsyncIterator[ScheduledEvent]:
        async for scheduled in any_iter(self.cancel_callback(verification)):
            yield scheduled

    async def get_complete_events(self, verification: Verification) -> AsyncIterator[ScheduledEvent]:
        async for scheduled in any_iter(self.complete_callback(verification)):
            yield scheduled

    async def check_complete(self, verification: Verification) -> bool:
        return self.complete


class TestEventsProcessor(BaseProcessor):
    """Simple processor which returns defined events."""

    def __init__(self, *, next_events: Optional[Iterable[ScheduledEvent]] = None,
                 cancel_events: Optional[Iterable[ScheduledEvent]] = None, **kwargs: Any):
        super().__init__(**kwargs)
        self.next_events = next_events
        self.cancel_events = cancel_events

    async def get_next_events(self, verification: Verification) -> AsyncIterator[ScheduledEvent]:
        for scheduled in self.next_events or []:
            yield scheduled

    async def get_cancel_events(self, verification: Verification, status: Optional[VerificationStatus] = None,
                                ) -> AsyncIterator[ScheduledEvent]:
        for scheduled in self.cancel_events or []:
            yield scheduled


def _get_processor(statuses: List[VerificationStatus], due_datetime: Optional[datetime] = None) -> TestProcessor:
    """Return processor which takes list of statuses and returns one event less on each call."""
    def get_next_events(verification: Verification) -> Iterable[ScheduledEvent]:
        scheduled = [TestEvent(verification, status=s, due_datetime=due_datetime, method='no-psirens')
                     for s in statuses]
        if statuses:
            statuses.pop(0)
        yield from scheduled

    return TestProcessor(next_callback=get_next_events, method='no-psirens')


class TestHandler(VerificationHandler):
    def __init__(self, processors: Iterable[BaseProcessor]):
        self._processors = {p.method: p for p in processors}
        super().__init__()

    @property
    def processors(self) -> Dict[str, BaseProcessor]:
        return self._processors


class TestCheckDomainHandler(VerificationHandler):
    """Simple handler for check domain tests."""

    def __init__(self):
        self.create_mock = Mock()

    async def create(self, session: AsyncSession, contact_id: ContactId, *, reason: str = '', reason_url: str = '',
                     ) -> Verification:
        # Do nothing, just record the call.
        self.create_mock(session, contact_id, reason=reason, reason_url=reason_url)
        return sentinel.verification  # type: ignore[no-any-return]


class VerificationHandlerTest(TransactionTestCase):
    contact_id = ContactId('2X4B')
    contact_handle = 'KRYTEN'
    new_contact_id = ContactId('ACE')
    new_contact_handle = 'RIMMER'

    async def asyncSetUp(self):
        APPLICATION.contact_client = AsyncMock(ContactClient)
        APPLICATION.domain_client = AsyncMock(DomainClient)
        APPLICATION.domain_client.get_domains_by_contact.return_value = {'domains': []}
        APPLICATION.domain_admin_client = AsyncMock(DomainAdminClient)
        APPLICATION.domain_admin_client.manage_domain_state_flags.return_value = []
        self.log_handler = LogCapture('verification', propagate=False)

    async def asyncTearDown(self):
        self.log_handler.uninstall()

    async def test_processors_empty(self):
        handler = VerificationHandler()

        with override_settings(processors={}):
            self.assertEqual(handler.processors, {})

    async def test_processors_no_class(self):
        handler = VerificationHandler()

        settings: Dict[str, Any] = {'no-psirens': {}}
        with override_settings(processors=settings):
            with self.assertRaisesRegex(RuntimeError, 'Processor no-psirens has no class defined.'):
                handler.processors

    async def test_processors_invalid(self):
        handler = VerificationHandler()

        settings = {'no-psirens': {'class': 'verification.tests.test_handler.InvalidProcessor'}}
        with override_settings(processors=settings):
            with self.assertRaisesRegex(RuntimeError, "Error in loading processor:"):
                handler.processors

    async def test_processors_no_args(self):
        handler = VerificationHandler()

        settings = {'no-psirens': {'class': 'verification.tests.test_handler.TestImportProcessor'}}
        processors = {'no-psirens': Comparison(TestImportProcessor, method='no-psirens', task='Play the guitar',
                                               partial=True)}
        with override_settings(processors=settings):
            self.assertEqual(handler.processors, processors)

    async def test_processors_args(self):
        handler = VerificationHandler()

        settings = {'no-psirens': {'class': 'verification.tests.test_handler.TestImportProcessor',
                                   'task': 'Trim your toe nails'}}
        processors = {'no-psirens': Comparison(TestImportProcessor, method='no-psirens', task='Trim your toe nails',
                                               partial=True)}
        with override_settings(processors=settings):
            self.assertEqual(handler.processors, processors)

    async def test_get_active_processors_no_processors(self):
        # Test active processors, when no are defined.
        verification = await make_verification(self.session, contact_id=self.contact_id,
                                               status=VerificationStatus.INITIALIZED)
        handler = TestHandler([])

        for reverse in (False, True):
            with self.subTest(reverse=reverse):
                self.assertEqual(await atuple(handler.get_active_processors(verification, reverse=reverse)), ())

    async def test_get_active_processors_default(self):
        # Test default active processors - verification has no method set up.
        verification = await make_verification(self.session, contact_id=self.contact_id,
                                               status=VerificationStatus.INITIALIZED)
        # Make two processors to check reversed returns the first default processors also on reverse.
        processor_1 = TestProcessor(lambda v: (), method='no-psirens')  # pragma: no branch # lambda
        processor_2 = TestProcessor(lambda v: (), method='no-replicants')  # pragma: no branch # lambda
        handler = TestHandler([processor_1, processor_2])

        for reverse in (False, True):
            with self.subTest(reverse=reverse):
                self.assertEqual(
                    await atuple(handler.get_active_processors(verification, reverse=reverse)), (processor_1, )
                )

    async def test_get_active_processors_default_unapplicable(self):
        # Test default active processors - verification has no method set up, unapplicable processor is not returned.
        verification = await make_verification(self.session, contact_id=self.contact_id,
                                               status=VerificationStatus.INITIALIZED)

        # Mock processor which is not applicable.
        mock = AsyncMock()
        mock.method = 'no-psirens'
        mock.is_applicable.return_value = False
        handler = TestHandler([mock])

        for reverse in (False, True):
            with self.subTest(reverse=reverse):
                self.assertEqual(await atuple(handler.get_active_processors(verification, reverse=reverse)), ())

    async def test_get_active_processors(self):
        # Test a single active processor.
        verification = await make_verification(self.session, contact_id=self.contact_id,
                                               status=VerificationStatus.INITIALIZED)
        await make_event(self.session, verification_id=verification.id, type=EventType.USE_METHOD, method='no-psirens')
        processor_1 = TestProcessor(lambda v: (), method='no-replicants')  # pragma: no branch # lambda
        processor_2 = TestProcessor(lambda v: (), method='no-psirens')  # pragma: no branch # lambda
        handler = TestHandler([processor_1, processor_2])

        for reverse in (False, True):
            with self.subTest(reverse=reverse):
                self.assertEqual(
                    await atuple(handler.get_active_processors(verification, reverse=reverse)), (processor_2, )
                )

    async def test_get_active_processors_multiple(self):
        # Test multiple active processors.
        verification = await make_verification(self.session, contact_id=self.contact_id,
                                               status=VerificationStatus.INITIALIZED)
        await make_event(self.session, verification_id=verification.id, type=EventType.USE_METHOD, method='no-psirens')
        await make_event(self.session, verification_id=verification.id, type=EventType.USE_METHOD,
                         method='no-replicants')
        processor_1 = TestProcessor(lambda v: (), method='no-replicants')  # pragma: no branch # lambda
        processor_2 = TestProcessor(lambda v: (), method='no-psirens')  # pragma: no branch # lambda
        handler = TestHandler([processor_1, processor_2])

        data = (
            # reverse, result
            (False, (processor_1, processor_2)),
            (True, (processor_2, processor_1)),
        )

        for reverse, result in data:
            with self.subTest(reverse=reverse):
                self.assertEqual(await atuple(handler.get_active_processors(verification, reverse=reverse)), result)

    async def test_get_next_events_no_processors(self):
        verification = await make_verification(self.session, contact_id=self.contact_id,
                                               status=VerificationStatus.INITIALIZED)
        handler = TestHandler([])

        self.assertEqual(await atuple(handler.get_next_events(verification)), ())

    async def test_get_next_events_empty(self):
        # Test no events returned by the default processor.
        verification = await make_verification(self.session, contact_id=self.contact_id,
                                               status=VerificationStatus.INITIALIZED)
        processor = TestEventsProcessor(next_events=(), method='no-psirens')
        handler = TestHandler([processor])

        self.assertEqual(await atuple(handler.get_next_events(verification)), ())

    async def test_get_next_events_default(self):
        # Test next event returned by the default processor.
        verification = await make_verification(self.session, contact_id=self.contact_id,
                                               status=VerificationStatus.INITIALIZED)
        processor = TestEventsProcessor(next_events=(sentinel.next_event, ), method='no-psirens')
        handler = TestHandler([processor])

        self.assertEqual(await atuple(handler.get_next_events(verification)), (sentinel.next_event, ))

    async def test_get_next_events_multiple(self):
        # Test multiple next events returned by the default processor.
        verification = await make_verification(self.session, contact_id=self.contact_id,
                                               status=VerificationStatus.INITIALIZED)
        processor = TestEventsProcessor(next_events=(sentinel.next_event, sentinel.other_event), method='no-psirens')
        handler = TestHandler([processor])

        self.assertEqual(await atuple(handler.get_next_events(verification)),
                         (sentinel.next_event, sentinel.other_event))

    async def test_get_next_events_deduplicate(self):
        # Test only the first copy of each event is returned.
        verification = await make_verification(self.session, contact_id=self.contact_id,
                                               status=VerificationStatus.INITIALIZED)
        processor = TestEventsProcessor(
            next_events=(sentinel.next_event, sentinel.other_event, sentinel.next_event), method='no-psirens')
        handler = TestHandler([processor])

        self.assertEqual(await atuple(handler.get_next_events(verification)),
                         (sentinel.next_event, sentinel.other_event))

    async def test_get_next_events_active(self):
        # Test events from a single active processor.
        verification = await make_verification(self.session, contact_id=self.contact_id,
                                               status=VerificationStatus.INITIALIZED)
        await make_event(self.session, verification_id=verification.id, type=EventType.USE_METHOD, method='no-psirens')
        processor_1 = TestEventsProcessor(next_events=(), method='no-replicants')
        processor_2 = TestEventsProcessor(next_events=(sentinel.next_event, ), method='no-psirens')
        handler = TestHandler([processor_1, processor_2])

        self.assertEqual(await atuple(handler.get_next_events(verification)), (sentinel.next_event, ))

    async def test_get_next_events_active_multiple(self):
        # Test events from multiple active processors.
        verification = await make_verification(self.session, contact_id=self.contact_id,
                                               status=VerificationStatus.INITIALIZED)
        await make_event(self.session, verification_id=verification.id, type=EventType.USE_METHOD, method='no-psirens')
        await make_event(self.session, verification_id=verification.id, type=EventType.USE_METHOD,
                         method='no-replicants')
        processor_1 = TestEventsProcessor(next_events=(sentinel.next_event, ), method='no-replicants')
        processor_2 = TestEventsProcessor(next_events=(sentinel.other_event, ), method='no-psirens')
        handler = TestHandler([processor_1, processor_2])

        self.assertEqual(await atuple(handler.get_next_events(verification)),
                         (sentinel.next_event, sentinel.other_event))

    async def test_process_no_event(self):
        verification = await make_verification(self.session, contact_id=self.contact_id)
        processor = _get_processor([])
        handler = TestHandler([processor])

        self.assertFalse(await handler.process(verification))

        # Check verification wasn't processed.
        verification = (await self.session.execute(select(Verification))).scalars().one()
        self.assertEqual(verification.status, VerificationStatus.CREATED)
        self.assertIsNone((await self.session.execute(select(Event))).scalar())

    async def test_process(self):
        verification = await make_verification(self.session, contact_id=self.contact_id)
        processor = _get_processor([VerificationStatus.PENDING])
        handler = TestHandler([processor])

        self.assertTrue(await handler.process(verification))

        # Check verification was processed.
        verification = (await self.session.execute(select(Verification))).scalars().one()
        self.assertEqual(verification.status, VerificationStatus.PENDING)
        event = (await self.session.execute(select(Event))).scalars().one()
        self.assertEqual(event.method, 'no-psirens')
        self.assertEqual(event.type, EventType.CANCEL)

    async def test_process_due_past(self):
        # Test an event with due datetime in past is processed.
        verification = await make_verification(self.session, contact_id=self.contact_id)
        past = datetime(2020, 5, 24, tzinfo=UTC)
        processor = _get_processor([VerificationStatus.PENDING], due_datetime=past)
        handler = TestHandler([processor])

        self.assertTrue(await handler.process(verification))

        # Check verification was processed.
        verification = (await self.session.execute(select(Verification))).scalars().one()
        self.assertEqual(verification.status, VerificationStatus.PENDING)
        event = (await self.session.execute(select(Event))).scalars().one()
        self.assertEqual(event.method, 'no-psirens')
        self.assertEqual(event.type, EventType.CANCEL)

    async def test_process_due_future(self):
        # Test an event with due datetime in future isn't processed.
        verification = await make_verification(self.session, contact_id=self.contact_id)
        future = datetime(2020, 5, 26, tzinfo=UTC)
        processor = _get_processor([VerificationStatus.PENDING], due_datetime=future)
        handler = TestHandler([processor])

        with freeze_time('2020-05-25'):
            self.assertFalse(await handler.process(verification))

        # Check verification wasn't processed.
        verification = (await self.session.execute(select(Verification))).scalars().one()
        self.assertEqual(verification.status, VerificationStatus.CREATED)
        self.assertIsNone((await self.session.execute(select(Event))).scalar())

    async def test_process_multiple_events(self):
        verification = await make_verification(self.session, contact_id=self.contact_id)
        processor = _get_processor([VerificationStatus.ACCESSED, VerificationStatus.PENDING])
        handler = TestHandler([processor])

        self.assertTrue(await handler.process(verification))

        # Check all events were processed.
        verification = (await self.session.execute(select(Verification))).scalars().one()
        self.assertEqual(verification.status, VerificationStatus.PENDING)
        result = tuple((await self.session.execute(select(Event))).scalars())
        self.assertEqual(len(result), 2)
        event_1, event_2 = result
        self.assertEqual(event_1.method, 'no-psirens')
        self.assertEqual(event_1.type, EventType.CANCEL)
        self.assertEqual(event_2.method, 'no-psirens')
        self.assertEqual(event_2.type, EventType.CANCEL)

    async def test_process_loop(self):
        # Test process running in an infinite loop.
        verification = await make_verification(self.session, contact_id=self.contact_id)
        # Processor always returns an event to process.
        processor = TestProcessor(lambda v: [TestEvent(v, method='no-psirens')],  # pragma: no branch # lambda
                                  method='no-psirens')
        handler = TestHandler([processor])

        self.assertTrue(await handler.process(verification))

        msg = StringComparison(r'Processing of .* exceeded the loop limit.')
        self.log_handler.check_present(('verification.handler', 'ERROR', msg))

        # Check up to limit events were processed.
        verification = (await self.session.execute(select(Verification))).scalars().one()
        self.assertEqual(verification.status, VerificationStatus.CANCELED)
        self.assertEqual(len(tuple((await self.session.execute(select(Event))).scalars())), 10)

    async def test_process_loop_events(self):
        # Test verification events are updated when process is running in an infinite loop.
        # Regression test for #235.
        verification = await make_verification(self.session, contact_id=self.contact_id)

        event_types = []
        # Processor always returns an event to process.
        async def next_events(verification: Verification) -> AsyncIterator[ScheduledEvent]:
            await verification.awaitable_attrs.events
            event_types.append([e.type for e in verification.events])
            if not verification.events:
                yield TestEvent(verification, status=VerificationStatus.CANCELED, method='no-psirens')

        processor = TestProcessor(next_events, method='no-psirens')
        handler = TestHandler([processor])

        self.assertTrue(await handler.process(verification))

        # Check only one event was processed
        verification = (await self.session.execute(select(Verification))).scalars().one()
        self.assertEqual(verification.status, VerificationStatus.CANCELED)
        event = (await self.session.execute(select(Event))).scalars().one()
        self.assertEqual(event.method, 'no-psirens')
        self.assertEqual(event.type, EventType.CANCEL)
        # Check verification events have been updated between each cycle of the process loop.
        self.assertEqual(event_types, [[], [EventType.CANCEL]])

    async def test_process_try_complete(self):
        # Test verification is closed when considered to be completed.
        verification = await make_verification(self.session, contact_id=self.contact_id)
        processor = self._get_complete_processor([VerificationStatus.SUCCEEDED], complete=True)
        # Set up a next event.
        processor.next_callback = _get_processor([VerificationStatus.CANCELED]).next_callback
        handler = TestHandler([processor])

        self.assertTrue(await handler.process(verification))

        # Check verification was successfully closed and then the next event was processed.
        verification = (await self.session.execute(select(Verification))).scalars().one()
        self.assertEqual(verification.status, VerificationStatus.CANCELED)
        event1, event2 = (await self.session.execute(select(Event).order_by(Event.id))).scalars()
        self.assertEqual(event1.method, 'no-psirens')
        self.assertEqual(event1.type, EventType.CANCEL)
        self.assertEqual(event2.method, 'no-psirens')
        self.assertEqual(event2.type, EventType.CANCEL)

    async def test_process_does_not_exist(self):
        verification = await make_verification(self.session, contact_id=self.contact_id)
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).side_effect = ContactDoesNotExist
        processor = BaseProcessor(method='no-psirens')
        handler = TestHandler([processor])

        self.assertTrue(await handler.process(verification))

        # Check verification was canceled without any processing.
        verification = (await self.session.execute(select(Verification))).scalars().one()
        self.assertEqual(verification.status, VerificationStatus.CANCELED)
        self.assertIsNotNone(verification.close_datetime)
        self.assertIsNone((await self.session.execute(select(Event))).scalar())

    async def test_process_does_not_exist_closed(self):
        # Test contact does not exist if verification is already closed.
        verification = await make_verification(self.session, contact_id=self.contact_id,
                                               status=VerificationStatus.FAILED)
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).side_effect = ContactDoesNotExist
        processor = BaseProcessor(method='no-psirens')
        handler = TestHandler([processor])

        self.assertFalse(await handler.process(verification))

        # Check verification wasn't changes
        verification = (await self.session.execute(select(Verification))).scalars().one()
        self.assertEqual(verification.status, VerificationStatus.FAILED)
        self.assertIsNone(verification.close_datetime)  # It was not set in fixtures and haven't changed.
        self.assertIsNone((await self.session.execute(select(Event))).scalar())

    async def test_process_check_domains(self):
        # Test domains are updated prior to event processing.
        verification = await make_verification(self.session, contact_id=self.contact_id)
        cast(AsyncMock, APPLICATION.domain_client.get_domains_by_contact).return_value = {
            'domains': [{'domain': DomainRef(id=DomainId('42'), fqdn='example.org')}]}

        processor = _get_processor([VerificationStatus.PENDING])
        handler = TestHandler([processor])

        self.assertTrue(await handler.process(verification))

        # Check domain stored
        domain = (await self.session.execute(select(Domain))).scalars().one()
        self.assertEqual(domain.domain_id, DomainId('42'))
        self.assertEqual(domain.fqdn, 'example.org')
        # Check verification was processed.
        verification = (await self.session.execute(select(Verification))).scalars().one()
        self.assertEqual(verification.status, VerificationStatus.PENDING)
        result = tuple((await self.session.execute(select(Event).order_by(Event.id))).scalars())
        self.assertEqual(len(result), 2)
        event_1, event_2 = result
        self.assertEqual(event_1.type, EventType.DOMAINS_CHANGED)
        self.assertEqual(event_2.type, EventType.CANCEL)

    async def test_process_check_domains_closed(self):
        # Test domains are not updated on closed verification
        verification = await make_verification(self.session, contact_id=self.contact_id,
                                               status=VerificationStatus.FAILED)
        cast(AsyncMock, APPLICATION.domain_client.get_domains_by_contact).return_value = {
            'domains': [{'domain': DomainRef(id=DomainId('42'), fqdn='example.org')}]}

        processor = _get_processor([VerificationStatus.PENDING])
        handler = TestHandler([processor])

        self.assertTrue(await handler.process(verification))

        # Check domain not stored
        self.assertIsNone((await self.session.execute(select(Domain))).scalar())
        # Check verification was processed.
        verification = (await self.session.execute(select(Verification))).scalars().one()
        self.assertEqual(verification.status, VerificationStatus.PENDING)
        event = (await self.session.execute(select(Event))).scalars().one()
        self.assertEqual(event.type, EventType.CANCEL)

    async def test_get_cancel_events_no_processors(self):
        verification = await make_verification(self.session, contact_id=self.contact_id,
                                               status=VerificationStatus.INITIALIZED)
        handler = TestHandler([])

        self.assertEqual(await atuple(handler.get_cancel_events(verification)), ())

    async def test_get_cancel_events_empty(self):
        # Test no events returned by the default processor.
        verification = await make_verification(self.session, contact_id=self.contact_id,
                                               status=VerificationStatus.INITIALIZED)
        processor = TestEventsProcessor(cancel_events=(), method='no-psirens')
        handler = TestHandler([processor])

        self.assertEqual(await atuple(handler.get_cancel_events(verification)), ())

    async def test_get_cancel_events_default(self):
        # Test cancel event returned by the default processor.
        verification = await make_verification(self.session, contact_id=self.contact_id,
                                               status=VerificationStatus.INITIALIZED)
        processor = TestEventsProcessor(cancel_events=(sentinel.cancel_event, ), method='no-psirens')
        handler = TestHandler([processor])

        self.assertEqual(await atuple(handler.get_cancel_events(verification)), (sentinel.cancel_event, ))

    async def test_get_cancel_events_multiple(self):
        # Test multiple cancel events returned by the default processor.
        verification = await make_verification(self.session, contact_id=self.contact_id,
                                               status=VerificationStatus.INITIALIZED)
        processor = TestEventsProcessor(cancel_events=(sentinel.cancel_event, sentinel.other_event),
                                        method='no-psirens')
        handler = TestHandler([processor])

        self.assertEqual(await atuple(handler.get_cancel_events(verification)),
                         (sentinel.cancel_event, sentinel.other_event))

    async def test_get_cancel_events_deduplicate(self):
        # Test only the first copy of each event is returned.
        verification = await make_verification(self.session, contact_id=self.contact_id,
                                               status=VerificationStatus.INITIALIZED)
        processor = TestEventsProcessor(
            cancel_events=(sentinel.cancel_event, sentinel.other_event, sentinel.cancel_event),
            method='no-psirens')
        handler = TestHandler([processor])

        self.assertEqual(await atuple(handler.get_cancel_events(verification)),
                         (sentinel.cancel_event, sentinel.other_event))

    async def test_get_cancel_events_active(self):
        # Test events from a single active processor.
        verification = await make_verification(self.session, contact_id=self.contact_id,
                                               status=VerificationStatus.INITIALIZED)
        await make_event(self.session, verification_id=verification.id, type=EventType.USE_METHOD, method='no-psirens')
        processor_1 = TestEventsProcessor(cancel_events=(), method='no-replicants')
        processor_2 = TestEventsProcessor(cancel_events=(sentinel.cancel_event, ), method='no-psirens')
        handler = TestHandler([processor_1, processor_2])

        self.assertEqual(await atuple(handler.get_cancel_events(verification)), (sentinel.cancel_event, ))

    async def test_get_cancel_events_active_multiple(self):
        # Test events from multiple active processors.
        verification = await make_verification(self.session, contact_id=self.contact_id,
                                               status=VerificationStatus.INITIALIZED)
        await make_event(self.session, verification_id=verification.id, type=EventType.USE_METHOD, method='no-psirens')
        await make_event(self.session, verification_id=verification.id, type=EventType.USE_METHOD,
                         method='no-replicants')
        processor_1 = TestEventsProcessor(cancel_events=(sentinel.cancel_event, ), method='no-replicants')
        processor_2 = TestEventsProcessor(cancel_events=(sentinel.other_event, ), method='no-psirens')
        handler = TestHandler([processor_1, processor_2])

        self.assertEqual(await atuple(handler.get_cancel_events(verification)),
                         (sentinel.other_event, sentinel.cancel_event))

    async def test_get_cancel_events_status(self):
        # Test status is passed to processor.
        verification = await make_verification(self.session, contact_id=self.contact_id)
        processor = AsyncMock(BaseProcessor)
        processor.method = 'no-psirens'
        processor.is_applicable.return_value = True
        processor.get_cancel_events.return_value = aiter((sentinel.cancel_event, ))
        handler = TestHandler([processor])

        self.assertEqual(await atuple(handler.get_cancel_events(verification, status=sentinel.status)),
                         (sentinel.cancel_event, ))
        # Check processor calls
        self.assertEqual(processor.mock_calls,
                         [call.is_applicable(verification), call.get_cancel_events(verification, sentinel.status)])
        self.assertEqual(mock_awaits(processor), [call.is_applicable(verification)])

    def _get_cancel_processor(self, statuses: List[VerificationStatus], complete: bool = False) -> TestProcessor:
        # Set up processor which takes list of statuses and returns one event less on each call.
        def get_cancel_events(verification: Verification) -> Iterable[ScheduledEvent]:
            scheduled = [TestEvent(verification, status=s, method='no-psirens') for s in statuses]
            if statuses:
                statuses.pop(0)
            yield from scheduled

        return TestProcessor(cancel_callback=get_cancel_events, complete=complete, method='no-psirens')

    async def test_cancel_no_event(self):
        verification = await make_verification(self.session, contact_id=self.contact_id)
        processor = self._get_cancel_processor([])
        handler = TestHandler([processor])

        await handler.cancel(verification)

        # Check verification wasn't processed.
        verification = (await self.session.execute(select(Verification))).scalars().one()
        self.assertEqual(verification.status, VerificationStatus.CREATED)
        self.assertIsNone((await self.session.execute(select(Event))).scalar())

    async def test_cancel(self):
        verification = await make_verification(self.session, contact_id=self.contact_id)
        processor = self._get_cancel_processor([VerificationStatus.CANCELED])
        handler = TestHandler([processor])

        await handler.cancel(verification)

        # Check verification was processed.
        verification = (await self.session.execute(select(Verification))).scalars().one()
        self.assertEqual(verification.status, VerificationStatus.CANCELED)
        event = (await self.session.execute(select(Event))).scalars().one()
        self.assertEqual(event.method, 'no-psirens')
        self.assertEqual(event.type, EventType.CANCEL)

    async def test_cancel_multiple_events(self):
        verification = await make_verification(self.session, contact_id=self.contact_id)
        processor = self._get_cancel_processor([VerificationStatus.PENDING, VerificationStatus.CANCELED])
        handler = TestHandler([processor])

        await handler.cancel(verification)

        # Check all events were processed.
        verification = (await self.session.execute(select(Verification))).scalars().one()
        self.assertEqual(verification.status, VerificationStatus.CANCELED)
        result = tuple((await self.session.execute(select(Event))).scalars())
        self.assertEqual(len(result), 2)
        event_1, event_2 = result
        self.assertEqual(event_1.method, 'no-psirens')
        self.assertEqual(event_1.type, EventType.CANCEL)
        self.assertEqual(event_2.method, 'no-psirens')
        self.assertEqual(event_2.type, EventType.CANCEL)

    async def test_cancel_loop(self):
        # Test cancel running in an infinite loop.
        verification = await make_verification(self.session, contact_id=self.contact_id)
        # Processor always returns an event to cancel.
        processor = TestProcessor(
            cancel_callback=lambda v: [TestEvent(v, method='no-psirens')],  # pragma: no branch # lambda
            method='no-psirens')
        handler = TestHandler([processor])

        await handler.cancel(verification)

        msg = StringComparison(r'Processing of .* exceeded the loop limit.')
        self.log_handler.check_present(('verification.handler', 'ERROR', msg))

        # Check up to limit events were processed.
        verification = (await self.session.execute(select(Verification))).scalars().one()
        self.assertEqual(verification.status, VerificationStatus.CANCELED)
        self.assertEqual(len(tuple((await self.session.execute(select(Event))).scalars())), 10)

    async def test_cancel_loop_events(self):
        # Test verification events are updated when cancel is running in an infinite loop.
        # Regression test for #235.
        verification = await make_verification(self.session, contact_id=self.contact_id)

        event_types = []
        # Processor always returns an event to process.
        async def cancel_events(verification: Verification) -> AsyncIterator[ScheduledEvent]:
            await verification.awaitable_attrs.events
            event_types.append([e.type for e in verification.events])
            if not verification.events:
                yield TestEvent(verification, status=VerificationStatus.CANCELED, method='no-psirens')

        processor = TestProcessor(cancel_callback=cancel_events, method='no-psirens')
        handler = TestHandler([processor])

        await handler.cancel(verification)

        # Check only one event was processed
        verification = (await self.session.execute(select(Verification))).scalars().one()
        self.assertEqual(verification.status, VerificationStatus.CANCELED)
        event = (await self.session.execute(select(Event))).scalars().one()
        self.assertEqual(event.method, 'no-psirens')
        self.assertEqual(event.type, EventType.CANCEL)
        # Check verification events have been updated between each cycle of the process loop.
        self.assertEqual(event_types, [[], [EventType.CANCEL]])

    async def test_cancel_status(self):
        verification = await make_verification(self.session, contact_id=self.contact_id)
        processor = AsyncMock(BaseProcessor)
        processor.method = 'no-psirens'
        processor.is_applicable.return_value = True
        processor.get_cancel_events.return_value = aiter(())
        handler = TestHandler([processor])

        await handler.cancel(verification, sentinel.status)

        # Check processor calls
        self.assertEqual(processor.mock_calls,
                         [call.is_applicable(verification), call.get_cancel_events(verification, sentinel.status)])
        self.assertEqual(mock_awaits(processor), [call.is_applicable(verification)])

    async def test_cancel_does_not_exist(self):
        verification = await make_verification(self.session, contact_id=self.contact_id)
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).side_effect = ContactDoesNotExist
        processor = BaseProcessor(method='no-psirens')
        handler = TestHandler([processor])

        await handler.cancel(verification)

        # Check verification was canceled without any processing.
        verification = (await self.session.execute(select(Verification))).scalars().one()
        self.assertEqual(verification.status, VerificationStatus.CANCELED)
        self.assertIsNotNone(verification.close_datetime)
        self.assertIsNone((await self.session.execute(select(Event))).scalar())

    async def test_cancel_check_domains(self):
        # Test domains are updated prior to event processing.
        verification = await make_verification(self.session, contact_id=self.contact_id)
        cast(AsyncMock, APPLICATION.domain_client.get_domains_by_contact).return_value = {
            'domains': [{'domain': DomainRef(id=DomainId('42'), fqdn='example.org')}]}

        processor = self._get_cancel_processor([VerificationStatus.CANCELED])
        handler = TestHandler([processor])

        await handler.cancel(verification)

        # Check domain stored
        domain = (await self.session.execute(select(Domain))).scalars().one()
        self.assertEqual(domain.domain_id, DomainId('42'))
        self.assertEqual(domain.fqdn, 'example.org')
        # Check verification was processed.
        verification = (await self.session.execute(select(Verification))).scalars().one()
        self.assertEqual(verification.status, VerificationStatus.CANCELED)
        result = tuple((await self.session.execute(select(Event).order_by(Event.id))).scalars())
        self.assertEqual(len(result), 2)
        event_1, event_2 = result
        self.assertEqual(event_1.type, EventType.DOMAINS_CHANGED)
        self.assertEqual(event_2.type, EventType.CANCEL)

    async def test_cancel_check_domains_closed(self):
        # Test domains are not updated on closed verification
        verification = await make_verification(self.session, contact_id=self.contact_id,
                                               status=VerificationStatus.FAILED)
        cast(AsyncMock, APPLICATION.domain_client.get_domains_by_contact).return_value = {
            'domains': [{'domain': DomainRef(id=DomainId('42'), fqdn='example.org')}]}

        processor = self._get_cancel_processor([VerificationStatus.CANCELED])
        handler = TestHandler([processor])

        await handler.cancel(verification)

        # Check domain not stored
        self.assertIsNone((await self.session.execute(select(Domain))).scalar())
        # Check verification was processed.
        verification = (await self.session.execute(select(Verification))).scalars().one()
        self.assertEqual(verification.status, VerificationStatus.CANCELED)
        event = (await self.session.execute(select(Event))).scalars().one()
        self.assertEqual(event.type, EventType.CANCEL)

    def _get_complete_processor(self, statuses: List[VerificationStatus], complete: bool = False) -> TestProcessor:
        # Set up processor which takes list of statuses and returns one event less on each call.
        def get_complete_events(verification: Verification) -> Iterable[ScheduledEvent]:
            scheduled = [TestEvent(verification, status=s, method='no-psirens') for s in statuses]
            if statuses:
                statuses.pop(0)
            yield from scheduled

        return TestProcessor(
            next_callback=lambda v: [TestErrorEvent(v, method='no-psirens')],  # pragma: no cover
            complete_callback=get_complete_events, complete=complete, method='no-psirens')

    async def test_complete_no_event(self):
        verification = await make_verification(self.session, contact_id=self.contact_id)
        processor = self._get_complete_processor([])
        handler = TestHandler([processor])

        await handler.complete(verification, processor)

        # Check verification wasn't processed.
        verification = (await self.session.execute(select(Verification))).scalars().one()
        self.assertEqual(verification.status, VerificationStatus.CREATED)
        self.assertIsNone((await self.session.execute(select(Event))).scalar())

    async def test_complete(self):
        verification = await make_verification(self.session, contact_id=self.contact_id)
        processor = self._get_complete_processor([VerificationStatus.SUCCEEDED])
        handler = TestHandler([processor])

        await handler.complete(verification, processor)

        # Check verification was processed.
        verification = (await self.session.execute(select(Verification))).scalars().one()
        self.assertEqual(verification.status, VerificationStatus.SUCCEEDED)
        event = (await self.session.execute(select(Event))).scalars().one()
        self.assertEqual(event.method, 'no-psirens')
        self.assertEqual(event.type, EventType.CANCEL)

    async def test_complete_multiple_events(self):
        verification = await make_verification(self.session, contact_id=self.contact_id)
        processor = self._get_complete_processor([VerificationStatus.PENDING, VerificationStatus.SUCCEEDED])
        handler = TestHandler([processor])

        await handler.complete(verification, processor)

        # Check all events were processed.
        verification = (await self.session.execute(select(Verification))).scalars().one()
        self.assertEqual(verification.status, VerificationStatus.SUCCEEDED)
        result = tuple((await self.session.execute(select(Event))).scalars())
        self.assertEqual(len(result), 2)
        event_1, event_2 = result
        self.assertEqual(event_1.method, 'no-psirens')
        self.assertEqual(event_1.type, EventType.CANCEL)
        self.assertEqual(event_2.method, 'no-psirens')
        self.assertEqual(event_2.type, EventType.CANCEL)

    async def test_complete_loop(self):
        # Test complete running in an infinite loop.
        verification = await make_verification(self.session, contact_id=self.contact_id)
        # Processor always returns an event to complete.
        processor = TestProcessor(
            complete_callback=lambda v: [TestEvent(v, method='no-psirens')],  # pragma: no branch # lambda
            method='no-psirens')
        handler = TestHandler([processor])

        await handler.complete(verification, processor)

        msg = StringComparison(r'Processing of .* exceeded the loop limit.')
        self.log_handler.check_present(('verification.handler', 'ERROR', msg))

        # Check up to limit events were processed.
        verification = (await self.session.execute(select(Verification))).scalars().one()
        self.assertEqual(verification.status, VerificationStatus.CANCELED)
        self.assertEqual(len(tuple((await self.session.execute(select(Event))).scalars())), 10)

    async def test_complete_loop_events(self):
        # Test verification events are updated when complete is running in an infinite loop.
        # Regression test for #235.
        verification = await make_verification(self.session, contact_id=self.contact_id)

        event_types = []
        # Processor always returns an event to process.
        async def complete_events(verification: Verification) -> AsyncIterator[ScheduledEvent]:
            await verification.awaitable_attrs.events
            event_types.append([e.type for e in verification.events])
            if not verification.events:
                yield TestEvent(verification, status=VerificationStatus.CANCELED, method='no-psirens')

        processor = TestProcessor(complete_callback=complete_events, method='no-psirens')
        handler = TestHandler([processor])

        await handler.complete(verification, processor)

        # Check only one event was processed
        verification = (await self.session.execute(select(Verification))).scalars().one()
        self.assertEqual(verification.status, VerificationStatus.CANCELED)
        event = (await self.session.execute(select(Event))).scalars().one()
        self.assertEqual(event.method, 'no-psirens')
        self.assertEqual(event.type, EventType.CANCEL)
        # Check verification events have been updated between each cycle of the process loop.
        self.assertEqual(event_types, [[], [EventType.CANCEL]])

    async def test_complete_does_not_exist(self):
        verification = await make_verification(self.session, contact_id=self.contact_id)
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).side_effect = ContactDoesNotExist
        processor = BaseProcessor(method='no-psirens')
        handler = TestHandler([processor])

        await handler.complete(verification, processor)

        # Check verification was canceled without any processing.
        verification = (await self.session.execute(select(Verification))).scalars().one()
        self.assertEqual(verification.status, VerificationStatus.CANCELED)
        self.assertIsNotNone(verification.close_datetime)
        self.assertIsNone((await self.session.execute(select(Event))).scalar())

    async def test_complete_check_domains(self):
        # Test domains are updated prior to event processing.
        verification = await make_verification(self.session, contact_id=self.contact_id)
        cast(AsyncMock, APPLICATION.domain_client.get_domains_by_contact).return_value = {
            'domains': [{'domain': DomainRef(id=DomainId('42'), fqdn='example.org')}]}

        processor = self._get_complete_processor([VerificationStatus.SUCCEEDED])
        handler = TestHandler([processor])

        await handler.complete(verification, processor)

        # Check domain stored
        domain = (await self.session.execute(select(Domain))).scalars().one()
        self.assertEqual(domain.domain_id, DomainId('42'))
        self.assertEqual(domain.fqdn, 'example.org')
        # Check verification was processed.
        verification = (await self.session.execute(select(Verification))).scalars().one()
        self.assertEqual(verification.status, VerificationStatus.SUCCEEDED)
        result = tuple((await self.session.execute(select(Event).order_by(Event.id))).scalars())
        self.assertEqual(len(result), 2)
        event_1, event_2 = result
        self.assertEqual(event_1.type, EventType.DOMAINS_CHANGED)
        self.assertEqual(event_2.type, EventType.CANCEL)

    async def test_complete_check_domains_closed(self):
        # Test domains are not updated on closed verification
        verification = await make_verification(self.session, contact_id=self.contact_id,
                                               status=VerificationStatus.FAILED)
        cast(AsyncMock, APPLICATION.domain_client.get_domains_by_contact).return_value = {
            'domains': [{'domain': DomainRef(id=DomainId('42'), fqdn='example.org')}]}

        processor = self._get_complete_processor([VerificationStatus.SUCCEEDED])
        handler = TestHandler([processor])

        await handler.complete(verification, processor)

        # Check domain not stored
        self.assertIsNone((await self.session.execute(select(Domain))).scalar())
        # Check verification was processed.
        verification = (await self.session.execute(select(Verification))).scalars().one()
        self.assertEqual(verification.status, VerificationStatus.SUCCEEDED)
        event = (await self.session.execute(select(Event))).scalars().one()
        self.assertEqual(event.type, EventType.CANCEL)

    async def test_try_complete(self):
        # Test verification is closed when considered to be completed.
        verification = await make_verification(self.session, contact_id=self.contact_id)
        processor = self._get_complete_processor([VerificationStatus.SUCCEEDED], complete=True)
        handler = TestHandler([processor])

        self.assertTrue(await handler.try_complete(verification))

        # Check verification was successfully closed.
        verification = (await self.session.execute(select(Verification))).scalars().one()
        self.assertEqual(verification.status, VerificationStatus.SUCCEEDED)
        event = (await self.session.execute(select(Event))).scalars().one()
        self.assertEqual(event.method, 'no-psirens')
        self.assertEqual(event.type, EventType.CANCEL)

    async def test_try_complete_multiple(self):
        # Test verification is closed when considered to be completed - multiple events.
        verification = await make_verification(self.session, contact_id=self.contact_id)
        processor = self._get_complete_processor([VerificationStatus.PENDING, VerificationStatus.SUCCEEDED],
                                                 complete=True)
        handler = TestHandler([processor])

        self.assertTrue(await handler.try_complete(verification))

        # Check verification was successfully closed and futile rolled back.
        verification = (await self.session.execute(select(Verification))).scalars().one()
        self.assertEqual(verification.status, VerificationStatus.SUCCEEDED)
        result = tuple((await self.session.execute(select(Event))).scalars())
        self.assertEqual(len(result), 2)
        event_1, event_2 = result
        self.assertEqual(event_1.method, 'no-psirens')
        self.assertEqual(event_1.type, EventType.CANCEL)
        self.assertEqual(event_2.method, 'no-psirens')
        self.assertEqual(event_2.type, EventType.CANCEL)

    async def test_try_complete_incomplete(self):
        # Test noop if verification is not complete.
        verification = await make_verification(self.session, contact_id=self.contact_id)
        processor = self._get_complete_processor([VerificationStatus.SUCCEEDED], complete=False)
        handler = TestHandler([processor])

        self.assertFalse(await handler.try_complete(verification))

        # Check verification wasn't changed.
        verification = (await self.session.execute(select(Verification))).scalars().one()
        self.assertEqual(verification.status, VerificationStatus.CREATED)
        self.assertIsNone((await self.session.execute(select(Event))).scalar())

    async def test_check_domains_empty(self):
        verification = await make_verification(self.session, contact_id=self.contact_id)
        handler = TestCheckDomainHandler()
        cast(AsyncMock, APPLICATION.domain_client.get_domains_by_contact).return_value = {'domains': []}

        self.assertFalse(await handler.check_domains(verification))

        # Check no domains
        self.assertIsNone((await self.session.execute(select(Domain))).scalar())
        # Check no event created
        self.assertIsNone((await self.session.execute(select(Event))).scalar())
        # Check domain call
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.domain_client)),
                         [call.get_domains_by_contact(self.contact_id, {DomainContactRole.HOLDER})])
        # Check create calls
        self.assertEqual(handler.create_mock.mock_calls, [])

    async def test_check_domains_added(self):
        verification = await make_verification(self.session, contact_id=self.contact_id)
        handler = TestCheckDomainHandler()
        cast(AsyncMock, APPLICATION.domain_client.get_domains_by_contact).return_value = {
            'domains': [{'domain': DomainRef(id=DomainId('42'), fqdn='example.org')}]}

        self.assertTrue(await handler.check_domains(verification))

        # Check domain stored
        domain = (await self.session.execute(select(Domain))).scalars().one()
        self.assertEqual(domain.verification_id, verification.id)
        self.assertEqual(domain.domain_id, DomainId('42'))
        self.assertEqual(domain.fqdn, 'example.org')
        # Check event created
        event = (await self.session.execute(select(Event))).scalars().one()
        self.assertIsNone(event.method)
        self.assertEqual(event.type, EventType.DOMAINS_CHANGED)
        self.assertEqual(event.details, {'added': ['42']})
        # Check domain call
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.domain_client)),
                         [call.get_domains_by_contact(self.contact_id, {DomainContactRole.HOLDER})])
        # Check create calls
        self.assertEqual(handler.create_mock.mock_calls, [])

    async def test_check_domains_deleted(self):
        # Test domain have been deleted.
        verification = await make_verification(self.session, contact_id=self.contact_id)
        await make_domain(self.session, verification_id=verification.id, domain_id='42', fqdn='example.org')
        handler = TestCheckDomainHandler()
        cast(AsyncMock, APPLICATION.domain_client.get_domains_by_contact).return_value = {'domains': []}
        cast(AsyncMock, APPLICATION.domain_client.get_domain_info).side_effect = DomainDoesNotExist

        self.assertTrue(await handler.check_domains(verification))

        # Check domain deleted
        self.assertIsNone((await self.session.execute(select(Domain))).scalar())
        # Check event created
        event = (await self.session.execute(select(Event))).scalars().one()
        self.assertIsNone(event.method)
        self.assertEqual(event.type, EventType.DOMAINS_CHANGED)
        self.assertEqual(event.details, {'removed': ['42']})
        # Check domain call
        domain_calls = [call.get_domains_by_contact(self.contact_id, {DomainContactRole.HOLDER}),
                        call.get_domain_info('42')]
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.domain_client)), domain_calls)
        # Check create calls
        self.assertEqual(handler.create_mock.mock_calls, [])

    async def test_check_domains_no_change(self):
        verification = await make_verification(self.session, contact_id=self.contact_id)
        await make_domain(self.session, verification_id=verification.id, domain_id='42', fqdn='example.org')
        handler = TestCheckDomainHandler()
        cast(AsyncMock, APPLICATION.domain_client.get_domains_by_contact).return_value = {
            'domains': [{'domain': DomainRef(id=DomainId('42'), fqdn='example.org')}]}

        self.assertFalse(await handler.check_domains(verification))

        # Check domain unchanged
        domain = (await self.session.execute(select(Domain))).scalars().one()
        self.assertEqual(domain.verification_id, verification.id)
        self.assertEqual(domain.domain_id, DomainId('42'))
        self.assertEqual(domain.fqdn, 'example.org')
        # Check no event created
        self.assertIsNone((await self.session.execute(select(Event))).scalar())
        # Check domain call
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.domain_client)),
                         [call.get_domains_by_contact(self.contact_id, {DomainContactRole.HOLDER})])
        # Check create calls
        self.assertEqual(handler.create_mock.mock_calls, [])

    async def _test_check_domains_removed(self, *, contact_calls: Sequence[_Call] = ()) -> None:
        # Test domain have been removed without a new verification created.
        verification = await make_verification(self.session, contact_id=self.contact_id)
        await make_domain(self.session, verification_id=verification.id, domain_id='42', fqdn='example.org')
        handler = TestCheckDomainHandler()
        cast(AsyncMock, APPLICATION.domain_client.get_domains_by_contact).return_value = {'domains': []}
        cast(AsyncMock, APPLICATION.domain_client.get_domain_info).return_value = _Domain(
            domain_id=DomainId('42'), fqdn='example.org',
            registrant=ContactRef(id=self.new_contact_id, handle=self.new_contact_handle))

        self.assertTrue(await handler.check_domains(verification))

        # Check no create calls
        self.assertEqual(handler.create_mock.mock_calls, [])
        # Check domain deleted
        self.assertIsNone((await self.session.execute(select(Domain))).scalar())
        # Check event created
        event = (await self.session.execute(select(Event))).scalars().one()
        self.assertIsNone(event.method)
        self.assertEqual(event.type, EventType.DOMAINS_CHANGED)
        self.assertEqual(event.details, {'removed': ['42']})
        # Check registry calls
        domain_calls = [call.get_domains_by_contact(self.contact_id, {DomainContactRole.HOLDER}),
                        call.get_domain_info('42')]
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.domain_client)), domain_calls)
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.contact_client)), contact_calls or [])
        self.assertEqual(cast(AsyncMock, APPLICATION.domain_admin_client).mock_calls,
                             [call.manage_domain_state_flags(close_requests=[])])

    async def _test_check_domains_dispersion(self, contact_calls: Sequence[_Call] = ()) -> None:
        # Test domain have been transferred to a new contact and its verification started.
        verification = await make_verification(self.session, contact_id=self.contact_id)
        await make_domain(self.session, verification_id=verification.id, domain_id='42', fqdn='example.org')
        handler = TestCheckDomainHandler()
        cast(AsyncMock, APPLICATION.domain_client.get_domains_by_contact).return_value = {'domains': []}
        cast(AsyncMock, APPLICATION.domain_client.get_domain_info).return_value = _Domain(
            domain_id=DomainId('42'), fqdn='example.org',
            registrant=ContactRef(id=self.new_contact_id, handle=self.new_contact_handle))

        self.assertTrue(await handler.check_domains(verification))

        # Check create calls
        self.assertEqual(handler.create_mock.mock_calls, [call(ANY, self.new_contact_id, reason='', reason_url='')])
        # Check domain deleted
        self.assertIsNone((await self.session.execute(select(Domain))).scalar())
        # Check event created
        event = (await self.session.execute(select(Event))).scalars().one()
        self.assertIsNone(event.method)
        self.assertEqual(event.type, EventType.DOMAINS_CHANGED)
        self.assertEqual(event.details, {'removed': ['42']})
        # Check registry calls
        calls = [call.get_domains_by_contact(self.contact_id, {DomainContactRole.HOLDER}),
                 call.get_domain_info('42')]
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.domain_client)), calls)
        contact_calls = contact_calls or [call.get_contact_state(self.new_contact_id)]
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.contact_client)), contact_calls)
        self.assertEqual(cast(AsyncMock, APPLICATION.domain_admin_client).mock_calls,
                             [call.manage_domain_state_flags(close_requests=[])])

    async def test_check_domains_dispersion_disabled(self):
        # Test domain have been transferred to another contact - dispersion disabled.
        with override_settings(dispersion=None):
            APPLICATION.contact_client = AsyncMock(ContactClient)
            APPLICATION.domain_client = AsyncMock(DomainClient)
            APPLICATION.domain_admin_client = AsyncMock(DomainAdminClient)
            APPLICATION.contact_client.get_contact_state.return_value = {}
            APPLICATION.domain_admin_client.manage_domain_state_flags.return_value = []

            await self._test_check_domains_removed()

    async def test_check_domains_dispersion(self):
        # Test domain have been transferred to another contact - dispersion enabled.
        with override_settings(dispersion={}):
            APPLICATION.contact_client = AsyncMock(ContactClient)
            APPLICATION.domain_client = AsyncMock(DomainClient)
            APPLICATION.domain_admin_client = AsyncMock(DomainAdminClient)
            APPLICATION.contact_client.get_contact_state.return_value = {}
            APPLICATION.domain_admin_client.manage_domain_state_flags.return_value = []

            await self._test_check_domains_dispersion()

    async def test_check_domains_dispersion_verification_open(self):
        # Test domain have been transferred to another contact with an open verification
        await make_verification(self.session, contact_id=self.new_contact_id, status=VerificationStatus.INITIALIZED)

        with override_settings(dispersion={}):
            APPLICATION.contact_client = AsyncMock(ContactClient)
            APPLICATION.domain_client = AsyncMock(DomainClient)
            APPLICATION.domain_admin_client = AsyncMock(DomainAdminClient)
            APPLICATION.contact_client.get_contact_state.return_value = {}
            APPLICATION.domain_admin_client.manage_domain_state_flags.return_value = []

            await self._test_check_domains_removed()

    async def test_check_domains_dispersion_verification_closed(self):
        # Test domain have been transferred to another contact with an unsuccessful verification
        await make_verification(self.session, contact_id=self.new_contact_id, status=VerificationStatus.FAILED,
                                close_datetime=datetime(1988, 9, 6, tzinfo=UTC))
        with override_settings(dispersion={'max_age': 365}):
            APPLICATION.contact_client = AsyncMock(ContactClient)
            APPLICATION.domain_client = AsyncMock(DomainClient)
            APPLICATION.domain_admin_client = AsyncMock(DomainAdminClient)
            APPLICATION.contact_client.get_contact_state.return_value = {}
            APPLICATION.domain_admin_client.manage_domain_state_flags.return_value = []

            with freeze_time('1989-01-01'):
                await self._test_check_domains_dispersion()

    async def test_check_domains_dispersion_verification_succeeded(self):
        # Test domain have been transferred to another contact with a recently successful verification
        await make_verification(self.session, contact_id=self.new_contact_id, status=VerificationStatus.SUCCEEDED,
                                close_datetime=datetime(1988, 9, 6, tzinfo=UTC))
        with override_settings(dispersion={'max_age': 365}):
            APPLICATION.contact_client = AsyncMock(ContactClient)
            APPLICATION.domain_client = AsyncMock(DomainClient)
            APPLICATION.domain_admin_client = AsyncMock(DomainAdminClient)
            APPLICATION.contact_client.get_contact_state.return_value = {}
            APPLICATION.domain_admin_client.manage_domain_state_flags.return_value = []

            with freeze_time('1989-01-01'):
                await self._test_check_domains_removed()

    async def test_check_domains_dispersion_verification_manual(self):
        # Test domain have been transferred to another contact with a recently manually closed verification
        await make_verification(self.session, contact_id=self.new_contact_id, status=VerificationStatus.MANUAL,
                                close_datetime=datetime(1988, 9, 6, tzinfo=UTC))
        with override_settings(dispersion={'max_age': 365}):
            APPLICATION.contact_client = AsyncMock(ContactClient)
            APPLICATION.domain_client = AsyncMock(DomainClient)
            APPLICATION.domain_admin_client = AsyncMock(DomainAdminClient)
            APPLICATION.contact_client.get_contact_state.return_value = {}
            APPLICATION.domain_admin_client.manage_domain_state_flags.return_value = []

            with freeze_time('1989-01-01'):
                await self._test_check_domains_removed()

    async def test_check_domains_dispersion_verification_permanent(self):
        # Test verifications are considered permanent.
        await make_verification(self.session, contact_id=self.new_contact_id, status=VerificationStatus.SUCCEEDED,
                                close_datetime=datetime(1988, 9, 6, tzinfo=UTC))
        with override_settings(dispersion={'max_age': 0}):
            APPLICATION.contact_client = AsyncMock(ContactClient)
            APPLICATION.domain_client = AsyncMock(DomainClient)
            APPLICATION.domain_admin_client = AsyncMock(DomainAdminClient)
            APPLICATION.contact_client.get_contact_state.return_value = {}
            APPLICATION.domain_admin_client.manage_domain_state_flags.return_value = []

            with freeze_time('1989-01-01'):
                await self._test_check_domains_removed()

    async def test_check_domains_dispersion_verification_old(self):
        # Test domain have been transferred to another contact with a not recently successful verification
        await make_verification(self.session, contact_id=self.new_contact_id, status=VerificationStatus.SUCCEEDED,
                                close_datetime=datetime(1988, 9, 6, tzinfo=UTC))
        with override_settings(dispersion={'max_age': 1}):
            APPLICATION.contact_client = AsyncMock(ContactClient)
            APPLICATION.domain_client = AsyncMock(DomainClient)
            APPLICATION.domain_admin_client = AsyncMock(DomainAdminClient)
            APPLICATION.contact_client.get_contact_state.return_value = {}
            APPLICATION.domain_admin_client.manage_domain_state_flags.return_value = []

            with freeze_time('1989-01-01'):
                await self._test_check_domains_dispersion()

    async def test_check_domains_dispersion_verification_canceled(self):
        # Test domain have been transferred to another contact with a recently canceled verification.
        await make_verification(self.session, contact_id=self.new_contact_id, status=VerificationStatus.CANCELED,
                                close_datetime=datetime(1988, 9, 6, tzinfo=UTC))
        with override_settings(dispersion={'max_age': 365}):
            APPLICATION.contact_client = AsyncMock(ContactClient)
            APPLICATION.domain_client = AsyncMock(DomainClient)
            APPLICATION.domain_admin_client = AsyncMock(DomainAdminClient)
            APPLICATION.contact_client.get_contact_state.return_value = {}
            APPLICATION.domain_admin_client.manage_domain_state_flags.return_value = []

            with freeze_time('1989-01-01'):
                await self._test_check_domains_dispersion()

    async def test_check_domains_dispersion_verification_canceled_ignored(self):
        # Test domain have been transferred to another contact with a recently successful verification
        # and a newer one canceled.
        await make_verification(self.session, contact_id=self.new_contact_id, status=VerificationStatus.SUCCEEDED,
                                close_datetime=datetime(1988, 9, 6, tzinfo=UTC))
        await make_verification(self.session, contact_id=self.new_contact_id, status=VerificationStatus.CANCELED,
                                close_datetime=datetime(1988, 9, 7, tzinfo=UTC))
        with override_settings(dispersion={'max_age': 365}):
            APPLICATION.contact_client = AsyncMock(ContactClient)
            APPLICATION.domain_client = AsyncMock(DomainClient)
            APPLICATION.domain_admin_client = AsyncMock(DomainAdminClient)
            APPLICATION.contact_client.get_contact_state.return_value = {}
            APPLICATION.domain_admin_client.manage_domain_state_flags.return_value = []

            with freeze_time('1989-01-01'):
                await self._test_check_domains_removed()

    async def test_check_domains_dispersion_verified(self):
        # Test domain have been transferred to a verified contact.
        with override_settings(dispersion={'verified_statuses': ['VERIFIED']}):
            APPLICATION.contact_client = AsyncMock(ContactClient)
            APPLICATION.domain_client = AsyncMock(DomainClient)
            APPLICATION.domain_admin_client = AsyncMock(DomainAdminClient)
            APPLICATION.contact_client.get_contact_state.return_value = {'VERIFIED': True}
            APPLICATION.domain_admin_client.manage_domain_state_flags.return_value = []

            contact_calls = [call.get_contact_state(self.new_contact_id)]
            await self._test_check_domains_removed(contact_calls=contact_calls)

    async def test_check_domains_dispersion_transient_new(self):
        # Test domain have been transferred to a recently transiently verified contact.
        with override_settings(dispersion={'transient_statuses': ['TRANSIENT'], 'max_age': 365}):
            APPLICATION.contact_client = AsyncMock(ContactClient)
            APPLICATION.domain_client = AsyncMock(DomainClient)
            APPLICATION.domain_admin_client = AsyncMock(DomainAdminClient)
            APPLICATION.contact_client.get_contact_state.return_value = {'TRANSIENT': True}
            APPLICATION.contact_client.get_contact_state_history.return_value = {
                    "timeline": [{"valid_from": datetime(1988, 9, 6, tzinfo=UTC), "flags": {"TRANSIENT": True}}]}
            APPLICATION.domain_admin_client.manage_domain_state_flags.return_value = []

            contact_calls = [call.get_contact_state(self.new_contact_id),
                             call.get_contact_state_history(self.new_contact_id)]
            with freeze_time('1989-01-01'):
                await self._test_check_domains_removed(contact_calls=contact_calls)

    async def test_check_domains_dispersion_transient_new_history(self):
        # Test domain have been transferred to a recently transiently verified contact with state history
        with override_settings(dispersion={'transient_statuses': ['TRANSIENT'], 'max_age': 365}):
            APPLICATION.contact_client = AsyncMock(ContactClient)
            APPLICATION.domain_client = AsyncMock(DomainClient)
            APPLICATION.domain_admin_client = AsyncMock(DomainAdminClient)
            APPLICATION.contact_client.get_contact_state.return_value = {'TRANSIENT': True}
            timeline = [{"valid_from": datetime(1987, 9, 6, tzinfo=UTC), "flags": {"TRANSIENT": False}},
                        {"valid_from": datetime(1988, 9, 6, tzinfo=UTC), "flags": {"TRANSIENT": True}}]
            APPLICATION.contact_client.get_contact_state_history.return_value = {"timeline": timeline}
            APPLICATION.domain_admin_client.manage_domain_state_flags.return_value = []

            contact_calls = [call.get_contact_state(self.new_contact_id),
                             call.get_contact_state_history(self.new_contact_id)]
            with freeze_time('1989-01-01'):
                await self._test_check_domains_removed(contact_calls=contact_calls)

    async def test_check_domains_dispersion_transient_old(self):
        # Test domain have been transferred to a not recently transiently verified contact.
        with override_settings(dispersion={'transient_statuses': ['TRANSIENT'], 'max_age': 1}):
            APPLICATION.contact_client = AsyncMock(ContactClient)
            APPLICATION.domain_client = AsyncMock(DomainClient)
            APPLICATION.domain_admin_client = AsyncMock(DomainAdminClient)
            APPLICATION.contact_client.get_contact_state.return_value = {'TRANSIENT': True}
            APPLICATION.contact_client.get_contact_state_history.return_value = {
                    "timeline": [{"valid_from": datetime(1988, 9, 6, tzinfo=UTC), "flags": {"TRANSIENT": True}}]}
            APPLICATION.domain_admin_client.manage_domain_state_flags.return_value = []

            contact_calls = [call.get_contact_state(self.new_contact_id),
                             call.get_contact_state_history(self.new_contact_id)]
            with freeze_time('1989-01-01'):
                await self._test_check_domains_dispersion(contact_calls=contact_calls)

    async def test_check_domains_dispersion_transient_permanent(self):
        # Test domain have been transferred to a not recently transiently verified contact.
        # Transient is considered permanent.
        with override_settings(dispersion={'transient_statuses': ['TRANSIENT'], 'max_age': 0}):
            APPLICATION.contact_client = AsyncMock(ContactClient)
            APPLICATION.domain_client = AsyncMock(DomainClient)
            APPLICATION.domain_admin_client = AsyncMock(DomainAdminClient)
            APPLICATION.contact_client.get_contact_state.return_value = {'TRANSIENT': True}
            APPLICATION.contact_client.get_contact_state_history.return_value = {
                    "timeline": [{"valid_from": datetime(1988, 9, 6, tzinfo=UTC), "flags": {"TRANSIENT": True}}]}
            APPLICATION.domain_admin_client.manage_domain_state_flags.return_value = []

            contact_calls = [call.get_contact_state(self.new_contact_id),
                             call.get_contact_state_history(self.new_contact_id)]
            with freeze_time('1989-01-01'):
                await self._test_check_domains_removed(contact_calls=contact_calls)

    async def test_check_domains_dispersion_transient_off(self):
        # Test transient statuses are set up, but contact doesn't have it.
        with override_settings(dispersion={'transient_statuses': ['TRANSIENT'], 'max_age': 365}):
            APPLICATION.contact_client = AsyncMock(ContactClient)
            APPLICATION.domain_client = AsyncMock(DomainClient)
            APPLICATION.domain_admin_client = AsyncMock(DomainAdminClient)
            APPLICATION.contact_client.get_contact_state.return_value = {'TRANSIENT': False}
            APPLICATION.domain_admin_client.manage_domain_state_flags.return_value = []

            with freeze_time('1989-01-01'):
                await self._test_check_domains_dispersion()

    async def _test_check_domains_outzone(self, verification: Verification, request_ids: List[str]) -> None:
        # Test outzone cancelation on domain transfer.
        await make_domain(self.session, verification_id=verification.id, domain_id='42', fqdn='example.org')
        handler = TestCheckDomainHandler()
        with override_settings(dispersion=None):
            APPLICATION.contact_client = AsyncMock(ContactClient)
            APPLICATION.contact_client.get_contact_state.return_value = {}
            APPLICATION.domain_client = AsyncMock(DomainClient)
            APPLICATION.domain_client.get_domains_by_contact.return_value = {'domains': []}
            APPLICATION.domain_client.get_domain_info.return_value = _Domain(
                domain_id=DomainId('42'), fqdn='example.org',
                registrant=ContactRef(id=self.new_contact_id, handle=self.new_contact_handle))
            APPLICATION.domain_admin_client = AsyncMock(DomainAdminClient)
            APPLICATION.domain_admin_client.manage_domain_state_flags.return_value = []

            self.assertTrue(await handler.check_domains(verification))

            # Check no create calls
            self.assertEqual(handler.create_mock.mock_calls, [])
            # Check registry calls
            domain_calls = [call.get_domains_by_contact(self.contact_id, {DomainContactRole.HOLDER}),
                            call.get_domain_info('42')]
            self.assertEqual(mock_awaits(APPLICATION.domain_client), domain_calls)
            self.assertEqual(mock_awaits(APPLICATION.contact_client), [])
            self.assertEqual(APPLICATION.domain_admin_client.mock_calls,
                             [call.manage_domain_state_flags(close_requests=request_ids)])

    async def test_check_domains_cancel_outzone(self):
        # Test outzone statuses have been canceled if present.
        verification = await make_verification(self.session, contact_id=self.contact_id)
        await make_event(self.session, verification_id=verification.id, type=EventType.OUTZONE,
                         details={'request_ids': {'42': ['196']}})
        await self._test_check_domains_outzone(verification, ['196'])

    async def test_check_domains_cancel_outzone_transferred(self):
        # Test only transfered domains are canceled.
        verification = await make_verification(self.session, contact_id=self.contact_id)
        await make_event(self.session, verification_id=verification.id, type=EventType.OUTZONE,
                         details={'request_ids': {'42': ['196'], '43': ['197']}})
        await self._test_check_domains_outzone(verification, ['196'])

    async def test_check_domains_cancel_outzone_last_event(self):
        # Test only last outzone event is canceled.
        verification = await make_verification(self.session, contact_id=self.contact_id)
        await make_event(self.session, verification_id=verification.id, type=EventType.OUTZONE,
                         details={'request_ids': {'42': ['196']}})
        await make_event(self.session, verification_id=verification.id, type=EventType.OUTZONE,
                         details={'request_ids': {'42': ['150']}}, event_datetime=datetime(1988, 9, 6, tzinfo=UTC))
        await self._test_check_domains_outzone(verification, ['196'])

    async def test_check_domains_cancel_outzone_canceled(self):
        # Test noop when last outzone is cancelation.
        verification = await make_verification(self.session, contact_id=self.contact_id)
        await make_event(self.session, verification_id=verification.id, type=EventType.OUTZONE,
                         details={'request_ids': {'42': ['196']}, 'canceled': True})
        await self._test_check_domains_outzone(verification, [])

    async def test_check_domains_cancel_outzone_no_requests(self):
        # Test noop when outzone event has no request IDs.
        verification = await make_verification(self.session, contact_id=self.contact_id)
        await make_event(self.session, verification_id=verification.id, type=EventType.OUTZONE, details={})
        await self._test_check_domains_outzone(verification, [])

    async def test_check_domains_cancel_outzone_no_event(self):
        # Test noop when outzone event have not occurred.
        verification = await make_verification(self.session, contact_id=self.contact_id)
        await self._test_check_domains_outzone(verification, [])

    async def test_check_domains_cancel_outzone_legacy(self):
        # Test noop on outzone events from verification < 1.5.0
        verification = await make_verification(self.session, contact_id=self.contact_id)
        await make_event(self.session, verification_id=verification.id, type=EventType.OUTZONE,
                         details={'request_ids': ['42']})
        await self._test_check_domains_outzone(verification, [])

    async def _test_create(self, reason: str = '', reason_url: str = '', domains: Sequence[Dict[str, DomainRef]] = (),
                           domain_ids: Sequence[str] = ()) -> None:
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = Contact(
            contact_id=self.contact_id, contact_handle=self.contact_handle)
        cast(AsyncMock, APPLICATION.domain_client.get_domains_by_contact).return_value = {'domains': domains or []}
        handler = TestHandler([])

        verification = await handler.create(self.session, self.contact_id, reason=reason, reason_url=reason_url)

        # Check verification was created.
        self.assertEqual(verification.contact_id, self.contact_id)
        self.assertEqual(verification.contact_handle, self.contact_handle)
        self.assertEqual(verification.status, VerificationStatus.CREATED)
        self.assertEqual(verification.reason, reason)
        self.assertEqual(verification.reason_url, reason_url)
        self.assertIsNone(verification.close_datetime)
        # Check event created
        event = (await self.session.execute(select(Event))).scalars().one()
        self.assertEqual(event.type, EventType.CREATE)
        self.assertEqual(event.details, {'domains': domain_ids or []})
        # Check domains created
        self.assertEqual(tuple((await self.session.execute(select(Domain.domain_id))).scalars()), tuple(domain_ids))
        # Check registry calls
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.contact_client)),
                         [call.get_contact_info(self.contact_id)])
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.domain_client)),
                         [call.get_domains_by_contact(self.contact_id, {DomainContactRole.HOLDER})])

    async def test_create(self):
        await self._test_create()

    async def test_create_domains(self):
        domains = [{'domain': DomainRef(id=DomainId('42'), fqdn='exaple.org')}]
        domain_ids = ['42']

        await self._test_create(domains=domains, domain_ids=domain_ids)

    async def test_create_reason(self):
        await self._test_create(reason='Smeghead!')

    async def test_create_reason_url(self):
        await self._test_create(reason_url='Smeghead!')

    async def test_create_contact_does_not_exist(self):
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).side_effect = ContactDoesNotExist
        handler = TestHandler([])

        with self.assertRaises(ContactDoesNotExist):
            await handler.create(self.session, self.contact_id)

        # Check nothing created
        self.assertIsNone((await self.session.execute(select(Verification))).scalar())
        self.assertIsNone((await self.session.execute(select(Event))).scalar())
        self.assertIsNone((await self.session.execute(select(Domain))).scalar())
        # Check registry calls
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.contact_client)),
                         [call.get_contact_info(self.contact_id)])
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.domain_client)), [])

    async def test_create_already_exist(self):
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = Contact(
            contact_id=self.contact_id, contact_handle=self.contact_handle)
        await make_verification(self.session, contact_id=self.contact_id)
        handler = TestHandler([])

        with self.assertRaises(VerificationAlreadyExists):
            await handler.create(self.session, self.contact_id)

        # Check nothing created
        self.assertEqual(len(tuple((await self.session.execute(select(Verification))).scalars())), 1)
        self.assertIsNone((await self.session.execute(select(Event))).scalar())
        self.assertIsNone((await self.session.execute(select(Domain))).scalar())
        # Check registry calls
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.contact_client)),
                         [call.get_contact_info(self.contact_id)])
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.domain_client)), [])


@override_database()
class VerificationHandlerProcessAllTest(IsolatedAsyncioTestCase):
    contact_id = 'KRYTEN'

    async def asyncSetUp(self):
        APPLICATION.contact_client = AsyncMock(ContactClient)
        APPLICATION.domain_client = AsyncMock(DomainClient)
        self.log_handler = LogCapture('verification', propagate=False)
        self.sigterm_handler = signal.getsignal(signal.SIGTERM)

    async def asyncTearDown(self):
        signal.signal(signal.SIGTERM, self.sigterm_handler)
        self.log_handler.uninstall()

    async def test_process_all_no_verification(self):
        handler = TestHandler([])

        self.assertEqual(await handler.process_all(), Result())

        async with test_session_cls()() as session:
            self.assertIsNone((await session.execute(select(Verification))).scalar())
            self.assertIsNone((await session.execute(select(Event))).scalar())

    async def test_process_all_no_processors(self):
        async with atomic() as session:
            await make_verification(session, contact_id=self.contact_id)
        handler = TestHandler([])

        self.assertEqual(await handler.process_all(), Result(skip=1))

        # Check verification wasn't processed.
        async with test_session_cls()() as session:
            verification = (await session.execute(select(Verification))).scalars().one()
            self.assertEqual(verification.status, VerificationStatus.CREATED)
            self.assertIsNone((await session.execute(select(Event))).scalar())

    async def test_process_all_no_event(self):
        async with atomic() as session:
            await make_verification(session, contact_id=self.contact_id)
        processor = _get_processor([])
        handler = TestHandler([processor])

        self.assertEqual(await handler.process_all(), Result(skip=1))

        # Check verification wasn't processed.
        async with test_session_cls()() as session:
            verification = (await session.execute(select(Verification))).scalars().one()
            self.assertEqual(verification.status, VerificationStatus.CREATED)
            self.assertIsNone((await session.execute(select(Event))).scalar())

    async def test_process_all(self):
        async with atomic() as session:
            await make_verification(session, contact_id=self.contact_id)
        processor = _get_processor([VerificationStatus.PENDING])
        handler = TestHandler([processor])

        self.assertEqual(await handler.process_all(), Result(success=1))

        # Check verification was processed.
        async with test_session_cls()() as session:
            verification = (await session.execute(select(Verification))).scalars().one()
            self.assertEqual(verification.status, VerificationStatus.PENDING)
            event = (await session.execute(select(Event))).scalars().one()
            self.assertEqual(event.method, 'no-psirens')
            self.assertEqual(event.type, EventType.CANCEL)

    async def test_process_all_closed_recent(self):
        # Test recently closed verification is also processed.
        async with atomic() as session:
            await make_verification(session, contact_id=self.contact_id, status=VerificationStatus.FAILED,
                                    close_datetime=datetime(1988, 2, 15, 21, 0, tzinfo=UTC))
        processor = _get_processor([VerificationStatus.PENDING])
        handler = TestHandler([processor])

        with freeze_time('1988-02-15T21:05:00'):
            self.assertEqual(await handler.process_all(closed_limit=600), Result(success=1))

        # Check verification was processed.
        async with test_session_cls()() as session:
            verification = (await session.execute(select(Verification))).scalars().one()
            self.assertEqual(verification.status, VerificationStatus.PENDING)
            event = (await session.execute(select(Event))).scalars().one()
            self.assertEqual(event.method, 'no-psirens')
            self.assertEqual(event.type, EventType.CANCEL)

    async def test_process_all_closed_old(self):
        # Test old closed verification is not processed.
        async with atomic() as session:
            await make_verification(session, contact_id=self.contact_id, status=VerificationStatus.FAILED,
                                    close_datetime=datetime(1988, 2, 15, 21, 0, tzinfo=UTC))
        processor = _get_processor([VerificationStatus.PENDING])
        handler = TestHandler([processor])

        with freeze_time('1988-02-15T21:15:00'):
            self.assertEqual(await handler.process_all(closed_limit=600), Result())

        # Check verification wasn't processed.
        async with test_session_cls()() as session:
            verification = (await session.execute(select(Verification))).scalars().one()
            self.assertEqual(verification.status, VerificationStatus.FAILED)
            self.assertIsNone((await session.execute(select(Event))).scalar())

    async def test_process_all_error(self):
        # Test an error occurs in event processing.
        async with atomic() as session:
            await make_verification(session, contact_id=self.contact_id)
        event = Mock()
        event.due_datetime = None
        event.process.side_effect = Exception('Gazpacho!')
        processor = TestProcessor(lambda v: [event], method='no-psirens')  # pragma: no branch # lambda
        handler = TestHandler([processor])

        self.assertEqual(await handler.process_all(), Result(error=1))

        # Check verification is unchanged.
        async with test_session_cls()() as session:
            verification = (await session.execute(select(Verification))).scalars().one()
            self.assertEqual(verification.status, VerificationStatus.CREATED)
            self.assertIsNone((await session.execute(select(Event))).scalar())
        # Check logs
        msg = StringComparison(r'An error in processing verification\[.*\]: .*Exception.*Gazpacho!.*')
        self.log_handler.check_present(('verification.handler', 'ERROR', msg))

    @skipIf(APPLICATION.settings.test_db_connection.startswith('sqlite'), 'Does not work on SQLite database.')
    async def test_process_all_skip_locked(self):
        # Test locked rows are skipped.
        async with atomic() as session:
            await make_verification(session, contact_id=self.contact_id)
        processor = TestProcessor(lambda v: [TestEvent(v, method='no-psirens')],  # pragma: no branch # lambda
                                  method='no-psirens')
        handler = TestHandler([processor])

        # Lock the verification in test transaction.
        session = test_session_cls()()
        query = select(Verification).with_for_update()
        await session.execute(query)

        self.assertEqual(await handler.process_all(), Result())

        # Release the lock.
        await session.rollback()

        # Check verification wasn't processed.
        async with test_session_cls()() as session:
            verification = (await session.execute(select(Verification))).scalars().one()
            self.assertEqual(verification.status, VerificationStatus.CREATED)
            self.assertIsNone((await session.execute(select(Event))).scalar())

    async def test_process_all_graceful_stop(self):
        # Test graceful stop.
        # Steps:
        # 1. Start processing of the event.
        step_1 = asyncio.Event()
        # 2. Send SIGTERM signal.
        step_2 = asyncio.Event()
        # 3. Continue processing of the event.
        # 4. End processing cycle.
        async with atomic() as session:
            await make_verification(session, contact_id=self.contact_id)
            other_contact_id = self.contact_id + '_2'
            await make_verification(session, contact_id=other_contact_id)

        # Custom process on the first event to trigger events in correct order.
        async def process(verification: Verification) -> None:
            step_1.set()
            await step_2.wait()
            verification.status = VerificationStatus.CANCELED

        # Create custom next_events to return what is needed.
        counter = []

        def next_events_callback(verification: Verification) -> Iterable[ScheduledEvent]:
            counter.append(sentinel.DEFAULT)
            if len(counter) == 1:
                scheduled = AsyncMock(due_datetime=None)
                scheduled.process.side_effect = partial(process, verification=verification)
                scheduled.to_event.return_value = Event(verification_id=verification.id, type=EventType.CANCEL)
                return [scheduled]
            elif len(counter) == 3:
                return [TestEvent(verification)]  # pragma: no cover
            else:
                return []

        processor = TestProcessor(next_events_callback, method='no-psirens')  # pragma: no branch # lambda
        handler = TestHandler([processor])

        async def send_signal():
            await step_1.wait()
            step_2.set()
            signal.raise_signal(signal.SIGTERM)

        results = await asyncio.gather(handler.process_all(), send_signal())
        self.assertEqual(results[0], Result(success=1))

        async with test_session_cls()() as session:
            # Check first verification is changed.
            query = select(Verification).filter_by(contact_id=self.contact_id)
            verification = (await session.execute(query)).scalars().one()
            self.assertEqual(verification.status, VerificationStatus.CANCELED)
            # Check other verification is unchanged.
            query = select(Verification).filter_by(contact_id=other_contact_id)
            verification = (await session.execute(query)).scalars().one()
            self.assertEqual(verification.status, VerificationStatus.CREATED)
