from datetime import date, datetime
from typing import Any, Dict, List, Optional, Tuple, cast
from unittest import TestCase
from unittest.mock import AsyncMock, call

from fred_types import ContactHistoryId, ContactId, DomainId, DomainRef
from pytz import UTC
from regal import (
    AdditionalIdentifier,
    Address,
    Contact,
    ContactClient,
    ContactIdentifierType,
    Domain as _Domain,
    DomainClient,
    DomainContactRole,
)
from regal.exceptions import DomainDoesNotExist
from sqlalchemy.ext.asyncio import AsyncSession

from verification.constants import EventType, MessageType, VerificationStatus
from verification.models import ContactData, Domain, Verification
from verification.models.verification import _random_token, compare_strings
from verification.settings import APPLICATION

from .utils import TransactionTestCase, make_domain, make_event, make_verification, mock_awaits


class VerificationTest(TransactionTestCase):
    async def asyncSetUp(self):
        APPLICATION.contact_client = AsyncMock(ContactClient)
        APPLICATION.domain_client = AsyncMock(DomainClient)

    async def test_contact(self):
        contact_id = ContactId('2X4B')
        contact_handle = 'KRYTEN'
        contact = Contact(contact_id=contact_id, contact_handle=contact_handle)
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = contact
        verification = Verification(contact_id=contact_id)

        self.assertEqual(await verification.contact, contact)

        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.contact_client)), [call.get_contact_info(contact_id)])

    async def test_contact_cached(self):
        contact_id = ContactId('2X4B')
        contact_handle = 'KRYTEN'
        contact = Contact(contact_id=contact_id, contact_handle=contact_handle)
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = contact
        verification = Verification(contact_id=contact_id)

        self.assertIs(await verification.contact, await verification.contact)

        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.contact_client)), [call.get_contact_info(contact_id)])

    async def test_get_events_empty(self):
        verification = await make_verification(self.session, contact_id='KRYTEN')

        self.assertEqual(await verification.get_events(), [])

    async def test_get_events(self):
        verification = await make_verification(self.session, contact_id='KRYTEN')
        event1 = await make_event(self.session, verification_id=verification.id, type=EventType.CANCEL)
        event2 = await make_event(self.session, verification_id=verification.id, type=EventType.COMPLETE)

        self.assertEqual(await verification.get_events(), [event2, event1])

    async def test_get_events_type(self):
        verification = await make_verification(self.session, contact_id='KRYTEN')
        event = await make_event(self.session, verification_id=verification.id, type=EventType.CANCEL)

        self.assertEqual(await verification.get_events(EventType.CANCEL), [event])

    async def test_get_events_mismatch(self):
        verification = await make_verification(self.session, contact_id='KRYTEN')
        await make_event(self.session, verification_id=verification.id, type=EventType.CANCEL)

        self.assertEqual(await verification.get_events(EventType.FAILED), [])

    async def test_get_events_multiple(self):
        verification = await make_verification(self.session, contact_id='KRYTEN')
        # Events have switched datetimes.
        event1 = await make_event(self.session, verification_id=verification.id, type=EventType.CANCEL,
                                  event_datetime=datetime(1988, 2, 15, 21, 30, tzinfo=UTC))
        event2 = await make_event(self.session, verification_id=verification.id, type=EventType.CANCEL,
                                  event_datetime=datetime(1988, 2, 15, 21, 0, tzinfo=UTC))

        self.assertEqual(await verification.get_events(EventType.CANCEL), [event2, event1])

    async def test_active_methods_empty(self):
        verification = await make_verification(self.session, contact_id='KRYTEN')

        self.assertEqual(await verification.active_methods, set())

    async def test_active_methods_other_events(self):
        verification = await make_verification(self.session, contact_id='KRYTEN')
        await make_event(self.session, verification_id=verification.id, type=EventType.CANCEL)

        self.assertEqual(await verification.active_methods, set())

    async def test_active_methods_single(self):
        # Test active methods with a single method activated.
        verification = await make_verification(self.session, contact_id='KRYTEN')
        await make_event(self.session, verification_id=verification.id, type=EventType.USE_METHOD,
                         method='no-psirens')

        self.assertEqual(await verification.active_methods, {'no-psirens'})

    async def test_active_methods_multiple(self):
        # Test active methods with a single method activated multiple times.
        verification = await make_verification(self.session, contact_id='KRYTEN')
        await make_event(self.session, verification_id=verification.id, type=EventType.USE_METHOD,
                         method='no-psirens')
        await make_event(self.session, verification_id=verification.id, type=EventType.USE_METHOD,
                         method='no-psirens')

        self.assertEqual(await verification.active_methods, {'no-psirens'})

    async def test_active_methods_various(self):
        # Test active methods with multiple different methods activated.
        verification = await make_verification(self.session, contact_id='KRYTEN')
        await make_event(self.session, verification_id=verification.id, type=EventType.USE_METHOD,
                         method='no-psirens')
        await make_event(self.session, verification_id=verification.id, type=EventType.USE_METHOD,
                         method='no-replicants')

        self.assertEqual(await verification.active_methods, {'no-psirens', 'no-replicants'})

    async def test_get_message_events_empty(self):
        verification = await make_verification(self.session, contact_id='KRYTEN')

        self.assertEqual(await verification.get_message_events(MessageType.INITIAL_EMAIL), [])

    async def test_get_message_events_typeless(self):
        # Test message event with no type set.
        verification = await make_verification(self.session, contact_id='KRYTEN')
        await make_event(self.session, verification_id=verification.id, type=EventType.SEND_MESSAGE)

        self.assertEqual(await verification.get_message_events(MessageType.INITIAL_EMAIL), [])

    async def test_get_message_events(self):
        verification = await make_verification(self.session, contact_id='KRYTEN')
        event = await make_event(self.session, verification_id=verification.id, type=EventType.SEND_MESSAGE,
                                 details={'message_type': MessageType.INITIAL_EMAIL})

        self.assertEqual(await verification.get_message_events(MessageType.INITIAL_EMAIL), [event])

    async def test_get_message_events_mismatch(self):
        verification = await make_verification(self.session, contact_id='KRYTEN')
        await make_event(self.session, verification_id=verification.id, type=EventType.SEND_MESSAGE,
                         details={'message_type': MessageType.INITIAL_EMAIL})

        self.assertEqual(await verification.get_message_events(MessageType.CONFIRMATION_LETTER), [])

    async def test_get_message_events_multiple(self):
        verification = await make_verification(self.session, contact_id='KRYTEN')
        event1 = await make_event(self.session, verification_id=verification.id, type=EventType.SEND_MESSAGE,
                                  details={'message_type': MessageType.INITIAL_EMAIL},
                                  event_datetime=datetime(1988, 2, 15, 21, 0, tzinfo=UTC))
        event2 = await make_event(self.session, verification_id=verification.id, type=EventType.SEND_MESSAGE,
                                  details={'message_type': MessageType.INITIAL_EMAIL},
                                  event_datetime=datetime(1988, 2, 15, 21, 30, tzinfo=UTC))

        self.assertEqual(await verification.get_message_events(MessageType.INITIAL_EMAIL), [event2, event1])

    async def _make_delay_event(self, session: AsyncSession, verification: Verification, status: VerificationStatus,
                                delay: int) -> None:
        details = {'status': status, 'delay': delay}
        await make_event(session, verification_id=verification.id, type=EventType.DELAY_STATUS, details=details)

    async def test_get_status_delay_none(self):
        # Test no delay event.
        verification = await make_verification(self.session, contact_id='KRYTEN')
        self.assertEqual(await verification.get_status_delay(VerificationStatus.FUTILE), 0)

    async def test_get_status_delay(self):
        # Test single delay event.
        verification = await make_verification(self.session, contact_id='KRYTEN')
        await self._make_delay_event(self.session, verification, VerificationStatus.FUTILE, 42)

        self.assertEqual(await verification.get_status_delay(VerificationStatus.FUTILE), 42)

    async def test_get_status_delay_negative(self):
        # Test single delay event with negative value.
        verification = await make_verification(self.session, contact_id='KRYTEN')
        await self._make_delay_event(self.session, verification, VerificationStatus.FUTILE, -42)

        self.assertEqual(await verification.get_status_delay(VerificationStatus.FUTILE), -42)

    async def test_get_status_delay_multiple(self):
        # Test multiple delay events.
        verification = await make_verification(self.session, contact_id='KRYTEN')
        await self._make_delay_event(self.session, verification, VerificationStatus.FUTILE, 42)
        await self._make_delay_event(self.session, verification, VerificationStatus.FUTILE, 196)

        self.assertEqual(await verification.get_status_delay(VerificationStatus.FUTILE), 238)

    async def test_get_status_delay_other(self):
        # Test delay event for other status.
        verification = await make_verification(self.session, contact_id='KRYTEN')
        await self._make_delay_event(self.session, verification, VerificationStatus.FAILED, 42)

        self.assertEqual(await verification.get_status_delay(VerificationStatus.FUTILE), 0)

    async def test_get_status_delay_missing_status(self):
        # Test delay event with missing status field.
        verification = await make_verification(self.session, contact_id='KRYTEN')
        await make_event(self.session, verification_id=verification.id, type=EventType.DELAY_STATUS,
                         details={'delay': 42})

        self.assertEqual(await verification.get_status_delay(VerificationStatus.FUTILE), 0)

    async def test_get_status_delay_missing_delay(self):
        # Test delay event with missing delay field.
        verification = await make_verification(self.session, contact_id='KRYTEN')
        details = {'status': VerificationStatus.FUTILE}
        await make_event(self.session, verification_id=verification.id, type=EventType.DELAY_STATUS, details=details)

        self.assertEqual(await verification.get_status_delay(VerificationStatus.FUTILE), 0)

    async def test_affected_domains_empty(self):
        verification = await make_verification(self.session, contact_id='KRYTEN')
        await make_event(self.session, verification_id=verification.id, type=EventType.CREATE,
                         details={'domains': ['42']})

        self.assertEqual(await verification.affected_domains, ())

        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.domain_client)), [])

    async def test_affected_domains(self):
        verification = await make_verification(self.session, contact_id='KRYTEN')
        await make_domain(self.session, verification_id=verification.id, domain_id='42', fqdn='example.org')

        self.assertEqual(await verification.affected_domains, (DomainRef(id=DomainId('42'), fqdn='example.org'), ))

        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.domain_client)), [])

    async def test_affected_domains_fallback_outzone(self):
        # Test affected domains are extracted from outzone if missing
        verification = await make_verification(self.session, contact_id='KRYTEN')
        await make_event(self.session, verification_id=verification.id, type=EventType.CREATE)
        await make_event(self.session, verification_id=verification.id, type=EventType.OUTZONE,
                         details={'domains': ['42']})
        cast(AsyncMock, APPLICATION.domain_client.get_domain_info).return_value = _Domain(domain_id=DomainId('42'),
                                                                                         fqdn='example.org')

        self.assertEqual(await verification.affected_domains, [DomainRef(id=DomainId('42'), fqdn='example.org')])

        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.domain_client)), [call.get_domain_info('42')])

    async def test_affected_domains_fallback_outzone_no_create(self):
        # Test affected domains are extracted from outzone if create event is missing
        verification = await make_verification(self.session, contact_id='KRYTEN')
        details = {'domains': ['42']}
        await make_event(self.session, verification_id=verification.id, type=EventType.OUTZONE, details=details)
        cast(AsyncMock, APPLICATION.domain_client.get_domain_info).return_value = _Domain(domain_id=DomainId('42'),
                                                                                         fqdn='example.org')

        self.assertEqual(await verification.affected_domains, [DomainRef(id=DomainId('42'), fqdn='example.org')])

        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.domain_client)), [call.get_domain_info('42')])

    async def test_affected_domains_fallback_outzone_create(self):
        # Test affected domains are not extracted from outzone if create event has proper structure
        verification = await make_verification(self.session, contact_id='KRYTEN')
        details = {'domains': ['42']}
        await make_event(self.session, verification_id=verification.id, type=EventType.CREATE, details=details)
        await make_event(self.session, verification_id=verification.id, type=EventType.OUTZONE, details=details)

        self.assertEqual(await verification.affected_domains, ())

        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.domain_client)), [])

    async def test_affected_domains_fallback_outzone_canceled(self):
        # Test affected domains are not extracted from outzone if its canceled
        verification = await make_verification(self.session, contact_id='KRYTEN')
        await make_event(self.session, verification_id=verification.id, type=EventType.CREATE,
                         details={'domains': ['42']})
        await make_event(self.session, verification_id=verification.id, type=EventType.OUTZONE,
                         details={'domains': ['42'], 'canceled': True})

        self.assertEqual(await verification.affected_domains, ())

        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.domain_client)), [])

    async def test_affected_domains_fallback_outzone_ignored(self):
        # Test outzone event is ignored if domains exist
        verification = await make_verification(self.session, contact_id='KRYTEN')
        await make_domain(self.session, verification_id=verification.id, domain_id='42', fqdn='example.org')
        details = {'domains': ['196']}
        await make_event(self.session, verification_id=verification.id, type=EventType.CREATE, details=details)
        await make_event(self.session, verification_id=verification.id, type=EventType.OUTZONE, details=details)

        self.assertEqual(await verification.affected_domains, (DomainRef(id=DomainId('42'), fqdn='example.org'), ))

        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.domain_client)), [])

    async def test_affected_domains_fallback_outzone_ignore_missing(self):
        # Test affected domains exclude any domains that no longer exist.
        verification = await make_verification(self.session, contact_id='KRYTEN')
        await make_event(self.session, verification_id=verification.id, type=EventType.CREATE)
        await make_event(self.session, verification_id=verification.id, type=EventType.OUTZONE,
                         details={'domains': ['42']})
        cast(AsyncMock, APPLICATION.domain_client.get_domain_info).side_effect = DomainDoesNotExist

        self.assertEqual(await verification.affected_domains, [])

        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.domain_client)), [call.get_domain_info('42')])

    async def test_affected_domains_fallback_registry(self):
        # Test affected domains are extracted from registry if missing
        verification = await make_verification(self.session, contact_id='KRYTEN')
        await make_event(self.session, verification_id=verification.id, type=EventType.CREATE)
        cast(AsyncMock, APPLICATION.domain_client.get_domains_by_contact).return_value = {
            'domains': [{'domain': DomainRef(id=DomainId('42'), fqdn='example.org')}]}

        self.assertEqual(await verification.affected_domains, (DomainRef(id=DomainId('42'), fqdn='example.org'), ))

        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.domain_client)),
                         [call.get_domains_by_contact('KRYTEN', {DomainContactRole.HOLDER})])

    async def test_affected_domains_fallback_registry_no_create(self):
        # Test affected domains are not extracted from registry if create event is missing
        verification = await make_verification(self.session, contact_id='KRYTEN')
        cast(AsyncMock, APPLICATION.domain_client.get_domains_by_contact).return_value = {
            'domains': [{'domain': DomainRef(id=DomainId('42'), fqdn='example.org')}]}

        self.assertEqual(await verification.affected_domains, (DomainRef(id=DomainId('42'), fqdn='example.org'), ))

        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.domain_client)),
                         [call.get_domains_by_contact('KRYTEN', {DomainContactRole.HOLDER})])

    async def test_affected_domains_fallback_registry_create(self):
        # Test affected domains are not extracted from registry if create event has proper structure
        verification = await make_verification(self.session, contact_id='KRYTEN')
        details = {'domains': ['42']}
        await make_event(self.session, verification_id=verification.id, type=EventType.CREATE, details=details)

        self.assertEqual(await verification.affected_domains, ())

        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.domain_client)), [])

    async def test_affected_domains_migrated(self):
        # Test affected domains on migrated constrained verification with no domains left. Regression for #257.
        verification = await make_verification(self.session, contact_id='KRYTEN')
        await make_event(self.session, verification_id=verification.id, type=EventType.CREATE)
        await make_event(self.session, verification_id=verification.id, type=EventType.OUTZONE,
                         details={'domains': ['42']})
        await make_event(self.session, verification_id=verification.id, type=EventType.DOMAINS_CHANGED,
                         details={'removed': ['42']})

        self.assertEqual(await verification.affected_domains, ())

        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.domain_client)), [])


class DomainTest(TestCase):
    def test_ref(self):
        domain = Domain(domain_id='42', fqdn='example.org')

        self.assertEqual(domain.ref, DomainRef(id=DomainId('42'), fqdn='example.org'))


class TokenTest(TestCase):
    def test_default(self):
        token = _random_token()

        self.assertEqual(len(token), 10)


class CompareStringsTest(TestCase):
    def test_compare_strings(self):
        data = (
            ('', '', True),
            ('', ' ', True),
            ('', ' \t\n\r      ', True),
            ('', '\u00a0', True),  # unbreakable space
            ('', 'Kryten', False),
            ('Kryten', 'Kryten', True),
            ('Kryten', '  \t \n \r Kryten', True),
            ('Kryten', 'Kryten  \t \n \r ', True),
            ('Kryten', '  \t \n \r Kryten  \t \n \r ', True),
            ('Kryten', 'KRYTEN', True),  # case insensitive
            ('Kryten', 'Křýťěň', True),  # acute insensitive
            ('Kryten', 'Rimmer', False),
            ('Arnold Rimmer', 'Arnold Rimmer', True),
            ('Arnold Rimmer', 'Arnold \n\t\r Rimmer', True),
            ('Arnold Rimmer', 'ArnoldRimmer', False),
            ('Arnold Rimmer', 'Arnold Rimer', False),
            ('Arnold Rimmer', 'Arnold Rimerr', False),
        )
        for first, second, result in data:
            with self.subTest(first=first, second=second):
                self.assertIs(compare_strings(first, second), result)

            with self.subTest(first=second, second=first):
                self.assertIs(compare_strings(second, first), result)


class ContactDataTest(TransactionTestCase):
    contact_id = ContactId('KRYTEN')
    contact_history_id = ContactHistoryId('STAR-BUG')

    async def asyncSetUp(self):
        APPLICATION.contact_client = AsyncMock(ContactClient)

    def test_compare_name(self):
        data = (
            ('Arnold Rimmer', 'Arnold Rimmer', True),
            ('Arnold Rimmer', 'Arnold\tRimmer\n', True),
            ('Arnold Rimmer', 'ARNOLD RIMMER', True),
            ('Arnold Rimmer', 'Åřňóľď Rimmer', True),
            ('Arnold Rimmer', 'Rimmer Arnold', True),  # order doesn't matter
            ('Arnold Judas Rimmer BSC SSC', 'BSC Arnold Judas SSC Rimmer', True),  # order doesn't matter
            ('Arnold Judas Rimmer BSC', 'Arnold Judas Rimmer SSC', False),
            ('Arnold Rimmer', 'Arnold Arnold Rimmer', False),
            ('Arnold Rimmer', 'Arnold "Ace" Rimmer', False),
        )
        for first, second, result in data:
            with self.subTest(provided=first, reference=second):
                contact_data = ContactData(name=first)
                self.assertEqual(contact_data._compare_name(second), result)

            with self.subTest(provided=second, reference=first):
                contact_data = ContactData(name=second)
                self.assertEqual(contact_data._compare_name(first), result)

    def test_compare_organization(self):
        data = (
            ('Jupiter Mining Corporation', 'Jupiter Mining Corporation', True),
            ('Jupiter Mining Corporation', 'Jupiter\tMining\nCorporation', True),
            ('Jupiter Mining Corporation', 'JUPITER MINING CORPORATION', True),
            ('Jupiter Mining Corporation', 'Jupiter Míňíňg Corporation', True),
            ('Jupiter Mining Corporation', 'Jupiter Corporation Mining', False),  # order matters
            ('Jupiter Mining Corporation', 'Jupiter Corporation', False),
        )
        for first, second, result in data:
            with self.subTest(provided=first, reference=second):
                contact_data = ContactData(organization=first)
                self.assertEqual(contact_data._compare_organization(second), result)

            with self.subTest(provided=second, reference=first):
                contact_data = ContactData(organization=second)
                self.assertEqual(contact_data._compare_organization(first), result)

    def test_compare_company_registration_number(self):
        data = (
            ('JMC', 'JMC', True),
            ('JMC', 'JMC \n\t', True),
            ('JMC', 'jmc', True),
            ('JMC', 'JMČ', True),
            ('JMC42', 'JMC 42', False),
        )
        for first, second, result in data:
            with self.subTest(provided=first, reference=second):
                contact_data = ContactData(company_registration_number=first)
                self.assertEqual(contact_data._compare_company_registration_number(second), result)

            with self.subTest(provided=second, reference=first):
                contact_data = ContactData(company_registration_number=second)
                self.assertEqual(contact_data._compare_company_registration_number(first), result)

    def test_compare_emails(self):
        data = (
            ('kryten@example.org', 'kryten@example.org', True),
            ('kryten@example.org', 'kryten@example.org  \n\t', True),
            ('kryten@example.org', 'kryten@eXaMpLe.oRg', False),
            ('kryten@example.org', 'kRyTeN@example.org', False),
            ('kryten@example.org', 'křýťěň@example.org', False),
            ('kryten@example.org', 'kryten@éxámpľě.org', False),
            ('kryten@example.org', 'kryten+bucket@example.org', False),
            ('kryten@example.org', 'kryten@example.com', False),
            ('kryten@example.org', 'rimmer@example.org', False),
        )
        for first, second, result in data:
            with self.subTest(provided=first, reference=second):
                contact_data = ContactData(emails=[first])
                self.assertEqual(contact_data._compare_emails([second]), result)

            with self.subTest(provided=second, reference=first):
                contact_data = ContactData(emails=[second])
                self.assertEqual(contact_data._compare_emails([first]), result)

    def test_compare_emails_multiple(self):
        # Test email comparison for email sets.
        data: Tuple[Tuple[List[str], List[str], bool], ...] = (
            (['kryten@example.org'], [], False),
            (['kryten@example.org'], ['kryten@example.org', 'kryten@example.org'], True),
            (['kryten@example.org'], ['kryten@example.org', 'rimmer@example.org'], True),
            (['rimmer@example.org'], ['kryten@example.org', 'rimmer@example.org'], True),
            (['kryten@example.org'], ['arnold@example.org', 'rimmer@example.org'], False),
            (['kryten@example.org', 'kryten@example.org'], ['kryten@example.org'], True),
            (['kryten@example.org', 'rimmer@example.org'], ['kryten@example.org'], False),
        )
        for provided, reference, result in data:
            with self.subTest(provided=provided, reference=reference):
                contact_data = ContactData(emails=provided)
                self.assertEqual(contact_data._compare_emails(reference), result)

    def test_compare_address(self):
        data: Tuple[Tuple[Dict[str, Any], Dict[str, Any], bool], ...] = (
            ({'street': []}, {'street': None}, True),
            ({'street': ['Deck 16']}, {'street': ['Deck 16']}, True),
            ({'street': ['Deck 16']}, {'street': ['Deck\n\t16']}, True),
            ({'street': ['Deck 16']}, {'street': ['DECK 16']}, True),
            ({'street': ['Deck 16']}, {'street': ['Ďéčk 16']}, True),
            ({'street': ['Deck 16']}, {'street': ['Deck', '16']}, False),
            ({'street': ['Deck', '16']}, {'street': ['Deck', '16']}, True),
            ({'street': ['Deck', '16']}, {'street': ['16', 'Deck']}, False),
            ({'street': ['Deck', '16']}, {'street': ['Deck']}, False),
            ({'city': ''}, {'city': None}, True),
            ({'city': 'Red Dwarf'}, {'city': 'Red Dwarf'}, True),
            ({'city': 'Red Dwarf'}, {'city': 'Red\n\tDwarf'}, True),
            ({'city': 'Red Dwarf'}, {'city': 'RED DWARF'}, True),
            ({'city': 'Red Dwarf'}, {'city': 'Red Ďwářf'}, True),
            ({'city': 'Red Dwarf'}, {'city': 'Nova 5'}, False),
            ({'state': ''}, {'state': None}, True),
            ({'state': 'Deep Space'}, {'state': 'Deep Space'}, True),
            ({'state': 'Deep Space'}, {'state': 'Deep\n\tSpace'}, True),
            ({'state': 'Deep Space'}, {'state': 'DEEP SPACE'}, True),
            ({'state': 'Deep Space'}, {'state': 'Deep Špáčé'}, True),
            ({'state': 'Deep Space'}, {'state': 'Earth'}, False),
            ({'postal_code': ''}, {'postal_code': None}, True),
            ({'postal_code': 'JMC'}, {'postal_code': 'JMC'}, True),
            ({'postal_code': 'JMC'}, {'postal_code': 'J M C'}, True),  # Ignore whitespaces completely.
            ({'postal_code': 'JMC'}, {'postal_code': 'JMC\n\r\t'}, True),
            ({'postal_code': 'JMC'}, {'postal_code': 'jmc'}, True),
            ({'postal_code': 'JMC'}, {'postal_code': 'JMČ'}, True),
            ({'postal_code': 'JMC'}, {'postal_code': 'DDI'}, False),
            ({'country_code': ''}, {'country_code': None}, True),
            ({'country_code': 'JU'}, {'country_code': 'JU'}, True),
            ({'country_code': 'JU'}, {'country_code': 'JU\n\t'}, True),
            ({'country_code': 'JU'}, {'country_code': 'ju'}, True),
            ({'country_code': 'JU'}, {'country_code': 'JŮ'}, True),
            ({'country_code': 'JU'}, {'country_code': 'EA'}, False),
            ({'street': ['Deck 16'], 'city': 'Red Dwarf', 'state': 'Deep Space', 'postal_code': 'JMC',
              'country_code': 'JU'},
             {'street': ['Deck 16'], 'city': 'Red Dwarf', 'state': 'Deep Space', 'postal_code': 'JMC',
              'country_code': 'JU'},
             True),
            ({'city': 'Red Dwarf'}, {'state': 'Red Dwarf'}, False),  # different parts of address
        )
        for first, second, result in data:
            with self.subTest(provided=first, reference=second):
                contact_data = ContactData(**{'address_' + n: v for n, v in first.items()})
                self.assertEqual(contact_data._compare_address(**second), result)

            with self.subTest(provided=second, reference=first):
                contact_data = ContactData(**{'address_' + n: v for n, v in second.items()})
                self.assertEqual(contact_data._compare_address(**first), result)

    async def test_verify_name(self):
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = Contact(
            contact_id=self.contact_id, contact_handle='2X4B', contact_history_id=self.contact_history_id,
            name='Kryten')
        verification = await make_verification(self.session, contact_id=self.contact_id)
        contact_data = ContactData(verification_id=verification.id, name='Kryten')
        self.session.add(contact_data)

        await contact_data.verify()

        self.assertEqual(contact_data.contact_history_id, self.contact_history_id)
        self.assertTrue(contact_data.verified_name)
        self.assertIsNone(contact_data.verified_organization)
        self.assertIsNone(contact_data.verified_birthdate)
        self.assertIsNone(contact_data.verified_company_registration_number)
        self.assertIsNone(contact_data.verified_emails)
        self.assertIsNone(contact_data.verified_address)

    async def test_verify_organization(self):
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = Contact(
            contact_id=self.contact_id, contact_handle='2X4B', contact_history_id=self.contact_history_id,
            organization='JMC')
        verification = await make_verification(self.session, contact_id=self.contact_id)
        contact_data = ContactData(verification_id=verification.id, organization='JMC')
        self.session.add(contact_data)

        await contact_data.verify()

        self.assertEqual(contact_data.contact_history_id, self.contact_history_id)
        self.assertIsNone(contact_data.verified_name)
        self.assertTrue(contact_data.verified_organization)
        self.assertIsNone(contact_data.verified_birthdate)
        self.assertIsNone(contact_data.verified_company_registration_number)
        self.assertIsNone(contact_data.verified_emails)
        self.assertIsNone(contact_data.verified_address)

    async def _test_verify_birthdate(self, ident: Optional[AdditionalIdentifier], result: bool) -> None:
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = Contact(
            contact_id=self.contact_id, contact_handle='2X4B', contact_history_id=self.contact_history_id,
            additional_identifier=ident)
        verification = await make_verification(self.session, contact_id=self.contact_id)
        contact_data = ContactData(verification_id=verification.id, birthdate=date(1988, 9, 6))
        self.session.add(contact_data)

        await contact_data.verify()

        self.assertEqual(contact_data.contact_history_id, self.contact_history_id)
        self.assertIsNone(contact_data.verified_name)
        self.assertIsNone(contact_data.verified_organization)
        self.assertIs(contact_data.verified_birthdate, result)
        self.assertIsNone(contact_data.verified_company_registration_number)
        self.assertIsNone(contact_data.verified_emails)
        self.assertIsNone(contact_data.verified_address)

    async def test_verify_birthdate(self):
        ident = AdditionalIdentifier(type=ContactIdentifierType.BIRTHDATE, value='1988-09-06')
        await self._test_verify_birthdate(ident, True)

    async def test_verify_birthdate_invalid_type(self):
        ident = AdditionalIdentifier(type=ContactIdentifierType.PASSPORT_NUMBER, value='1988-09-06')
        await self._test_verify_birthdate(ident, False)

    async def test_verify_birthdate_invalid_value(self):
        ident = AdditionalIdentifier(type=ContactIdentifierType.BIRTHDATE, value='JUNK')
        await self._test_verify_birthdate(ident, False)

    async def test_verify_birthdate_mismatch(self):
        ident = AdditionalIdentifier(type=ContactIdentifierType.BIRTHDATE, value='1988-09-07')
        await self._test_verify_birthdate(ident, False)

    async def test_verify_birthdate_no_ident(self):
        await self._test_verify_birthdate(None, False)

    async def _test_verify_company_registration_number(self, ident: Optional[AdditionalIdentifier], result: bool,
                                                       ) -> None:
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = Contact(
            contact_id=self.contact_id, contact_handle='2X4B', contact_history_id=self.contact_history_id,
            additional_identifier=ident)
        verification = await make_verification(self.session, contact_id=self.contact_id)
        contact_data = ContactData(verification_id=verification.id, company_registration_number='JMC')
        self.session.add(contact_data)

        await contact_data.verify()

        self.assertEqual(contact_data.contact_history_id, self.contact_history_id)
        self.assertIsNone(contact_data.verified_name)
        self.assertIsNone(contact_data.verified_organization)
        self.assertIsNone(contact_data.verified_birthdate)
        self.assertIs(contact_data.verified_company_registration_number, result)
        self.assertIsNone(contact_data.verified_emails)
        self.assertIsNone(contact_data.verified_address)

    async def test_verify_company_registration_number(self):
        ident = AdditionalIdentifier(type=ContactIdentifierType.COMPANY_REGISTRATION_NUMBER, value='JMC')
        await self._test_verify_company_registration_number(ident, True)

    async def test_verify_company_registration_number_invalid_type(self):
        ident = AdditionalIdentifier(type=ContactIdentifierType.PASSPORT_NUMBER, value='JMC')
        await self._test_verify_company_registration_number(ident, False)

    async def test_verify_company_registration_number_mismatch(self):
        ident = AdditionalIdentifier(type=ContactIdentifierType.COMPANY_REGISTRATION_NUMBER, value='Diva Droid')
        await self._test_verify_company_registration_number(ident, False)

    async def test_verify_company_registration_number_no_ident(self):
        await self._test_verify_company_registration_number(None, False)

    async def test_verify_emails(self):
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = Contact(
            contact_id=self.contact_id, contact_handle='2X4B', contact_history_id=self.contact_history_id,
            emails=['kryten@example.org'])
        verification = await make_verification(self.session, contact_id=self.contact_id)
        contact_data = ContactData(verification_id=verification.id, emails=['kryten@example.org'])
        self.session.add(contact_data)

        await contact_data.verify()

        self.assertEqual(contact_data.contact_history_id, self.contact_history_id)
        self.assertIsNone(contact_data.verified_name)
        self.assertIsNone(contact_data.verified_organization)
        self.assertIsNone(contact_data.verified_birthdate)
        self.assertIsNone(contact_data.verified_company_registration_number)
        self.assertTrue(contact_data.verified_emails)
        self.assertIsNone(contact_data.verified_address)

    async def test_verify_address(self):
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = Contact(
            contact_id=self.contact_id, contact_handle='2X4B', contact_history_id=self.contact_history_id,
            place=Address(street=['Deck 16'], city='Red Dwarf'))
        verification = await make_verification(self.session, contact_id=self.contact_id)
        contact_data = ContactData(verification_id=verification.id, address_street=['Deck 16'],
                                   address_city='Red Dwarf')
        self.session.add(contact_data)

        await contact_data.verify()

        self.assertEqual(contact_data.contact_history_id, self.contact_history_id)
        self.assertIsNone(contact_data.verified_name)
        self.assertIsNone(contact_data.verified_organization)
        self.assertIsNone(contact_data.verified_birthdate)
        self.assertIsNone(contact_data.verified_company_registration_number)
        self.assertIsNone(contact_data.verified_emails)
        self.assertTrue(contact_data.verified_address)

    async def test_verify_address_empty(self):
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = Contact(
            contact_id=self.contact_id, contact_handle='2X4B', contact_history_id=self.contact_history_id)
        verification = await make_verification(self.session, contact_id=self.contact_id)
        contact_data = ContactData(verification_id=verification.id, address_street=['Deck 16'],
                                   address_city='Red Dwarf')
        self.session.add(contact_data)

        await contact_data.verify()

        self.assertEqual(contact_data.contact_history_id, self.contact_history_id)
        self.assertIsNone(contact_data.verified_name)
        self.assertIsNone(contact_data.verified_organization)
        self.assertIsNone(contact_data.verified_birthdate)
        self.assertIsNone(contact_data.verified_company_registration_number)
        self.assertIsNone(contact_data.verified_emails)
        self.assertFalse(contact_data.verified_address)
