from datetime import datetime
from typing import Any, AsyncIterator, Callable, Dict, List, Optional, Sequence, Tuple, Type, cast
from unittest import IsolatedAsyncioTestCase
from unittest.mock import AsyncMock, _Call, call, patch, sentinel
from uuid import UUID

import requests
import responses
from asyncstdlib import iter as aiter, list as alist
from fred_types import ContactHistoryId, ContactId, ContactRef, DomainId, DomainRef
from freezegun import freeze_time
from hermes import (
    Email,
    EmailEnvelope,
    EmailMessengerClient,
    LetterMessengerClient,
    MessageEnvelope,
    SmsMessengerClient,
)
from hermes.message import Reference
from pytz import UTC
from regal import Contact, ContactClient, Domain, DomainAdminClient, DomainClient
from regal.domain import DomainContactRole
from regal.exceptions import ContactDoesNotExist
from responses.matchers import header_matcher, query_param_matcher, request_kwargs_matcher
from sqlalchemy import select
from testfixtures import Comparison

from verification.constants import EventType, MessageType, ObjectState, Template, TransportMethod, VerificationStatus
from verification.handler import VerificationHandler
from verification.models import Event, Token, Verification
from verification.processors.base import (
    BaseProcessor,
    CancelFutileEvent,
    CancelOutzoneEvent,
    CancelVerificationEvent,
    CheckDomainsEvent,
    CompleteEvent,
    DelayFutileEvent,
    DelayOutzoneEvent,
    FailedEvent,
    FutileEvent,
    FutileReminderEvent,
    InitialEmailEvent,
    NotifyCancelEvent,
    NotifyCompleteEvent,
    NotifyContactsEvent,
    NotifyFailedEvent,
    OutzoneEvent,
    OutzoneReminderEvent,
    ScheduledEvent,
    SendMessageBatchEvent,
    SendMessageEvent,
)
from verification.settings import APPLICATION

from ..constants import ReferenceType
from .utils import (
    TransactionTestCase,
    make_contact_data,
    make_domain,
    make_event,
    make_verification,
    mock_awaits,
    override_settings,
)


class TestScheduledEvent(ScheduledEvent):
    """Simple event for testing ScheduledEvent."""

    type = EventType.CANCEL

    async def process(self) -> None:  # pragma: no cover
        pass


class ScheduledEventTest(IsolatedAsyncioTestCase):
    def test_repr(self):
        data = (
            (TestScheduledEvent(sentinel.verification),
             '<TestScheduledEvent sentinel.verification method=None, due_datetime=None, details={}>'),
            (TestScheduledEvent(sentinel.verification, method=sentinel.method, due_datetime=sentinel.due_datetime,
                                details=sentinel.details),
             ('<TestScheduledEvent sentinel.verification method=sentinel.method, due_datetime=sentinel.due_datetime, '
              'details=sentinel.details>')),
        )
        for scheduled, repr_str in data:
            with self.subTest(scheduled=scheduled):
                self.assertEqual(repr(scheduled), repr_str)

    def test_equality(self):
        data = (
            # scheduled, other, equal
            (TestScheduledEvent(sentinel.verification), TestScheduledEvent(sentinel.verification), True),
            (TestScheduledEvent(sentinel.verification), TestScheduledEvent(sentinel.other), False),
            # Same event type, but different class.
            (TestScheduledEvent(sentinel.verification), CancelVerificationEvent(sentinel.verification), False),

            (TestScheduledEvent(sentinel.verification, method=sentinel.method),
             TestScheduledEvent(sentinel.verification, method=sentinel.method), True),
            (TestScheduledEvent(sentinel.verification, method=sentinel.method),
             TestScheduledEvent(sentinel.other, method=sentinel.method), False),
            (TestScheduledEvent(sentinel.verification, method=sentinel.method),
             TestScheduledEvent(sentinel.verification), False),
            (TestScheduledEvent(sentinel.verification, method=sentinel.method),
             TestScheduledEvent(sentinel.verification, method=sentinel.other), False),

            (TestScheduledEvent(sentinel.verification, due_datetime=sentinel.due),
             TestScheduledEvent(sentinel.verification, due_datetime=sentinel.due), True),
            (TestScheduledEvent(sentinel.verification, due_datetime=sentinel.due),
             TestScheduledEvent(sentinel.other, due_datetime=sentinel.due), False),
            (TestScheduledEvent(sentinel.verification, due_datetime=sentinel.due),
             TestScheduledEvent(sentinel.verification, due_datetime=sentinel.other), False),
            (TestScheduledEvent(sentinel.verification, due_datetime=sentinel.due),
             TestScheduledEvent(sentinel.verification), False),

            (TestScheduledEvent(sentinel.verification, details=sentinel.details),
             TestScheduledEvent(sentinel.verification, details=sentinel.details), True),
            (TestScheduledEvent(sentinel.verification, details=sentinel.details),
             TestScheduledEvent(sentinel.other, details=sentinel.details), False),
            (TestScheduledEvent(sentinel.verification, details=sentinel.details),
             TestScheduledEvent(sentinel.verification, details=sentinel.other), False),
            (TestScheduledEvent(sentinel.verification, details=sentinel.details),
             TestScheduledEvent(sentinel.verification), False),
            # Test utilities
            (TestScheduledEvent(sentinel.verification), Comparison(TestScheduledEvent), True),
        )
        for scheduled, other, equal in data:
            with self.subTest(scheduled=scheduled, other=other):
                self.assertEqual(scheduled == other, equal)
                self.assertEqual(scheduled != other, not equal)

    async def test_get_details(self):
        verification = Verification(id=sentinel.id)
        data: Tuple[Tuple[TestScheduledEvent, dict], ...] = (
            # scheduled, details
            (TestScheduledEvent(verification), {}),
            (TestScheduledEvent(verification, details={'key': sentinel.value}), {'key': sentinel.value}),
        )
        for scheduled, details in data:
            with self.subTest(scheduled=scheduled):
                self.assertEqual(await scheduled.get_details(), details)

    async def test_get_details_copy(self):
        # Test get_details result is a copy of the scheduled event details.
        verification = Verification(id=sentinel.id)
        scheduled = TestScheduledEvent(verification, details={'key': sentinel.value})

        # Modify the get_details() result
        details = await scheduled.get_details()
        details['update'] = sentinel.value

        # Check the scheduled event details are intact by the modification
        self.assertEqual(await scheduled.get_details(), {'key': sentinel.value})

    async def test_to_event(self):
        verification = Verification(id=sentinel.id)
        scheduled = TestScheduledEvent(verification)

        event = await scheduled.to_event()

        self.assertEqual(event.type, EventType.CANCEL)
        self.assertEqual(event.verification_id, sentinel.id)
        self.assertIsNone(event.method)
        self.assertEqual(event.details, {})

    async def test_to_event_method(self):
        verification = Verification(id=sentinel.id)
        scheduled = TestScheduledEvent(verification, method=sentinel.method)

        event = await scheduled.to_event()

        self.assertEqual(event.type, EventType.CANCEL)
        self.assertEqual(event.verification_id, sentinel.id)
        self.assertEqual(event.method, sentinel.method)
        self.assertEqual(event.details, {})

    async def test_to_event_details(self):
        verification = Verification(id=sentinel.id)
        scheduled = TestScheduledEvent(verification, details={'key': sentinel.value})

        event = await scheduled.to_event()

        self.assertEqual(event.type, EventType.CANCEL)
        self.assertEqual(event.verification_id, sentinel.id)
        self.assertIsNone(event.method)
        self.assertEqual(event.details, {'key': sentinel.value})


class TestSendMessageEvent(SendMessageEvent):
    """Simple event for testing SendMessageEvent."""

    message_type = sentinel.message_type

    # Add a `_method` attribute for testing purposes.
    def __init__(self, *args: Any, _method: TransportMethod, _status: Optional[VerificationStatus] = None,
                 **kwargs: Any):
        super().__init__(*args, **kwargs)
        self._method = _method
        self.status = _status

    @property
    def transport_method(self) -> TransportMethod:
        return self._method

    async def _get_message(self):
        return sentinel.message


class SendMessageEventTest(TransactionTestCase):
    contact_id = 'KRYTEN'
    uuid = UUID(int=42)

    async def asyncSetUp(self):
        APPLICATION.email_client = AsyncMock(EmailMessengerClient)
        APPLICATION.letter_client = AsyncMock(LetterMessengerClient)
        APPLICATION.sms_client = AsyncMock(SmsMessengerClient)

    async def test_get_context(self):
        initial_context = {'context': sentinel.context}
        scheduled = TestSendMessageEvent(
            sentinel.verification,
            context=initial_context,
            _method=sentinel.method)

        context = await scheduled.get_context()
        self.assertEqual(context, {'context': sentinel.context})
        # Check context is a different instance.
        self.assertIsNot(context, initial_context)

    async def test_get_details(self):
        scheduled = TestSendMessageEvent(sentinel.verification, _method=sentinel.method)

        self.assertEqual(await scheduled.get_details(),
                         {'method': sentinel.method, 'message_type': sentinel.message_type})

    async def _test_process(self, method: TransportMethod, *, email_calls: Optional[List[_Call]] = None,
                            letter_calls: Optional[List[_Call]] = None, sms_calls: Optional[List[_Call]] = None,
                            status: Optional[VerificationStatus] = None, resend: bool = False,
                            final_status: VerificationStatus = VerificationStatus.CREATED) -> None:
        verification = await make_verification(self.session, contact_id=self.contact_id, uuid=self.uuid)
        scheduled = TestSendMessageEvent(verification, details={'recipient': sentinel.recipient}, _method=method,
                                         _status=status, resend=resend)

        await scheduled.process()

        details = {'method': method, 'message_type': sentinel.message_type, 'recipient': sentinel.recipient,
                   'message_id': sentinel.message_id}
        self.assertEqual(await scheduled.get_details(), details)
        # Check final verification status.
        self.assertEqual(verification.status, final_status)
        # Check messenger calls
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.email_client)), email_calls or [])
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.letter_client)), letter_calls or [])
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.sms_client)), sms_calls or [])

    async def test_process_email(self):
        cast(AsyncMock, APPLICATION.email_client.send).return_value = sentinel.message_id
        calls = [call.send(sentinel.message,
                           references=[Reference(type=ReferenceType.CONTACT, value=self.contact_id),
                                       Reference(type=ReferenceType.VERIFICATION, value=str(self.uuid))])]
        await self._test_process(TransportMethod.EMAIL, email_calls=calls)

    async def test_process_letter(self):
        cast(AsyncMock, APPLICATION.letter_client.send).return_value = sentinel.message_id
        calls = [call.send(sentinel.message,
                           references=[Reference(type=ReferenceType.CONTACT, value=self.contact_id),
                                       Reference(type=ReferenceType.VERIFICATION, value=str(self.uuid))])]
        await self._test_process(TransportMethod.LETTER, letter_calls=calls)

    async def test_process_sms(self):
        cast(AsyncMock, APPLICATION.sms_client.send).return_value = sentinel.message_id
        calls = [call.send(sentinel.message,
                           references=[Reference(type=ReferenceType.CONTACT, value=self.contact_id),
                                       Reference(type=ReferenceType.VERIFICATION, value=str(self.uuid))])]
        await self._test_process(TransportMethod.SMS, sms_calls=calls)

    async def test_process_status(self):
        # Test change of status
        cast(AsyncMock, APPLICATION.email_client.send).return_value = sentinel.message_id
        calls = [call.send(sentinel.message,
                           references=[Reference(type=ReferenceType.CONTACT, value=self.contact_id),
                                       Reference(type=ReferenceType.VERIFICATION, value=str(self.uuid))])]
        await self._test_process(TransportMethod.EMAIL, email_calls=calls, status=VerificationStatus.PENDING,
                                 final_status=VerificationStatus.PENDING)

    async def test_process_status_resend(self):
        # Test status is not changed on resend messages.
        cast(AsyncMock, APPLICATION.email_client.send).return_value = sentinel.message_id
        calls = [call.send(sentinel.message,
                           references=[Reference(type=ReferenceType.CONTACT, value=self.contact_id),
                                       Reference(type=ReferenceType.VERIFICATION, value=str(self.uuid))])]
        await self._test_process(TransportMethod.EMAIL, email_calls=calls, status=VerificationStatus.PENDING,
                                 resend=True)


class TestSendMessageBatchEvent(SendMessageBatchEvent):
    """Simple event for testing SendMessageEvent."""

    message_type = sentinel.message_type

    # Add a `_method` attribute for testing purposes.
    def __init__(self, *args: Any, _status: Optional[VerificationStatus] = None, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.status = _status

    async def _get_envelopes(self):
        return [sentinel.envelope, sentinel.envelope2]


class SendMessageBatchEventTest(TransactionTestCase):
    uuid = UUID(int=42)

    async def asyncSetUp(self):
        APPLICATION.email_client = AsyncMock(EmailMessengerClient)

    async def test_get_context(self):
        initial_context = {'context': sentinel.context}
        scheduled = TestSendMessageBatchEvent(
            sentinel.verification,
            context=initial_context)

        context = await scheduled.get_context()
        self.assertEqual(context, {'context': sentinel.context})
        # Check context is a different instance.
        self.assertIsNot(context, initial_context)

    async def test_get_details(self):
        scheduled = TestSendMessageBatchEvent(sentinel.verification)

        self.assertEqual(await scheduled.get_details(),
                         {'method': TransportMethod.EMAIL, 'message_type': sentinel.message_type})

    async def _test_process(self, *, status: Optional[VerificationStatus] = None, resend: bool = False,
                            final_status: VerificationStatus = VerificationStatus.CREATED) -> None:
        cast(AsyncMock, APPLICATION.email_client.batch_send).return_value = aiter(
            [sentinel.message_id, sentinel.message_id2])

        verification = await make_verification(self.session, contact_id='KRYTEN', uuid=self.uuid)
        scheduled = TestSendMessageBatchEvent(verification, details={'recipient': sentinel.recipient},
                                              _status=status, resend=resend)

        await scheduled.process()

        details = {'method': TransportMethod.EMAIL, 'message_type': sentinel.message_type,
                   'recipient': sentinel.recipient, 'message_ids': [sentinel.message_id, sentinel.message_id2]}
        self.assertEqual(await scheduled.get_details(), details)
        # Check final verification status.
        self.assertEqual(verification.status, final_status)
        # Check messenger calls
        calls = [call.batch_send([sentinel.envelope, sentinel.envelope2])]
        self.assertEqual(cast(AsyncMock, APPLICATION.email_client).mock_calls, calls)

    async def test_process(self):
        await self._test_process()

    async def test_process_status(self):
        # Test change of status
        await self._test_process(status=VerificationStatus.PENDING, final_status=VerificationStatus.PENDING)

    async def test_process_status_resend(self):
        # Test status is not changed on resend messages.
        await self._test_process(status=VerificationStatus.PENDING, resend=True)


class TestNotifyContactsEvent(NotifyContactsEvent):
    """Simple event for testing NotifyContactsEvent."""

    # Ignore that there aren't enums used in here.
    message_type = 'quagaars'  # type: ignore[assignment]
    subject_template = 'subject.txt'  # type: ignore[assignment]
    body_template = 'body.txt'  # type: ignore[assignment]
    admin_body_template = 'admin-body.txt'  # type: ignore[assignment]


class NotifyContactsEventTest(TransactionTestCase):
    uuid = UUID(int=42)
    contact_id = ContactId('KRYTEN')
    handle = '2X4B'
    email = 'kryten@example.org'

    async def asyncSetUp(self):
        APPLICATION.contact_client = AsyncMock(ContactClient)
        APPLICATION.domain_client = AsyncMock(DomainClient)
        APPLICATION.email_client = AsyncMock(EmailMessengerClient)

        # Patch secrets to set access_token
        secrets_patcher = patch('secrets.choice', return_value='9')
        self.addCleanup(secrets_patcher.stop)
        secrets_patcher.start()

    async def test_get_context(self):
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = Contact(
            contact_id=self.contact_id, contact_handle=self.handle)

        initial_context = {'context': sentinel.context}
        verification = await make_verification(self.session, contact_id=self.contact_id, uuid=self.uuid)
        scheduled = TestNotifyContactsEvent(verification, context=initial_context)

        context = await scheduled.get_context()

        self.assertEqual(context, {'context': sentinel.context, 'uuid': str(self.uuid), 'handle': self.handle})

    async def test_get_details(self):
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = Contact(
            contact_id=self.contact_id, contact_handle=self.handle, emails=[self.email])

        verification = await make_verification(self.session, contact_id=self.contact_id, uuid=self.uuid)
        scheduled = TestNotifyContactsEvent(verification)

        self.assertEqual(await scheduled.get_details(),
                         {'method': TransportMethod.EMAIL, 'message_type': 'quagaars', 'recipient': [self.email]})
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.contact_client)),
                         [call.get_contact_info(self.contact_id)])

    async def test_get_details_recipient(self):
        scheduled = TestNotifyContactsEvent(sentinel.verification, details={'recipient': sentinel.recipient})

        self.assertEqual(await scheduled.get_details(),
                         {'method': TransportMethod.EMAIL, 'message_type': 'quagaars', 'recipient': sentinel.recipient})
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.contact_client)), [])

    def _get_envelope(self, recipient: str, contact_id: ContactId, context: Dict[str, Any], template: str = 'body.txt',
                      ) -> EmailEnvelope:
        message = Email(recipient=recipient, type='quagaars', subject_template='subject.txt', body_template=template,
                        context=context)
        references = [Reference(type=ReferenceType.CONTACT, value=contact_id),
                      Reference(type=ReferenceType.VERIFICATION, value=str(self.uuid))]
        return EmailEnvelope(message_data=message, references=references)

    async def _test_process(self, recipient: Any, envelopes: Sequence[MessageEnvelope], *, resend: bool = False
                            ) -> None:
        cast(AsyncMock, APPLICATION.email_client.batch_send).return_value = aiter([sentinel.message_id])

        verification = await make_verification(self.session, contact_id=self.contact_id, uuid=self.uuid)
        scheduled = TestNotifyContactsEvent(verification, resend=resend)

        await scheduled.process()

        details = {'method': TransportMethod.EMAIL, 'message_type': 'quagaars',
                   'recipient': recipient, 'message_ids': [sentinel.message_id]}
        self.assertEqual(await scheduled.get_details(), details)
        # Check messenger calls
        calls = [call.batch_send(envelopes)]
        self.assertEqual(cast(AsyncMock, APPLICATION.email_client).mock_calls, calls)

    async def test_process(self):
        # Test with a single email
        email = 'kryten@example.org'
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = Contact(
            contact_id=self.contact_id, contact_handle=self.handle, emails=[email])
        cast(AsyncMock, APPLICATION.domain_client.get_domains_by_contact).return_value = {'domains': []}

        context = {'uuid': str(self.uuid), 'handle': self.handle, 'access_token': '9999999999', 'domains': []}
        envelopes = [self._get_envelope(email, self.contact_id, context)]
        await self._test_process([email], envelopes)

        # Check registry calls
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.contact_client)),
                         [call.get_contact_info(self.contact_id)])
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.domain_client)),
                         [call.get_domains_by_contact(self.contact_id, {DomainContactRole.HOLDER})])

    async def test_process_multi_mails(self):
        # Test with a multiple emails
        emails = ['kryten@example.org', 'rimmer@example.org']
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = Contact(
            contact_id=self.contact_id, contact_handle=self.handle, emails=emails)
        cast(AsyncMock, APPLICATION.domain_client.get_domains_by_contact).return_value = {'domains': []}

        context = {'uuid': str(self.uuid), 'handle': self.handle, 'access_token': '9999999999', 'domains': []}
        envelopes = [self._get_envelope('kryten@example.org, rimmer@example.org', self.contact_id, context)]
        await self._test_process(emails, envelopes)

        # Check registry calls
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.contact_client)),
                         [call.get_contact_info(self.contact_id)])
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.domain_client)),
                         [call.get_domains_by_contact(self.contact_id, {DomainContactRole.HOLDER})])

    async def test_process_domains(self):
        # Test 'domains' context
        email = 'kryten@example.org'
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = Contact(
            contact_id=self.contact_id, contact_handle=self.handle, emails=[email])
        cast(AsyncMock, APPLICATION.domain_client.get_domains_by_contact).return_value = {
            'domains': [{'domain': DomainRef(id=DomainId('42'), fqdn='kryten.example.org')},
                        {'domain': DomainRef(id=DomainId('43'), fqdn='rimmer.example.org')}]}
        cast(AsyncMock, APPLICATION.domain_client.get_domain_info).side_effect = [
            Domain(domain_id=DomainId('42'), fqdn='kryten.example.org'),
            Domain(domain_id=DomainId('43'), fqdn='rimmer.example.org')]

        context = {'uuid': str(self.uuid), 'handle': self.handle, 'access_token': '9999999999',
                   'domains': ['kryten.example.org', 'rimmer.example.org']}
        envelopes = [self._get_envelope(email, self.contact_id, context)]
        await self._test_process([email], envelopes)

        # Check registry calls
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.contact_client)),
                         [call.get_contact_info(self.contact_id)])
        domain_calls = [call.get_domains_by_contact(self.contact_id, {DomainContactRole.HOLDER}),
                        call.get_domain_info(DomainId('42')), call.get_domain_info(DomainId('43'))]
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.domain_client)), domain_calls)

    async def test_process_admin(self):
        # Test email send to a domain admin contact.
        email = 'kryten@example.org'
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).side_effect = [
            Contact(contact_id=self.contact_id, contact_handle=self.handle, emails=[email]),
            Contact(contact_id=ContactId('196'), contact_handle='RIMMER', emails=['rimmer@example.org'])]
        cast(AsyncMock, APPLICATION.domain_client.get_domains_by_contact).return_value = {
            'domains': [{'domain': DomainRef(id=DomainId('42'), fqdn='kryten.example.org')},
                        {'domain': DomainRef(id=DomainId('43'), fqdn='rimmer.example.org')}]}
        cast(AsyncMock, APPLICATION.domain_client.get_domain_info).side_effect = [
            Domain(domain_id=DomainId('42'), fqdn='kryten.example.org'),
            Domain(domain_id=DomainId('43'), fqdn='rimmer.example.org',
                   administrative_contacts=[ContactRef(id=ContactId('196'), handle='RIMMER')])]

        context = {'uuid': str(self.uuid), 'handle': self.handle, 'access_token': '9999999999',
                   'domains': ['kryten.example.org', 'rimmer.example.org']}
        context2 = {'uuid': str(self.uuid), 'handle': self.handle, 'domains': ['rimmer.example.org']}
        envelopes = [self._get_envelope(email, self.contact_id, context),
                     self._get_envelope('rimmer@example.org', ContactId('196'), context2, template='admin-body.txt')]
        await self._test_process([email], envelopes)

        # Check registry calls
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.contact_client)),
                         [call.get_contact_info(self.contact_id), call.get_contact_info(ContactId('196'))])
        domain_calls = [call.get_domains_by_contact(self.contact_id, {DomainContactRole.HOLDER}),
                        call.get_domain_info(DomainId('42')), call.get_domain_info(DomainId('43'))]
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.domain_client)), domain_calls)

    async def test_process_resend(self):
        # Test status is not changed on resend messages.
        email = 'kryten@example.org'
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = Contact(
            contact_id=self.contact_id, contact_handle=self.handle, emails=[email])
        cast(AsyncMock, APPLICATION.domain_client.get_domains_by_contact).return_value = {
            'domains': [{'domain': DomainRef(id=DomainId('42'), fqdn='kryten.example.org')}]}

        context = {'uuid': str(self.uuid), 'handle': self.handle, 'access_token': '9999999999',
                   'domains': ['kryten.example.org']}
        envelopes = [self._get_envelope(email, self.contact_id, context)]
        await self._test_process([email], envelopes, resend=True)

        # Check registry calls
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.contact_client)),
                         [call.get_contact_info(self.contact_id)])
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.domain_client)),
                         [call.get_domains_by_contact(self.contact_id, {DomainContactRole.HOLDER})])


class InitialEmailEventTest(TransactionTestCase):
    uuid = UUID(int=42)
    contact_id = ContactId('KRYTEN')
    contact_history_id = ContactHistoryId('42')
    handle = '2X4B'
    email = 'kryten@example.org'

    async def asyncSetUp(self):
        APPLICATION.contact_client = AsyncMock(ContactClient)
        APPLICATION.domain_client = AsyncMock(DomainClient)

        APPLICATION.email_client = AsyncMock(EmailMessengerClient)
        cast(AsyncMock, APPLICATION.email_client.batch_send).return_value = aiter([sentinel.message_id])
        # Patch secrets to set access_token
        secrets_patcher = patch('secrets.choice', return_value='9')
        self.addCleanup(secrets_patcher.stop)
        secrets_patcher.start()

    def _get_envelope(self, recipient: str, contact_id: ContactId, context: Dict[str, Any],
                      template: Template = Template.INITIAL_EMAIL_BODY) -> EmailEnvelope:
        message = Email(recipient=recipient, type=MessageType.INITIAL_EMAIL,
                        subject_template=Template.INITIAL_EMAIL_SUBJECT, body_template=template, context=context)
        references = [Reference(type=ReferenceType.CONTACT, value=contact_id),
                      Reference(type=ReferenceType.VERIFICATION, value=str(self.uuid))]
        return EmailEnvelope(message_data=message, references=references)

    async def _test_process(self, contact_calls: List[_Call], *, resend: bool = False,
                            status: VerificationStatus = VerificationStatus.INITIALIZED) -> None:
        # Test with a single email
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = Contact(
            contact_id=self.contact_id, contact_handle=self.handle, contact_history_id=self.contact_history_id,
            emails=[self.email])
        cast(AsyncMock, APPLICATION.domain_client.get_domains_by_contact).return_value = {'domains': []}

        verification = await make_verification(self.session, contact_id=self.contact_id, uuid=self.uuid)
        scheduled = InitialEmailEvent(verification, details={'recipient': [self.email]}, resend=resend)

        await scheduled.process()

        details = {'method': TransportMethod.EMAIL, 'message_type': 'verification_initial',
                   'recipient': [self.email], 'message_ids': [sentinel.message_id]}
        self.assertEqual(await scheduled.get_details(), details)
        # Check final verification status.
        self.assertEqual(verification.status, status)
        # Check messenger calls
        context = {'uuid': str(self.uuid), 'access_token': '9999999999', 'handle': self.handle, 'domains': []}
        envelopes = [self._get_envelope(self.email, self.contact_id, context)]
        self.assertEqual(cast(AsyncMock, APPLICATION.email_client).mock_calls, [call.batch_send(envelopes)])
        # Check registry calls
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.contact_client)), contact_calls)
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.domain_client)),
                         [call.get_domains_by_contact(self.contact_id, {DomainContactRole.HOLDER})])

    async def test_process(self):
        flags = {ObjectState.CONTACT_IN_MANUAL_VERIFICATION: True,
                 ObjectState.CONTACT_PASSED_MANUAL_VERIFICATION: False,
                 ObjectState.CONTACT_FAILED_MANUAL_VERIFICATION: False}
        calls = [call.get_contact_info(self.contact_id),
                 call.update_contact_state(self.contact_id, self.contact_history_id, flags)]
        await self._test_process(calls)

    async def test_process_resend(self):
        # Test resend doesn't change contact state.
        calls = [call.get_contact_info(self.contact_id)]
        await self._test_process(calls, resend=True, status=VerificationStatus.CREATED)


class CheckDomainsEventTest(TransactionTestCase):
    uuid = UUID(int=42)
    contact_id = ContactId('2X4B')
    handle = 'KRYTEN'
    name = 'Kryten'
    organization = 'JMC'

    async def setup_mocks(self):
        # Can't be in asyncSetUp, due to conflicts with override_settings.
        APPLICATION.contact_client = AsyncMock(ContactClient)
        APPLICATION.contact_client.get_contact_info.return_value = Contact(
            contact_id=self.contact_id, contact_handle=self.handle, name=self.name, organization=self.organization)
        APPLICATION.domain_client = AsyncMock(DomainClient)
        APPLICATION.domain_client.get_domains_by_contact.return_value = {
            'domains': [{'domain': DomainRef(id=DomainId('42'), fqdn='example.org')}]}

        APPLICATION.email_client = AsyncMock(EmailMessengerClient)

    async def _test_process(self, *, json: Any = None, count: int = 1, ranked_domains: Optional[List[str]] = None,
                            match: Optional[Sequence[Callable]] = None,
                            status: VerificationStatus = VerificationStatus.CREATED,
                            email_calls: Optional[List[_Call]] = None) -> None:
        with responses.RequestsMock() as rsps:
            rsps.add(responses.GET, "http://example.org/", json=(json or []), match=(match or []))

            verification = await make_verification(self.session, contact_id=self.contact_id, uuid=self.uuid)
            scheduled = CheckDomainsEvent(verification)

            with freeze_time('1988-02-15'):
                await scheduled.process()

            self.assertEqual(verification.status, status)
            # Check event details
            self.assertEqual(await scheduled.get_details(),
                             {'count': count, 'ranked_domains': ranked_domains or []})

        # Check domain call
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.domain_client)),
                         [call.get_domains_by_contact(self.contact_id, {DomainContactRole.HOLDER})])
        # Check messenger call
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.email_client)), email_calls or [])

    async def test_process(self):
        settings = {'rank_query': 'http://example.org/'}
        with override_settings(check_domains=settings):
            await self.setup_mocks()
            await self._test_process(match=[query_param_matcher({})])

    async def test_process_query(self):
        settings = {'rank_query': 'http://example.org/?query=yes'}
        with override_settings(check_domains=settings):
            await self.setup_mocks()
            await self._test_process(match=[query_param_matcher({'query': 'yes'})])

    async def test_process_query_format(self):
        settings = {'rank_query': 'http://example.org/?last_month={last_month}&handle={handle}'}
        with override_settings(check_domains=settings):
            await self.setup_mocks()
            await self._test_process(match=[query_param_matcher({'last_month': '1988-01-01', 'handle': self.handle})])

    async def test_process_headers(self):
        settings = {'rank_query': 'http://example.org/', 'rank_headers': {'X-Header': 'Gazpacho!'}}
        with override_settings(check_domains=settings):
            await self.setup_mocks()
            await self._test_process(match=[header_matcher({'X-Header': 'Gazpacho!'})])

    async def test_process_headers_json(self):
        settings = {"rank_query": "http://example.org/", "rank_headers": '{"X-Header": "Gazpacho!"}'}
        with override_settings(check_domains=settings):
            await self.setup_mocks()
            await self._test_process(match=[header_matcher({'X-Header': 'Gazpacho!'})])

    async def test_process_timeout(self):
        settings = {'rank_query': 'http://example.org/', 'rank_timeout': 42}
        with override_settings(check_domains=settings):
            await self.setup_mocks()
            await self._test_process(match=[request_kwargs_matcher({'timeout': 42})])

    async def test_process_domains_found(self):
        settings = {'rank_query': 'http://example.org/'}
        with override_settings(check_domains=settings):
            await self.setup_mocks()
            await self._test_process(json=[{"domain": "example.org"}],
                                     ranked_domains=["example.org"],
                                     status=VerificationStatus.PAUSED)

    async def test_process_filter_owned(self):
        # Test ranked_domains are filtered to domains actually owned by the contact.
        settings = {'rank_query': 'http://example.org/'}
        with override_settings(check_domains=settings):
            await self.setup_mocks()
            cast(AsyncMock, APPLICATION.domain_client.get_domains_by_contact).return_value = {
                'domains': [{'domain': DomainRef(id=DomainId('42'), fqdn='kryten.example.org')}]}

            await self._test_process(json=[{"domain": "kryten.example.org"}, {"domain": "arnold.example.org"}],
                                     ranked_domains=["kryten.example.org"],
                                     status=VerificationStatus.PAUSED)

    async def test_process_error(self):
        settings = {'rank_query': 'http://example.org/'}
        with override_settings(check_domains=settings):
            await self.setup_mocks()
            with responses.RequestsMock() as rsps:
                rsps.add(responses.GET, "http://example.org/", status=500)

                verification = await make_verification(self.session, contact_id=self.contact_id, uuid=self.uuid)
                scheduled = CheckDomainsEvent(verification)

                with self.assertRaisesRegex(requests.HTTPError, '500 Server Error:'):
                    await scheduled.process()

    async def test_process_count_limit_over(self):
        settings = {'rank_query': 'http://example.org/', 'count_limit': 1}
        with override_settings(check_domains=settings):
            await self.setup_mocks()
            await self._test_process(status=VerificationStatus.PAUSED)

    async def test_process_count_limit_under(self):
        settings = {'count_limit': 2, 'rank_query': 'http://example.org/'}
        with override_settings(check_domains=settings):
            await self.setup_mocks()
            await self._test_process()

    async def test_process_notification(self):
        settings = {'rank_query': 'http://example.org/', 'notification_email': 'holly@example.org'}
        with override_settings(check_domains=settings):
            await self.setup_mocks()
            cast(AsyncMock, APPLICATION.domain_client.get_domains_by_contact).return_value = {
                'domains': [{'domain': DomainRef(id=DomainId('42'), fqdn='ranked.example.org')},
                            {'domain': DomainRef(id=DomainId('43'), fqdn='other.example.org')}]}

            context = {
                'uuid': str(self.uuid), 'handle': self.handle, 'name': self.name, 'organization': self.organization,
                'domains': ['ranked.example.org', 'other.example.org'], 'ranked_domains': ["ranked.example.org"]}
            message = Email(recipient='holly@example.org', subject_template=Template.PAUSED_NOTIFICATION_SUBJECT,
                            body_template=Template.PAUSED_NOTIFICATION_BODY, context=context,
                            type=MessageType.PAUSED_NOTIFICATION)
            email_call = call.send(message,
                                   references=[Reference(type=ReferenceType.VERIFICATION, value=str(self.uuid))])

            await self._test_process(json=[{"domain": "ranked.example.org"}], status=VerificationStatus.PAUSED,
                                     ranked_domains=["ranked.example.org"], email_calls=[email_call], count=2)


class OutzoneEventTest(TransactionTestCase):
    fqdn = 'kryten.example.org'

    async def asyncSetUp(self):
        APPLICATION.contact_client = AsyncMock(ContactClient)
        APPLICATION.domain_admin_client = AsyncMock(DomainAdminClient)
        APPLICATION.domain_client = AsyncMock(DomainClient)
        APPLICATION.email_client = AsyncMock(EmailMessengerClient)

    async def _test_process(self, emails: List[str], recipient: str, futile_datetime: str = '') -> None:
        contact_id = ContactId('KRYTEN')
        handle = '2X4B'
        uuid = UUID(int=42)
        create_datetime = datetime(2019, 12, 11, 10, 9, 8, tzinfo=UTC)
        cast(AsyncMock, APPLICATION.domain_client.get_domains_by_contact).return_value = {
            'domains': [{'domain': DomainRef(id=DomainId('42'), fqdn=self.fqdn)}]}
        cast(AsyncMock, APPLICATION.domain_admin_client.manage_domain_state_flags).return_value = aiter([
            {'domain_id': '42', 'request_id': '196'}])
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = Contact(
            contact_id=contact_id, contact_handle=handle, contact_history_id=ContactHistoryId('42'), emails=emails)

        verification = await make_verification(self.session, uuid=uuid, contact_id=contact_id,
                                               create_datetime=create_datetime, access_token='Gazpacho!')
        scheduled = OutzoneEvent(verification, details={})

        await scheduled.process()

        self.assertEqual(await scheduled.get_details(), {'domains': ['42'], 'request_ids': {'42': ['196']}})
        self.assertEqual(verification.status, VerificationStatus.CONSTRAINED)
        # Check domain call
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.domain_client)),
                         [call.get_domains_by_contact(contact_id, {DomainContactRole.HOLDER})])
        # Check outzone call
        flags = {ObjectState.SERVER_BLOCKED: None, ObjectState.SERVER_OUTZONE_MANUAL: None}
        outzone_call = call.manage_domain_state_flags(domains=[Domain(domain_id=DomainId('42'), fqdn=self.fqdn)],
                                                      state_flags=flags)
        self.assertEqual(cast(AsyncMock, APPLICATION.domain_admin_client).mock_calls, [outzone_call])
        # Check contact call
        contact_calls = [
            call.get_contact_info(contact_id),
            call.update_contact_state(contact_id, ContactHistoryId('42'), {ObjectState.SERVER_LINK_PROHIBITED: True}),
        ]
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.contact_client)), contact_calls)
        # Check messenger calls
        context = {'uuid': str(uuid), 'access_token': 'Gazpacho!',
                   'initial_email_datetime': create_datetime.isoformat(), 'handle': handle,
                   'domains': [self.fqdn], 'futile_datetime': futile_datetime}
        message = Email(recipient=recipient, subject_template=Template.OUTZONE_SUBJECT,
                        body_template=Template.OUTZONE_BODY, context=context, type=MessageType.OUTZONE)
        calls = [call.send(message, references=[Reference(type=ReferenceType.CONTACT, value=contact_id),
                                                Reference(type=ReferenceType.VERIFICATION, value=str(uuid))])]
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.email_client)), calls)

    async def test_process(self):
        # Test with a single email
        APPLICATION.handler = AsyncMock(VerificationHandler)
        APPLICATION.handler.get_active_processors.return_value = aiter([])
        email = 'kryten@example.org'
        await self._test_process([email], email)

    async def test_process_multi_mails(self):
        # Test with a multiple emails
        APPLICATION.handler = AsyncMock(VerificationHandler)
        APPLICATION.handler.get_active_processors.return_value = aiter([])
        emails = ['kryten@example.org', 'rimmer@example.org']
        recipient = 'kryten@example.org, rimmer@example.org'
        await self._test_process(emails, recipient)

    async def test_process_futile(self):
        # Test futile_datetime context variable
        APPLICATION.handler = AsyncMock(VerificationHandler)
        processor = AsyncMock(BaseProcessor)
        processor._get_futile_due.return_value = datetime(2020, 1, 24, 10, 9, 8, tzinfo=UTC)
        APPLICATION.handler.get_active_processors.return_value = aiter([processor])
        email = 'kryten@example.org'
        await self._test_process([email], email, '2020-01-24T10:09:08+00:00')


class DelayOutzoneEventTest(TransactionTestCase):
    contact_id = 'KRYTEN'

    async def test_process(self):
        delay = 42
        verification = await make_verification(self.session, contact_id=self.contact_id)
        scheduled = DelayOutzoneEvent(verification, details={'delay': delay})

        await scheduled.process()

        # Check status haven't changed.
        self.assertEqual(verification.status, VerificationStatus.CREATED)
        # Check delay event was created.
        event = (await self.session.execute(select(Event))).scalars().one()
        self.assertIsNone(event.method)
        self.assertEqual(event.type, EventType.DELAY_STATUS)
        self.assertEqual(event.details, {'delay': delay, 'status': VerificationStatus.FUTILE})

    async def test_to_event(self):
        delay = 42
        verification = await make_verification(self.session, contact_id=self.contact_id)
        scheduled = DelayOutzoneEvent(verification, details={'delay': delay})

        event = await scheduled.to_event()

        self.assertIsNone(event.method)
        self.assertEqual(event.type, EventType.DELAY_STATUS)
        self.assertEqual(event.details, {'delay': delay, 'status': VerificationStatus.CONSTRAINED})


class FutileEventTest(TransactionTestCase):
    async def test_process(self):
        contact_id = 'KRYTEN'

        verification = await make_verification(self.session, contact_id=contact_id)
        scheduled = FutileEvent(verification)

        await scheduled.process()

        self.assertEqual(verification.status, VerificationStatus.FUTILE)


class DelayFutileEventTest(TransactionTestCase):
    contact_id = 'KRYTEN'

    async def test_process(self):
        delay = 42
        verification = await make_verification(self.session, contact_id=self.contact_id)
        scheduled = DelayFutileEvent(verification, details={'delay': delay})

        await scheduled.process()

        # Check status haven't changed.
        self.assertEqual(verification.status, VerificationStatus.CREATED)
        # Check no event was created.
        self.assertIsNone((await self.session.execute(select(Event))).scalar())

    async def test_to_event(self):
        delay = 42
        verification = await make_verification(self.session, contact_id=self.contact_id)
        scheduled = DelayFutileEvent(verification, details={'delay': delay})

        event = await scheduled.to_event()

        self.assertIsNone(event.method)
        self.assertEqual(event.type, EventType.DELAY_STATUS)
        self.assertEqual(event.details, {'delay': delay, 'status': VerificationStatus.FUTILE})


class CompleteEventTest(TransactionTestCase):
    async def asyncSetUp(self):
        APPLICATION.contact_client = AsyncMock(ContactClient)

    async def test_process(self):
        contact_id = ContactId('KRYTEN')
        handle = '2X4B'
        uuid = UUID(int=42)
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = Contact(
            contact_id=contact_id, contact_handle=handle, contact_history_id=ContactHistoryId('42'))

        verification = await make_verification(self.session, uuid=uuid, contact_id=contact_id)

        scheduled = CompleteEvent(verification)

        await scheduled.process()

        self.assertIsNotNone(verification.close_datetime)
        self.assertEqual(verification.status, VerificationStatus.SUCCEEDED)

        # Check contact call
        flags = {ObjectState.CONTACT_IN_MANUAL_VERIFICATION: False,
                 ObjectState.CONTACT_PASSED_MANUAL_VERIFICATION: True,
                 ObjectState.CONTACT_FAILED_MANUAL_VERIFICATION: False}
        calls = [call.get_contact_info(contact_id),
                 call.update_contact_state(contact_id, ContactHistoryId('42'), flags)]
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.contact_client)), calls)


class FailedEventTest(TransactionTestCase):
    async def asyncSetUp(self):
        APPLICATION.contact_client = AsyncMock(ContactClient)

    async def test_process(self):
        contact_id = ContactId('KRYTEN')
        handle = '2X4B'
        uuid = UUID(int=42)
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = Contact(
            contact_id=contact_id, contact_handle=handle, contact_history_id=ContactHistoryId('42'))

        verification = await make_verification(self.session, uuid=uuid, contact_id=contact_id)

        scheduled = FailedEvent(verification)

        await scheduled.process()

        self.assertIsNotNone(verification.close_datetime)
        self.assertEqual(verification.status, VerificationStatus.FAILED)

        # Check contact call
        flags = {ObjectState.CONTACT_IN_MANUAL_VERIFICATION: False,
                 ObjectState.CONTACT_PASSED_MANUAL_VERIFICATION: False,
                 ObjectState.CONTACT_FAILED_MANUAL_VERIFICATION: True}
        calls = [call.get_contact_info(contact_id),
                 call.update_contact_state(contact_id, ContactHistoryId('42'), flags)]
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.contact_client)), calls)


class NotifyFailedEventTest(TransactionTestCase):
    uuid = UUID(int=42)
    contact_id = ContactId('KRYTEN')
    handle = '2X4B'

    async def asyncSetUp(self):
        APPLICATION.contact_client = AsyncMock(ContactClient)
        APPLICATION.domain_client = AsyncMock(DomainClient)
        APPLICATION.email_client = AsyncMock(EmailMessengerClient)

    async def test_get_context_no_event(self):
        # Test context in case there is no FAILED event.
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = Contact(
            contact_id=self.contact_id, contact_handle=self.handle)

        initial_context = {'context': sentinel.context}
        verification = await make_verification(self.session, contact_id=self.contact_id)
        scheduled = NotifyFailedEvent(verification, context=initial_context)

        context = await scheduled.get_context()

        self.assertEqual(context, {'context': sentinel.context, 'handle': self.handle, 'domains': []})
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.domain_client)), [])

    async def test_get_context_no_domains(self):
        # Test context in case FAILED event has no domains.
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = Contact(
            contact_id=self.contact_id, contact_handle=self.handle)

        initial_context = {'context': sentinel.context}
        verification = await make_verification(self.session, contact_id=self.contact_id)
        await make_event(self.session, verification_id=verification.id, type=EventType.FAILED, details={})
        scheduled = NotifyFailedEvent(verification, context=initial_context)

        context = await scheduled.get_context()

        self.assertEqual(context, {'context': sentinel.context, 'handle': self.handle, 'domains': []})
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.domain_client)), [])

    async def test_get_context_empty_domains(self):
        # Test context in case FAILED event has empty domain list.
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = Contact(
            contact_id=self.contact_id, contact_handle=self.handle)

        initial_context = {'context': sentinel.context}
        verification = await make_verification(self.session, contact_id=self.contact_id)
        await make_event(self.session, verification_id=verification.id, type=EventType.FAILED, details={'domains': []})
        scheduled = NotifyFailedEvent(verification, context=initial_context)

        context = await scheduled.get_context()

        self.assertEqual(context, {'context': sentinel.context, 'handle': self.handle, 'domains': []})
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.domain_client)), [])

    async def test_get_context_domains(self):
        # Test context in case FAILED event has domain list.
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = Contact(
            contact_id=self.contact_id, contact_handle=self.handle)
        cast(AsyncMock, APPLICATION.domain_client.get_domains_by_contact).return_value = {
            'domains': [{'domain': DomainRef(id=DomainId('42'), fqdn='example.org')}]}
        cast(AsyncMock, APPLICATION.domain_client.get_domain_info).return_value = Domain(domain_id=DomainId('42'),
                                                                                         fqdn='example.org')

        initial_context = {'context': sentinel.context}
        verification = await make_verification(self.session, contact_id=self.contact_id)
        await make_event(self.session, verification_id=verification.id, type=EventType.FAILED,
                         details={'domains': ['42']})
        scheduled = NotifyFailedEvent(verification, context=initial_context)

        context = await scheduled.get_context()

        self.assertEqual(context, {'context': sentinel.context, 'handle': self.handle, 'domains': ['example.org']})
        self.assertEqual(
            mock_awaits(cast(AsyncMock, APPLICATION.domain_client)),
            [call.get_domains_by_contact(self.contact_id, {DomainContactRole.HOLDER}, include_deleted=True)])

    async def test_get_details(self):
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = Contact(
            contact_id=self.contact_id, contact_handle=self.handle, emails=['kryten@example.org'])

        verification = await make_verification(self.session, contact_id=self.contact_id)
        scheduled = NotifyFailedEvent(verification)

        details = {'method': TransportMethod.EMAIL, 'message_type': MessageType.FAILED,
                   'recipient': ['kryten@example.org']}
        self.assertEqual(await scheduled.get_details(), details)

    async def _test_process(self, recipient: Any, email: Email) -> None:
        cast(AsyncMock, APPLICATION.email_client.send).return_value = sentinel.message_id

        verification = await make_verification(self.session, contact_id=self.contact_id, uuid=self.uuid)
        scheduled = NotifyFailedEvent(verification)

        await scheduled.process()

        details = {'method': TransportMethod.EMAIL, 'message_type': MessageType.FAILED,
                   'recipient': recipient, 'message_id': sentinel.message_id}
        self.assertEqual(await scheduled.get_details(), details)
        # Check messenger calls
        calls = [call.send(email, references=[Reference(type=ReferenceType.CONTACT, value=self.contact_id),
                                              Reference(type=ReferenceType.VERIFICATION, value=str(self.uuid))])]
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.email_client)), calls)

    async def test_process(self):
        # Test with a single email
        email = 'kryten@example.org'
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = Contact(
            contact_id=self.contact_id, contact_handle=self.handle, emails=[email])

        context = {'handle': self.handle, 'domains': []}
        email_msg = Email(recipient=email, type=MessageType.FAILED, subject_template=Template.FAILED_SUBJECT,
                          body_template=Template.FAILED_BODY, context=context)
        await self._test_process([email], email_msg)

        # Check registry calls
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.contact_client)),
                         [call.get_contact_info(self.contact_id)])

    async def test_process_multi_mails(self):
        # Test with a multiple emails
        emails = ['kryten@example.org', 'rimmer@example.org']
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = Contact(
            contact_id=self.contact_id, contact_handle=self.handle, emails=emails)

        context = {'handle': self.handle, 'domains': []}
        email_msg = Email(recipient='kryten@example.org, rimmer@example.org', type=MessageType.FAILED,
                          subject_template=Template.FAILED_SUBJECT, body_template=Template.FAILED_BODY,
                          context=context)
        await self._test_process(emails, email_msg)

        # Check registry calls
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.contact_client)),
                         [call.get_contact_info(self.contact_id)])


class CancelVerificationEventTest(TransactionTestCase):
    async def asyncSetUp(self):
        APPLICATION.contact_client = AsyncMock(ContactClient)

    async def test_process(self):
        contact_id = ContactId('KRYTEN')
        handle = '2X4B'
        uuid = UUID(int=42)
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = Contact(
            contact_id=contact_id, contact_handle=handle, contact_history_id=ContactHistoryId('42'))

        verification = await make_verification(self.session, uuid=uuid, contact_id=contact_id)
        scheduled = CancelVerificationEvent(verification)

        await scheduled.process()

        self.assertIsNotNone(verification.close_datetime)
        self.assertEqual(verification.status, VerificationStatus.CANCELED)

        # Check contact call
        flags = {ObjectState.CONTACT_IN_MANUAL_VERIFICATION: False}
        calls = [call.get_contact_info(contact_id),
                 call.update_contact_state(contact_id, ContactHistoryId('42'), flags)]
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.contact_client)), calls)


class CancelOutzoneEventTest(TransactionTestCase):
    contact_id = ContactId('KRYTEN')

    async def asyncSetUp(self):
        APPLICATION.contact_client = AsyncMock(ContactClient)
        APPLICATION.domain_admin_client = AsyncMock(DomainAdminClient)
        APPLICATION.domain_admin_client.manage_domain_state_flags.return_value = []

    async def test_get_details(self):
        scheduled = CancelOutzoneEvent(sentinel.verification)

        self.assertEqual(await scheduled.get_details(), {'canceled': True})

    async def _test_process(self, request_ids: Dict[str, List[str]], status: VerificationStatus,
                            close_requests: List[str]) -> None:
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = Contact(
            contact_id=self.contact_id, contact_handle='2X4B', contact_history_id=ContactHistoryId('42'))

        verification = (await self.session.execute(select(Verification))).scalars().one()
        scheduled = CancelOutzoneEvent(verification)

        await scheduled.process()

        self.assertEqual(await scheduled.get_details(), {'canceled': True, 'request_ids': request_ids})
        self.assertEqual(verification.status, status)
        # Check outzone call
        calls = [call.manage_domain_state_flags(close_requests=close_requests)]
        self.assertEqual(cast(AsyncMock, APPLICATION.domain_admin_client).mock_calls, calls)
        # Check contact call
        contact_calls = [
            call.get_contact_info(self.contact_id),
            call.update_contact_state(self.contact_id, ContactHistoryId('42'),
                                      {ObjectState.SERVER_LINK_PROHIBITED: False}),
        ]
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.contact_client)), contact_calls)

    async def test_process(self):
        verification = await make_verification(self.session, contact_id=self.contact_id)
        await make_domain(self.session, verification_id=verification.id, domain_id=DomainId('42'), fqdn='example.org')
        await make_event(self.session, verification_id=verification.id, type=EventType.OUTZONE,
                         details={'request_ids': {'42': ['196']}})

        await self._test_process({'42': ['196']}, VerificationStatus.INITIALIZED, ['196'])

    async def test_process_affected(self):
        # Test only affected domains are canceled.
        verification = await make_verification(self.session, contact_id=self.contact_id)
        await make_domain(self.session, verification_id=verification.id, domain_id=DomainId('42'), fqdn='example.org')
        await make_event(self.session, verification_id=verification.id, type=EventType.OUTZONE,
                         details={'request_ids': {'42': ['196'], '43': ['197']}})

        await self._test_process({'42': ['196']}, VerificationStatus.INITIALIZED, ['196'])

    async def test_process_accessed(self):
        verification = await make_verification(self.session, contact_id=self.contact_id,
                                               last_access_datetime=datetime(1988, 9, 6, tzinfo=UTC))
        await make_domain(self.session, verification_id=verification.id, domain_id=DomainId('42'), fqdn='example.org')
        await make_event(self.session, verification_id=verification.id, type=EventType.OUTZONE,
                         details={'request_ids': {'42': ['196']}})

        await self._test_process({'42': ['196']}, VerificationStatus.ACCESSED, ['196'])

    async def test_process_last_outzone(self):
        # Test only last outzone event is canceled.
        verification = await make_verification(self.session, contact_id=self.contact_id)
        await make_domain(self.session, verification_id=verification.id, domain_id=DomainId('42'), fqdn='example.org')
        await make_event(self.session, verification_id=verification.id, type=EventType.OUTZONE,
                         details={'request_ids': {'42': ['196']}})
        await make_event(self.session, verification_id=verification.id, type=EventType.OUTZONE,
                         details={'request_ids': {'42': ['150']}}, event_datetime=datetime(1988, 9, 6, tzinfo=UTC))

        await self._test_process({'42': ['196']}, VerificationStatus.INITIALIZED, ['196'])

    async def test_process_last_canceled(self):
        # Test noop when last outzone is cancelation.
        verification = await make_verification(self.session, contact_id=self.contact_id)
        await make_domain(self.session, verification_id=verification.id, domain_id=DomainId('42'), fqdn='example.org')
        await make_event(self.session, verification_id=verification.id, type=EventType.OUTZONE,
                         details={'request_ids': {'42': ['196']}, 'canceled': True})

        await self._test_process({}, VerificationStatus.INITIALIZED, [])

    async def test_process_no_requests(self):
        # Test noop when outzone event has no request IDs.
        verification = await make_verification(self.session, contact_id=self.contact_id)
        await make_event(self.session, verification_id=verification.id, type=EventType.OUTZONE, details={})

        await self._test_process({}, VerificationStatus.INITIALIZED, [])

    async def test_process_no_outzone(self):
        # Test noop when outzone event have not occurred.
        await make_verification(self.session, contact_id=self.contact_id)

        await self._test_process({}, VerificationStatus.INITIALIZED, [])

    async def _test_process_legacy(self, request_ids: List[str], status: VerificationStatus) -> None:
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = Contact(
            contact_id=self.contact_id, contact_handle='2X4B', contact_history_id=ContactHistoryId('42'))

        verification = (await self.session.execute(select(Verification))).scalars().one()
        scheduled = CancelOutzoneEvent(verification)

        await scheduled.process()

        self.assertEqual(await scheduled.get_details(), {'canceled': True, 'request_ids': request_ids})
        self.assertEqual(verification.status, status)
        # Check outzone call
        calls = [call.manage_domain_state_flags(close_requests=request_ids)]
        self.assertEqual(cast(AsyncMock, APPLICATION.domain_admin_client).mock_calls, calls)
        # Check contact call
        contact_calls = [
            call.get_contact_info(self.contact_id),
            call.update_contact_state(self.contact_id, ContactHistoryId('42'),
                                      {ObjectState.SERVER_LINK_PROHIBITED: False}),
        ]
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.contact_client)), contact_calls)

    async def test_process_legacy(self):
        # Test cancel outzone on events from verification < 1.5.0
        verification = await make_verification(self.session, contact_id=self.contact_id)
        await make_event(self.session, verification_id=verification.id, type=EventType.OUTZONE,
                         details={'request_ids': ['42']})

        await self._test_process_legacy(['42'], VerificationStatus.INITIALIZED)


class CancelFutileEventTest(TransactionTestCase):
    async def test_get_details(self):
        scheduled = CancelFutileEvent(sentinel.verification)

        self.assertEqual(await scheduled.get_details(), {'canceled': True})

    async def test_process(self):
        contact_id = 'KRYTEN'

        verification = await make_verification(self.session, contact_id=contact_id)
        scheduled = CancelFutileEvent(verification)

        await scheduled.process()

        self.assertEqual(await scheduled.get_details(), {'canceled': True})
        self.assertEqual(verification.status, VerificationStatus.CONSTRAINED)


class TestProcessor(BaseProcessor):
    """Simple processor for testing BaseProcessor."""

    def __init__(self, *, next_event: Optional[ScheduledEvent] = None, cancel_event: Optional[ScheduledEvent] = None,
                 complete_event: Optional[CompleteEvent] = None, **kwargs: Any):
        super().__init__(**kwargs)
        self.next_event = next_event
        self.cancel_event = cancel_event
        self.complete_event = complete_event

    async def _get_next_events(self, verification: Verification) -> AsyncIterator[ScheduledEvent]:
        result = []
        if self.next_event:  # pragma: no cover
            result = [self.next_event]
        for scheduled in result:
            yield scheduled

    async def _get_cancel_events(self, verification: Verification, status: Optional[VerificationStatus],
                                 ) -> AsyncIterator[ScheduledEvent]:
        result = []
        if self.cancel_event:  # pragma: no cover
            result = [self.cancel_event]
        for scheduled in result:
            yield scheduled

    async def _get_complete_events(self, verification: Verification) -> AsyncIterator[ScheduledEvent]:
        result = []
        if self.complete_event:  # pragma: no cover
            result = [self.complete_event]
        for scheduled in result:
            yield scheduled


class BaseProcessorTest(TransactionTestCase):
    method = 'ANALYSIS'
    contact_id = ContactId('KRYTEN')
    contact_history_id = ContactHistoryId('42')
    handle = '2X4B'
    email = 'kryten@example.org'
    create_datetime = datetime(2019, 12, 11, 10, 9, 8, tzinfo=UTC)

    async def asyncSetUp(self):
        APPLICATION.contact_client = AsyncMock(ContactClient)
        APPLICATION.contact_client.get_contact_info.return_value = Contact(
            contact_id=self.contact_id, contact_handle=self.handle, contact_history_id=self.contact_history_id,
            emails=[self.email])
        APPLICATION.domain_client = AsyncMock(DomainClient)
        # Return empty list, unless stated otherwise.
        APPLICATION.domain_client.get_domains_by_contact.return_value = {'domains': []}

    def test_repr(self):
        processor = BaseProcessor(method=self.method)
        self.assertEqual(repr(processor), '<BaseProcessor of method ANALYSIS>')

    async def test_is_applicable(self):
        processor = BaseProcessor(method=self.method)
        verification = Verification()

        self.assertTrue(await processor.is_applicable(verification))

    async def test_get_next_event(self):
        processor = BaseProcessor(method=self.method)
        data: Tuple[Tuple[VerificationStatus, List[ScheduledEvent]], ...] = (
            # status, events
            (VerificationStatus.PAUSED, []),
            (VerificationStatus.MANUAL, []),
        )
        for status, events in data:
            with self.subTest(status=status):
                verification = Verification(status=status, create_datetime=self.create_datetime)
                self.assertEqual(await alist(processor.get_next_events(verification)), events)

    async def test_get_next_event_created(self):
        # Test get_next_event for verification in CREATED status.
        processor = BaseProcessor(method=self.method)
        verification = await make_verification(self.session, status=VerificationStatus.CREATED,
                                               contact_id=self.contact_id, create_datetime=self.create_datetime)

        result = processor.get_next_events(verification)

        details = {'contact_history_id': self.contact_history_id}
        context = {'outzone_datetime': '2020-01-10T10:09:08+00:00', 'futile_datetime': '2020-01-24T10:09:08+00:00'}
        initial = Comparison(InitialEmailEvent, due_datetime=None, method=None, _details=details, context=context,
                             partial=True)
        context = {'initial_email_datetime': self.create_datetime.isoformat(),
                   'outzone_datetime': '2020-01-10T10:09:08+00:00', 'futile_datetime': '2020-01-24T10:09:08+00:00'}
        outzone_rem = Comparison(OutzoneReminderEvent, due_datetime=datetime(2020, 1, 3, 10, 9, 8, tzinfo=UTC),
                                 _details={}, context=context, partial=True)
        outzone = Comparison(OutzoneEvent, due_datetime=datetime(2020, 1, 10, 10, 9, 8, tzinfo=UTC), partial=True)
        context = {'initial_email_datetime': self.create_datetime.isoformat(),
                   'futile_datetime': '2020-01-24T10:09:08+00:00'}
        futile_rem = Comparison(FutileReminderEvent, due_datetime=datetime(2020, 1, 17, 10, 9, 8, tzinfo=UTC),
                                _details={}, context=context, partial=True)
        futile = Comparison(FutileEvent, due_datetime=datetime(2020, 1, 24, 10, 9, 8, tzinfo=UTC), partial=True)
        failed = Comparison(FailedEvent, due_datetime=datetime(2020, 1, 24, 10, 9, 8, tzinfo=UTC), partial=True)
        self.assertEqual(await alist(result), [initial, outzone_rem, outzone, futile_rem, futile, failed])
        # Check registry calls
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.contact_client)),
                         [call.get_contact_info(self.contact_id)])

    async def _test_get_next_event_init(self, status: VerificationStatus) -> None:
        # Test get_next_event for verification in INITIALIZED, ACCESSED and PENDING status.
        processor = BaseProcessor(method=self.method)
        verification = await make_verification(self.session, status=status, contact_id=self.contact_id,
                                               create_datetime=self.create_datetime)

        result = processor.get_next_events(verification)

        context = {'initial_email_datetime': self.create_datetime.isoformat(),
                   'outzone_datetime': '2020-01-10T10:09:08+00:00', 'futile_datetime': '2020-01-24T10:09:08+00:00'}
        outzone_rem = Comparison(OutzoneReminderEvent, due_datetime=datetime(2020, 1, 3, 10, 9, 8, tzinfo=UTC),
                                 _details={}, context=context, partial=True)
        outzone = Comparison(OutzoneEvent, due_datetime=datetime(2020, 1, 10, 10, 9, 8, tzinfo=UTC), partial=True)
        context = {'initial_email_datetime': self.create_datetime.isoformat(),
                   'futile_datetime': '2020-01-24T10:09:08+00:00'}
        futile_rem = Comparison(FutileReminderEvent, due_datetime=datetime(2020, 1, 17, 10, 9, 8, tzinfo=UTC),
                                _details={}, context=context, partial=True)
        futile = Comparison(FutileEvent, due_datetime=datetime(2020, 1, 24, 10, 9, 8, tzinfo=UTC), partial=True)
        failed = Comparison(FailedEvent, due_datetime=datetime(2020, 1, 24, 10, 9, 8, tzinfo=UTC), partial=True)
        self.assertEqual(await alist(result), [outzone_rem, outzone, futile_rem, futile, failed])

    async def test_get_next_event_initialized(self):
        await self._test_get_next_event_init(VerificationStatus.INITIALIZED)

    async def test_get_next_event_accessed(self):
        await self._test_get_next_event_init(VerificationStatus.ACCESSED)

    async def test_get_next_event_pending(self):
        await self._test_get_next_event_init(VerificationStatus.PENDING)

    async def test_get_next_event_check_domains(self):
        # Test CheckDomainsEvent is properly scheduled.
        with override_settings(check_domains={'rank_query': 'http://example.org'}):
            APPLICATION.contact_client = AsyncMock(ContactClient)
            APPLICATION.contact_client.get_contact_info.return_value = Contact(
                contact_id=self.contact_id, contact_handle=self.handle, contact_history_id=self.contact_history_id,
                emails=[self.email])
            APPLICATION.domain_client = AsyncMock(DomainClient)
            APPLICATION.domain_client.get_domains_by_contact.return_value = {'domains': []}
            processor = BaseProcessor(method=self.method)
            verification = await make_verification(self.session, status=VerificationStatus.INITIALIZED,
                                                   contact_id=self.contact_id, create_datetime=self.create_datetime)

            result = processor.get_next_events(verification)

            check_domains = Comparison(CheckDomainsEvent, due_datetime=datetime(2020, 1, 10, 10, 9, 8, tzinfo=UTC),
                                       method=None, partial=True)
            outzone = Comparison(OutzoneEvent, due_datetime=datetime(2020, 1, 10, 10, 9, 8, tzinfo=UTC), method=None,
                                 partial=True)
            next_events = [Comparison(OutzoneReminderEvent), check_domains, outzone, Comparison(FutileReminderEvent),
                           Comparison(FutileEvent), Comparison(FailedEvent)]
            self.assertEqual(await alist(result), next_events)

    async def test_get_next_event_check_domains_processed(self):
        # Test CheckDomainsEvent is not scheduled, if already processed.
        with override_settings(check_domains={'rank_query': 'http://example.org'}):
            APPLICATION.contact_client = AsyncMock(ContactClient)
            APPLICATION.contact_client.get_contact_info.return_value = Contact(
                contact_id=self.contact_id, contact_handle=self.handle, contact_history_id=self.contact_history_id,
                emails=[self.email])
            APPLICATION.domain_client = AsyncMock(DomainClient)
            APPLICATION.domain_client.get_domains_by_contact.return_value = {'domains': []}
            processor = BaseProcessor(method=self.method)
            verification = await make_verification(self.session, status=VerificationStatus.INITIALIZED,
                                                   contact_id=self.contact_id, create_datetime=self.create_datetime)
            await make_event(self.session, verification_id=verification.id, type=EventType.CHECK_DOMAINS)

            result = processor.get_next_events(verification)

            next_events = [Comparison(OutzoneReminderEvent), Comparison(OutzoneEvent), Comparison(FutileReminderEvent),
                           Comparison(FutileEvent), Comparison(FailedEvent)]
            self.assertEqual(await alist(result), next_events)

    async def test_get_next_event_constrained(self):
        # Test get_next_event for verification in CONSTRAINED status.
        processor = BaseProcessor(method=self.method)
        verification = await make_verification(self.session, status=VerificationStatus.CONSTRAINED,
                                               contact_id=self.contact_id, create_datetime=self.create_datetime)

        result = processor.get_next_events(verification)

        context = {'initial_email_datetime': self.create_datetime.isoformat(),
                   'futile_datetime': '2020-01-24T10:09:08+00:00'}
        futile_rem = Comparison(FutileReminderEvent, due_datetime=datetime(2020, 1, 17, 10, 9, 8, tzinfo=UTC),
                                _details={}, context=context, partial=True)
        futile = Comparison(FutileEvent, due_datetime=datetime(2020, 1, 24, 10, 9, 8, tzinfo=UTC), partial=True)
        failed = Comparison(FailedEvent, due_datetime=datetime(2020, 1, 24, 10, 9, 8, tzinfo=UTC), partial=True)
        self.assertEqual(await alist(result), [futile_rem, futile, failed])

    async def test_get_next_event_delay_constrained(self):
        processor = BaseProcessor(method=self.method)
        verification = await make_verification(self.session, contact_id=self.contact_id,
                                               status=VerificationStatus.INITIALIZED,
                                               create_datetime=self.create_datetime)
        details = {'status': VerificationStatus.CONSTRAINED, 'delay': 42}
        await make_event(self.session, verification_id=verification.id, type=EventType.DELAY_STATUS, details=details)

        context = {'initial_email_datetime': self.create_datetime.isoformat(),
                   'outzone_datetime': '2020-01-10T10:09:50+00:00', 'futile_datetime': '2020-01-24T10:09:08+00:00'}
        outzone_rem = Comparison(OutzoneReminderEvent, due_datetime=datetime(2020, 1, 3, 10, 9, 50, tzinfo=UTC),
                                 _details={}, context=context, partial=True)
        outzone = Comparison(OutzoneEvent, due_datetime=datetime(2020, 1, 10, 10, 9, 50, tzinfo=UTC), partial=True)
        context = {'initial_email_datetime': self.create_datetime.isoformat(),
                   'futile_datetime': '2020-01-24T10:09:08+00:00'}
        futile_rem = Comparison(FutileReminderEvent, due_datetime=datetime(2020, 1, 17, 10, 9, 8, tzinfo=UTC),
                                _details={}, context=context, partial=True)
        futile = Comparison(FutileEvent, due_datetime=datetime(2020, 1, 24, 10, 9, 8, tzinfo=UTC), partial=True)
        failed = Comparison(FailedEvent, due_datetime=datetime(2020, 1, 24, 10, 9, 8, tzinfo=UTC), partial=True)
        self.assertEqual(await alist(processor.get_next_events(verification)),
                         [outzone_rem, outzone, futile_rem, futile, failed])

    async def test_get_next_event_delay_constrained_limit(self):
        # Check outzone can be delayed to at most futile due.
        processor = BaseProcessor(method=self.method)
        verification = await make_verification(self.session, contact_id=self.contact_id,
                                               status=VerificationStatus.INITIALIZED,
                                               create_datetime=self.create_datetime)
        details = {'status': VerificationStatus.CONSTRAINED, 'delay': 30 * 86400}  # delay by 30 days
        await make_event(self.session, verification_id=verification.id, type=EventType.DELAY_STATUS, details=details)

        context = {'initial_email_datetime': self.create_datetime.isoformat(),
                   'outzone_datetime': '2020-01-24T10:09:08+00:00', 'futile_datetime': '2020-01-24T10:09:08+00:00'}
        outzone_rem = Comparison(OutzoneReminderEvent, due_datetime=datetime(2020, 1, 17, 10, 9, 8, tzinfo=UTC),
                                 _details={}, context=context, partial=True)
        outzone = Comparison(OutzoneEvent, due_datetime=datetime(2020, 1, 24, 10, 9, 8, tzinfo=UTC), partial=True)
        context = {'initial_email_datetime': self.create_datetime.isoformat(),
                   'futile_datetime': '2020-01-24T10:09:08+00:00'}
        futile_rem = Comparison(FutileReminderEvent, due_datetime=datetime(2020, 1, 17, 10, 9, 8, tzinfo=UTC),
                                _details={}, context=context, partial=True)
        futile = Comparison(FutileEvent, due_datetime=datetime(2020, 1, 24, 10, 9, 8, tzinfo=UTC), partial=True)
        failed = Comparison(FailedEvent, due_datetime=datetime(2020, 1, 24, 10, 9, 8, tzinfo=UTC), partial=True)
        self.assertEqual(await alist(processor.get_next_events(verification)),
                         [outzone_rem, outzone, futile_rem, futile, failed])

    async def test_get_next_event_delay_futile(self):
        processor = BaseProcessor(method=self.method)
        verification = await make_verification(self.session, contact_id=self.contact_id,
                                               status=VerificationStatus.INITIALIZED,
                                               create_datetime=self.create_datetime)
        details = {'status': VerificationStatus.FUTILE, 'delay': 42}
        await make_event(self.session, verification_id=verification.id, type=EventType.DELAY_STATUS, details=details)

        context = {'initial_email_datetime': self.create_datetime.isoformat(),
                   'outzone_datetime': '2020-01-10T10:09:08+00:00', 'futile_datetime': '2020-01-24T10:09:50+00:00'}
        outzone_rem = Comparison(OutzoneReminderEvent, due_datetime=datetime(2020, 1, 3, 10, 9, 8, tzinfo=UTC),
                                 _details={}, context=context, partial=True)
        outzone = Comparison(OutzoneEvent, due_datetime=datetime(2020, 1, 10, 10, 9, 8, tzinfo=UTC), partial=True)
        context = {'initial_email_datetime': self.create_datetime.isoformat(),
                   'futile_datetime': '2020-01-24T10:09:50+00:00'}
        futile_rem = Comparison(FutileReminderEvent, due_datetime=datetime(2020, 1, 17, 10, 9, 50, tzinfo=UTC),
                                _details={}, context=context, partial=True)
        futile = Comparison(FutileEvent, due_datetime=datetime(2020, 1, 24, 10, 9, 50, tzinfo=UTC), partial=True)
        failed = Comparison(FailedEvent, due_datetime=datetime(2020, 1, 24, 10, 9, 50, tzinfo=UTC), partial=True)
        self.assertEqual(await alist(processor.get_next_events(verification)),
                         [outzone_rem, outzone, futile_rem, futile, failed])

    async def test_get_next_event_outzone_sent(self):
        # Test get_next_event when outzone reminder was already sent.
        processor = BaseProcessor(method=self.method)
        verification = await make_verification(self.session, status=VerificationStatus.INITIALIZED,
                                               contact_id=self.contact_id, create_datetime=self.create_datetime)
        details: Dict[str, Any] = {'message_type': MessageType.OUTZONE_REMINDER}
        await make_event(self.session, verification_id=verification.id, type=EventType.SEND_MESSAGE, details=details)

        result = processor.get_next_events(verification)

        outzone = Comparison(OutzoneEvent, due_datetime=datetime(2020, 1, 10, 10, 9, 8, tzinfo=UTC), partial=True)
        context = {'initial_email_datetime': self.create_datetime.isoformat(),
                   'futile_datetime': '2020-01-24T10:09:08+00:00'}
        futile_rem = Comparison(FutileReminderEvent, due_datetime=datetime(2020, 1, 17, 10, 9, 8, tzinfo=UTC),
                                _details={}, context=context, partial=True)
        futile = Comparison(FutileEvent, due_datetime=datetime(2020, 1, 24, 10, 9, 8, tzinfo=UTC), partial=True)
        failed = Comparison(FailedEvent, due_datetime=datetime(2020, 1, 24, 10, 9, 8, tzinfo=UTC), partial=True)
        self.assertEqual(await alist(result), [outzone, futile_rem, futile, failed])

    async def test_get_next_event_futile_sent(self):
        # Test get_next_event when futile reminder was already sent.
        processor = BaseProcessor(method=self.method)
        verification = await make_verification(self.session, status=VerificationStatus.INITIALIZED,
                                               contact_id=self.contact_id, create_datetime=self.create_datetime)
        details: Dict[str, Any] = {'message_type': MessageType.FUTILE_REMINDER}
        await make_event(self.session, verification_id=verification.id, type=EventType.SEND_MESSAGE, details=details)

        result = processor.get_next_events(verification)

        context = {'initial_email_datetime': self.create_datetime.isoformat(),
                   'outzone_datetime': '2020-01-10T10:09:08+00:00', 'futile_datetime': '2020-01-24T10:09:08+00:00'}
        outzone_rem = Comparison(OutzoneReminderEvent, due_datetime=datetime(2020, 1, 3, 10, 9, 8, tzinfo=UTC),
                                 _details={}, context=context, partial=True)
        outzone = Comparison(OutzoneEvent, due_datetime=datetime(2020, 1, 10, 10, 9, 8, tzinfo=UTC), partial=True)
        futile = Comparison(FutileEvent, due_datetime=datetime(2020, 1, 24, 10, 9, 8, tzinfo=UTC), partial=True)
        failed = Comparison(FailedEvent, due_datetime=datetime(2020, 1, 24, 10, 9, 8, tzinfo=UTC), partial=True)
        self.assertEqual(await alist(result), [outzone_rem, outzone, futile, failed])

    async def test_get_next_event_manual_check(self):
        # Test get_next_event for verification with request_manual_check pending.
        processor = BaseProcessor(method=self.method, manual_check_delay=42)
        verification = await make_verification(self.session, status=VerificationStatus.CREATED,
                                               contact_id=self.contact_id, create_datetime=self.create_datetime,
                                               manual_check_request=True)

        result = processor.get_next_events(verification)

        details = {'contact_history_id': self.contact_history_id}
        initial = Comparison(InitialEmailEvent, due_datetime=None, method=None, _details=details, partial=True)
        context = {'initial_email_datetime': self.create_datetime.isoformat(),
                   'outzone_datetime': '2020-01-10T10:09:08+00:00', 'futile_datetime': '2020-01-24T10:09:08+00:00'}
        outzone_rem = Comparison(OutzoneReminderEvent, due_datetime=datetime(2020, 1, 3, 10, 9, 8, tzinfo=UTC),
                                 _details={}, context=context, partial=True)
        outzone = Comparison(DelayOutzoneEvent, due_datetime=datetime(2020, 1, 10, 10, 9, 8, tzinfo=UTC),
                             _details={'delay': 42}, partial=True)
        context = {'initial_email_datetime': self.create_datetime.isoformat(),
                   'futile_datetime': '2020-01-24T10:09:08+00:00'}
        futile_rem = Comparison(FutileReminderEvent, due_datetime=datetime(2020, 1, 17, 10, 9, 8, tzinfo=UTC),
                                _details={}, context=context, partial=True)
        futile = Comparison(DelayFutileEvent, due_datetime=datetime(2020, 1, 24, 10, 9, 8, tzinfo=UTC),
                            _details={'delay': 42}, partial=True)
        failed = Comparison(FailedEvent, due_datetime=datetime(2020, 1, 24, 10, 9, 8, tzinfo=UTC), partial=True)
        self.assertEqual(await alist(result), [initial, outzone_rem, outzone, futile_rem, futile, failed])
        # Check registry calls
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.contact_client)),
                         [call.get_contact_info(self.contact_id)])

    async def test_get_next_event_futile(self):
        # Test get_next_event for verification in FUTILE status.
        processor = BaseProcessor(method=self.method)
        verification = await make_verification(self.session, status=VerificationStatus.FUTILE,
                                               contact_id=self.contact_id, create_datetime=self.create_datetime)

        result = processor.get_next_events(verification)

        failed = Comparison(FailedEvent, due_datetime=datetime(2020, 1, 24, 10, 9, 8, tzinfo=UTC), partial=True)
        self.assertEqual(await alist(result), [failed])

    async def test_get_next_event_futile_domains(self):
        # Test get_next_event for verification in FUTILE status if contact has domains registered.
        processor = BaseProcessor(method=self.method)
        verification = await make_verification(self.session, status=VerificationStatus.FUTILE,
                                               contact_id=self.contact_id)
        cast(AsyncMock, APPLICATION.domain_client.get_domains_by_contact).return_value = {
            'domains': [{'domain': DomainRef(id=DomainId('42'), fqdn='kryten.example.org')}]}

        result = processor.get_next_events(verification)

        self.assertEqual(await alist(result), [])

    async def _test_get_next_event_notify(self, status: VerificationStatus, notify_cls: Type[ScheduledEvent]) -> None:
        # Test get_next_event for verification in status which triggers notification.
        processor = BaseProcessor(method=self.method)
        verification = await make_verification(self.session, status=status, contact_id=self.contact_id)

        result = processor.get_next_events(verification)

        notify = Comparison(notify_cls, due_datetime=None, method=None, _details={}, partial=True)
        self.assertEqual(await alist(result), [notify])

    async def _test_get_next_event_notify_sent(self, status: VerificationStatus, message_type: MessageType) -> None:
        # Test get_next_event for verification in status which triggers notification. Notification is already sent.
        processor = BaseProcessor(method=self.method)
        verification = await make_verification(self.session, status=status, contact_id=self.contact_id)
        details: Dict[str, Any] = {'message_type': message_type}
        await make_event(self.session, verification_id=verification.id, type=EventType.SEND_MESSAGE, details=details)

        result = processor.get_next_events(verification)

        self.assertEqual(await alist(result), [])

    async def _test_get_next_event_notify_contact_does_not_exist(self, status: VerificationStatus) -> None:
        # Test get_next_event for verification in status which triggers notification. Contact doesn't exist any more.
        processor = BaseProcessor(method=self.method)
        verification = await make_verification(self.session, status=status, contact_id=self.contact_id)

        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).side_effect = ContactDoesNotExist
        result = processor.get_next_events(verification)

        self.assertEqual(await alist(result), [])

    async def test_get_next_event_succeeded(self):
        # Test get_next_event for verification in SUCCEEDED status.
        await self._test_get_next_event_notify(VerificationStatus.SUCCEEDED, NotifyCompleteEvent)

    async def test_get_next_event_succeeded_sent(self):
        # Test get_next_event for verification in SUCCEEDED status with notification already sent.
        await self._test_get_next_event_notify_sent(VerificationStatus.SUCCEEDED, MessageType.SUCCEEDED)

    async def test_get_next_event_succeeded_contact_does_not_exist(self):
        # Test get_next_event for verification in SUCCEEDED status with contact not existing any more.
        await self._test_get_next_event_notify_contact_does_not_exist(VerificationStatus.SUCCEEDED)

    async def test_get_next_event_failed(self):
        # Test get_next_event for verification in FAILED status.
        await self._test_get_next_event_notify(VerificationStatus.FAILED, NotifyFailedEvent)

    async def test_get_next_event_failed_sent(self):
        # Test get_next_event for verification in FAILED status with notification already sent.
        await self._test_get_next_event_notify_sent(VerificationStatus.FAILED, MessageType.FAILED)

    async def test_get_next_event_failed_contact_does_not_exist(self):
        # Test get_next_event for verification in FAILED status with contact not existing any more.
        await self._test_get_next_event_notify_contact_does_not_exist(VerificationStatus.FAILED)

    async def test_get_next_event_canceled(self):
        # Test get_next_event for verification in CANCELED status.
        await self._test_get_next_event_notify(VerificationStatus.CANCELED, NotifyCancelEvent)

    async def test_get_next_event_canceled_sent(self):
        # Test get_next_event for verification in CANCELED status with notification already sent.
        await self._test_get_next_event_notify_sent(VerificationStatus.CANCELED, MessageType.CANCELED)

    async def test_get_next_event_canceled_contact_does_not_exist(self):
        # Test get_next_event for verification in CANCELED status with contact not existing any more.
        await self._test_get_next_event_notify_contact_does_not_exist(VerificationStatus.CANCELED)

    async def test_get_next_event_custom(self):
        # Test next event on a custom processor.
        processor = TestProcessor(method=self.method, next_event=sentinel.next)
        data: Tuple[Tuple[VerificationStatus, List[ScheduledEvent]], ...] = (
            # status, events
            (VerificationStatus.PAUSED, []),
            (VerificationStatus.MANUAL, []),
        )
        for status, events in data:
            with self.subTest(status=status):
                verification = Verification(status=status, create_datetime=self.create_datetime)
                self.assertEqual(await alist(processor.get_next_events(verification)), events)

    async def test_get_next_event_custom_created(self):
        # Test get_next_event for verification in CREATED status - custom test processor.
        processor = TestProcessor(method=self.method, next_event=sentinel.next)
        verification = await make_verification(self.session, status=VerificationStatus.CREATED,
                                               contact_id=self.contact_id, create_datetime=self.create_datetime)

        result = processor.get_next_events(verification)

        details = {'contact_history_id': self.contact_history_id}
        initial = Comparison(InitialEmailEvent, due_datetime=None, method=None, _details=details, partial=True)
        context = {'initial_email_datetime': self.create_datetime.isoformat(),
                   'outzone_datetime': '2020-01-10T10:09:08+00:00', 'futile_datetime': '2020-01-24T10:09:08+00:00'}
        outzone_rem = Comparison(OutzoneReminderEvent, due_datetime=datetime(2020, 1, 3, 10, 9, 8, tzinfo=UTC),
                                 _details={}, context=context, partial=True)
        outzone = Comparison(OutzoneEvent, due_datetime=datetime(2020, 1, 10, 10, 9, 8, tzinfo=UTC), partial=True)
        context = {'initial_email_datetime': self.create_datetime.isoformat(),
                   'futile_datetime': '2020-01-24T10:09:08+00:00'}
        futile_rem = Comparison(FutileReminderEvent, due_datetime=datetime(2020, 1, 17, 10, 9, 8, tzinfo=UTC),
                                _details={}, context=context, partial=True)
        futile = Comparison(FutileEvent, due_datetime=datetime(2020, 1, 24, 10, 9, 8, tzinfo=UTC), partial=True)
        failed = Comparison(FailedEvent, due_datetime=datetime(2020, 1, 24, 10, 9, 8, tzinfo=UTC), partial=True)
        self.assertEqual(await alist(result),
                         [initial, sentinel.next, outzone_rem, outzone, futile_rem, futile, failed])
        # Check registry calls
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.contact_client)),
                         [call.get_contact_info(self.contact_id)])

    async def _test_get_next_event_custom_init(self, status: VerificationStatus) -> None:
        # Test get_next_event for verification in INITIALIZED, ACCESSED and PENDING status.
        processor = TestProcessor(method=self.method, next_event=sentinel.next)
        verification = await make_verification(self.session, status=status, contact_id=self.contact_id,
                                               create_datetime=self.create_datetime)

        result = processor.get_next_events(verification)

        context = {'initial_email_datetime': self.create_datetime.isoformat(),
                   'outzone_datetime': '2020-01-10T10:09:08+00:00', 'futile_datetime': '2020-01-24T10:09:08+00:00'}
        outzone_rem = Comparison(OutzoneReminderEvent, due_datetime=datetime(2020, 1, 3, 10, 9, 8, tzinfo=UTC),
                                 _details={}, context=context, partial=True)
        outzone = Comparison(OutzoneEvent, due_datetime=datetime(2020, 1, 10, 10, 9, 8, tzinfo=UTC), partial=True)
        context = {'initial_email_datetime': self.create_datetime.isoformat(),
                   'futile_datetime': '2020-01-24T10:09:08+00:00'}
        futile_rem = Comparison(FutileReminderEvent, due_datetime=datetime(2020, 1, 17, 10, 9, 8, tzinfo=UTC),
                                _details={}, context=context, partial=True)
        futile = Comparison(FutileEvent, due_datetime=datetime(2020, 1, 24, 10, 9, 8, tzinfo=UTC), partial=True)
        failed = Comparison(FailedEvent, due_datetime=datetime(2020, 1, 24, 10, 9, 8, tzinfo=UTC), partial=True)
        self.assertEqual(await alist(result), [sentinel.next, outzone_rem, outzone, futile_rem, futile, failed])

    async def test_get_next_event_custom_initialized(self):
        await self._test_get_next_event_custom_init(VerificationStatus.INITIALIZED)

    async def test_get_next_event_custom_accessed(self):
        await self._test_get_next_event_custom_init(VerificationStatus.ACCESSED)

    async def test_get_next_event_custom_pending(self):
        await self._test_get_next_event_custom_init(VerificationStatus.PENDING)

    async def test_get_next_event_custom_check_domains(self):
        # Test CheckDomainsEvent is properly scheduled.
        with override_settings(check_domains={'rank_query': 'http://example.org'}):
            APPLICATION.contact_client = AsyncMock(ContactClient)
            APPLICATION.contact_client.get_contact_info.return_value = Contact(
                contact_id=self.contact_id, contact_handle=self.handle, contact_history_id=self.contact_history_id,
                emails=[self.email])
            APPLICATION.domain_client = AsyncMock(DomainClient)
            APPLICATION.domain_client.get_domains_by_contact.return_value = {'domains': []}
            processor = TestProcessor(method=self.method, next_event=sentinel.next)
            verification = await make_verification(self.session, status=VerificationStatus.INITIALIZED,
                                                   contact_id=self.contact_id, create_datetime=self.create_datetime)

            result = processor.get_next_events(verification)

            next_events = [sentinel.next, Comparison(OutzoneReminderEvent), Comparison(CheckDomainsEvent),
                           Comparison(OutzoneEvent), Comparison(FutileReminderEvent), Comparison(FutileEvent),
                           Comparison(FailedEvent)]
            self.assertEqual(await alist(result), next_events)

    async def test_get_next_event_custom_constrained(self):
        # Test get_next_event for verification in CONSTRAINED status.
        processor = TestProcessor(method=self.method, next_event=sentinel.next)
        verification = await make_verification(self.session, status=VerificationStatus.CONSTRAINED,
                                               contact_id=self.contact_id, create_datetime=self.create_datetime)

        result = processor.get_next_events(verification)

        context = {'initial_email_datetime': self.create_datetime.isoformat(),
                   'futile_datetime': '2020-01-24T10:09:08+00:00'}
        futile_rem = Comparison(FutileReminderEvent, due_datetime=datetime(2020, 1, 17, 10, 9, 8, tzinfo=UTC),
                                _details={}, context=context, partial=True)
        futile = Comparison(FutileEvent, due_datetime=datetime(2020, 1, 24, 10, 9, 8, tzinfo=UTC), partial=True)
        failed = Comparison(FailedEvent, due_datetime=datetime(2020, 1, 24, 10, 9, 8, tzinfo=UTC), partial=True)
        self.assertEqual(await alist(result), [sentinel.next, futile_rem, futile, failed])

    async def test_get_next_event_custom_futile(self):
        # Test get_next_event for verification in FUTILE status.
        processor = TestProcessor(method=self.method, next_event=sentinel.next)
        verification = await make_verification(self.session, status=VerificationStatus.FUTILE,
                                               contact_id=self.contact_id, create_datetime=self.create_datetime)

        result = processor.get_next_events(verification)

        failed = Comparison(FailedEvent, due_datetime=datetime(2020, 1, 24, 10, 9, 8, tzinfo=UTC), partial=True)
        self.assertEqual(await alist(result), [sentinel.next, failed])

    async def _test_get_next_event_custom_notify(self, status: VerificationStatus, notify_cls: Type[ScheduledEvent],
                                                 ) -> None:
        # Test get_next_event for verification in SUCCEEDED status.
        processor = TestProcessor(method=self.method, next_event=sentinel.next)
        verification = await make_verification(self.session, status=status, contact_id=self.contact_id)

        result = processor.get_next_events(verification)

        complete = Comparison(notify_cls, due_datetime=None, method=None, _details={}, partial=True)
        self.assertEqual(await alist(result), [complete])

    async def test_get_next_event_custom_succeeded(self):
        # Test get_next_event for verification in SUCCEEDED status.
        await self._test_get_next_event_custom_notify(VerificationStatus.SUCCEEDED, NotifyCompleteEvent)

    async def test_get_next_event_custom_failed(self):
        # Test get_next_event for verification in FAILED status.
        await self._test_get_next_event_custom_notify(VerificationStatus.FAILED, NotifyFailedEvent)

    async def test_get_next_event_custom_canceled(self):
        # Test get_next_event for verification in CANCELED status.
        await self._test_get_next_event_custom_notify(VerificationStatus.CANCELED, NotifyCancelEvent)

    async def test_get_resend_event_initial(self):
        processor = BaseProcessor(method=self.method)
        verification = await make_verification(self.session, contact_id=self.contact_id,
                                               create_datetime=self.create_datetime)

        scheduled = await processor.get_resend_event(verification, MessageType.INITIAL_EMAIL)

        self.assertIsInstance(scheduled, InitialEmailEvent)
        self.assertTrue(scheduled.resend)
        self.assertIsNone(scheduled.due_datetime)
        self.assertIsNone(scheduled.method)
        details = {'recipient': [self.email], 'contact_history_id': self.contact_history_id,
                   'method': TransportMethod.EMAIL, 'message_type': MessageType.INITIAL_EMAIL}
        self.assertEqual(await scheduled.get_details(), details)
        self.assertEqual(
            scheduled.context,
            {'outzone_datetime': '2020-01-10T10:09:08+00:00', 'futile_datetime': '2020-01-24T10:09:08+00:00'})
        # Check registry calls
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.contact_client)),
                         [call.get_contact_info(self.contact_id)])

    async def test_get_resend_event_unknown(self):
        processor = BaseProcessor(method=self.method)
        verification = await make_verification(self.session, contact_id=self.contact_id)

        with self.assertRaisesRegex(ValueError, "Message type 'unknown' not found"):
            await processor.get_resend_event(verification, 'unknown')

    async def test_get_cancel_events(self):
        processor = BaseProcessor(method=self.method)
        cancel = Comparison(CancelVerificationEvent, due_datetime=None, method=None, partial=True)
        outzone = Comparison(CancelOutzoneEvent, due_datetime=None, method=None, partial=True)
        futile = Comparison(CancelFutileEvent, due_datetime=None, method=None, partial=True)
        data: Tuple[Tuple[VerificationStatus, Optional[VerificationStatus], List[ScheduledEvent]], ...] = (
            # status, canceled, events
            # Full cancelation
            (VerificationStatus.CREATED, None, [cancel]),
            (VerificationStatus.INITIALIZED, None, [cancel]),
            (VerificationStatus.ACCESSED, None, [cancel]),
            (VerificationStatus.PENDING, None, [cancel]),
            (VerificationStatus.CONSTRAINED, None, [outzone, cancel]),
            (VerificationStatus.FUTILE, None, [futile, outzone, cancel]),
            (VerificationStatus.PAUSED, None, [futile, outzone, cancel]),
            (VerificationStatus.SUCCEEDED, None, []),
            (VerificationStatus.FAILED, None, []),
            (VerificationStatus.CANCELED, None, []),
            (VerificationStatus.MANUAL, None, []),
            # Partial cancelation
            (VerificationStatus.CREATED, VerificationStatus.CREATED, []),
            (VerificationStatus.INITIALIZED, VerificationStatus.CREATED, []),
            (VerificationStatus.ACCESSED, VerificationStatus.CREATED, []),
            (VerificationStatus.PENDING, VerificationStatus.CREATED, []),
            (VerificationStatus.CONSTRAINED, VerificationStatus.CREATED, []),
            (VerificationStatus.FUTILE, VerificationStatus.CREATED, []),
            (VerificationStatus.PAUSED, VerificationStatus.CREATED, []),
            (VerificationStatus.SUCCEEDED, VerificationStatus.CREATED, []),
            (VerificationStatus.FAILED, VerificationStatus.CREATED, []),
            (VerificationStatus.CANCELED, VerificationStatus.CREATED, []),
            (VerificationStatus.MANUAL, VerificationStatus.CREATED, []),
            # Important partial cancelations.
            (VerificationStatus.FUTILE, VerificationStatus.FUTILE, [futile]),
            (VerificationStatus.FUTILE, VerificationStatus.CONSTRAINED, []),
            (VerificationStatus.CONSTRAINED, VerificationStatus.CONSTRAINED, [outzone]),
            # Forward cancelations
            (VerificationStatus.CREATED, VerificationStatus.FUTILE, []),
            (VerificationStatus.CREATED, VerificationStatus.CONSTRAINED, []),
            (VerificationStatus.CONSTRAINED, VerificationStatus.FUTILE, []),
        )
        for i, (status, canceled, events) in enumerate(data):
            with self.subTest(status=status, canceled=canceled):
                verification = await make_verification(self.session, contact_id=self.contact_id + str(i), status=status,
                                                       create_datetime=self.create_datetime)
                with freeze_time(self.create_datetime):
                    self.assertEqual(await alist(processor.get_cancel_events(verification, canceled)), events)

    async def test_get_cancel_events_due_datetimes(self):
        # Test cancel events dependent on due datetimes.
        processor = BaseProcessor(method=self.method)
        cancel = Comparison(CancelVerificationEvent, due_datetime=None, method=None, partial=True)
        outzone = Comparison(CancelOutzoneEvent, due_datetime=None, method=None, partial=True)
        futile = Comparison(CancelFutileEvent, due_datetime=None, method=None, partial=True)
        # On '2019-12-12' are events to cancel not due yet. On '2020-06-01' they are.
        data: Tuple[Tuple[VerificationStatus, Optional[VerificationStatus], str, List[ScheduledEvent]], ...] = (
            # status, canceled, now, events
            (VerificationStatus.CONSTRAINED, None, '2019-12-12', [outzone, cancel]),
            (VerificationStatus.CONSTRAINED, None, '2020-06-01', [outzone, cancel]),
            (VerificationStatus.CONSTRAINED, VerificationStatus.CONSTRAINED, '2019-12-12', [outzone]),
            (VerificationStatus.CONSTRAINED, VerificationStatus.CONSTRAINED, '2020-06-01', []),
            (VerificationStatus.FUTILE, None, '2019-12-12', [futile, outzone, cancel]),
            (VerificationStatus.FUTILE, None, '2020-06-01', [futile, outzone, cancel]),
            (VerificationStatus.FUTILE, VerificationStatus.FUTILE, '2019-12-12', [futile]),
            (VerificationStatus.FUTILE, VerificationStatus.FUTILE, '2020-06-01', []),
            (VerificationStatus.PAUSED, None, '2019-12-12', [futile, outzone, cancel]),
            (VerificationStatus.PAUSED, None, '2020-06-01', [futile, outzone, cancel]),
            # Cancel CONSTRAINED is a noop if the verification is FUTILE.
            (VerificationStatus.FUTILE, VerificationStatus.CONSTRAINED, '2019-12-12', []),
            (VerificationStatus.FUTILE, VerificationStatus.CONSTRAINED, '2020-06-01', []),
            # Cancel CONSTRAINED or FUTILE is a noop if the verification is PAUSED.
            (VerificationStatus.PAUSED, VerificationStatus.CONSTRAINED, '2019-12-12', []),
            (VerificationStatus.PAUSED, VerificationStatus.CONSTRAINED, '2020-06-01', []),
            (VerificationStatus.PAUSED, VerificationStatus.FUTILE, '2019-12-12', []),
            (VerificationStatus.PAUSED, VerificationStatus.FUTILE, '2020-06-01', []),
        )
        for i, (status, canceled, now, events) in enumerate(data):
            with self.subTest(status=status, canceled=canceled, now=now):
                verification = await make_verification(self.session, contact_id=self.contact_id + str(i), status=status,
                                                       create_datetime=self.create_datetime)
                with freeze_time(now):
                    self.assertEqual(await alist(processor.get_cancel_events(verification, canceled)), events)

    async def test_get_cancel_events_custom(self):
        # Test cancel events on a custom processor.
        processor = TestProcessor(method=self.method, cancel_event=sentinel.cancel)
        cancel = Comparison(CancelVerificationEvent, due_datetime=None, method=None, partial=True)
        outzone = Comparison(CancelOutzoneEvent, due_datetime=None, method=None, partial=True)
        futile = Comparison(CancelFutileEvent, due_datetime=None, method=None, partial=True)
        # Only test full cancelation. We only check the order of events.
        data: Tuple[Tuple[VerificationStatus, List[ScheduledEvent]], ...] = (
            # status, events
            (VerificationStatus.CREATED, [sentinel.cancel, cancel]),
            (VerificationStatus.INITIALIZED, [sentinel.cancel, cancel]),
            (VerificationStatus.ACCESSED, [sentinel.cancel, cancel]),
            (VerificationStatus.PENDING, [sentinel.cancel, cancel]),
            (VerificationStatus.CONSTRAINED, [outzone, sentinel.cancel, cancel]),
            (VerificationStatus.FUTILE, [futile, outzone, sentinel.cancel, cancel]),
            (VerificationStatus.PAUSED, [futile, outzone, sentinel.cancel, cancel]),
            (VerificationStatus.SUCCEEDED, []),
            (VerificationStatus.FAILED, []),
            (VerificationStatus.CANCELED, []),
            (VerificationStatus.MANUAL, []),
        )
        for i, (status, events) in enumerate(data):
            with self.subTest(status=status):
                verification = await make_verification(self.session, contact_id=self.contact_id + str(i), status=status,
                                                       create_datetime=self.create_datetime)
                with freeze_time(self.create_datetime):
                    self.assertEqual(await alist(processor.get_cancel_events(verification)), events)

    async def test_get_cancel_events_closed(self):
        # Test cancel events of closed statuses raise error.
        processor = BaseProcessor(method=self.method)
        for status in (VerificationStatus.SUCCEEDED, VerificationStatus.FAILED, VerificationStatus.CANCELED,
                       VerificationStatus.MANUAL):
            with self.subTest(status=status):
                with self.assertRaisesRegex(ValueError, "Can't cancel closed status."):
                    await alist(processor.get_cancel_events(sentinel.verification, status))

    async def test_get_complete_events(self):
        processor = BaseProcessor(method=self.method)
        complete = Comparison(CompleteEvent, due_datetime=None, method=self.method, partial=True)
        outzone = Comparison(CancelOutzoneEvent, due_datetime=None, method=None, partial=True)
        futile = Comparison(CancelFutileEvent, due_datetime=None, method=None, partial=True)
        data: Tuple[Tuple[VerificationStatus, List[ScheduledEvent]], ...] = (
            # status, events
            (VerificationStatus.CREATED, [complete]),
            (VerificationStatus.INITIALIZED, [complete]),
            (VerificationStatus.ACCESSED, [complete]),
            (VerificationStatus.PENDING, [complete]),
            (VerificationStatus.CONSTRAINED, [outzone, complete]),
            (VerificationStatus.FUTILE, [futile, outzone, complete]),
            (VerificationStatus.PAUSED, [futile, outzone, complete]),
            (VerificationStatus.SUCCEEDED, []),
            (VerificationStatus.FAILED, []),
            (VerificationStatus.CANCELED, []),
            (VerificationStatus.MANUAL, []),
        )
        for i, (status, events) in enumerate(data):
            with self.subTest(status=status):
                verification = await make_verification(self.session, contact_id=self.contact_id + str(i), status=status)
                self.assertEqual(await alist(processor.get_complete_events(verification)), events)

    async def test_get_complete_events_custom(self):
        # Test complete events on a custom processor.
        processor = TestProcessor(method=self.method, complete_event=sentinel.complete)
        complete = Comparison(CompleteEvent, due_datetime=None, method=self.method, partial=True)
        outzone = Comparison(CancelOutzoneEvent, due_datetime=None, method=None, partial=True)
        futile = Comparison(CancelFutileEvent, due_datetime=None, method=None, partial=True)
        data: Tuple[Tuple[VerificationStatus, List[ScheduledEvent]], ...] = (
            # status, events
            (VerificationStatus.CREATED, [sentinel.complete, complete]),
            (VerificationStatus.INITIALIZED, [sentinel.complete, complete]),
            (VerificationStatus.ACCESSED, [sentinel.complete, complete]),
            (VerificationStatus.PENDING, [sentinel.complete, complete]),
            (VerificationStatus.CONSTRAINED, [outzone, sentinel.complete, complete]),
            (VerificationStatus.FUTILE, [futile, outzone, sentinel.complete, complete]),
            (VerificationStatus.PAUSED, [futile, outzone, sentinel.complete, complete]),
            (VerificationStatus.SUCCEEDED, []),
            (VerificationStatus.FAILED, []),
            (VerificationStatus.CANCELED, []),
            (VerificationStatus.MANUAL, []),
        )
        for i, (status, events) in enumerate(data):
            with self.subTest(status=status):
                verification = await make_verification(self.session, contact_id=self.contact_id + str(i), status=status)
                self.assertEqual(await alist(processor.get_complete_events(verification)), events)

    async def test_check_complete_nothing(self):
        # Test no verified data.
        processor = BaseProcessor(method=self.method)
        verification = await make_verification(self.session, contact_id=self.contact_id)

        self.assertFalse(await processor.check_complete(verification))

    async def test_check_complete_incomplete(self):
        # Test verified data are not complete.
        processor = BaseProcessor(method=self.method)
        verification = await make_verification(self.session, contact_id=self.contact_id)
        await make_contact_data(self.session, verification_id=verification.id,
                                contact_history_id=self.contact_history_id, name='', verified_name=True)

        self.assertFalse(await processor.check_complete(verification))

    async def test_check_complete_person(self):
        # Test verified data for a person.
        processor = BaseProcessor(method=self.method)
        verification = await make_verification(self.session, contact_id=self.contact_id)
        await make_contact_data(
            self.session, verification_id=verification.id, contact_history_id=self.contact_history_id,
            name='', verified_name=True, organization='', verified_organization=True,
            address_city='', verified_address=True)

        self.assertTrue(await processor.check_complete(verification))

    async def test_check_complete_person_unknown(self):
        # Test verified data for a person is missing verified organization (we don't know it's a person).
        processor = BaseProcessor(method=self.method)
        verification = await make_verification(self.session, contact_id=self.contact_id)
        await make_contact_data(
            self.session, verification_id=verification.id, contact_history_id=self.contact_history_id,
            name='', verified_name=True, address_city='', verified_address=True)

        self.assertFalse(await processor.check_complete(verification))

    async def test_check_complete_person_no_name(self):
        # Test verified data for a person is missing an address.
        processor = BaseProcessor(method=self.method)
        verification = await make_verification(self.session, contact_id=self.contact_id)
        await make_contact_data(
            self.session, verification_id=verification.id, contact_history_id=self.contact_history_id,
            organization='', verified_organization=True, address_city='', verified_address=True)

        self.assertFalse(await processor.check_complete(verification))

    async def test_check_complete_person_no_address(self):
        # Test verified data for a person is missing an address.
        processor = BaseProcessor(method=self.method)
        verification = await make_verification(self.session, contact_id=self.contact_id)
        await make_contact_data(
            self.session, verification_id=verification.id, contact_history_id=self.contact_history_id,
            name='', verified_name=True, organization='', verified_organization=True)

        self.assertFalse(await processor.check_complete(verification))

    async def test_check_complete_company(self):
        # Test verified data for a company.
        processor = BaseProcessor(method=self.method)
        verification = await make_verification(self.session, contact_id=self.contact_id)
        await make_contact_data(
            self.session, verification_id=verification.id, contact_history_id=self.contact_history_id,
            organization='JMC', verified_organization=True, address_city='', verified_address=True)

        self.assertTrue(await processor.check_complete(verification))

    async def test_check_complete_company_unknown(self):
        # Test verified data for a company is missing verified organization (we don't know it's a company).
        processor = BaseProcessor(method=self.method)
        verification = await make_verification(self.session, contact_id=self.contact_id)
        await make_contact_data(
            self.session, verification_id=verification.id, contact_history_id=self.contact_history_id,
            address_city='', verified_address=True)

        self.assertFalse(await processor.check_complete(verification))

    async def test_check_complete_company_no_address(self):
        # Test verified data for a company is missing company reg. number.
        processor = BaseProcessor(method=self.method)
        verification = await make_verification(self.session, contact_id=self.contact_id)
        await make_contact_data(
            self.session, verification_id=verification.id, contact_history_id=self.contact_history_id,
            organization='JMC', verified_organization=True)

        self.assertFalse(await processor.check_complete(verification))

    async def test_check_complete_split(self):
        # Test verified data are split among several records.
        processor = BaseProcessor(method=self.method)
        verification = await make_verification(self.session, contact_id=self.contact_id)
        await make_contact_data(self.session, verification_id=verification.id,
                                contact_history_id=self.contact_history_id)
        await make_contact_data(
            self.session, verification_id=verification.id, contact_history_id=self.contact_history_id,
            name='', verified_name=True, organization='', verified_organization=True,
            address_city='', verified_address=True)

        self.assertTrue(await processor.check_complete(verification))

    async def test_check_complete_method(self):
        # Test verified data are provided for another method.
        processor = BaseProcessor(method=self.method)
        verification = await make_verification(self.session, contact_id=self.contact_id)
        await make_contact_data(
            self.session, verification_id=verification.id, contact_history_id=self.contact_history_id,
            name='', verified_name=True, organization='', verified_organization=True,
            address_city='', verified_address=True, method='ANOTHER')

        self.assertFalse(await processor.check_complete(verification))

    async def test_token_verified(self):
        processor = BaseProcessor(method=self.method)

        # It's a noop.
        self.assertIsNone(await processor.handle_token_verified(Token()))  # type: ignore[func-returns-value]
