from datetime import datetime
from typing import List, Optional, cast
from unittest.mock import AsyncMock, MagicMock, Mock, call, patch, sentinel
from uuid import UUID

from asyncstdlib import list as alist
from filed import Storage
from fred_types import ContactHistoryId, ContactId
from hermes import Letter, LetterMessengerClient
from hermes.letter import Address
from hermes.message import Reference
from pytz import UTC
from regal import Address as ContactAddress, Contact, ContactClient, DomainClient
from regal.exceptions import ContactDoesNotExist
from sqlalchemy import select
from testfixtures import Comparison

from verification.constants import EventType, MessageType, Template, TransportMethod, VerificationStatus
from verification.models import ContactData, Token, Verification
from verification.processors import (
    ConfirmationLetterEvent,
    FailedEvent,
    FutileEvent,
    FutileReminderEvent,
    LetterProcessor,
    OutzoneEvent,
    OutzoneReminderEvent,
)
from verification.processors.base import ScheduledEvent
from verification.settings import APPLICATION

from ..constants import ReferenceType
from .utils import TransactionTestCase, make_event, make_token, make_verification, mock_awaits


class ConfirmationLetterEventTest(TransactionTestCase):
    uuid = UUID(int=42)
    contact_id = ContactId('KRYTEN')
    contact_history_id = ContactHistoryId('42')
    handle = '2X4B'
    method = 'ANALYSIS'

    async def asyncSetUp(self):
        APPLICATION.contact_client = AsyncMock(ContactClient)
        APPLICATION.contact_client.get_contact_info.return_value = Contact(
            contact_id=self.contact_id, contact_handle=self.handle, contact_history_id=self.contact_history_id)

        APPLICATION.file_storage = MagicMock(Storage)
        APPLICATION.letter_client = AsyncMock(LetterMessengerClient)
        APPLICATION.secretary_client = Mock(spec=('render_pdf', ))

        patcher = patch('secrets.choice', side_effect='Gazpacho!!' * 2)
        self.addCleanup(patcher.stop)
        self.mock_secrets = patcher.start()

    async def _test_process(self):
        APPLICATION.file_storage.create.return_value.__enter__.return_value.uuid = UUID(int=196)
        cast(AsyncMock, APPLICATION.letter_client.send).return_value = sentinel.message_id
        cast(Mock, APPLICATION.secretary_client).render_pdf.return_value = sentinel.content

        address = {'name': 'Kryten', 'organization': 'Jupiter Mining Corporation', 'street': 'Deck 16',
                   'city': 'Red Dwarf', 'state_or_province': 'Deep Space', 'postal_code': 'JMC', 'country': 'JU'}

        query = select(Verification).filter_by(uuid=self.uuid)
        verification = (await self.session.execute(query)).scalars().one()
        scheduled = ConfirmationLetterEvent(verification, details={'recipient': address}, method=self.method)

        await scheduled.process()

        # Check token created
        token = (await self.session.execute(select(Token))).scalars().one()
        self.assertEqual(token.method, self.method)
        self.assertEqual(token.token, 'Gazpacho!!')
        self.assertEqual(token.verification_id, verification.id)

        details = {'method': TransportMethod.LETTER, 'message_type': MessageType.CONFIRMATION_LETTER,
                   'recipient': address, 'message_id': sentinel.message_id, 'token': str(token.uuid)}
        self.assertEqual(await scheduled.get_details(), details)
        self.assertEqual(verification.status, VerificationStatus.CREATED)
        # Check secretary calls
        context = {'address': address, 'token': 'Gazpacho!!', 'handle': self.handle}
        self.assertEqual(cast(Mock, APPLICATION.secretary_client).mock_calls,
                         [call.render_pdf(Template.CONFIRMATION_LETTER, context=context)])
        # Check fileman calls
        file_calls = [call.create(name='verification-letter.pdf', mimetype='application/pdf'),
                      call.create().__enter__(),
                      call.create().__enter__().write(sentinel.content),
                      call.create().__exit__(None, None, None)]
        self.assertEqual(APPLICATION.file_storage.mock_calls, file_calls)
        # Check messenger calls
        letter = Letter(recipient=Address(**address), file=str(UUID(int=196)), type=MessageType.CONFIRMATION_LETTER)
        calls = [call.send(letter, references=[Reference(type=ReferenceType.CONTACT, value=self.contact_id),
                                               Reference(type=ReferenceType.VERIFICATION, value=str(self.uuid))])]
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.letter_client)), calls)

    async def test_process(self):
        await make_verification(self.session, contact_id=self.contact_id, uuid=self.uuid)
        await self._test_process()

    async def test_process_token_conflict(self):
        # Test existing token is deleted before a new one is created.
        verification = await make_verification(self.session, contact_id=self.contact_id, uuid=self.uuid)
        await make_token(self.session, verification_id=verification.id, token='Original!', method=self.method)
        await self._test_process()

    async def _test_process_letter_limit_bonus(self, resend: bool = False, final_delay: int = 86400) -> None:
        APPLICATION.file_storage.create.return_value.__enter__.return_value.uuid = UUID(int=196)
        address = {'name': 'Kryten', 'organization': 'Jupiter Mining Corporation', 'street': 'Deck 16',
                   'city': 'Red Dwarf', 'state_or_province': 'Deep Space', 'postal_code': 'JMC', 'country': 'JU'}

        verification = await make_verification(self.session, contact_id=self.contact_id)
        days_delay = 1
        scheduled = ConfirmationLetterEvent(
            verification, details={'recipient': address}, method=self.method, letter_limit_bonus=days_delay,
            resend=resend,
        )

        await scheduled.process()

        # Check status delays
        self.assertEqual(await verification.get_status_delay(VerificationStatus.CONSTRAINED), final_delay)
        self.assertEqual(await verification.get_status_delay(VerificationStatus.FUTILE), final_delay)

    async def test_process_letter_limit_bonus(self):
        # Test status delays are prolonged on letter send.
        await self._test_process_letter_limit_bonus()

    async def test_process_letter_limit_bonus_resend(self):
        # Test status delays are not prolonged on letter resend.
        await self._test_process_letter_limit_bonus(resend=True, final_delay=0)


class LetterProcessorTest(TransactionTestCase):
    contact_id = 'KRYTEN'
    contact_history_id = ContactHistoryId('NOVA5')
    method = 'ANALYSIS'
    name = 'Kryten'
    organization = 'Jupiter Mining Corporation'
    place = ContactAddress(street=['Deck 16'], city='Red Dwarf', state_or_province='Deep Space', postal_code='JMC',
                           country_code='JU')
    email = 'kryten@example.org'
    create_datetime = datetime(2019, 12, 11, 10, 9, 8, tzinfo=UTC)

    async def asyncSetUp(self):
        APPLICATION.contact_client = AsyncMock(ContactClient)
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = Contact(
            contact_id=sentinel.contact_id, contact_history_id=self.contact_history_id, contact_handle='2X4B',
            name=self.name, organization=self.organization, place=self.place, emails=[self.email])
        APPLICATION.domain_client = AsyncMock(DomainClient)
        APPLICATION.domain_client.get_domains_by_contact.return_value = {'domains': []}

    async def _test_get_next_event_letter(self, verification: Verification,
                                          other_events: Optional[List[ScheduledEvent]] = None) -> None:
        # Test cases when letter is generated.
        processor = LetterProcessor(method=self.method)

        result = processor.get_next_events(verification)

        address = {'name': self.name, 'organization': self.organization, 'street': 'Deck 16', 'city': 'Red Dwarf',
                   'state_or_province': 'Deep Space', 'postal_code': 'JMC', 'country': 'JU'}
        details = {'recipient': address, 'contact_history_id': self.contact_history_id}
        context = {'initial_email_datetime': self.create_datetime.isoformat()}
        letter = Comparison(ConfirmationLetterEvent, due_datetime=None, method=self.method, _details=details,
                            context=context, partial=True)
        scheduled = [letter, Comparison(OutzoneReminderEvent), Comparison(OutzoneEvent),
                     Comparison(FutileReminderEvent), Comparison(FutileEvent), Comparison(FailedEvent)]
        self.assertEqual(await alist(result), scheduled)

    async def test_get_next_event(self):
        verification = await make_verification(self.session, status=VerificationStatus.INITIALIZED,
                                               contact_id=self.contact_id, create_datetime=self.create_datetime)
        await self._test_get_next_event_letter(verification)

    async def test_get_next_event_letter_sent(self):
        # Test letter is not sent if it was already sent.
        verification = await make_verification(self.session, status=VerificationStatus.INITIALIZED,
                                               contact_id=self.contact_id, create_datetime=self.create_datetime)
        details = {'message_type': MessageType.CONFIRMATION_LETTER}
        await make_event(self.session, verification_id=verification.id, type=EventType.SEND_MESSAGE,
                         details=details)
        processor = LetterProcessor(method=self.method)

        result = processor.get_next_events(verification)

        scheduled = [Comparison(OutzoneReminderEvent), Comparison(OutzoneEvent), Comparison(FutileReminderEvent),
                     Comparison(FutileEvent), Comparison(FailedEvent)]
        self.assertEqual(await alist(result), scheduled)

    async def test_get_next_event_other_message(self):
        # Test letter is sent if other message was sent.
        verification = await make_verification(self.session, status=VerificationStatus.INITIALIZED,
                                               contact_id=self.contact_id, create_datetime=self.create_datetime)
        details = {'message_type': MessageType.INITIAL_EMAIL}
        await make_event(self.session, verification_id=verification.id, type=EventType.SEND_MESSAGE,
                         details=details)
        await self._test_get_next_event_letter(verification)

    async def test_get_next_event_other_event(self):
        # Test letter is sent if other message was sent.
        verification = await make_verification(self.session, status=VerificationStatus.INITIALIZED,
                                               contact_id=self.contact_id, create_datetime=self.create_datetime)
        details = {'message_type': MessageType.CONFIRMATION_LETTER}
        await make_event(self.session, verification_id=verification.id, type=EventType.CANCEL, details=details)
        await self._test_get_next_event_letter(verification)

    async def test_get_resend_event_letter(self):
        processor = LetterProcessor(method=self.method)
        verification = await make_verification(self.session, contact_id=self.contact_id,
                                               create_datetime=self.create_datetime)

        scheduled = await processor.get_resend_event(verification, MessageType.CONFIRMATION_LETTER)

        self.assertIsInstance(scheduled, ConfirmationLetterEvent)
        self.assertTrue(scheduled.resend)
        self.assertIsNone(scheduled.due_datetime)
        self.assertEqual(scheduled.method, self.method)
        address = {'name': self.name, 'organization': self.organization, 'street': 'Deck 16', 'city': 'Red Dwarf',
                   'state_or_province': 'Deep Space', 'postal_code': 'JMC', 'country': 'JU'}
        details = {'recipient': address, 'contact_history_id': self.contact_history_id,
                   'method': TransportMethod.LETTER, 'message_type': MessageType.CONFIRMATION_LETTER}
        self.assertEqual(await scheduled.get_details(), details)
        self.assertEqual(scheduled.context, {'initial_email_datetime': self.create_datetime.isoformat()})

    async def test_get_resend_event_unknown(self):
        processor = LetterProcessor(method=self.method)
        verification = await make_verification(self.session, contact_id=self.contact_id)

        with self.assertRaisesRegex(ValueError, "Message type 'unknown' not found"):
            await processor.get_resend_event(verification, 'unknown')

    async def test_handle_token_verified(self):
        verification = await make_verification(self.session, contact_id='Kryten')
        token = await make_token(self.session, verification_id=verification.id, method=self.method)
        details = {'recipient': {'name': 'Kryten', 'organization': 'Jupiter Mining Corporation',
                                 'street': ['Deck 16'], 'city': 'Red Dwarf', 'state_or_province': 'Deep Space',
                                 'postal_code': 'JMC', 'country': 'JU'},
                   'contact_history_id': self.contact_history_id,
                   'token': str(token.uuid)}
        await make_event(self.session, verification_id=verification.id, type=EventType.SEND_MESSAGE, details=details,
                         method=self.method)
        processor = LetterProcessor(method=self.method)

        await processor.handle_token_verified(token)

        # Check contact data created.
        contact_data = (await self.session.execute(select(ContactData))).scalars().one()
        self.assertEqual(contact_data.method, self.method)
        self.assertEqual(contact_data.contact_history_id, self.contact_history_id)
        self.assertEqual(contact_data.name, 'Kryten')
        self.assertTrue(contact_data.verified_name)
        self.assertEqual(contact_data.organization, 'Jupiter Mining Corporation')
        self.assertTrue(contact_data.verified_organization)
        self.assertEqual(contact_data.address_street, ['Deck 16'])
        self.assertEqual(contact_data.address_city, 'Red Dwarf')
        self.assertEqual(contact_data.address_state, 'Deep Space')
        self.assertEqual(contact_data.address_postal_code, 'JMC')
        self.assertEqual(contact_data.address_country_code, 'JU')
        self.assertTrue(contact_data.verified_address)

    async def test_handle_token_verified_event_mismatch(self):
        # Check callback is not triggered on mismatching event.
        verification = await make_verification(self.session, contact_id='Kryten')
        token = await make_token(self.session, verification_id=verification.id, method=self.method)
        await make_event(self.session, verification_id=verification.id, type=EventType.SEND_MESSAGE, method=self.method,
                         details={'token': str(UUID(int=196))})
        processor = LetterProcessor(method=self.method)

        await processor.handle_token_verified(token)

        # Check contact data not created.
        self.assertIsNone((await self.session.execute(select(ContactData))).scalar())

    async def test_handle_token_verified_event_invalid_token(self):
        # Check callback is not triggered on mismatching event.
        verification = await make_verification(self.session, contact_id='Kryten')
        token = await make_token(self.session, verification_id=verification.id, method=self.method)
        await make_event(self.session, verification_id=verification.id, type=EventType.SEND_MESSAGE, method=self.method,
                         details={'token': 'invalid'})
        processor = LetterProcessor(method=self.method)

        await processor.handle_token_verified(token)

        # Check contact data not created.
        self.assertIsNone((await self.session.execute(select(ContactData))).scalar())

    async def test_handle_token_verified_event_missing(self):
        # Check callback doesn't fail if the event is missing.
        verification = await make_verification(self.session, contact_id='Kryten')
        token = await make_token(self.session, verification_id=verification.id, method=self.method)
        processor = LetterProcessor(method=self.method)

        await processor.handle_token_verified(token)

        # Check contact data not created.
        self.assertIsNone((await self.session.execute(select(ContactData))).scalar())

    async def test_is_applicable_no_address(self):
        # address not available
        verification = await make_verification(self.session, contact_id='Kryten')
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = Contact(
            contact_id=ContactId('Kryten'), contact_handle='chnd:1'
        )
        processor = LetterProcessor(method=self.method)

        self.assertFalse(await processor.is_applicable(verification))

        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).assert_awaited_once_with(ContactId('Kryten'))

    async def test_is_applicable_country_not_limited(self):
        # countries not limited
        verification = await make_verification(self.session, contact_id='Kryten')
        address = ContactAddress(street=['Deck 16'], city='Red Dwarf', country_code='JU')
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = Contact(
            contact_id=ContactId('Kryten'), contact_handle='chnd:1', place=address
        )
        processor = LetterProcessor(method=self.method)

        self.assertTrue(await processor.is_applicable(verification))

    async def test_is_applicable_country_not_allowed(self):
        # limited countries, country not allowed
        verification = await make_verification(self.session, contact_id='Kryten')
        address = ContactAddress(street=['Deck 16'], city='Red Dwarf', country_code='JU')
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = Contact(
            contact_id=ContactId('Kryten'), contact_handle='chnd:1', place=address
        )
        processor = LetterProcessor(method=self.method, allowed_countries=['ME', 'VE', 'EA', 'MA'])

        self.assertFalse(await processor.is_applicable(verification))

    async def test_is_applicable_country_allowed(self):
        # limited countries, country allowed
        verification = await make_verification(self.session, contact_id='Kryten')
        address = ContactAddress(street=['Deck 16'], city='Red Dwarf', country_code='JU')
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = Contact(
            contact_id=ContactId('Kryten'), contact_handle='chnd:1', place=address
        )
        processor = LetterProcessor(method=self.method, allowed_countries=['ME', 'VE', 'EA', 'MA', 'JU'])

        self.assertTrue(await processor.is_applicable(verification))

    async def test_is_applicable_contact_does_not_exist(self):
        # Test method is not applicable if contact no longer exist.
        verification = await make_verification(self.session, contact_id='Kryten')
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).side_effect = ContactDoesNotExist
        processor = LetterProcessor(method=self.method)

        self.assertFalse(await processor.is_applicable(verification))

    async def test_letter_limit_bonus(self):
        # Check propagation of bonus to ConfirmationLetterEvent
        processor = LetterProcessor(method=self.method, letter_limit_bonus=3)
        verification = await make_verification(self.session, contact_id=self.contact_id)

        event = await processor._get_confirmation_letter_event(verification)

        self.assertEqual(event.letter_limit_bonus, 3)
