from datetime import datetime
from typing import cast
from unittest.mock import AsyncMock, call, sentinel
from uuid import UUID

from asyncstdlib import list as alist
from fred_types import ContactHistoryId, ContactId, RepresentativeId
from pytz import UTC
from regal import Address as ContactAddress, Contact, ContactClient, DomainClient, Representative, RepresentativeClient
from regal.exceptions import ContactDoesNotExist, RepresentativeDoesNotExist
from testfixtures import Comparison

from verification.constants import EventType, MessageType, TransportMethod, VerificationStatus
from verification.database import object_session
from verification.models import Verification
from verification.processors import (
    FailedEvent,
    FutileEvent,
    FutileReminderEvent,
    OutzoneEvent,
    OutzoneReminderEvent,
    RepresentativeLetterEvent,
    RepresentativeProcessor,
)
from verification.settings import APPLICATION

from .utils import TransactionTestCase, make_event, make_token, make_verification, mock_awaits


class RepresentativeProcessorTest(TransactionTestCase):
    contact_id = ContactId('KRYTEN')
    contact_history_id = ContactHistoryId('NOVA5')
    method = 'ANALYSIS'
    name = 'Kryten'
    organization = 'Jupiter Mining Corporation'
    address_data = {'street': 'Deck 16', 'city': 'Red Dwarf', 'state_or_province': 'Deep Space', 'postal_code': 'JMC',
                    'country': 'JU'}
    address = ContactAddress(street=['Deck 16'], city='Red Dwarf', state_or_province='Deep Space', postal_code='JMC',
                             country_code='JU')
    telephone = 'Arnie!'
    email = 'arnold@example.org'
    create_datetime = datetime(2019, 12, 11, 10, 9, 8, tzinfo=UTC)

    async def asyncSetUp(self):
        APPLICATION.contact_client = AsyncMock(ContactClient)
        APPLICATION.domain_client = AsyncMock(DomainClient)
        APPLICATION.domain_client.get_domains_by_contact.return_value = {'domains': []}
        APPLICATION.representative_client = AsyncMock(RepresentativeClient)

    async def test_is_applicable_no_address(self):
        verification = await make_verification(self.session, contact_id='Kryten')
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = Contact(
            contact_id=ContactId('Kryten'), contact_handle='chnd:1')
        processor = RepresentativeProcessor(method=self.method)

        self.assertFalse(await processor.is_applicable(verification))

        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.contact_client)),
                         [call.get_contact_info(ContactId('Kryten'))])

    async def test_is_applicable_unlimited(self):
        # Countries not limited
        verification = await make_verification(self.session, contact_id='Kryten')
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = Contact(
            contact_id=ContactId('Kryten'), contact_handle='chnd:1',
            place=ContactAddress(street=['Deck 16'], city='Red Dwarf', country_code='JU'))
        processor = RepresentativeProcessor(method=self.method)

        self.assertTrue(await processor.is_applicable(verification))

        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.contact_client)),
                         [call.get_contact_info(ContactId('Kryten'))])

    async def test_is_applicable(self):
        # Contact not in disallowed country
        verification = await make_verification(self.session, contact_id='Kryten')
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = Contact(
            contact_id=ContactId('Kryten'), contact_handle='chnd:1',
            place=ContactAddress(street=['Deck 16'], city='Red Dwarf', country_code='JU'))
        processor = RepresentativeProcessor(method=self.method, disallowed_countries=['ME', 'VE', 'EA', 'MA'])

        self.assertTrue(await processor.is_applicable(verification))

        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.contact_client)),
                         [call.get_contact_info(ContactId('Kryten'))])

    async def test_is_applicable_not(self):
        # Contact in disallowed country
        verification = await make_verification(self.session, contact_id='Kryten')
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = Contact(
            contact_id=ContactId('Kryten'), contact_handle='chnd:1',
            place=ContactAddress(street=['Deck 16'], city='Red Dwarf', country_code='JU'))
        processor = RepresentativeProcessor(method=self.method, disallowed_countries=['ME', 'VE', 'EA', 'MA', 'JU'])

        self.assertFalse(await processor.is_applicable(verification))

        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.contact_client)),
                         [call.get_contact_info(ContactId('Kryten'))])

    async def test_is_applicable_contact_does_not_exist(self):
        # Test method is not applicable if contact no longer exist.
        verification = await make_verification(self.session, contact_id='Kryten')
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).side_effect = ContactDoesNotExist
        processor = RepresentativeProcessor(method=self.method)

        self.assertFalse(await processor.is_applicable(verification))

    async def test_get_next_event_no_representative(self):
        # Test letter is not sent if representative is not set.
        verification = await make_verification(self.session, status=VerificationStatus.INITIALIZED,
                                               contact_id=self.contact_id)
        processor = RepresentativeProcessor(method=self.method)

        result = processor.get_next_events(verification)

        scheduled = [Comparison(OutzoneReminderEvent), Comparison(OutzoneEvent), Comparison(FutileReminderEvent),
                     Comparison(FutileEvent), Comparison(FailedEvent)]
        self.assertEqual(await alist(result), scheduled)

    async def _test_get_next_event_letter(self, verification: Verification) -> None:
        # Test cases when letter is generated.
        session = object_session(verification)
        details = {'name': self.name, 'organization': self.organization, 'address': self.address.dict(),
                   'telephone': self.telephone, 'email': self.email}
        await make_event(session, verification_id=verification.id, type=EventType.SET_CONTACT_REPRESENTATIVE,
                         details=details)

        processor = RepresentativeProcessor(method=self.method, letter_limit_bonus=sentinel.bonus)

        result = processor.get_next_events(verification)

        address = dict(self.address_data, name=self.name, organization=self.organization)
        details = {'recipient': address, 'telephone': self.telephone, 'email': self.email}
        context = {'initial_email_datetime': self.create_datetime.isoformat()}
        letter = Comparison(RepresentativeLetterEvent, due_datetime=None, method=self.method, _details=details,
                            context=context, letter_limit_bonus=sentinel.bonus, partial=True)
        scheduled = [letter, Comparison(OutzoneReminderEvent), Comparison(OutzoneEvent),
                     Comparison(FutileReminderEvent), Comparison(FutileEvent), Comparison(FailedEvent)]
        self.assertEqual(await alist(result), scheduled)

    async def test_get_next_event_letter(self):
        verification = await make_verification(self.session, status=VerificationStatus.INITIALIZED,
                                               contact_id=self.contact_id, create_datetime=self.create_datetime)
        await self._test_get_next_event_letter(verification)

    async def test_get_next_event_letter_sent(self):
        # Test letter is not sent if it was already sent.
        verification = await make_verification(self.session, status=VerificationStatus.INITIALIZED,
                                               contact_id=self.contact_id, create_datetime=self.create_datetime)
        details = {'message_type': MessageType.REPRESENTATIVE_LETTER}
        await make_event(self.session, verification_id=verification.id, type=EventType.SEND_MESSAGE,
                         details=details)
        await make_event(self.session, verification_id=verification.id, type=EventType.SET_CONTACT_REPRESENTATIVE,
                         details={'name': self.name, 'address': self.address.dict()})
        processor = RepresentativeProcessor(method=self.method)

        result = processor.get_next_events(verification)

        scheduled = [Comparison(OutzoneReminderEvent), Comparison(OutzoneEvent), Comparison(FutileReminderEvent),
                     Comparison(FutileEvent), Comparison(FailedEvent)]
        self.assertEqual(await alist(result), scheduled)

    async def test_get_next_event_other_message(self):
        # Test letter is sent if other message was sent.
        verification = await make_verification(self.session, status=VerificationStatus.INITIALIZED,
                                               contact_id=self.contact_id, create_datetime=self.create_datetime)
        await make_event(self.session, verification_id=verification.id, type=EventType.SEND_MESSAGE,
                         details={'message_type': MessageType.INITIAL_EMAIL})

        await self._test_get_next_event_letter(verification)

    async def test_get_next_event_other_event(self):
        # Test letter is sent if other event was processed.
        verification = await make_verification(self.session, status=VerificationStatus.INITIALIZED,
                                               contact_id=self.contact_id, create_datetime=self.create_datetime)
        await make_event(self.session, verification_id=verification.id, type=EventType.CANCEL,
                         details={'message_type': MessageType.REPRESENTATIVE_LETTER})

        await self._test_get_next_event_letter(verification)

    async def test_get_resend_event_letter(self):
        processor = RepresentativeProcessor(method=self.method, letter_limit_bonus=sentinel.bonus)
        verification = await make_verification(self.session, contact_id=self.contact_id,
                                               create_datetime=self.create_datetime)
        details = {'name': self.name, 'organization': self.organization, 'address': self.address.dict(),
                   'telephone': self.telephone, 'email': self.email}
        await make_event(self.session, verification_id=verification.id, type=EventType.SET_CONTACT_REPRESENTATIVE,
                         details=details)

        scheduled = await processor.get_resend_event(verification, MessageType.REPRESENTATIVE_LETTER)

        self.assertIsInstance(scheduled, RepresentativeLetterEvent)
        self.assertTrue(scheduled.resend)
        self.assertEqual(cast(RepresentativeLetterEvent, scheduled).letter_limit_bonus, sentinel.bonus)
        self.assertIsNone(scheduled.due_datetime)
        self.assertEqual(scheduled.method, self.method)
        address = dict(self.address_data, name=self.name, organization=self.organization)
        details = {'recipient': address, 'telephone': self.telephone, 'email': self.email,
                   'method': TransportMethod.LETTER, 'message_type': MessageType.REPRESENTATIVE_LETTER}
        self.assertEqual(await scheduled.get_details(), details)
        self.assertEqual(scheduled.context, {'initial_email_datetime': self.create_datetime.isoformat()})

    async def test_get_resend_event_no_representative(self):
        processor = RepresentativeProcessor(method=self.method, letter_limit_bonus=sentinel.bonus)
        verification = await make_verification(self.session, contact_id=self.contact_id,
                                               create_datetime=self.create_datetime)

        with self.assertRaisesRegex(ValueError, "Message type 'verification_representative_letter' can't be resent"):
            await processor.get_resend_event(verification, MessageType.REPRESENTATIVE_LETTER.value)

    async def test_get_resend_event_unknown(self):
        processor = RepresentativeProcessor(method=self.method)
        verification = await make_verification(self.session, contact_id=self.contact_id)

        with self.assertRaisesRegex(ValueError, "Message type 'unknown' not found"):
            await processor.get_resend_event(verification, 'unknown')

    async def test_check_complete(self):
        verification = await make_verification(self.session, contact_id=self.contact_id)
        await make_event(self.session, verification_id=verification.id, type=EventType.VERIFY_TOKEN, method=self.method)
        processor = RepresentativeProcessor(method=self.method)

        self.assertTrue(await processor.check_complete(verification))

    async def test_check_complete_incomplete(self):
        verification = await make_verification(self.session, contact_id=self.contact_id)
        processor = RepresentativeProcessor(method=self.method)

        self.assertFalse(await processor.check_complete(verification))

    async def test_check_complete_other_method(self):
        # Test check_complete doesn't pass on token from other method.
        verification = await make_verification(self.session, contact_id=self.contact_id)
        await make_event(self.session, verification_id=verification.id, type=EventType.VERIFY_TOKEN,
                         method='no-psirens')
        processor = RepresentativeProcessor(method=self.method)

        self.assertFalse(await processor.check_complete(verification))

    async def test_handle_token_verified(self):
        # A new representative is created.
        cast(AsyncMock, APPLICATION.representative_client.get).side_effect = RepresentativeDoesNotExist
        verification = await make_verification(self.session, contact_id=self.contact_id)
        token = await make_token(self.session, verification_id=verification.id, method=self.method)
        details = {'recipient': dict(**self.address_data, name=self.name, organization=self.organization),
                   'telephone': self.telephone, 'email': self.email, 'token': str(token.uuid)}
        await make_event(self.session, verification_id=verification.id, type=EventType.SEND_MESSAGE, details=details,
                         method=self.method)
        processor = RepresentativeProcessor(method=self.method)

        await processor.handle_token_verified(token)

        # Check representative created.
        representative = Representative(id=RepresentativeId('0'), contact_id=self.contact_id, name=self.name,
                                        organization=self.organization, place=self.address, telephone=self.telephone,
                                        email=self.email)
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.representative_client)),
                         [call.get(self.contact_id), call.create(representative)])

    async def test_handle_token_verified_update(self):
        # The existing representative is updated.
        cast(AsyncMock, APPLICATION.representative_client.get).return_value = Representative(
            id=RepresentativeId('42'), contact_id=self.contact_id)
        verification = await make_verification(self.session, contact_id=self.contact_id)
        token = await make_token(self.session, verification_id=verification.id, method=self.method)
        details = {'recipient': dict(**self.address_data, name=self.name, organization=self.organization),
                   'telephone': self.telephone, 'email': self.email, 'token': str(token.uuid)}
        await make_event(self.session, verification_id=verification.id, type=EventType.SEND_MESSAGE, details=details,
                         method=self.method)
        processor = RepresentativeProcessor(method=self.method)

        await processor.handle_token_verified(token)

        # Check representative created.
        representative = Representative(id=RepresentativeId('42'), contact_id=self.contact_id, name=self.name,
                                        organization=self.organization, place=self.address, telephone=self.telephone,
                                        email=self.email)
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.representative_client)),
                         [call.get(self.contact_id), call.update(representative)])

    async def test_handle_token_verified_event_mismatch(self):
        # Check callback is not triggered on mismatching event.
        verification = await make_verification(self.session, contact_id=self.contact_id)
        token = await make_token(self.session, verification_id=verification.id, method=self.method)
        await make_event(self.session, verification_id=verification.id, type=EventType.SEND_MESSAGE, method=self.method,
                         details={'token': str(UUID(int=196))})
        processor = RepresentativeProcessor(method=self.method)

        await processor.handle_token_verified(token)

        # Check no calls to representative client.
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.representative_client)), [])

    async def test_handle_token_verified_event_invalid_token(self):
        # Check callback is not triggered on mismatching event.
        verification = await make_verification(self.session, contact_id=self.contact_id)
        token = await make_token(self.session, verification_id=verification.id, method=self.method)
        await make_event(self.session, verification_id=verification.id, type=EventType.SEND_MESSAGE, method=self.method,
                         details={'token': 'invalid'})
        processor = RepresentativeProcessor(method=self.method)

        await processor.handle_token_verified(token)

        # Check no calls to representative client.
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.representative_client)), [])

    async def test_handle_token_verified_event_missing(self):
        # Check callback doesn't fail if the event is missing.
        verification = await make_verification(self.session, contact_id=self.contact_id)
        token = await make_token(self.session, verification_id=verification.id, method=self.method)
        processor = RepresentativeProcessor(method=self.method)

        await processor.handle_token_verified(token)

        # Check no calls to representative client.
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.representative_client)), [])
