from datetime import date, datetime
from typing import Any, Callable, Dict, Iterable, List, Mapping, Optional, cast
from unittest import IsolatedAsyncioTestCase, TestCase, skipUnless
from unittest.mock import AsyncMock, MagicMock, call, sentinel
from uuid import UUID

from asyncstdlib import iter as aiter, tuple as atuple
from fred_api.verification.common_types_pb2 import (
    ContactDoesNotExist as ProtoContactDoesNotExist,
    Event as ProtoEvent,
    IncorrectStatus,
    InvalidData,
    MethodNotApplicable,
    VerificationAlreadyExists,
    VerificationDoesNotExist,
    VerificationInfo,
    VerificationResult,
    VerificationStatus as ProtoVerificationStatus,
)
from fred_api.verification.service_verification_grpc_pb2 import (
    ApproveFutileReply,
    ApproveFutileRequest,
    ApproveManualReply,
    ApproveManualRequest,
    CancelReply,
    CancelRequest,
    CreateReply,
    CreateRequest,
    DelayStatusReply,
    DelayStatusRequest,
    GetMethodsReply,
    GetMethodsRequest,
    GetReply,
    GetRequest,
    ListReply,
    ListRequest,
    ListVerificationResultsReply,
    ListVerificationResultsRequest,
    RequestManualCheckReply,
    RequestManualCheckRequest,
    ResendMessageReply,
    ResendMessageRequest,
    ResolveManualCheckReply,
    ResolveManualCheckRequest,
    ResumeReply,
    ResumeRequest,
    SetContactRepresentativeReply,
    SetContactRepresentativeRequest,
    UseMethodReply,
    UseMethodRequest,
    VerifyContactDataReply,
    VerifyContactDataRequest,
    VerifyTokenReply,
    VerifyTokenRequest,
)
from fred_types import ContactHistoryId, ContactId, DomainHistoryId, DomainId, DomainRef
from freezegun import freeze_time
from google.protobuf.message import Message
from grpc import StatusCode
from hermes import Email, EmailMessengerClient
from hermes.message import Reference
from pytz import UTC
from regal import (
    AdditionalIdentifier,
    Address,
    Contact,
    ContactClient,
    ContactIdentifierType,
    Domain,
    DomainAdminClient,
    DomainClient,
    NotDeletedReason,
)
from regal.domain import DomainContactRole
from regal.exceptions import ContactDoesNotExist
from sqlalchemy import select, text
from sqlalchemy.ext.asyncio import AsyncSession
from testfixtures import Comparison, LogCapture, StringComparison

from verification.constants import EventType, MessageType, ObjectState, Template, VerificationStatus
from verification.database import atomic, get_engine
from verification.handler import VerificationHandler
from verification.models import ContactData, Event, Token, Verification
from verification.processors.base import BaseProcessor
from verification.service import VerificationServicer, _reverse_order_by
from verification.settings import APPLICATION

from ..constants import ReferenceType
from .test_handler import TestEvent, TestEventsProcessor, TestHandler
from .utils import (
    make_contact_data,
    make_domain,
    make_event,
    make_token,
    make_verification,
    mock_awaits,
    override_database,
    override_settings,
    test_session_cls,
)


class ReverseOrderByTest(TestCase):
    def test_reverse(self):
        data = (
            (select(Verification), select(Verification)),
            (select(Verification).order_by(Verification.id), select(Verification).order_by(Verification.id.desc())),
            (select(Verification).order_by(Verification.id.asc()),
             select(Verification).order_by(Verification.id.desc())),
            (select(Verification).order_by(Verification.id.desc()),
             select(Verification).order_by(Verification.id.asc())),
            (select(Verification).order_by(Verification.uuid), select(Verification).order_by(Verification.uuid.desc())),
            # Multiple columns
            (select(Verification).order_by(Verification.uuid, Verification.id),
             select(Verification).order_by(Verification.uuid.desc(), Verification.id.desc())),
        )
        for query, result in data:
            with self.subTest(query=query):
                # We can only compare string representations of queries.
                self.assertEqual(str(_reverse_order_by(query)), str(result))


class TestTokenProcessor(BaseProcessor):
    async def handle_token_verified(self, token: Token) -> None:
        # Change the verification, so we can check the callback was called.
        token.verification.status = VerificationStatus.SUCCEEDED


class TestProcessor(BaseProcessor):
    """Simple test processor with custom applicability."""

    def __init__(self, *, method: str, is_applicable: bool = True):
        self._is_applicable = is_applicable

    async def is_applicable(self, verification: Verification) -> bool:
        return self._is_applicable


@override_database()
class VerificationServicerTest(IsolatedAsyncioTestCase):
    uuid = UUID(int=42)
    create_datetime = datetime(2019, 12, 11, 10, 9, 8, tzinfo=UTC)
    close_datetime = datetime(2019, 12, 11, 11, 39, 8, tzinfo=UTC)
    contact_id = ContactId('KRYTEN')
    event_datetime = datetime(1997, 2, 21, 20, 8, 20, tzinfo=UTC)
    method = 'ANALYSIS'
    contact_history_id = ContactHistoryId('STAR-BUG')

    async def asyncSetUp(self):
        self.service = VerificationServicer()
        self.log_handler = LogCapture('verification', propagate=False)

        APPLICATION.contact_client = AsyncMock(ContactClient)
        APPLICATION.domain_client = AsyncMock(DomainClient)
        APPLICATION.domain_admin_client = AsyncMock(DomainAdminClient)
        APPLICATION.email_client = AsyncMock(EmailMessengerClient)

    async def asyncTearDown(self):
        self.log_handler.uninstall()

    async def _make_verification(self, session: AsyncSession, **kwargs: Any) -> Verification:
        model_kwargs = dict({'uuid': self.uuid, 'create_datetime': self.create_datetime, 'contact_id': self.contact_id},
                            **(kwargs or {}))
        return await make_verification(session, **model_kwargs)

    def _get_verification_info(
            self, *, id: str = str(uuid), contact_id: str = contact_id,
            create_datetime: datetime = create_datetime, close_datetime: Optional[datetime] = None,
            last_access_datetime: Optional[datetime] = None,
            status: ProtoVerificationStatus = ProtoVerificationStatus.created, manual_check_request: bool = False,
            reason: str = '', reason_url: str = '',
            events: Optional[Iterable[ProtoEvent]] = None, next_events: Optional[Iterable[ProtoEvent]] = None,
            ) -> VerificationInfo:
        verification = VerificationInfo()
        verification.verification_id.value = str(id)
        verification.contact_id.value = contact_id
        verification.create_datetime.FromDatetime(create_datetime)
        if close_datetime:
            verification.close_datetime.FromDatetime(close_datetime)
        if last_access_datetime:
            verification.last_access_datetime.FromDatetime(last_access_datetime)
        verification.status = status
        verification.manual_check_request = manual_check_request
        verification.reason = reason
        verification.reason_url = reason_url
        if events:
            verification.events.extend(events)
        if next_events:
            verification.next_events.extend(next_events)
        return verification

    async def _test_db_error(self, method: Callable[[Message, Any], Any], request: Message) -> None:
        # Test DB error occurs on database level - dropped table
        sql = "DROP TABLE {}".format(Verification.__tablename__)
        if not APPLICATION.settings.db_connection.startswith('sqlite'):
            sql += ' CASCADE'
        async with get_engine().begin() as conn:
            await conn.execute(text(sql))
        context = AsyncMock(name='context')
        context.abort.side_effect = Exception(sentinel.abort)

        with self.assertRaises(Exception) as catcher:
            await method(request, context)

        self.assertEqual(catcher.exception.args, (sentinel.abort, ))
        self.assertEqual(mock_awaits(context),
                         [call.abort(StatusCode.FAILED_PRECONDITION, StringComparison("Database error: .*"))])

    async def test_create_db_error(self):
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = Contact(contact_id=self.contact_id,
                                                                                            contact_handle='2X4B')
        request = CreateRequest()
        request.contact_id.value = self.contact_id
        await self._test_db_error(self.service.create, request)

    async def test_create_unknown(self):
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).side_effect = ContactDoesNotExist
        request = CreateRequest()
        request.contact_id.value = self.contact_id

        result = await self.service.create(request, sentinel.context)

        reply = CreateReply()
        reply.exception.contact_does_not_exist.CopyFrom(ProtoContactDoesNotExist())
        self.assertEqual(result, reply)

        # Check registry calls
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.contact_client)),
                         [call.get_contact_info(self.contact_id)])

    async def test_create_already_exist(self):
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = Contact(contact_id=self.contact_id,
                                                                                            contact_handle='2X4B')
        async with atomic() as session:
            await self._make_verification(session)
        request = CreateRequest()
        request.contact_id.value = self.contact_id

        result = await self.service.create(request, sentinel.context)

        reply = CreateReply()
        reply.exception.verification_already_exists.CopyFrom(VerificationAlreadyExists())
        self.assertEqual(result, reply)

    async def _test_create(self, reason: str = '', reason_url: str = '') -> None:
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = Contact(contact_id=self.contact_id,
                                                                                            contact_handle='2X4B')
        request = CreateRequest()
        request.contact_id.value = self.contact_id
        request.reason = reason
        request.reason_url = reason_url

        result = await self.service.create(request, sentinel.context)

        async with test_session_cls()() as session:
            verification = (await session.execute(select(Verification))).scalars().one()
            event = (await session.execute(select(Event))).scalars().one()
        reply = CreateReply()
        proto_event = ProtoEvent()
        proto_event.event_datetime.FromDatetime(event.event_datetime)
        proto_event.type = EventType.CREATE
        proto_event.details.update({'domains': []})
        info = self._get_verification_info(
            id=str(verification.uuid), create_datetime=verification.create_datetime,
            status=ProtoVerificationStatus.created, reason=reason, reason_url=reason_url, events=[proto_event])
        reply.data.verification.CopyFrom(info)
        self.assertEqual(result, reply)

        # Check verification is created.
        self.assertEqual(verification.contact_id, self.contact_id)
        self.assertEqual(verification.status, VerificationStatus.CREATED)
        self.assertEqual(verification.reason, reason)
        self.assertEqual(verification.reason_url, reason_url)

    async def test_create(self):
        await self._test_create()

    async def test_create_reason(self):
        await self._test_create(reason='Smeghead!')

    async def test_create_reason_url(self):
        await self._test_create(reason_url='Smeghead!')

    async def test_get_db_error(self):
        request = GetRequest()
        request.verification_id.value = str(UUID(int=42))
        await self._test_db_error(self.service.get, request)

    async def _test_get_not_exist(self, request: GetRequest) -> None:
        result = await self.service.get(request, sentinel.context)

        reply = GetReply()
        reply.exception.verification_does_not_exist.CopyFrom(VerificationDoesNotExist())
        self.assertEqual(result, reply)

    async def test_get_empty(self):
        request = GetRequest()
        await self._test_get_not_exist(request)

    async def test_get_invalid(self):
        request = GetRequest()
        request.verification_id.value = 'INVALID'
        await self._test_get_not_exist(request)

    async def test_get_unknown(self):
        request = GetRequest()
        request.verification_id.value = str(UUID(int=42))
        await self._test_get_not_exist(request)

    async def _test_get(self, reply_kwargs: Optional[Mapping[str, Any]] = None,
                        *, access_token: Optional[str] = None) -> None:
        request = GetRequest()
        request.verification_id.value = str(self.uuid)
        if access_token is not None:
            request.access_token = access_token
        result = await self.service.get(request, sentinel.context)

        reply = GetReply()
        reply.data.verification.CopyFrom(self._get_verification_info(**(reply_kwargs or {})))
        self.assertEqual(result, reply)

    async def test_get(self):
        async with atomic() as session:
            await self._make_verification(session)
        await self._test_get()

    async def test_get_close_datetime(self):
        async with atomic() as session:
            close_datetime = datetime(1997, 2, 21, 20, tzinfo=UTC)
            await self._make_verification(session, close_datetime=close_datetime)
        await self._test_get({'close_datetime': close_datetime})

    async def test_get_last_access_datetime(self):
        async with atomic() as session:
            last_access_datetime = datetime(1997, 2, 21, 20, tzinfo=UTC)
            await self._make_verification(session, last_access_datetime=last_access_datetime)
        await self._test_get({'last_access_datetime': last_access_datetime})

    async def test_get_status(self):
        async with atomic() as session:
            await self._make_verification(session, status=VerificationStatus.SUCCEEDED)
        await self._test_get({'status': ProtoVerificationStatus.succeeded})

    async def test_get_request_manual_check(self):
        async with atomic() as session:
            await self._make_verification(session, manual_check_request=True)
        await self._test_get({'manual_check_request': True})

    async def test_get_reason(self):
        async with atomic() as session:
            await self._make_verification(session, reason='Smeghead!')
        await self._test_get({'reason': 'Smeghead!'})

    async def test_get_reason_url(self):
        async with atomic() as session:
            await self._make_verification(session, reason_url='Smeghead!')
        await self._test_get({'reason_url': 'Smeghead!'})

    async def test_get_events(self):
        type = 'Lobster'
        details = {'ketchup': 'brown'}
        async with atomic() as session:
            verification = await self._make_verification(session)
            await make_event(session, verification_id=verification.id, event_datetime=self.event_datetime, type=type,
                             details=details)
        event = ProtoEvent(type=type)
        event.event_datetime.FromDatetime(self.event_datetime)
        event.details.update(details)
        await self._test_get({'events': [event]})

    async def test_get_events_method(self):
        type = 'Lobster'
        details = {'ketchup': 'brown'}
        async with atomic() as session:
            verification = await self._make_verification(session)
            await make_event(session, verification_id=verification.id, event_datetime=self.event_datetime,
                             method=self.method, type=type, details=details)
        event = ProtoEvent(type=type, method=self.method)
        event.event_datetime.FromDatetime(self.event_datetime)
        event.details.update(details)
        await self._test_get({'events': [event]})

    async def test_get_next_events(self):
        details = {'ketchup': 'brown'}
        async with atomic() as session:
            verification = await self._make_verification(session)
        scheduled = TestEvent(verification, details=details)
        processor = TestEventsProcessor(next_events=[scheduled], method=self.method)
        APPLICATION.handler = TestHandler([processor])
        next_event = ProtoEvent(type=EventType.CANCEL)
        next_event.details.update(details)
        await self._test_get({'next_events': [next_event]})

    async def test_get_next_events_method(self):
        details = {'ketchup': 'brown'}
        async with atomic() as session:
            verification = await self._make_verification(session)
        scheduled = TestEvent(verification, method=self.method, details=details)
        processor = TestEventsProcessor(next_events=[scheduled], method=self.method)
        APPLICATION.handler = TestHandler([processor])
        next_event = ProtoEvent(type=EventType.CANCEL, method=self.method)
        next_event.details.update(details)
        await self._test_get({'next_events': [next_event]})

    async def test_get_next_events_due_datetime(self):
        details = {'ketchup': 'brown'}
        async with atomic() as session:
            verification = await self._make_verification(session)
        scheduled = TestEvent(verification, details=details, due_datetime=self.event_datetime)
        processor = TestEventsProcessor(next_events=[scheduled], method=self.method)
        APPLICATION.handler = TestHandler([processor])
        next_event = ProtoEvent(type=EventType.CANCEL)
        next_event.event_datetime.FromDatetime(self.event_datetime)
        next_event.details.update(details)
        await self._test_get({'next_events': [next_event]})

    async def test_get_token(self):
        async with atomic() as session:
            await self._make_verification(session, access_token='Gazpacho!', status=VerificationStatus.PENDING)
        request = GetRequest()
        request.verification_id.value = str(self.uuid)
        request.access_token = 'Gazpacho!'

        result = await self.service.get(request, sentinel.context)

        # Check verification access is updated.
        async with test_session_cls()() as session:
            verification = (await session.execute(select(Verification))).scalars().one()
            self.assertIsNotNone(verification.last_access_datetime)

        reply = GetReply()
        reply.data.verification.CopyFrom(
            self._get_verification_info(last_access_datetime=verification.last_access_datetime,
                                        status=ProtoVerificationStatus.pending))
        self.assertEqual(result, reply)

    async def test_get_token_empty(self):
        async with atomic() as session:
            await self._make_verification(session, access_token='Gazpacho!')
        await self._test_get(access_token='')

    async def test_get_token_invalid(self):
        async with atomic() as session:
            await self._make_verification(session, access_token='Gazpacho!')
        request = GetRequest()
        request.verification_id.value = str(self.uuid)
        request.access_token = 'Quagaars!'

        result = await self.service.get(request, sentinel.context)

        reply = GetReply()
        reply.exception.verification_does_not_exist.CopyFrom(VerificationDoesNotExist())
        self.assertEqual(result, reply)

    async def _test_get_accessed(self, status: VerificationStatus,
                                 final_status: VerificationStatus = VerificationStatus.ACCESSED) -> None:
        # Test switches to accessed status.
        async with atomic() as session:
            await self._make_verification(session, access_token='Gazpacho!', status=status)
        request = GetRequest()
        request.verification_id.value = str(self.uuid)
        request.access_token = 'Gazpacho!'

        # Ignore reply, check only verification status.
        result = await self.service.get(request, sentinel.context)

        # Check verification status is updated.
        async with test_session_cls()() as session:
            verification = (await session.execute(select(Verification))).scalars().one()
            self.assertEqual(verification.status, final_status)

        reply = GetReply()
        reply.data.verification.CopyFrom(
            self._get_verification_info(last_access_datetime=verification.last_access_datetime,
                                        status=ProtoVerificationStatus.Value(final_status)))
        self.assertEqual(result, reply)

    async def test_get_accessed(self):
        # Test verification is set to accessed if INITIALIZED.
        await self._test_get_accessed(VerificationStatus.INITIALIZED)

    async def test_get_accessed_created(self):
        # Test verification is set to accessed if INITIALIZED.
        await self._test_get_accessed(VerificationStatus.CREATED)

    async def test_get_accessed_other_status(self):
        # Test verification is not set to accessed if not INITIALIZED.
        await self._test_get_accessed(VerificationStatus.PENDING, VerificationStatus.PENDING)

    async def test_get_accessed_invalid(self):
        # Test verification status is unchanged if token is invalid.
        async with atomic() as session:
            await self._make_verification(session, access_token='Gazpacho!', status=VerificationStatus.INITIALIZED)
        request = GetRequest()
        request.verification_id.value = str(self.uuid)
        request.access_token = 'Quagaars!'

        # Ignore reply, check only verification status.
        result = await self.service.get(request, sentinel.context)

        # Check verification status is not updated.
        async with test_session_cls()() as session:
            verification = (await session.execute(select(Verification))).scalars().one()
            self.assertEqual(verification.status, VerificationStatus.INITIALIZED)

        reply = GetReply()
        reply.exception.verification_does_not_exist.CopyFrom(VerificationDoesNotExist())
        self.assertEqual(result, reply)

    async def test_get_verify_emails_none(self):
        # Check noop if initial email event is not found.
        async with atomic() as session:
            await self._make_verification(session, access_token='Gazpacho!')
        request = GetRequest()
        request.verification_id.value = str(self.uuid)
        request.access_token = 'Gazpacho!'

        # Ignore reply, test only email verification.
        await self.service.get(request, sentinel.context)

        async with test_session_cls()() as session:
            self.assertIsNone((await session.execute(select(ContactData))).scalar())

    async def test_get_verify_emails_other(self):
        # Check noop if only other message event is found.
        async with atomic() as session:
            verification = await self._make_verification(session, access_token='Gazpacho!')
            email_details = {'message_type': 'other'}
            await make_event(session, verification_id=verification.id, event_datetime=self.event_datetime,
                             type=EventType.SEND_MESSAGE, details=email_details)
        request = GetRequest()
        request.verification_id.value = str(self.uuid)
        request.access_token = 'Gazpacho!'

        # Ignore reply, test only email verification.
        await self.service.get(request, sentinel.context)

        async with test_session_cls()() as session:
            self.assertIsNone((await session.execute(select(ContactData))).scalar())

    async def test_get_verify_emails(self):
        # Check emails from initial email are marked verified on access.
        async with atomic() as session:
            verification = await self._make_verification(session, access_token='Gazpacho!')
            email_details = {'message_type': MessageType.INITIAL_EMAIL, 'recipient': ['kryten@example.org'],
                             'contact_history_id': self.contact_history_id}
            await make_event(session, verification_id=verification.id, event_datetime=self.event_datetime,
                             type=EventType.SEND_MESSAGE, details=email_details)
        request = GetRequest()
        request.verification_id.value = str(self.uuid)
        request.access_token = 'Gazpacho!'

        # Ignore reply, test only email verification.
        await self.service.get(request, sentinel.context)

        async with test_session_cls()() as session:
            verification = (await session.execute(select(Verification))).scalars().one()
            contact_data = (await session.execute(select(ContactData))).scalars().one()
            self.assertIsNone(contact_data.method)
            self.assertEqual(contact_data.contact_history_id, self.contact_history_id)
            self.assertEqual(contact_data.emails, ['kryten@example.org'])
            self.assertTrue(contact_data.verified_emails)

    async def test_get_verify_emails_verified(self):
        # Check noop if emails are already verified.
        async with atomic() as session:
            verification = await self._make_verification(session, access_token='Gazpacho!')
            await make_contact_data(session, verification_id=verification.id, verify_datetime=self.event_datetime,
                                    method=None, contact_history_id=self.contact_history_id,
                                    emails=['kryten@example.org'], verified_emails=True)
        request = GetRequest()
        request.verification_id.value = str(self.uuid)
        request.access_token = 'Gazpacho!'

        # Ignore reply, test only email verification.
        await self.service.get(request, sentinel.context)

        async with test_session_cls()() as session:
            verification = (await session.execute(select(Verification))).scalars().one()
            # Check no contact data were added.
            query = select(ContactData)
            self.assertEqual(len(tuple((await session.execute(query)).scalars())), 1)

    async def test_get_verify_emails_multiple(self):
        # Check noop if multiple initial email events found.
        async with atomic() as session:
            verification = await self._make_verification(session, access_token='Gazpacho!')
            email_details = {'message_type': MessageType.INITIAL_EMAIL, 'recipient': ['kryten@example.org'],
                             'contact_history_id': self.contact_history_id}
            await make_event(session, verification_id=verification.id, event_datetime=self.event_datetime,
                             type=EventType.SEND_MESSAGE, details=email_details)
            await make_event(session, verification_id=verification.id, event_datetime=self.event_datetime,
                             type=EventType.SEND_MESSAGE, details=email_details)
        request = GetRequest()
        request.verification_id.value = str(self.uuid)
        request.access_token = 'Gazpacho!'

        # Ignore reply, test only email verification.
        await self.service.get(request, sentinel.context)

        async with test_session_cls()() as session:
            self.assertIsNone((await session.execute(select(ContactData))).scalar())

    async def test_get_verify_emails_invalid(self):
        # Check emails are not verified on invalid token.
        async with atomic() as session:
            verification = await self._make_verification(session, access_token='Gazpacho!')
            email_details = {'message_type': MessageType.INITIAL_EMAIL, 'recipient': ['kryten@example.org'],
                             'contact_history_id': self.contact_history_id}
            await make_event(session, verification_id=verification.id, event_datetime=self.event_datetime,
                             type=EventType.SEND_MESSAGE, details=email_details)
        request = GetRequest()
        request.verification_id.value = str(self.uuid)
        request.access_token = 'Quagaars!'

        # Ignore reply, test only email verification.
        await self.service.get(request, sentinel.context)

        async with test_session_cls()() as session:
            verification = (await session.execute(select(Verification))).scalars().one()
            self.assertIsNone((await session.execute(select(ContactData))).scalar())

    async def test_cancel_db_error(self):
        request = CancelRequest()
        request.verification_id.value = str(UUID(int=42))
        await self._test_db_error(self.service.cancel, request)

    async def _test_cancel_not_exist(self, request: CancelRequest) -> None:
        result = await self.service.cancel(request, sentinel.context)

        reply = CancelReply()
        reply.exception.verification_does_not_exist.CopyFrom(VerificationDoesNotExist())
        self.assertEqual(result, reply)

    async def test_cancel_empty(self):
        request = CancelRequest()
        await self._test_cancel_not_exist(request)

    async def test_cancel_invalid(self):
        request = CancelRequest()
        request.verification_id.value = 'INVALID'
        await self._test_cancel_not_exist(request)

    async def test_cancel_unknown(self):
        request = CancelRequest()
        request.verification_id.value = str(UUID(int=42))
        await self._test_cancel_not_exist(request)

    async def _test_cancel_incorrect_status(self, status: VerificationStatus) -> None:
        async with atomic() as session:
            await self._make_verification(session, status=status)
        request = CancelRequest()
        request.verification_id.value = str(self.uuid)

        result = await self.service.cancel(request, sentinel.context)

        reply = CancelReply()
        reply.exception.incorrect_status.CopyFrom(IncorrectStatus())
        self.assertEqual(result, reply)

    async def test_cancel_incorrect_succeeded(self):
        await self._test_cancel_incorrect_status(VerificationStatus.SUCCEEDED)

    async def test_cancel_incorrect_failed(self):
        await self._test_cancel_incorrect_status(VerificationStatus.FAILED)

    async def test_cancel_incorrect_canceled(self):
        await self._test_cancel_incorrect_status(VerificationStatus.CANCELED)

    async def test_cancel_incorrect_manual(self):
        await self._test_cancel_incorrect_status(VerificationStatus.MANUAL)

    async def test_cancel(self):
        async with atomic() as session:
            await self._make_verification(session)
        request = CancelRequest()
        request.verification_id.value = str(self.uuid)

        APPLICATION.handler = AsyncMock(VerificationHandler)
        with freeze_time(self.close_datetime):
            result = await self.service.cancel(request, sentinel.context)

        self.assertEqual(result, CancelReply())

        async with test_session_cls()() as session:
            # Check verification was canceled.
            # XXX: Attach 'verification' argument to a session so it can be checked.
            session.add(APPLICATION.handler.mock_calls[0][1][0])
            await session.refresh(APPLICATION.handler.mock_calls[0][1][0])
        calls = [call.cancel(Comparison(Verification, uuid=self.uuid, partial=True))]
        self.assertEqual(mock_awaits(APPLICATION.handler), calls)

    async def test_resume_db_error(self):
        request = ResumeRequest()
        request.verification_id.value = str(UUID(int=42))
        await self._test_db_error(self.service.resume, request)

    async def _test_resume_not_exist(self, request: ResumeRequest) -> None:
        result = await self.service.resume(request, sentinel.context)

        reply = ResumeReply()
        reply.exception.verification_does_not_exist.CopyFrom(VerificationDoesNotExist())
        self.assertEqual(result, reply)

    async def test_resume_empty(self):
        request = ResumeRequest()
        await self._test_resume_not_exist(request)

    async def test_resume_invalid(self):
        request = ResumeRequest()
        request.verification_id.value = 'INVALID'
        await self._test_resume_not_exist(request)

    async def test_resume_unknown(self):
        request = ResumeRequest()
        request.verification_id.value = str(UUID(int=42))
        await self._test_resume_not_exist(request)

    async def _test_resume_incorrect_status(self, status: VerificationStatus) -> None:
        async with atomic() as session:
            await self._make_verification(session, status=status)
        request = ResumeRequest()
        request.verification_id.value = str(self.uuid)

        result = await self.service.resume(request, sentinel.context)

        reply = ResumeReply()
        reply.exception.incorrect_status.CopyFrom(IncorrectStatus())
        self.assertEqual(result, reply)

    async def test_resume_incorrect_created(self):
        await self._test_resume_incorrect_status(VerificationStatus.CREATED)

    async def test_resume_incorrect_initialized(self):
        await self._test_resume_incorrect_status(VerificationStatus.INITIALIZED)

    async def test_resume_incorrect_accessed(self):
        await self._test_resume_incorrect_status(VerificationStatus.ACCESSED)

    async def test_resume_incorrect_pending(self):
        await self._test_resume_incorrect_status(VerificationStatus.PENDING)

    async def test_resume_incorrect_constrained(self):
        await self._test_resume_incorrect_status(VerificationStatus.CONSTRAINED)

    async def test_resume_incorrect_futile(self):
        await self._test_resume_incorrect_status(VerificationStatus.FUTILE)

    async def test_resume_incorrect_succeeded(self):
        await self._test_resume_incorrect_status(VerificationStatus.SUCCEEDED)

    async def test_resume_incorrect_failed(self):
        await self._test_resume_incorrect_status(VerificationStatus.FAILED)

    async def test_resume_incorrect_canceled(self):
        await self._test_resume_incorrect_status(VerificationStatus.CANCELED)

    async def test_resume_incorrect_manual(self):
        await self._test_resume_incorrect_status(VerificationStatus.MANUAL)

    async def _test_resume(self, events: List[Comparison]) -> None:
        request = ResumeRequest()
        request.verification_id.value = str(self.uuid)

        APPLICATION.handler = AsyncMock(VerificationHandler)
        with freeze_time('1997-02-21 21:08:20'):  # 1 hour after self.event_datetime
            result = await self.service.resume(request, sentinel.context)

        self.assertEqual(result, ResumeReply())

        async with test_session_cls()() as session:
            # Check verification was resumed.
            verification = (await session.execute(select(Verification))).scalars().one()
            self.assertEqual(verification.status, VerificationStatus.PENDING)
            # Check events created.
            query = select(Event).order_by('id')
            self.assertEqual(list((await session.execute(query)).scalars()), events)

    async def test_resume_no_event(self):
        # Test resume with no event.
        async with atomic() as session:
            await self._make_verification(session, status=VerificationStatus.PAUSED)
        events = [Comparison(Event, type=EventType.RESUME, method=None, details={}, partial=True)]
        await self._test_resume(events)

    async def test_resume_delay(self):
        # Test resume with delay since CHECK_DOMAINS event.
        async with atomic() as session:
            verification = await self._make_verification(session, status=VerificationStatus.PAUSED)
            await make_event(session, verification_id=verification.id, event_datetime=self.event_datetime,
                             type=EventType.CHECK_DOMAINS)
        events = [Comparison(Event, type=EventType.CHECK_DOMAINS, partial=True),
                  Comparison(Event, type=EventType.RESUME, method=None, details={}, partial=True),
                  Comparison(Event, type=EventType.DELAY_STATUS, method=None,
                             details={'status': VerificationStatus.FUTILE, 'delay': 3600}, partial=True)]
        await self._test_resume(events)

    async def test_list_db_error(self):
        # Test DB error occurs on database level - dropped table
        sql = "DROP TABLE {}".format(Verification.__tablename__)
        if not APPLICATION.settings.db_connection.startswith('sqlite'):
            sql += ' CASCADE'
        async with get_engine().begin() as conn:
            await conn.execute(text(sql))
        request = ListRequest()
        context = AsyncMock(name='context')
        context.abort.side_effect = Exception(sentinel.abort)

        with self.assertRaises(Exception) as catcher:
            await atuple(self.service.list(request, context))

        self.assertEqual(catcher.exception.args, (sentinel.abort, ))
        self.assertEqual(mock_awaits(context),
                         [call.abort(StatusCode.FAILED_PRECONDITION, StringComparison("Database error: .*"))])

    async def test_list_empty(self):
        request = ListRequest()

        result = self.service.list(request, sentinel.context)

        self.assertEqual(await atuple(result), ())

    async def test_list(self):
        async with atomic() as session:
            await self._make_verification(session)
        request = ListRequest()

        result = self.service.list(request, sentinel.context)

        reply = ListReply()
        reply.data.verifications.append(self._get_verification_info())
        self.assertEqual(await atuple(result), (reply, ))

    async def test_list_contact_id(self):
        async with atomic() as session:
            await self._make_verification(session)
            # Make another verification for another contact.
            await self._make_verification(session, contact_id='RIMMER', uuid=UUID(int=196))
        request = ListRequest()
        request.contact_id.value = self.contact_id

        result = self.service.list(request, sentinel.context)

        reply = ListReply()
        reply.data.verifications.append(self._get_verification_info())
        self.assertEqual(await atuple(result), (reply, ))

    async def test_list_status(self):
        async with atomic() as session:
            await self._make_verification(session, status=VerificationStatus.PENDING)
            # Make another verification for another contact.
            await self._make_verification(session, contact_id='RIMMER', uuid=UUID(int=196))
        request = ListRequest()
        request.statuses.append(ProtoVerificationStatus.pending)

        result = self.service.list(request, sentinel.context)

        reply = ListReply()
        reply.data.verifications.append(self._get_verification_info(status=ProtoVerificationStatus.pending))
        self.assertEqual(await atuple(result), (reply, ))

    async def test_list_status_unknown(self):
        # Test list called with unknown status in filter.
        request = ListRequest()
        request.statuses.append(ProtoVerificationStatus.unknown)
        context = AsyncMock(name='context')
        context.abort.side_effect = Exception(sentinel.abort)

        with self.assertRaises(Exception) as catcher:
            await atuple(self.service.list(request, context))

        self.assertEqual(catcher.exception.args, (sentinel.abort, ))
        self.assertEqual(mock_awaits(context),
                         [call.abort(StatusCode.INVALID_ARGUMENT, StringComparison("Unknown status 'unknown'"))])

    async def test_list_manual_check_request_true(self):
        async with atomic() as session:
            await self._make_verification(session, manual_check_request=True)
            # Make another verification for another contact.
            await self._make_verification(session, contact_id='RIMMER', uuid=UUID(int=196))
        request = ListRequest()
        request.manual_check_request = True

        result = self.service.list(request, sentinel.context)

        reply = ListReply()
        reply.data.verifications.append(self._get_verification_info(manual_check_request=True))
        self.assertEqual(await atuple(result), (reply, ))

    async def test_list_manual_check_request_false(self):
        async with atomic() as session:
            await self._make_verification(session)
            # Make another verification for another contact.
            await self._make_verification(session, contact_id='RIMMER', uuid=UUID(int=196), manual_check_request=True)
        request = ListRequest()
        request.manual_check_request = False

        result = self.service.list(request, sentinel.context)

        reply = ListReply()
        reply.data.verifications.append(self._get_verification_info())
        self.assertEqual(await atuple(result), (reply, ))

    async def test_list_manual_check_request_unset(self):
        async with atomic() as session:
            await self._make_verification(session, manual_check_request=True)
            # Make another verification for another contact.
            await self._make_verification(session, contact_id='RIMMER', uuid=UUID(int=196), manual_check_request=False)
        request = ListRequest()

        result = self.service.list(request, sentinel.context)

        reply_1 = ListReply()
        reply_1.data.verifications.append(self._get_verification_info(manual_check_request=True))
        reply_2 = ListReply()
        reply_2.data.verifications.append(self._get_verification_info(contact_id='RIMMER', id=str(UUID(int=196))))
        self.assertCountEqual(await atuple(result), (reply_1, reply_2))

    async def test_list_domain_id(self):
        # Filter verifications by domain ID.
        domain_id = 'the-domain-id'
        async with atomic() as session:
            verification = await self._make_verification(session)
            await make_domain(session, verification_id=verification.id, domain_id=domain_id, fqdn='example.org')
            # Make another verification for another contact.
            await self._make_verification(session, contact_id='RIMMER', uuid=UUID(int=196))
        request = ListRequest()
        request.domain_id.value = domain_id

        result = self.service.list(request, sentinel.context)

        reply = ListReply()
        reply.data.verifications.append(self._get_verification_info())
        self.assertEqual(await atuple(result), (reply, ))

    async def _test_list_order_by(self, order_by: str, ordered_contacts: List[str]) -> None:
        request = ListRequest()
        request.order_by = order_by

        result = self.service.list(request, sentinel.context)

        self.assertEqual(
            [v.contact_id.value for r in await atuple(result) for v in cast(ListReply, r).data.verifications],
            ordered_contacts)

    async def test_list_order_by_default(self):
        async with atomic() as session:
            await self._make_verification(session, contact_id='KRYTEN',
                                          create_datetime=datetime(2019, 12, 7, tzinfo=UTC))
            await self._make_verification(session, contact_id='RIMMER', uuid=UUID(int=43),
                                          create_datetime=datetime(2019, 12, 8, tzinfo=UTC))
        await self._test_list_order_by('', ['RIMMER', 'KRYTEN'])

    async def test_list_order_by_create_datetime(self):
        async with atomic() as session:
            await self._make_verification(session, contact_id='KRYTEN',
                                          create_datetime=datetime(2019, 12, 7, tzinfo=UTC))
            await self._make_verification(session, contact_id='RIMMER', uuid=UUID(int=43),
                                          create_datetime=datetime(2019, 12, 8, tzinfo=UTC))
        data = (
            ('create_datetime', ['KRYTEN', 'RIMMER']),
            ('+create_datetime', ['KRYTEN', 'RIMMER']),
            ('-create_datetime', ['RIMMER', 'KRYTEN']),
        )
        for order_by, ordered in data:
            with self.subTest(order_by=order_by):
                await self._test_list_order_by(order_by, ordered)

    @skipUnless(APPLICATION.settings.test_db_connection.startswith('sqlite'), 'SQLite database test.')
    async def test_list_order_by_close_datetime_sqlite(self):
        # Null is ordered differently in SQLite than in postgres.
        # Ignore the problem, just test the ordering works.
        async with atomic() as session:
            await self._make_verification(session, contact_id='KRYTEN',
                                          close_datetime=datetime(2019, 12, 7, tzinfo=UTC))
            await self._make_verification(session, contact_id='RIMMER', uuid=UUID(int=43),
                                          close_datetime=datetime(2019, 12, 8, tzinfo=UTC))
            # unclosed verification
            await self._make_verification(session, contact_id='LISTER', uuid=UUID(int=44))
        data = (
            ('close_datetime', ['LISTER', 'KRYTEN', 'RIMMER']),
            ('+close_datetime', ['LISTER', 'KRYTEN', 'RIMMER']),
            ('-close_datetime', ['RIMMER', 'KRYTEN', 'LISTER']),
        )
        for order_by, ordered in data:
            with self.subTest(order_by=order_by):
                await self._test_list_order_by(order_by, ordered)

    @skipUnless(APPLICATION.settings.test_db_connection.startswith('postgres'), 'PostgreSQL database test.')
    async def test_list_order_by_close_datetime_postgres(self):
        # Test ordering works correctly in postgres.
        async with atomic() as session:
            await self._make_verification(session, contact_id='KRYTEN',
                                          close_datetime=datetime(2019, 12, 7, tzinfo=UTC))
            await self._make_verification(session, contact_id='RIMMER', uuid=UUID(int=43),
                                          close_datetime=datetime(2019, 12, 8, tzinfo=UTC))
            # unclosed verification
            await self._make_verification(session, contact_id='LISTER', uuid=UUID(int=44))
        data = (
            ('close_datetime', ['KRYTEN', 'RIMMER', 'LISTER']),
            ('+close_datetime', ['KRYTEN', 'RIMMER', 'LISTER']),
            ('-close_datetime', ['LISTER', 'RIMMER', 'KRYTEN']),
        )
        for order_by, ordered in data:
            with self.subTest(order_by=order_by):
                await self._test_list_order_by(order_by, ordered)

    async def test_list_order_by_status(self):
        async with atomic() as session:
            await self._make_verification(session, contact_id='KRYTEN', status=VerificationStatus.PENDING)
            await self._make_verification(session, contact_id='RIMMER', uuid=UUID(int=43),
                                          status=VerificationStatus.CREATED)
        data = (
            ('status', ['RIMMER', 'KRYTEN']),
            ('+status', ['RIMMER', 'KRYTEN']),
            ('-status', ['KRYTEN', 'RIMMER']),
        )
        for order_by, ordered in data:
            with self.subTest(order_by=order_by):
                await self._test_list_order_by(order_by, ordered)

    async def test_list_order_by_junk(self):
        request = ListRequest()
        request.order_by = 'unknown'
        context = AsyncMock(name='context')
        context.abort.side_effect = Exception(sentinel.abort)

        with self.assertRaises(Exception) as catcher:
            await atuple(self.service.list(request, context))

        self.assertEqual(catcher.exception.args, (sentinel.abort, ))
        self.assertEqual(mock_awaits(context),
                         [call.abort(StatusCode.INVALID_ARGUMENT, StringComparison("Invalid order_by value: .*"))])

    async def test_list_page_size(self):
        # Test page_size is defined.
        async with atomic() as session:
            await self._make_verification(session)
            # Make the second validation older than the first.
            await self._make_verification(session, contact_id='RIMMER', uuid=UUID(int=196),
                                          create_datetime=datetime(2019, 12, 8, tzinfo=UTC))
        request = ListRequest()
        request.pagination.page_size = 1

        result = self.service.list(request, sentinel.context)

        reply = ListReply()
        reply.data.verifications.append(self._get_verification_info())
        reply.data.pagination.next_page_token = str(self.uuid)
        reply.data.pagination.items_left = 1
        self.assertEqual(await atuple(result), (reply, ))

    async def test_list_page_size_negative(self):
        # Test page_size is defined.
        async with atomic() as session:
            await self._make_verification(session)
            # Make the second validation older than the first.
            old_uuid = UUID(int=196)
            old_datetime = datetime(2019, 12, 8, tzinfo=UTC)
            await self._make_verification(session, contact_id='RIMMER', uuid=old_uuid, create_datetime=old_datetime)
        request = ListRequest()
        request.pagination.page_size = -1
        request.pagination.page_token = str(old_uuid)

        result = self.service.list(request, sentinel.context)

        reply = ListReply()
        # On page_size -1, it returns the last item from the previous page,
        # i.e. the verification referenced by the token itself.
        info = self._get_verification_info(id=str(old_uuid), contact_id='RIMMER', create_datetime=old_datetime)
        reply.data.verifications.append(info)
        # And the next token points to the newer verification.
        reply.data.pagination.next_page_token = str(self.uuid)
        reply.data.pagination.items_left = 1
        self.assertEqual(await atuple(result), (reply, ))

    async def test_list_page_size_large(self):
        # Test page_size is larger than number of results
        async with atomic() as session:
            await self._make_verification(session)
        request = ListRequest()
        request.pagination.page_size = 100

        result = self.service.list(request, sentinel.context)

        reply = ListReply()
        reply.data.verifications.append(self._get_verification_info())
        reply.data.pagination.next_page_token = ''
        reply.data.pagination.items_left = 0
        self.assertEqual(await atuple(result), (reply, ))

    async def test_list_page_token(self):
        # Test page_token is defined.
        async with atomic() as session:
            await self._make_verification(session)
            # Make the second validation older than the first.
            old_uuid = UUID(int=196)
            old_datetime = datetime(2019, 12, 8, tzinfo=UTC)
            await self._make_verification(session, contact_id='RIMMER', uuid=old_uuid, create_datetime=old_datetime)
        request = ListRequest()
        request.pagination.page_size = 1
        request.pagination.page_token = str(self.uuid)

        result = self.service.list(request, sentinel.context)

        reply = ListReply()
        info = self._get_verification_info(id=str(old_uuid), contact_id='RIMMER', create_datetime=old_datetime)
        reply.data.verifications.append(info)
        reply.data.pagination.next_page_token = ''
        reply.data.pagination.items_left = 0
        self.assertEqual(await atuple(result), (reply, ))

    async def test_list_page_token_invalid(self):
        # Test page_token is defined but invalid.
        request = ListRequest()
        request.pagination.page_token = 'invalid'
        context = AsyncMock(name='context')
        context.abort.side_effect = Exception(sentinel.abort)

        with self.assertRaises(Exception) as catcher:
            await atuple(self.service.list(request, context))

        self.assertEqual(catcher.exception.args, (sentinel.abort, ))
        self.assertEqual(mock_awaits(context),
                         [call.abort(StatusCode.INVALID_ARGUMENT, StringComparison("Invalid page token"))])

    async def test_list_page_token_unknown(self):
        # Test page_token is valid, but unknown.
        request = ListRequest()
        request.pagination.page_token = str(self.uuid)
        context = AsyncMock(name='context')
        context.abort.side_effect = Exception(sentinel.abort)

        with self.assertRaises(Exception) as catcher:
            await atuple(self.service.list(request, context))

        self.assertEqual(catcher.exception.args, (sentinel.abort, ))
        self.assertEqual(mock_awaits(context),
                         [call.abort(StatusCode.INVALID_ARGUMENT, StringComparison("Invalid page token"))])

    async def test_list_pagination_last(self):
        # Test pagination is in the last reply.
        async with atomic() as session:
            await self._make_verification(session)
            await self._make_verification(session, contact_id='RIMMER', uuid=UUID(int=196),
                                          create_datetime=datetime(2019, 12, 8, tzinfo=UTC))
            old_uuid = UUID(int=256)
            old_datetime = datetime(2019, 12, 9, tzinfo=UTC)
            await self._make_verification(session, contact_id='LISTER', uuid=old_uuid, create_datetime=old_datetime)
        request = ListRequest()
        request.pagination.page_size = 2

        result = self.service.list(request, sentinel.context)

        reply_1 = ListReply()
        reply_1.data.verifications.append(self._get_verification_info())
        reply_2 = ListReply()
        info = self._get_verification_info(id=str(old_uuid), contact_id='LISTER', create_datetime=old_datetime)
        reply_2.data.verifications.append(info)
        reply_2.data.pagination.next_page_token = str(old_uuid)
        reply_2.data.pagination.items_left = 1
        self.assertEqual(await atuple(result), (reply_1, reply_2))

    async def test_list_pagination_reproducible(self):
        # Test pagination is reproducible for verifications with the same create datetime.
        async with atomic() as session:
            # Create four items in shuffled order.
            await self._make_verification(session, contact_id='THIRD', uuid=UUID(int=3), id=3)
            await self._make_verification(session, contact_id='FIRST', uuid=UUID(int=1), id=1)
            await self._make_verification(session, contact_id='FOURTH', uuid=UUID(int=4), id=4)
            await self._make_verification(session, contact_id='SECOND', uuid=UUID(int=2), id=2)
        # Request a page slice from the third item (returned with descending IDs).
        request = ListRequest()
        request.pagination.page_size = 2
        request.pagination.page_token = str(UUID(int=3))

        result = self.service.list(request, sentinel.context)

        reply_1 = ListReply()
        reply_1.data.verifications.append(self._get_verification_info(contact_id='SECOND', id=str(UUID(int=2))))
        reply_2 = ListReply()
        reply_2.data.verifications.append(self._get_verification_info(contact_id='FIRST', id=str(UUID(int=1))))
        reply_2.data.pagination.next_page_token = ''
        reply_2.data.pagination.items_left = 0
        self.assertEqual(await atuple(result), (reply_1, reply_2))

    async def test_list_pagination_reproducible_backward(self):
        # Test pagination is reproducible for verifications with the same create datetime - backward direction.
        async with atomic() as session:
            # Create four items in shuffled order.
            await self._make_verification(session, contact_id='THIRD', uuid=UUID(int=3), id=3)
            await self._make_verification(session, contact_id='FIRST', uuid=UUID(int=1), id=1)
            await self._make_verification(session, contact_id='FOURTH', uuid=UUID(int=4), id=4)
            await self._make_verification(session, contact_id='SECOND', uuid=UUID(int=2), id=2)
        # Request a page slice from the third item (returned with descending IDs).
        request = ListRequest()
        request.pagination.page_size = -2
        request.pagination.page_token = str(UUID(int=3))

        result = self.service.list(request, sentinel.context)

        reply_1 = ListReply()
        reply_1.data.verifications.append(self._get_verification_info(contact_id='THIRD', id=str(UUID(int=3))))
        reply_2 = ListReply()
        reply_2.data.verifications.append(self._get_verification_info(contact_id='FOURTH', id=str(UUID(int=4))))
        reply_2.data.pagination.next_page_token = ''
        reply_2.data.pagination.items_left = 0
        self.assertEqual(await atuple(result), (reply_1, reply_2))

    async def test_get_methods_db_error(self):
        request = GetMethodsRequest()
        request.verification_id.value = str(UUID(int=42))
        await self._test_db_error(self.service.get_methods, request)

    async def _test_get_methods_not_exist(self, request: GetMethodsRequest) -> None:
        result = await self.service.get_methods(request, sentinel.context)

        reply = GetMethodsReply()
        reply.exception.verification_does_not_exist.CopyFrom(VerificationDoesNotExist())
        self.assertEqual(result, reply)

    async def test_get_methods_empty(self):
        request = GetMethodsRequest()
        await self._test_get_methods_not_exist(request)

    async def test_get_methods_invalid(self):
        request = GetMethodsRequest()
        request.verification_id.value = 'INVALID'
        await self._test_get_methods_not_exist(request)

    async def test_get_methods_unknown(self):
        request = GetMethodsRequest()
        request.verification_id.value = str(UUID(int=42))
        await self._test_get_methods_not_exist(request)

    async def test_get_methods_no_processors(self):
        async with atomic() as session:
            await self._make_verification(session)
        request = GetMethodsRequest()
        request.verification_id.value = str(self.uuid)

        with override_settings(processors={}):
            result = await self.service.get_methods(request, sentinel.context)

        self.assertEqual(result, GetMethodsReply())

    async def test_get_methods(self):
        async with atomic() as session:
            await self._make_verification(session)
        request = GetMethodsRequest()
        request.verification_id.value = str(self.uuid)

        processors = {'no-psirens': {'class': 'verification.tests.test_service.TestProcessor'},
                      'no-replicants': {'class': 'verification.tests.test_service.TestProcessor'}}
        with override_settings(processors=processors):
            result = await self.service.get_methods(request, sentinel.context)

        reply = GetMethodsReply()
        reply.data.methods[:] = ['no-psirens', 'no-replicants']
        self.assertEqual(result, reply)

    async def test_get_methods_inapplicable(self):
        async with atomic() as session:
            await self._make_verification(session)
        request = GetMethodsRequest()
        request.verification_id.value = str(self.uuid)

        processors = {'no-psirens': {'class': 'verification.tests.test_service.TestProcessor', 'is_applicable': False}}
        with override_settings(processors=processors):
            result = await self.service.get_methods(request, sentinel.context)

        self.assertEqual(result, GetMethodsReply())

    async def test_use_method_db_error(self):
        request = UseMethodRequest()
        request.verification_id.value = str(UUID(int=42))
        await self._test_db_error(self.service.use_method, request)

    async def _test_use_method_not_exist(self, request: UseMethodRequest) -> None:
        result = await self.service.use_method(request, sentinel.context)

        reply = UseMethodReply()
        reply.exception.verification_does_not_exist.CopyFrom(VerificationDoesNotExist())
        self.assertEqual(result, reply)

    async def test_use_method_empty(self):
        request = UseMethodRequest()
        await self._test_use_method_not_exist(request)

    async def test_use_method_invalid(self):
        request = UseMethodRequest()
        request.verification_id.value = 'INVALID'
        await self._test_use_method_not_exist(request)

    async def test_use_method_unknown(self):
        request = UseMethodRequest()
        request.verification_id.value = str(UUID(int=42))
        await self._test_use_method_not_exist(request)

    async def _test_use_method_incorrect_status(self, status: VerificationStatus) -> None:
        async with atomic() as session:
            await self._make_verification(session, status=status)
        request = UseMethodRequest()
        request.verification_id.value = str(self.uuid)

        result = await self.service.use_method(request, sentinel.context)

        reply = UseMethodReply()
        reply.exception.incorrect_status.CopyFrom(IncorrectStatus())
        self.assertEqual(result, reply)

    async def test_use_method_incorrect_succeeded(self):
        await self._test_use_method_incorrect_status(VerificationStatus.SUCCEEDED)

    async def test_use_method_incorrect_failed(self):
        await self._test_use_method_incorrect_status(VerificationStatus.FAILED)

    async def test_use_method_incorrect_canceled(self):
        await self._test_use_method_incorrect_status(VerificationStatus.CANCELED)

    async def test_use_method_incorrect_manual(self):
        await self._test_use_method_incorrect_status(VerificationStatus.MANUAL)

    async def test_use_method_no_processors(self):
        async with atomic() as session:
            await self._make_verification(session)
        request = UseMethodRequest(method='no-psirens')
        request.verification_id.value = str(self.uuid)

        with override_settings(processors={}):
            result = await self.service.use_method(request, sentinel.context)

        reply = UseMethodReply()
        reply.exception.method_not_applicable.CopyFrom(MethodNotApplicable())
        self.assertEqual(result, reply)

    async def _test_use_method(self, status: VerificationStatus, final_status: VerificationStatus) -> None:
        async with atomic() as session:
            await self._make_verification(session, status=status)
        request = UseMethodRequest(method='no-psirens')
        request.verification_id.value = str(self.uuid)

        processor = TestEventsProcessor(next_events=[], method='no-psirens')
        APPLICATION.handler = TestHandler([processor])
        result = await self.service.use_method(request, sentinel.context)

        async with test_session_cls()() as session:
            event_timestamp = (await session.execute(select(Event))).scalars().one().event_datetime
            verification = (await session.execute(select(Verification))).scalars().one()
            reply = UseMethodReply()
            proto_event = ProtoEvent()
            proto_event.event_datetime.FromDatetime(event_timestamp)
            proto_event.type = EventType.USE_METHOD
            proto_event.method = 'no-psirens'
            info = self._get_verification_info(id=str(verification.uuid), create_datetime=verification.create_datetime,
                                               status=ProtoVerificationStatus.Value(final_status), events=[proto_event])
            reply.data.verification.CopyFrom(info)
            self.assertEqual(result, reply)

            # Check verification is the same, but active methods changed.
            self.assertEqual(verification.status, final_status)
            self.assertIsNone(verification.close_datetime)
            self.assertEqual(await verification.active_methods, {'no-psirens'})
            event = (await session.execute(select(Event))).scalars().one()
            self.assertEqual(event.type, EventType.USE_METHOD)
            self.assertEqual(event.details, {})

    async def test_use_method(self):
        await self._test_use_method(VerificationStatus.CREATED, VerificationStatus.PENDING)

    async def test_use_method_constrained(self):
        # Test verification status is unchanged if status is further than PENDING.
        await self._test_use_method(VerificationStatus.CONSTRAINED, VerificationStatus.CONSTRAINED)

    async def test_use_method_repeated(self):
        async with atomic() as session:
            verification = await self._make_verification(session)
            await make_event(session, verification_id=verification.id, type=EventType.USE_METHOD, method='no-psirens',
                             event_datetime=self.event_datetime)
        request = UseMethodRequest(method='no-psirens')
        request.verification_id.value = str(self.uuid)

        processor = TestEventsProcessor(next_events=[], method='no-psirens')
        APPLICATION.handler = TestHandler([processor])
        result = await self.service.use_method(request, sentinel.context)

        async with test_session_cls()() as session:
            verification = (await session.execute(select(Verification))).scalars().one()
            reply = UseMethodReply()
            proto_event = ProtoEvent()
            proto_event.event_datetime.FromDatetime(self.event_datetime)
            proto_event.type = EventType.USE_METHOD
            proto_event.method = 'no-psirens'
            info = self._get_verification_info(id=str(verification.uuid), create_datetime=verification.create_datetime,
                                               status=ProtoVerificationStatus.created, events=[proto_event])
            reply.data.verification.CopyFrom(info)
            self.assertEqual(result, reply)

            # Check nothing changed.
            self.assertEqual(verification.status, VerificationStatus.CREATED)
            self.assertIsNone(verification.close_datetime)
            self.assertEqual(await verification.active_methods, {'no-psirens'})
            self.assertEqual(len(await verification.awaitable_attrs.events), 1)

    async def test_use_method_inapplicable(self):
        async with atomic() as session:
            await self._make_verification(session)
        request = UseMethodRequest(method='no-psirens')
        request.verification_id.value = str(self.uuid)

        processors = {'no-psirens': {'class': 'verification.tests.test_service.TestProcessor', 'is_applicable': False}}
        with override_settings(processors=processors):
            result = await self.service.use_method(request, sentinel.context)

        reply = UseMethodReply()
        reply.exception.method_not_applicable.CopyFrom(MethodNotApplicable())
        self.assertEqual(result, reply)

    async def test_verify_contact_data_db_error(self):
        request = VerifyContactDataRequest()
        request.verification_id.value = str(UUID(int=42))
        await self._test_db_error(self.service.verify_contact_data, request)

    async def _test_verify_contact_data_not_exist(self, request: VerifyContactDataRequest) -> None:
        result = await self.service.verify_contact_data(request, sentinel.context)

        reply = VerifyContactDataReply()
        reply.exception.verification_does_not_exist.CopyFrom(VerificationDoesNotExist())
        self.assertEqual(result, reply)

    async def test_verify_contact_data_empty(self):
        request = VerifyContactDataRequest()
        await self._test_verify_contact_data_not_exist(request)

    async def test_verify_contact_data_invalid(self):
        request = VerifyContactDataRequest()
        request.verification_id.value = 'INVALID'
        await self._test_verify_contact_data_not_exist(request)

    async def test_verify_contact_data_unknown(self):
        request = VerifyContactDataRequest()
        request.verification_id.value = str(UUID(int=42))
        await self._test_verify_contact_data_not_exist(request)

    async def _test_verify_contact_data_incorrect_status(self, status: VerificationStatus) -> None:
        async with atomic() as session:
            await self._make_verification(session, status=status)
        request = VerifyContactDataRequest()
        request.verification_id.value = str(self.uuid)

        result = await self.service.verify_contact_data(request, sentinel.context)

        reply = VerifyContactDataReply()
        reply.exception.incorrect_status.CopyFrom(IncorrectStatus())
        self.assertEqual(result, reply)

    async def test_verify_contact_data_incorrect_succeeded(self):
        await self._test_verify_contact_data_incorrect_status(VerificationStatus.SUCCEEDED)

    async def test_verify_contact_data_incorrect_failed(self):
        await self._test_verify_contact_data_incorrect_status(VerificationStatus.FAILED)

    async def test_verify_contact_data_incorrect_canceled(self):
        await self._test_verify_contact_data_incorrect_status(VerificationStatus.CANCELED)

    async def test_verify_contact_data_incorrect_manual(self):
        await self._test_verify_contact_data_incorrect_status(VerificationStatus.MANUAL)

    async def _test_verify_contact_data_method_invalid(self, request: VerifyContactDataRequest) -> None:
        async with atomic() as session:
            await self._make_verification(session)

        result = await self.service.verify_contact_data(request, sentinel.context)

        reply = VerifyContactDataReply()
        reply.exception.method_not_applicable.CopyFrom(MethodNotApplicable())
        self.assertEqual(result, reply)

    async def test_verify_contact_data_method_missing(self):
        request = VerifyContactDataRequest()
        request.verification_id.value = str(self.uuid)

        await self._test_verify_contact_data_method_invalid(request)

    async def test_verify_contact_data_method_unknown(self):
        request = VerifyContactDataRequest()
        request.verification_id.value = str(self.uuid)
        request.method = 'no-psirens'

        with override_settings(processors={}):
            await self._test_verify_contact_data_method_invalid(request)

    async def test_verify_contact_data_method_unapplicable(self):
        request = VerifyContactDataRequest()
        request.verification_id.value = str(self.uuid)
        request.method = 'no-psirens'

        processors = {'no-psirens': {'class': 'verification.tests.test_service.TestProcessor', 'is_applicable': False}}
        with override_settings(processors=processors):
            await self._test_verify_contact_data_method_invalid(request)

    async def _test_verify_contact_data(self, request: VerifyContactDataRequest, reply: VerifyContactDataReply,
                                        details: Dict[str, Any], contact: Contact) -> ContactData:
        async with atomic() as session:
            await self._make_verification(session)

        request.verification_id.value = str(self.uuid)
        request.method = 'no-psirens'

        processor = TestEventsProcessor(next_events=[], method='no-psirens')
        APPLICATION.handler = TestHandler([processor])
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = contact

        result = await self.service.verify_contact_data(request, sentinel.context)

        async with test_session_cls()() as session:
            # These timestamps may differ on SQLite DB.
            timestamp = (await session.execute(select(ContactData))).scalars().one().verify_datetime
            event_timestamp = (await session.execute(select(Event))).scalars().one().event_datetime
            proto_event = ProtoEvent()
            proto_event.event_datetime.FromDatetime(event_timestamp)
            proto_event.type = EventType.VERIFY_CONTACT_DATA
            proto_event.method = 'no-psirens'
            proto_event.details.update(details)
            reply.data.verification.CopyFrom(self._get_verification_info(events=[proto_event]))
            reply.data.result.verify_datetime.FromDatetime(timestamp)
            reply.data.result.method = 'no-psirens'
            reply.data.result.contact_history_id.value = self.contact_history_id
            self.assertEqual(result, reply)

            # Check verification unchanged.
            verification = (await session.execute(select(Verification))).scalars().one()
            self.assertIsNone(verification.close_datetime)
            self.assertEqual(verification.status, VerificationStatus.CREATED)
            # Check event created.
            event = (await session.execute(select(Event))).scalars().one()
            self.assertEqual(event.type, EventType.VERIFY_CONTACT_DATA)
            self.assertEqual(event.details, details)
            # Check contact data verified and stored.
            contact_data = (await session.execute(select(ContactData))).scalars().one()
            self.assertEqual(contact_data.method, 'no-psirens')
            self.assertEqual(contact_data.contact_history_id, self.contact_history_id)

        return contact_data

    async def test_verify_contact_data_name(self):
        request = VerifyContactDataRequest()
        request.data.name = 'Kryten'

        reply = VerifyContactDataReply()
        reply.data.result.data.name = 'Kryten'
        reply.data.result.name = True
        details = {'name': 'Kryten'}

        contact = Contact(contact_id=self.contact_id, contact_handle='2X4B', contact_history_id=self.contact_history_id,
                          name='Kryten')

        contact_data = await self._test_verify_contact_data(request, reply, details, contact)

        self.assertEqual(contact_data.name, 'Kryten')
        self.assertTrue(contact_data.verified_name)

    async def test_verify_contact_data_organization(self):
        request = VerifyContactDataRequest()
        request.data.organization = 'JMC'

        reply = VerifyContactDataReply()
        reply.data.result.data.organization = 'JMC'
        reply.data.result.organization = True
        details = {'organization': 'JMC'}

        contact = Contact(contact_id=self.contact_id, contact_handle='2X4B', contact_history_id=self.contact_history_id,
                          organization='JMC')

        contact_data = await self._test_verify_contact_data(request, reply, details, contact)

        self.assertEqual(contact_data.organization, 'JMC')
        self.assertTrue(contact_data.verified_organization)

    async def test_verify_contact_data_birthdate(self):
        request = VerifyContactDataRequest()
        request.data.birthdate = '1988-09-06'

        reply = VerifyContactDataReply()
        reply.data.result.data.birthdate = '1988-09-06'
        reply.data.result.birthdate = True
        details = {'birthdate': '1988-09-06'}

        ident = AdditionalIdentifier(type=ContactIdentifierType.BIRTHDATE, value='1988-09-06')
        contact = Contact(contact_id=self.contact_id, contact_handle='2X4B', contact_history_id=self.contact_history_id,
                          additional_identifier=ident)

        contact_data = await self._test_verify_contact_data(request, reply, details, contact)

        self.assertEqual(contact_data.birthdate, date(1988, 9, 6))
        self.assertTrue(contact_data.verified_birthdate)

    async def test_verify_contact_data_birthdate_invalid(self):
        async with atomic() as session:
            await self._make_verification(session)

        request = VerifyContactDataRequest()
        request.verification_id.value = str(self.uuid)
        request.method = 'no-psirens'
        request.data.birthdate = 'JUNK'

        processors = {'no-psirens': {'class': 'verification.tests.test_service.TestProcessor'}}
        with override_settings(processors=processors):
            result = await self.service.verify_contact_data(request, sentinel.context)

        reply = VerifyContactDataReply()
        reply.exception.invalid_data.CopyFrom(InvalidData(fields=['birthdate']))
        self.assertEqual(result, reply)

    async def test_verify_contact_data_company_registration_number(self):
        request = VerifyContactDataRequest()
        request.data.company_registration_number = 'JMC42'

        reply = VerifyContactDataReply()
        reply.data.result.data.company_registration_number = 'JMC42'
        reply.data.result.company_registration_number = True
        details = {'company_registration_number': 'JMC42'}

        ident = AdditionalIdentifier(type=ContactIdentifierType.COMPANY_REGISTRATION_NUMBER, value='JMC42')
        contact = Contact(contact_id=self.contact_id, contact_handle='2X4B', contact_history_id=self.contact_history_id,
                          additional_identifier=ident)

        contact_data = await self._test_verify_contact_data(request, reply, details, contact)

        self.assertEqual(contact_data.company_registration_number, 'JMC42')
        self.assertTrue(contact_data.verified_company_registration_number)

    async def test_verify_contact_data_address(self):
        request = VerifyContactDataRequest()
        request.data.address.street.append('Deck 16')
        request.data.address.city = 'Red Dwarf'
        request.data.address.state_or_province = 'Deep Space'
        request.data.address.postal_code = 'JMC'
        request.data.address.country_code = 'JU'

        reply = VerifyContactDataReply()
        reply.data.result.data.address.street.append('Deck 16')
        reply.data.result.data.address.city = 'Red Dwarf'
        reply.data.result.data.address.state_or_province = 'Deep Space'
        reply.data.result.data.address.postal_code = 'JMC'
        reply.data.result.data.address.country_code = 'JU'
        reply.data.result.address = True
        details = {'address': {'street': ['Deck 16'], 'city': 'Red Dwarf', 'state_or_province': 'Deep Space',
                               'postal_code': 'JMC', 'country_code': 'JU'}}

        address = Address(street=['Deck 16'], city='Red Dwarf', state_or_province='Deep Space', postal_code='JMC',
                          country_code='JU')
        contact = Contact(contact_id=self.contact_id, contact_handle='2X4B', contact_history_id=self.contact_history_id,
                          place=address)

        contact_data = await self._test_verify_contact_data(request, reply, details, contact)

        self.assertEqual(contact_data.address_street, ['Deck 16'])
        self.assertEqual(contact_data.address_city, 'Red Dwarf')
        self.assertEqual(contact_data.address_state, 'Deep Space')
        self.assertEqual(contact_data.address_postal_code, 'JMC')
        self.assertEqual(contact_data.address_country_code, 'JU')
        self.assertTrue(contact_data.verified_address)

    async def _test_verify_contact_data_complete(self, status: VerificationStatus) -> None:
        async with atomic() as session:
            await self._make_verification(session)
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = Contact(
            contact_id=self.contact_id, contact_handle='2X4B', contact_history_id=self.contact_history_id,
            name='Kryten')

        request = VerifyContactDataRequest()
        request.data.name = 'Kryten'
        request.verification_id.value = str(self.uuid)
        request.method = 'no-psirens'

        result = await self.service.verify_contact_data(request, sentinel.context)

        async with test_session_cls()() as session:
            # These timestamps may differ on SQLite DB.
            timestamp = (await session.execute(select(ContactData))).scalars().one().verify_datetime
            event_timestamp = (await session.execute(select(Event))).scalars().one().event_datetime
            proto_event = ProtoEvent()
            proto_event.event_datetime.FromDatetime(event_timestamp)
            proto_event.type = EventType.VERIFY_CONTACT_DATA
            proto_event.method = 'no-psirens'
            proto_event.details.update({'name': 'Kryten'})
            reply = VerifyContactDataReply()
            reply.data.result.data.name = 'Kryten'
            reply.data.result.name = True
            reply.data.verification.CopyFrom(self._get_verification_info(events=[proto_event],
                                                                         status=ProtoVerificationStatus.Value(status)))
            reply.data.result.verify_datetime.FromDatetime(timestamp)
            reply.data.result.method = 'no-psirens'
            reply.data.result.contact_history_id.value = self.contact_history_id
            self.assertEqual(result, reply)

            # Check verification status.
            verification = (await session.execute(select(Verification))).scalars().one()
            self.assertEqual(verification.status, status)
            del verification
            # Check event created.
            event = (await session.execute(select(Event))).scalars().one()
            self.assertEqual(event.type, EventType.VERIFY_CONTACT_DATA)
            self.assertEqual(event.details, {'name': 'Kryten'})
            # Check contact data verified and stored.
            contact_data = (await session.execute(select(ContactData))).scalars().one()
            self.assertEqual(contact_data.method, 'no-psirens')
            self.assertEqual(contact_data.contact_history_id, self.contact_history_id)
            self.assertEqual(contact_data.name, 'Kryten')
            self.assertTrue(contact_data.verified_name)
            # Check complete attempted
            # XXX: Attach 'verification' argument to a session so it can be checked.
            handler_mock = cast(AsyncMock, APPLICATION.handler)
            session.add(handler_mock.mock_calls[0][1][0])
            await session.refresh(handler_mock.mock_calls[0][1][0])
            self.assertEqual(mock_awaits(handler_mock),
                             [call.try_complete(Comparison(Verification, uuid=self.uuid, partial=True))])

    async def test_verify_contact_data_complete(self):
        APPLICATION.handler = AsyncMock(VerificationHandler)
        APPLICATION.handler.processors = {'no-psirens': TestEventsProcessor(next_events=[], method='no-psirens')}

        def try_complete(verification: Verification) -> bool:
            verification.status = VerificationStatus.SUCCEEDED
            return True

        APPLICATION.handler.try_complete.side_effect = try_complete

        await self._test_verify_contact_data_complete(VerificationStatus.SUCCEEDED)

    async def test_verify_contact_data_complete_error(self):
        APPLICATION.handler = AsyncMock(VerificationHandler)
        APPLICATION.handler.processors = {'no-psirens': TestEventsProcessor(next_events=[], method='no-psirens')}
        APPLICATION.handler.try_complete.side_effect = Exception('Smeghead!')

        await self._test_verify_contact_data_complete(VerificationStatus.CREATED)

    async def test_request_manual_check_db_error(self):
        request = RequestManualCheckRequest()
        request.verification_id.value = str(UUID(int=42))
        await self._test_db_error(self.service.request_manual_check, request)

    async def _test_request_manual_check_not_exist(self, request: RequestManualCheckRequest) -> None:
        result = await self.service.request_manual_check(request, sentinel.context)

        reply = RequestManualCheckReply()
        reply.exception.verification_does_not_exist.CopyFrom(VerificationDoesNotExist())
        self.assertEqual(result, reply)

    async def test_request_manual_check_empty(self):
        request = RequestManualCheckRequest()
        await self._test_request_manual_check_not_exist(request)

    async def test_request_manual_check_invalid(self):
        request = RequestManualCheckRequest()
        request.verification_id.value = 'INVALID'
        await self._test_request_manual_check_not_exist(request)

    async def test_request_manual_check_unknown(self):
        request = RequestManualCheckRequest()
        request.verification_id.value = str(UUID(int=42))
        await self._test_request_manual_check_not_exist(request)

    async def _test_request_manual_check_incorrect_status(self, status: VerificationStatus) -> None:
        async with atomic() as session:
            await self._make_verification(session, status=status)
        request = RequestManualCheckRequest()
        request.verification_id.value = str(self.uuid)

        result = await self.service.request_manual_check(request, sentinel.context)

        reply = RequestManualCheckReply()
        reply.exception.incorrect_status.CopyFrom(IncorrectStatus())
        self.assertEqual(result, reply)

    async def test_request_manual_check_incorrect_succeeded(self):
        await self._test_request_manual_check_incorrect_status(VerificationStatus.SUCCEEDED)

    async def test_request_manual_check_incorrect_failed(self):
        await self._test_request_manual_check_incorrect_status(VerificationStatus.FAILED)

    async def test_request_manual_check_incorrect_canceled(self):
        await self._test_request_manual_check_incorrect_status(VerificationStatus.CANCELED)

    async def test_request_manual_check_incorrect_manual(self):
        await self._test_request_manual_check_incorrect_status(VerificationStatus.MANUAL)

    async def test_request_manual_check(self):
        async with atomic() as session:
            verification = await self._make_verification(session)
            verification_id = verification.id
        request = RequestManualCheckRequest()
        request.verification_id.value = str(self.uuid)

        result = await self.service.request_manual_check(request, sentinel.context)

        self.assertEqual(result, RequestManualCheckReply())
        # Check event created
        async with test_session_cls()() as session:
            event = (await session.execute(select(Event))).scalars().one()
            self.assertEqual(event.verification_id, verification_id)
            self.assertIsNone(event.method)
            self.assertEqual(event.type, EventType.REQUEST_MANUAL_CHECK)
            self.assertEqual(event.details, {})

    async def test_resolve_manual_check_db_error(self):
        request = ResolveManualCheckRequest()
        request.verification_id.value = str(UUID(int=42))
        await self._test_db_error(self.service.resolve_manual_check, request)

    async def _test_resolve_manual_check_not_exist(self, request: ResolveManualCheckRequest) -> None:
        result = await self.service.resolve_manual_check(request, sentinel.context)

        reply = ResolveManualCheckReply()
        reply.exception.verification_does_not_exist.CopyFrom(VerificationDoesNotExist())
        self.assertEqual(result, reply)

    async def test_resolve_manual_check_empty(self):
        request = ResolveManualCheckRequest()
        await self._test_resolve_manual_check_not_exist(request)

    async def test_resolve_manual_check_invalid(self):
        request = ResolveManualCheckRequest()
        request.verification_id.value = 'INVALID'
        await self._test_resolve_manual_check_not_exist(request)

    async def test_resolve_manual_check_unknown(self):
        request = ResolveManualCheckRequest()
        request.verification_id.value = str(UUID(int=42))
        await self._test_resolve_manual_check_not_exist(request)

    async def _test_resolve_manual_check_incorrect_status(self, status: VerificationStatus) -> None:
        async with atomic() as session:
            await self._make_verification(session, status=status)
        request = ResolveManualCheckRequest()
        request.verification_id.value = str(self.uuid)

        result = await self.service.resolve_manual_check(request, sentinel.context)

        reply = ResolveManualCheckReply()
        reply.exception.incorrect_status.CopyFrom(IncorrectStatus())
        self.assertEqual(result, reply)

    async def test_resolve_manual_check_incorrect_succeeded(self):
        await self._test_resolve_manual_check_incorrect_status(VerificationStatus.SUCCEEDED)

    async def test_resolve_manual_check_incorrect_failed(self):
        await self._test_resolve_manual_check_incorrect_status(VerificationStatus.FAILED)

    async def test_resolve_manual_check_incorrect_canceled(self):
        await self._test_resolve_manual_check_incorrect_status(VerificationStatus.CANCELED)

    async def test_resolve_manual_check_incorrect_manual(self):
        await self._test_resolve_manual_check_incorrect_status(VerificationStatus.MANUAL)

    async def _test_resolve_manual_check(self, request: ResolveManualCheckRequest, details: Dict[str, Any],
                                         ) -> ContactData:
        async with atomic() as session:
            await self._make_verification(session, manual_check_request=True)
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = Contact(
            contact_id=ContactId('KRYTEN'), contact_handle='Chnd:1'
        )  # no emails

        result = await self.service.resolve_manual_check(request, sentinel.context)

        self.assertEqual(result, ResolveManualCheckReply())
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).assert_awaited_once_with('KRYTEN')
        cast(AsyncMock, APPLICATION.email_client.send).assert_not_called()  # no recipient provided

        async with test_session_cls()() as session:
            # Check manual request flag dropped.
            verification = (await session.execute(select(Verification))).scalars().one()
            self.assertFalse(verification.manual_check_request)
            # Check event created
            event = (await session.execute(select(Event))).scalars().one()
            self.assertEqual(event.verification_id, verification.id)
            self.assertIsNone(event.method)
            self.assertEqual(event.type, EventType.RESOLVE_MANUAL_CHECK)
            self.assertEqual(event.details, details)
            # Check verified data.
            contact_data = (await session.execute(select(ContactData))).scalars().one()
            self.assertEqual(contact_data.verification_id, verification.id)
            self.assertIsNone(contact_data.method)
            self.assertEqual(contact_data.contact_history_id, self.contact_history_id)
        return contact_data

    async def test_resolve_manual_check_name(self):
        request = ResolveManualCheckRequest()
        request.verification_id.value = str(self.uuid)
        request.verified = True
        request.data.name = 'Kryten'
        request.contact_history_id.value = self.contact_history_id

        contact_data = await self._test_resolve_manual_check(request, {'name': 'Kryten', 'verified': True})

        self.assertEqual(contact_data.name, 'Kryten')
        self.assertTrue(contact_data.verified_name)

    async def test_resolve_manual_check_name_false(self):
        request = ResolveManualCheckRequest()
        request.verification_id.value = str(self.uuid)
        request.verified = False
        request.data.name = 'Kryten'
        request.contact_history_id.value = self.contact_history_id

        contact_data = await self._test_resolve_manual_check(request, {'name': 'Kryten', 'verified': False})

        self.assertEqual(contact_data.name, 'Kryten')
        self.assertFalse(contact_data.verified_name)

    async def test_resolve_manual_check_organization(self):
        request = ResolveManualCheckRequest()
        request.verification_id.value = str(self.uuid)
        request.verified = True
        request.data.organization = 'JMC'
        request.contact_history_id.value = self.contact_history_id

        contact_data = await self._test_resolve_manual_check(request, {'organization': 'JMC', 'verified': True})

        self.assertEqual(contact_data.organization, 'JMC')
        self.assertTrue(contact_data.verified_organization)

    async def test_resolve_manual_check_organization_false(self):
        request = ResolveManualCheckRequest()
        request.verification_id.value = str(self.uuid)
        request.verified = False
        request.data.organization = 'JMC'
        request.contact_history_id.value = self.contact_history_id

        contact_data = await self._test_resolve_manual_check(request, {'organization': 'JMC', 'verified': False})

        self.assertEqual(contact_data.organization, 'JMC')
        self.assertFalse(contact_data.verified_organization)

    async def test_resolve_manual_check_birthdate(self):
        request = ResolveManualCheckRequest()
        request.verification_id.value = str(self.uuid)
        request.verified = True
        request.data.birthdate = '1988-09-06'
        request.contact_history_id.value = self.contact_history_id

        contact_data = await self._test_resolve_manual_check(request, {'birthdate': '1988-09-06', 'verified': True})

        self.assertEqual(contact_data.birthdate, date(1988, 9, 6))
        self.assertTrue(contact_data.verified_birthdate)

    async def test_resolve_manual_check_birthdate_false(self):
        request = ResolveManualCheckRequest()
        request.verification_id.value = str(self.uuid)
        request.verified = False
        request.data.birthdate = '1988-09-06'
        request.contact_history_id.value = self.contact_history_id

        contact_data = await self._test_resolve_manual_check(request, {'birthdate': '1988-09-06', 'verified': False})

        self.assertEqual(contact_data.birthdate, date(1988, 9, 6))
        self.assertFalse(contact_data.verified_birthdate)

    async def test_resolve_manual_check_birthdate_invalid(self):
        async with atomic() as session:
            await self._make_verification(session, manual_check_request=True)
        request = ResolveManualCheckRequest()
        request.verification_id.value = str(self.uuid)
        request.verified = False
        request.data.birthdate = 'JUNK'
        request.contact_history_id.value = self.contact_history_id

        result = await self.service.resolve_manual_check(request, sentinel.context)

        reply = ResolveManualCheckReply()
        reply.exception.invalid_data.CopyFrom(InvalidData(fields=['birthdate']))
        self.assertEqual(result, reply)
        # Check manual request flag not dropped.
        async with test_session_cls()() as session:
            verification = (await session.execute(select(Verification))).scalars().one()
        self.assertTrue(verification.manual_check_request)

    async def test_resolve_manual_check_company_registration_number(self):
        request = ResolveManualCheckRequest()
        request.verification_id.value = str(self.uuid)
        request.verified = True
        request.data.company_registration_number = '2X4B'
        request.contact_history_id.value = self.contact_history_id

        contact_data = await self._test_resolve_manual_check(
            request, {'company_registration_number': '2X4B', 'verified': True})

        self.assertEqual(contact_data.company_registration_number, '2X4B')
        self.assertTrue(contact_data.verified_company_registration_number)

    async def test_resolve_manual_check_company_registration_number_false(self):
        request = ResolveManualCheckRequest()
        request.verification_id.value = str(self.uuid)
        request.verified = False
        request.data.company_registration_number = '2X4B'
        request.contact_history_id.value = self.contact_history_id

        contact_data = await self._test_resolve_manual_check(
            request, {'company_registration_number': '2X4B', 'verified': False})

        self.assertEqual(contact_data.company_registration_number, '2X4B')
        self.assertFalse(contact_data.verified_company_registration_number)

    async def test_resolve_manual_check_street(self):
        request = ResolveManualCheckRequest()
        request.verification_id.value = str(self.uuid)
        request.verified = True
        request.data.address.street.append('Deck 16')
        request.contact_history_id.value = self.contact_history_id

        details = {'address': {'street': ['Deck 16'], 'city': '', 'state_or_province': '', 'postal_code': '',
                               'country_code': ''},
                   'verified': True}
        contact_data = await self._test_resolve_manual_check(request, details)

        self.assertEqual(contact_data.address_street, ['Deck 16'])
        self.assertTrue(contact_data.verified_address)

    async def test_resolve_manual_check_city(self):
        request = ResolveManualCheckRequest()
        request.verification_id.value = str(self.uuid)
        request.verified = True
        request.data.address.city = 'Red Dwarf'
        request.contact_history_id.value = self.contact_history_id

        details = {'address': {'street': [], 'city': 'Red Dwarf', 'state_or_province': '', 'postal_code': '',
                               'country_code': ''},
                   'verified': True}
        contact_data = await self._test_resolve_manual_check(request, details)

        self.assertEqual(contact_data.address_city, 'Red Dwarf')
        self.assertTrue(contact_data.verified_address)

    async def test_resolve_manual_check_state(self):
        request = ResolveManualCheckRequest()
        request.verification_id.value = str(self.uuid)
        request.verified = True
        request.data.address.state_or_province = 'Deep Space'
        request.contact_history_id.value = self.contact_history_id

        details = {'address': {'street': [], 'city': '', 'state_or_province': 'Deep Space', 'postal_code': '',
                               'country_code': ''},
                   'verified': True}
        contact_data = await self._test_resolve_manual_check(request, details)

        self.assertEqual(contact_data.address_state, 'Deep Space')
        self.assertTrue(contact_data.verified_address)

    async def test_resolve_manual_check_postal_code(self):
        request = ResolveManualCheckRequest()
        request.verification_id.value = str(self.uuid)
        request.verified = True
        request.data.address.postal_code = 'JMC'
        request.contact_history_id.value = self.contact_history_id

        details = {'address': {'street': [], 'city': '', 'state_or_province': '', 'postal_code': 'JMC',
                               'country_code': ''},
                   'verified': True}
        contact_data = await self._test_resolve_manual_check(request, details)

        self.assertEqual(contact_data.address_postal_code, 'JMC')
        self.assertTrue(contact_data.verified_address)

    async def test_resolve_manual_check_country_code(self):
        request = ResolveManualCheckRequest()
        request.verification_id.value = str(self.uuid)
        request.verified = True
        request.data.address.country_code = 'JU'
        request.contact_history_id.value = self.contact_history_id

        details = {'address': {'street': [], 'city': '', 'state_or_province': '', 'postal_code': '',
                               'country_code': 'JU'},
                   'verified': True}
        contact_data = await self._test_resolve_manual_check(request, details)

        self.assertEqual(contact_data.address_country_code, 'JU')
        self.assertTrue(contact_data.verified_address)

    async def test_resolve_manual_check_address_false(self):
        # Single test with unverified address.
        request = ResolveManualCheckRequest()
        request.verification_id.value = str(self.uuid)
        request.verified = False
        request.data.address.street.append('Deck 16')
        request.data.address.city = 'Red Dwarf'
        request.data.address.state_or_province = 'Deep Space'
        request.data.address.postal_code = 'JMC'
        request.data.address.country_code = 'JU'
        request.contact_history_id.value = self.contact_history_id

        details = {'address': {'street': ['Deck 16'], 'city': 'Red Dwarf', 'state_or_province': 'Deep Space',
                               'postal_code': 'JMC', 'country_code': 'JU'},
                   'verified': False}
        contact_data = await self._test_resolve_manual_check(request, details)

        self.assertEqual(contact_data.address_street, ['Deck 16'])
        self.assertEqual(contact_data.address_city, 'Red Dwarf')
        self.assertEqual(contact_data.address_state, 'Deep Space')
        self.assertEqual(contact_data.address_postal_code, 'JMC')
        self.assertEqual(contact_data.address_country_code, 'JU')
        self.assertFalse(contact_data.verified_address)

    async def _test_resolve_manual_check_email(self, futile_datetime: str) -> None:
        cast(AsyncMock, APPLICATION.email_client.send).return_value = "msg:1"
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = Contact(
            contact_id=ContactId('KRYTEN'), contact_handle='2X4B', emails=['kryten@example.org'],
            sponsoring_registrar='HOLLY',
        )  # with emails
        async with atomic() as session:
            await self._make_verification(session, manual_check_request=True, access_token='Gazpacho!')
        request = ResolveManualCheckRequest()
        request.verified = True
        request.data.name = 'Kryten'
        request.verification_id.value = str(self.uuid)

        result = await self.service.resolve_manual_check(request, sentinel.context)

        self.assertEqual(result, ResolveManualCheckReply())
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).assert_awaited_once_with('KRYTEN')
        async with test_session_cls()() as session:
            verification = (await session.execute(select(Verification))).scalars().one()
        context = {'uuid': str(verification.uuid), 'access_token': 'Gazpacho!',
                   'contact_data': {'name': 'Kryten', 'verified': True},
                   'verified': True, 'handle': '2X4B', 'registrar_handle': 'HOLLY', 'futile_datetime': futile_datetime}
        cast(AsyncMock, APPLICATION.email_client.send).assert_awaited_once_with(
            Email(
                recipient='kryten@example.org',
                subject_template=Template.MANUAL_CHECK_RESOLUTION_SUBJECT,
                body_template=Template.MANUAL_CHECK_RESOLUTION_BODY,
                type=MessageType.MANUAL_CHECK_RESOLUTION,
                context=context,
            ),
            references=[Reference(type=ReferenceType.CONTACT, value=verification.contact_id),
                        Reference(type=ReferenceType.VERIFICATION, value=str(verification.uuid))]
        )

    async def test_resolve_manual_check_email(self):
        APPLICATION.handler = AsyncMock(VerificationHandler)
        APPLICATION.handler.get_active_processors.return_value = aiter([])
        await self._test_resolve_manual_check_email('')

    async def test_resolve_manual_check_email_futile(self):
        APPLICATION.handler = AsyncMock(VerificationHandler)
        processor = AsyncMock(BaseProcessor)
        processor._get_futile_due.return_value = datetime(2020, 1, 24, 10, 9, 8, tzinfo=UTC)
        APPLICATION.handler.get_active_processors.return_value = aiter([processor])
        await self._test_resolve_manual_check_email('2020-01-24T10:09:08+00:00')

    async def _test_resolve_manual_check_complete(self) -> None:
        request = ResolveManualCheckRequest()
        request.verification_id.value = str(self.uuid)
        request.verified = True
        request.data.name = 'Kryten'
        request.contact_history_id.value = self.contact_history_id

        await self._test_resolve_manual_check(request, {'name': 'Kryten', 'verified': True})

        # Check complete attempted
        async with test_session_cls()() as session:
            # XXX: Attach 'verification' argument to a session so it can be checked.
            handler_mock = cast(AsyncMock, APPLICATION.handler)
            session.add(handler_mock.mock_calls[0][1][0])
            await session.refresh(handler_mock.mock_calls[0][1][0])
            self.assertEqual(mock_awaits(handler_mock),
                             [call.try_complete(Comparison(Verification, uuid=self.uuid, partial=True))])

    async def test_resolve_manual_check_complete(self):
        APPLICATION.handler = AsyncMock(VerificationHandler)

        await self._test_resolve_manual_check_complete()

    async def test_resolve_manual_check_complete_error(self):
        APPLICATION.handler = AsyncMock(VerificationHandler)
        APPLICATION.handler.try_complete.side_effect = Exception('Smeghead!')

        await self._test_resolve_manual_check_complete()

    async def test_list_verification_results_db_error(self):
        request = ListVerificationResultsRequest()
        request.verification_id.value = str(UUID(int=42))
        await self._test_db_error(self.service.list_verification_results, request)

    async def _test_list_verification_results_not_exist(self, request: ListVerificationResultsRequest) -> None:
        result = await self.service.list_verification_results(request, sentinel.context)

        reply = ListVerificationResultsReply()
        reply.exception.verification_does_not_exist.CopyFrom(VerificationDoesNotExist())
        self.assertEqual(result, reply)

    async def test_list_verification_results_empty(self):
        request = ListVerificationResultsRequest()
        await self._test_list_verification_results_not_exist(request)

    async def test_list_verification_results_invalid(self):
        request = ListVerificationResultsRequest()
        request.verification_id.value = 'INVALID'
        await self._test_list_verification_results_not_exist(request)

    async def test_list_verification_results_unknown(self):
        request = ListVerificationResultsRequest()
        request.verification_id.value = str(UUID(int=42))
        await self._test_list_verification_results_not_exist(request)

    async def test_list_verification_results_result_empty(self):
        async with atomic() as session:
            await self._make_verification(session)
        request = ListVerificationResultsRequest()
        request.verification_id.value = str(self.uuid)

        with freeze_time(self.close_datetime):
            result = await self.service.list_verification_results(request, sentinel.context)

        self.assertEqual(result, ListVerificationResultsReply())

    async def _test_list_verification_results(self, v_result: VerificationResult) -> None:
        request = ListVerificationResultsRequest()
        request.verification_id.value = str(self.uuid)

        with freeze_time(self.close_datetime):
            result = await self.service.list_verification_results(request, sentinel.context)

        reply = ListVerificationResultsReply()
        reply.data.results.append(v_result)
        self.assertEqual(result, reply)

    async def test_list_verification_results_name(self):
        async with atomic() as session:
            verification = await self._make_verification(session)
            await make_contact_data(session, verification_id=verification.id, verify_datetime=self.event_datetime,
                                    method=self.method, contact_history_id=self.contact_history_id,
                                    name='Kryten', verified_name=True)
        v_result = VerificationResult(method=self.method)
        v_result.contact_history_id.value = self.contact_history_id
        v_result.verify_datetime.FromDatetime(self.event_datetime)
        v_result.data.name = 'Kryten'
        v_result.name = True
        await self._test_list_verification_results(v_result)

    async def test_list_verification_results_name_false(self):
        async with atomic() as session:
            verification = await self._make_verification(session)
            await make_contact_data(session, verification_id=verification.id, verify_datetime=self.event_datetime,
                                    method=self.method, contact_history_id=self.contact_history_id,
                                    name='Kryten', verified_name=False)
        v_result = VerificationResult(method=self.method)
        v_result.contact_history_id.value = self.contact_history_id
        v_result.verify_datetime.FromDatetime(self.event_datetime)
        v_result.data.name = 'Kryten'
        v_result.name = False
        await self._test_list_verification_results(v_result)

    async def test_list_verification_results_organization(self):
        async with atomic() as session:
            verification = await self._make_verification(session)
            await make_contact_data(session, verification_id=verification.id, verify_datetime=self.event_datetime,
                                    method=self.method, contact_history_id=self.contact_history_id,
                                    organization='JMC', verified_organization=True)
        v_result = VerificationResult(method=self.method)
        v_result.contact_history_id.value = self.contact_history_id
        v_result.verify_datetime.FromDatetime(self.event_datetime)
        v_result.data.organization = 'JMC'
        v_result.organization = True
        await self._test_list_verification_results(v_result)

    async def test_list_verification_results_organization_false(self):
        async with atomic() as session:
            verification = await self._make_verification(session)
            await make_contact_data(session, verification_id=verification.id, verify_datetime=self.event_datetime,
                                    method=self.method, contact_history_id=self.contact_history_id,
                                    organization='JMC', verified_organization=False)
        v_result = VerificationResult(method=self.method)
        v_result.contact_history_id.value = self.contact_history_id
        v_result.verify_datetime.FromDatetime(self.event_datetime)
        v_result.data.organization = 'JMC'
        v_result.organization = False
        await self._test_list_verification_results(v_result)

    async def test_list_verification_results_birthdate(self):
        async with atomic() as session:
            verification = await self._make_verification(session)
            await make_contact_data(session, verification_id=verification.id, verify_datetime=self.event_datetime,
                                    method=self.method, contact_history_id=self.contact_history_id,
                                    birthdate=date(1988, 9, 6), verified_birthdate=True)
        v_result = VerificationResult(method=self.method)
        v_result.contact_history_id.value = self.contact_history_id
        v_result.verify_datetime.FromDatetime(self.event_datetime)
        v_result.data.birthdate = '1988-09-06'
        v_result.birthdate = True
        await self._test_list_verification_results(v_result)

    async def test_list_verification_results_birthdate_false(self):
        async with atomic() as session:
            verification = await self._make_verification(session)
            await make_contact_data(session, verification_id=verification.id, verify_datetime=self.event_datetime,
                                    method=self.method, contact_history_id=self.contact_history_id,
                                    birthdate=date(1988, 9, 6), verified_birthdate=False)
        v_result = VerificationResult(method=self.method)
        v_result.contact_history_id.value = self.contact_history_id
        v_result.verify_datetime.FromDatetime(self.event_datetime)
        v_result.data.birthdate = '1988-09-06'
        v_result.birthdate = False
        await self._test_list_verification_results(v_result)

    async def test_list_verification_results_company_registration_number(self):
        async with atomic() as session:
            verification = await self._make_verification(session)
            await make_contact_data(session, verification_id=verification.id, verify_datetime=self.event_datetime,
                                    method=self.method, contact_history_id=self.contact_history_id,
                                    company_registration_number='2X4B', verified_company_registration_number=True)
        v_result = VerificationResult(method=self.method)
        v_result.contact_history_id.value = self.contact_history_id
        v_result.verify_datetime.FromDatetime(self.event_datetime)
        v_result.data.company_registration_number = '2X4B'
        v_result.company_registration_number = True
        await self._test_list_verification_results(v_result)

    async def test_list_verification_results_company_registration_number_false(self):
        async with atomic() as session:
            verification = await self._make_verification(session)
            await make_contact_data(session, verification_id=verification.id, verify_datetime=self.event_datetime,
                                    method=self.method, contact_history_id=self.contact_history_id,
                                    company_registration_number='2X4B', verified_company_registration_number=False)
        v_result = VerificationResult(method=self.method)
        v_result.contact_history_id.value = self.contact_history_id
        v_result.verify_datetime.FromDatetime(self.event_datetime)
        v_result.data.company_registration_number = '2X4B'
        v_result.company_registration_number = False
        await self._test_list_verification_results(v_result)

    async def test_list_verification_results_street(self):
        async with atomic() as session:
            verification = await self._make_verification(session)
            await make_contact_data(session, verification_id=verification.id, verify_datetime=self.event_datetime,
                                    method=self.method, contact_history_id=self.contact_history_id,
                                    address_street=['Deck 16'], verified_address=True)
        v_result = VerificationResult(method=self.method)
        v_result.contact_history_id.value = self.contact_history_id
        v_result.verify_datetime.FromDatetime(self.event_datetime)
        v_result.data.address.street.append('Deck 16')
        v_result.address = True
        await self._test_list_verification_results(v_result)

    async def test_list_verification_results_city(self):
        async with atomic() as session:
            verification = await self._make_verification(session)
            await make_contact_data(session, verification_id=verification.id, verify_datetime=self.event_datetime,
                                    method=self.method, contact_history_id=self.contact_history_id,
                                    address_city='Red Dwarf', verified_address=True)
        v_result = VerificationResult(method=self.method)
        v_result.contact_history_id.value = self.contact_history_id
        v_result.verify_datetime.FromDatetime(self.event_datetime)
        v_result.data.address.city = 'Red Dwarf'
        v_result.address = True
        await self._test_list_verification_results(v_result)

    async def test_list_verification_results_state(self):
        async with atomic() as session:
            verification = await self._make_verification(session)
            await make_contact_data(session, verification_id=verification.id, verify_datetime=self.event_datetime,
                                    method=self.method, contact_history_id=self.contact_history_id,
                                    address_state='Deep Space', verified_address=True)
        v_result = VerificationResult(method=self.method)
        v_result.contact_history_id.value = self.contact_history_id
        v_result.verify_datetime.FromDatetime(self.event_datetime)
        v_result.data.address.state_or_province = 'Deep Space'
        v_result.address = True
        await self._test_list_verification_results(v_result)

    async def test_list_verification_results_postal_code(self):
        async with atomic() as session:
            verification = await self._make_verification(session)
            await make_contact_data(session, verification_id=verification.id, verify_datetime=self.event_datetime,
                                    method=self.method, contact_history_id=self.contact_history_id,
                                    address_postal_code='JMC', verified_address=True)
        v_result = VerificationResult(method=self.method)
        v_result.contact_history_id.value = self.contact_history_id
        v_result.verify_datetime.FromDatetime(self.event_datetime)
        v_result.data.address.postal_code = 'JMC'
        v_result.address = True
        await self._test_list_verification_results(v_result)

    async def test_list_verification_results_country_code(self):
        async with atomic() as session:
            verification = await self._make_verification(session)
            await make_contact_data(session, verification_id=verification.id, verify_datetime=self.event_datetime,
                                    method=self.method, contact_history_id=self.contact_history_id,
                                    address_country_code='JU', verified_address=True)
        v_result = VerificationResult(method=self.method)
        v_result.contact_history_id.value = self.contact_history_id
        v_result.verify_datetime.FromDatetime(self.event_datetime)
        v_result.data.address.country_code = 'JU'
        v_result.address = True
        await self._test_list_verification_results(v_result)

    async def test_list_verification_results_address_false(self):
        # Single test with unverified address.
        async with atomic() as session:
            verification = await self._make_verification(session)
            await make_contact_data(session, verification_id=verification.id, verify_datetime=self.event_datetime,
                                    method=self.method, contact_history_id=self.contact_history_id,
                                    address_street=['Deck 16'], address_city='Red Dwarf', address_state='Deep Space',
                                    address_postal_code='JMC', address_country_code='JU', verified_address=False)
        v_result = VerificationResult(method=self.method)
        v_result.contact_history_id.value = self.contact_history_id
        v_result.verify_datetime.FromDatetime(self.event_datetime)
        v_result.data.address.street.append('Deck 16')
        v_result.data.address.city = 'Red Dwarf'
        v_result.data.address.state_or_province = 'Deep Space'
        v_result.data.address.postal_code = 'JMC'
        v_result.data.address.country_code = 'JU'
        v_result.address = False
        await self._test_list_verification_results(v_result)

    async def test_list_verification_results_emails(self):
        async with atomic() as session:
            verification = await self._make_verification(session)
            await make_contact_data(session, verification_id=verification.id, verify_datetime=self.event_datetime,
                                    method=self.method, contact_history_id=self.contact_history_id,
                                    emails=['kryten@example.org'], verified_emails=True)
        v_result = VerificationResult(method=self.method)
        v_result.contact_history_id.value = self.contact_history_id
        v_result.verify_datetime.FromDatetime(self.event_datetime)
        v_result.data.emails.append('kryten@example.org')
        v_result.emails = True
        await self._test_list_verification_results(v_result)

    async def test_list_verification_results_emails_false(self):
        async with atomic() as session:
            verification = await self._make_verification(session)
            await make_contact_data(session, verification_id=verification.id, verify_datetime=self.event_datetime,
                                    method=self.method, contact_history_id=self.contact_history_id,
                                    emails=['kryten@example.org'], verified_emails=False)
        v_result = VerificationResult(method=self.method)
        v_result.contact_history_id.value = self.contact_history_id
        v_result.verify_datetime.FromDatetime(self.event_datetime)
        v_result.data.emails.append('kryten@example.org')
        v_result.emails = False
        await self._test_list_verification_results(v_result)

    async def test_verify_token_db_error(self):
        request = VerifyTokenRequest()
        request.verification_id.value = str(UUID(int=42))
        await self._test_db_error(self.service.verify_token, request)

    async def _test_verify_token_not_exist(self, request: VerifyTokenRequest) -> None:
        result = await self.service.verify_token(request, sentinel.context)

        reply = VerifyTokenReply()
        reply.exception.verification_does_not_exist.CopyFrom(VerificationDoesNotExist())
        self.assertEqual(result, reply)

    async def test_verify_token_empty(self):
        request = VerifyTokenRequest()
        await self._test_verify_token_not_exist(request)

    async def test_verify_token_invalid(self):
        request = VerifyTokenRequest()
        request.verification_id.value = 'INVALID'
        await self._test_verify_token_not_exist(request)

    async def test_verify_token_unknown(self):
        request = VerifyTokenRequest()
        request.verification_id.value = str(UUID(int=42))
        await self._test_verify_token_not_exist(request)

    async def _test_verify_token_incorrect_status(self, status: VerificationStatus) -> None:
        async with atomic() as session:
            await self._make_verification(session, status=status)
        request = VerifyTokenRequest()
        request.verification_id.value = str(self.uuid)

        result = await self.service.verify_token(request, sentinel.context)

        reply = VerifyTokenReply()
        reply.exception.incorrect_status.CopyFrom(IncorrectStatus())
        self.assertEqual(result, reply)

    async def test_verify_token_incorrect_succeeded(self):
        await self._test_verify_token_incorrect_status(VerificationStatus.SUCCEEDED)

    async def test_verify_token_incorrect_failed(self):
        await self._test_verify_token_incorrect_status(VerificationStatus.FAILED)

    async def test_verify_token_incorrect_canceled(self):
        await self._test_verify_token_incorrect_status(VerificationStatus.CANCELED)

    async def test_verify_token_incorrect_manual(self):
        await self._test_verify_token_incorrect_status(VerificationStatus.MANUAL)

    async def _test_verify_token(self, model_token: str, token: str, matches: bool, *, check: bool = False) -> None:
        async with atomic() as session:
            verification = await self._make_verification(session)
            await make_token(session, verification_id=verification.id, token=model_token, method=self.method)
        request = VerifyTokenRequest()
        request.verification_id.value = str(self.uuid)
        request.token = token
        request.check = check

        with freeze_time(self.close_datetime):
            result = await self.service.verify_token(request, sentinel.context)

        reply = VerifyTokenReply()
        reply.data.token_matches = matches
        self.assertEqual(result, reply)

    async def test_verify_token_empty_token(self):
        await self._test_verify_token('Gazpacho!', '', False)

        async with test_session_cls()() as session:
            # Check whether the token was not deleted.
            self.assertIsNotNone((await session.execute(select(Token))).scalar())
            # Check verification not changed.
            verification = (await session.execute(select(Verification))).scalars().one()
            self.assertEqual(verification.status, VerificationStatus.CREATED)
            self.assertIsNone((await session.execute(select(Event))).scalar())

    async def test_verify_token(self):
        await self._test_verify_token('Gazpacho!', 'Gazpacho!', True)

        async with test_session_cls()() as session:
            # Check whether the token was deleted.
            self.assertIsNone((await session.execute(select(Token))).scalar())
            # Check verification not changed, but event created.
            verification = (await session.execute(select(Verification))).scalars().one()
            self.assertEqual(verification.status, VerificationStatus.CREATED)
            event = (await session.execute(select(Event))).scalars().one()
            self.assertEqual(event.type, EventType.VERIFY_TOKEN)
            self.assertEqual(event.method, self.method)
            self.assertEqual(event.details, {})

    async def test_verify_token_case_insensitive(self):
        await self._test_verify_token('Gazpacho!', 'gazPaCHo!', True)

        async with test_session_cls()() as session:
            # Check whether the token was deleted.
            self.assertIsNone((await session.execute(select(Token))).scalar())
            # Check verification not changed, but event created.
            verification = (await session.execute(select(Verification))).scalars().one()
            self.assertEqual(verification.status, VerificationStatus.CREATED)
            event = (await session.execute(select(Event))).scalars().one()
            self.assertEqual(event.type, EventType.VERIFY_TOKEN)
            self.assertEqual(event.method, self.method)
            self.assertEqual(event.details, {})

    async def test_verify_token_mismatch(self):
        await self._test_verify_token('Gazpacho!', 'Quagaars!', False)

        async with test_session_cls()() as session:
            # Check whether the token was not deleted.
            self.assertIsNotNone((await session.execute(select(Token))).scalar())
            # Check verification not changed.
            verification = (await session.execute(select(Verification))).scalars().one()
            self.assertEqual(verification.status, VerificationStatus.CREATED)
            self.assertIsNone((await session.execute(select(Event))).scalar())

    async def test_verify_token_mismatch_id(self):
        # Test token matches, but the verification ID doesn't.
        other_uuid = UUID(int=196)
        async with atomic() as session:
            await self._make_verification(session, uuid=other_uuid, contact_id='RIMMER')
            verification = await self._make_verification(session)
            await make_token(session, verification_id=verification.id, token='Gazpacho!', method=self.method)
        request = VerifyTokenRequest()
        request.verification_id.value = str(other_uuid)
        request.token = 'Gazpacho!'

        with freeze_time(self.close_datetime):
            result = await self.service.verify_token(request, sentinel.context)

        reply = VerifyTokenReply()
        reply.data.token_matches = False
        self.assertEqual(result, reply)

        async with test_session_cls()() as session:
            # Check whether the token was not deleted.
            self.assertIsNotNone((await session.execute(select(Token))).scalar())
            # Check verification not changed.
            verification = (await session.execute(select(Verification).filter_by(uuid=other_uuid))).scalars().one()
            self.assertEqual(verification.status, VerificationStatus.CREATED)
            self.assertIsNone((await session.execute(select(Event))).scalar())

    async def test_verify_token_callback(self):
        # Test callback after token is verified.
        processors = {self.method: {'class': 'verification.tests.test_service.TestTokenProcessor'}}
        with override_settings(processors=processors):
            await self._test_verify_token('Gazpacho!', 'Gazpacho!', True)

        async with test_session_cls()() as session:
            # Check whether the token was deleted.
            self.assertIsNone((await session.execute(select(Token))).scalar())
            # Check verification changed by the callback and event created.
            verification = (await session.execute(select(Verification))).scalars().one()
            self.assertEqual(verification.status, VerificationStatus.SUCCEEDED)
            event = (await session.execute(select(Event))).scalars().one()
            self.assertEqual(event.type, EventType.VERIFY_TOKEN)
            self.assertEqual(event.method, self.method)
            self.assertEqual(event.details, {})

    async def test_verify_token_method_missing(self):
        # Test callback is not found - processor is missing.
        with override_settings(processors={}):
            await self._test_verify_token('Gazpacho!', 'Gazpacho!', True)

        async with test_session_cls()() as session:
            # Check whether the token was deleted.
            self.assertIsNone((await session.execute(select(Token))).scalar())
            # Check verification not changed, but event created.
            verification = (await session.execute(select(Verification))).scalars().one()
            self.assertEqual(verification.status, VerificationStatus.CREATED)
            event = (await session.execute(select(Event))).scalars().one()
            self.assertEqual(event.type, EventType.VERIFY_TOKEN)
            self.assertEqual(event.method, self.method)
            self.assertEqual(event.details, {})
        # Check warning
        msg = StringComparison("Token's .* method ANALYSIS not found in processors.")
        # Can't use log_handler.check_present due to a inability to render log from the verification handler.
        self.assertEqual(self.log_handler.records[1].getMessage(), msg)

    async def test_verify_token_check(self):
        await self._test_verify_token('Gazpacho!', 'Gazpacho!', True, check=True)

        async with test_session_cls()() as session:
            # Check whether the token was not deleted.
            self.assertIsNotNone((await session.execute(select(Token))).scalar())
            # Check verification not changed.
            verification = (await session.execute(select(Verification))).scalars().one()
            self.assertEqual(verification.status, VerificationStatus.CREATED)
            self.assertIsNone((await session.execute(select(Event))).scalar())

    async def _test_verify_token_complete(self) -> None:
        await self._test_verify_token('Gazpacho!', 'Gazpacho!', True)

        async with test_session_cls()() as session:
            # Check whether the token was deleted.
            self.assertIsNone((await session.execute(select(Token))).scalar())
            # Check verification not changed, but event created.
            verification = (await session.execute(select(Verification))).scalars().one()
            self.assertEqual(verification.status, VerificationStatus.CREATED)
            del verification
            event = (await session.execute(select(Event))).scalars().one()
            self.assertEqual(event.type, EventType.VERIFY_TOKEN)
            self.assertEqual(event.method, self.method)
            self.assertEqual(event.details, {})
            # Check complete attempted
            # XXX: Attach 'verification' argument to a session so it can be checked.
            handler_mock = cast(AsyncMock, APPLICATION.handler)
            session.add(handler_mock.mock_calls[0][1][0])
            await session.refresh(handler_mock.mock_calls[0][1][0])
            self.assertEqual(mock_awaits(handler_mock),
                             [call.try_complete(Comparison(Verification, uuid=self.uuid, partial=True))])

    async def test_verify_token_complete(self):
        APPLICATION.handler = AsyncMock(VerificationHandler)
        APPLICATION.handler.processors = {}
        await self._test_verify_token_complete()

    async def test_verify_token_complete_error(self):
        APPLICATION.handler = AsyncMock(VerificationHandler)
        APPLICATION.handler.processors = {}
        APPLICATION.handler.try_complete.side_effect = Exception('Smeghead!')
        await self._test_verify_token_complete()

    async def test_delay_status_db_error(self):
        request = DelayStatusRequest()
        request.verification_id.value = str(UUID(int=42))
        await self._test_db_error(self.service.delay_status, request)

    async def _test_delay_status_not_exist(self, request: DelayStatusRequest) -> None:
        result = await self.service.delay_status(request, sentinel.context)

        reply = DelayStatusReply()
        reply.exception.verification_does_not_exist.CopyFrom(VerificationDoesNotExist())
        self.assertEqual(result, reply)

    async def test_delay_status_empty(self):
        request = DelayStatusRequest()
        await self._test_delay_status_not_exist(request)

    async def test_delay_status_invalid(self):
        request = DelayStatusRequest()
        request.verification_id.value = 'INVALID'
        await self._test_delay_status_not_exist(request)

    async def test_delay_status_unknown(self):
        request = DelayStatusRequest()
        request.verification_id.value = str(UUID(int=42))
        await self._test_delay_status_not_exist(request)

    async def test_delay_status_status_unknown(self):
        # Test delay_status called with an unknown status.
        async with atomic() as session:
            await self._make_verification(session, status=VerificationStatus.PENDING)
        request = DelayStatusRequest()
        request.verification_id.value = str(self.uuid)
        request.status = ProtoVerificationStatus.unknown
        context = AsyncMock(name='context')
        context.abort.side_effect = Exception(sentinel.abort)

        with self.assertRaises(Exception) as catcher:
            await self.service.delay_status(request, context)

        self.assertEqual(catcher.exception.args, (sentinel.abort, ))
        self.assertEqual(mock_awaits(context),
                         [call.abort(StatusCode.INVALID_ARGUMENT, StringComparison("Unknown status 'unknown'"))])

    async def _test_delay_status_incorrect_status(self, status: VerificationStatus) -> None:
        async with atomic() as session:
            await self._make_verification(session, status=status)
        request = DelayStatusRequest()
        request.verification_id.value = str(self.uuid)

        result = await self.service.delay_status(request, sentinel.context)

        reply = DelayStatusReply()
        reply.exception.incorrect_status.CopyFrom(IncorrectStatus())
        self.assertEqual(result, reply)

    async def test_delay_status_incorrect_succeeded(self):
        await self._test_delay_status_incorrect_status(VerificationStatus.SUCCEEDED)

    async def test_delay_status_incorrect_failed(self):
        await self._test_delay_status_incorrect_status(VerificationStatus.FAILED)

    async def test_delay_status_incorrect_canceled(self):
        await self._test_delay_status_incorrect_status(VerificationStatus.CANCELED)

    async def test_delay_status_incorrect_manual(self):
        await self._test_delay_status_incorrect_status(VerificationStatus.MANUAL)

    async def _test_delay_status(self, status: VerificationStatus, delay: int) -> None:
        async with atomic() as session:
            await self._make_verification(session)
        handler = AsyncMock()
        handler.attach_mock(MagicMock(), 'get_next_events')
        handler.get_next_events.return_value = aiter(())
        APPLICATION.handler = handler
        request = DelayStatusRequest()
        request.verification_id.value = str(self.uuid)
        request.status = ProtoVerificationStatus.Value(status)
        request.delay = delay

        result = await self.service.delay_status(request, sentinel.context)

        async with test_session_cls()() as session:
            event = (await session.execute(select(Event))).scalars().one()
            proto_event = ProtoEvent()
            proto_event.event_datetime.FromDatetime(event.event_datetime)
            proto_event.type = EventType.DELAY_STATUS
            proto_event.details.update({'status': status, 'delay': delay})
            reply = DelayStatusReply()
            reply.data.verification.CopyFrom(
                self._get_verification_info(events=[proto_event]))
            self.assertEqual(result, reply)
            # Check event created
            self.assertIsNone(event.method)
            self.assertEqual(event.type, EventType.DELAY_STATUS)
            self.assertEqual(event.details, {'delay': delay, 'status': status})
            # Check handler cancel called
            # XXX: Attach 'verification' argument to a session so it can be checked.
            session.add(handler.mock_calls[0][1][0])
            await session.refresh(handler.mock_calls[0][1][0])
            calls = [call.cancel(Comparison(Verification, uuid=self.uuid, partial=True), status),
                     call.get_next_events(Comparison(Verification))]
            self.assertEqual(handler.mock_calls, calls)
            self.assertEqual(mock_awaits(handler), [call.cancel(Comparison(Verification), status)])

    async def test_delay_status(self):
        await self._test_delay_status(VerificationStatus.FUTILE, 42)

    async def test_delay_status_negative(self):
        await self._test_delay_status(VerificationStatus.FUTILE, -42)

    async def test_delay_status_other(self):
        await self._test_delay_status(VerificationStatus.CONSTRAINED, 42)

    async def test_resend_message_db_error(self):
        request = ResendMessageRequest()
        request.verification_id.value = str(UUID(int=42))
        await self._test_db_error(self.service.resend_message, request)

    async def _test_resend_message_not_exist(self, request: ResendMessageRequest) -> None:
        result = await self.service.resend_message(request, sentinel.context)

        reply = ResendMessageReply()
        reply.exception.verification_does_not_exist.CopyFrom(VerificationDoesNotExist())
        self.assertEqual(result, reply)

    async def test_resend_message_empty(self):
        request = ResendMessageRequest()
        await self._test_resend_message_not_exist(request)

    async def test_resend_message_invalid(self):
        request = ResendMessageRequest()
        request.verification_id.value = 'INVALID'
        await self._test_resend_message_not_exist(request)

    async def test_resend_message_unknown(self):
        request = ResendMessageRequest()
        request.verification_id.value = str(UUID(int=42))
        await self._test_resend_message_not_exist(request)

    async def _test_resend_message_incorrect_status(self, status: VerificationStatus) -> None:
        async with atomic() as session:
            await self._make_verification(session, status=status)
        request = ResendMessageRequest()
        request.verification_id.value = str(self.uuid)

        result = await self.service.resend_message(request, sentinel.context)

        reply = ResendMessageReply()
        reply.exception.incorrect_status.CopyFrom(IncorrectStatus())
        self.assertEqual(result, reply)

    async def test_resend_message_incorrect_created(self):
        await self._test_resend_message_incorrect_status(VerificationStatus.CREATED)

    async def test_resend_message_incorrect_succeeded(self):
        await self._test_resend_message_incorrect_status(VerificationStatus.SUCCEEDED)

    async def test_resend_message_incorrect_failed(self):
        await self._test_resend_message_incorrect_status(VerificationStatus.FAILED)

    async def test_resend_message_incorrect_canceled(self):
        await self._test_resend_message_incorrect_status(VerificationStatus.CANCELED)

    async def test_resend_message_incorrect_manual(self):
        await self._test_resend_message_incorrect_status(VerificationStatus.MANUAL)

    async def test_resend_message_invalid_message_type(self):
        async with atomic() as session:
            await self._make_verification(session, status=VerificationStatus.INITIALIZED)
        processor = BaseProcessor(method='no-psirens')
        APPLICATION.handler = TestHandler([processor])
        request = ResendMessageRequest()
        request.verification_id.value = str(self.uuid)
        request.message_type = 'unknown'
        context = AsyncMock(name='context')
        context.abort.side_effect = Exception(sentinel.abort)

        with self.assertRaises(Exception) as catcher:
            await self.service.resend_message(request, context)

        self.assertEqual(catcher.exception.args, (sentinel.abort, ))
        self.assertEqual(mock_awaits(context),
                         [call.abort(StatusCode.FAILED_PRECONDITION, StringComparison("Message type not found"))])

    async def test_resend_message(self):
        async with atomic() as session:
            verification = await self._make_verification(session, status=VerificationStatus.INITIALIZED)
            processor = AsyncMock(BaseProcessor)
            processor.method = 'no-psirens'
            processor.get_resend_event.return_value.to_event.return_value = Event(
                verification_id=verification.id, type=EventType.SEND_MESSAGE, details={'key': 'Gazpacho!'})
        APPLICATION.handler = TestHandler([processor])
        request = ResendMessageRequest()
        request.verification_id.value = str(self.uuid)
        request.message_type = MessageType.INITIAL_EMAIL

        result = await self.service.resend_message(request, sentinel.context)

        self.assertEqual(result, ResendMessageReply())

        async with test_session_cls()() as session:
            event = (await session.execute(select(Event))).scalars().one()
            self.assertEqual(event.details, {'key': 'Gazpacho!'})
            # Check processor calls
            # XXX: Attach 'verification' argument to a session so it can be checked.
            session.add(processor.mock_calls[0][1][0])
            await session.refresh(processor.mock_calls[0][1][0])
            calls = [
                call.is_applicable(Comparison(Verification)),
                call.get_resend_event(Comparison(Verification, uuid=self.uuid, partial=True),
                                      MessageType.INITIAL_EMAIL),
                call.get_resend_event().process(),
                call.get_resend_event().to_event()]
            self.assertEqual(mock_awaits(processor), calls)

    async def test_approve_futile_db_error(self):
        request = ApproveFutileRequest()
        request.verification_id.value = str(UUID(int=42))
        await self._test_db_error(self.service.approve_futile, request)

    async def _test_approve_futile_not_exist(self, request: ApproveFutileRequest) -> None:
        result = await self.service.approve_futile(request, sentinel.context)

        reply = ApproveFutileReply()
        reply.exception.verification_does_not_exist.CopyFrom(VerificationDoesNotExist())
        self.assertEqual(result, reply)

    async def test_approve_futile_empty(self):
        request = ApproveFutileRequest()
        await self._test_approve_futile_not_exist(request)

    async def test_approve_futile_invalid(self):
        request = ApproveFutileRequest()
        request.verification_id.value = 'INVALID'
        await self._test_approve_futile_not_exist(request)

    async def test_approve_futile_unknown(self):
        request = ApproveFutileRequest()
        request.verification_id.value = str(UUID(int=42))
        await self._test_approve_futile_not_exist(request)

    async def _test_approve_futile_incorrect_status(self, status: VerificationStatus) -> None:
        async with atomic() as session:
            await self._make_verification(session, status=status)
        request = ApproveFutileRequest()
        request.verification_id.value = str(self.uuid)

        result = await self.service.approve_futile(request, sentinel.context)

        reply = ApproveFutileReply()
        reply.exception.incorrect_status.CopyFrom(IncorrectStatus())
        self.assertEqual(result, reply)

    async def test_approve_futile_incorrect_created(self):
        await self._test_approve_futile_incorrect_status(VerificationStatus.CREATED)

    async def test_approve_futile_incorrect_initialized(self):
        await self._test_approve_futile_incorrect_status(VerificationStatus.INITIALIZED)

    async def test_approve_futile_incorrect_accessed(self):
        await self._test_approve_futile_incorrect_status(VerificationStatus.ACCESSED)

    async def test_approve_futile_incorrect_pending(self):
        await self._test_approve_futile_incorrect_status(VerificationStatus.PENDING)

    async def test_approve_futile_incorrect_constrained(self):
        await self._test_approve_futile_incorrect_status(VerificationStatus.CONSTRAINED)

    async def test_approve_futile_incorrect_succeeded(self):
        await self._test_approve_futile_incorrect_status(VerificationStatus.SUCCEEDED)

    async def test_approve_futile_incorrect_paused(self):
        await self._test_approve_futile_incorrect_status(VerificationStatus.PAUSED)

    async def test_approve_futile_incorrect_failed(self):
        await self._test_approve_futile_incorrect_status(VerificationStatus.FAILED)

    async def test_approve_futile_incorrect_canceled(self):
        await self._test_approve_futile_incorrect_status(VerificationStatus.CANCELED)

    async def test_approve_futile_incorrect_manual(self):
        await self._test_approve_futile_incorrect_status(VerificationStatus.MANUAL)

    async def test_approve_futile_manual_check(self):
        async with atomic() as session:
            await self._make_verification(session, status=VerificationStatus.FUTILE, manual_check_request=True)
        request = ApproveFutileRequest()
        request.verification_id.value = str(self.uuid)

        result = await self.service.approve_futile(request, sentinel.context)

        reply = ApproveFutileReply()
        reply.exception.incorrect_status.CopyFrom(IncorrectStatus())
        self.assertEqual(result, reply)

    async def test_approve_futile(self) -> None:
        fqdn = 'kryten.example.org'
        async with atomic() as session:
            await self._make_verification(session, status=VerificationStatus.FUTILE)
        request = ApproveFutileRequest()
        request.verification_id.value = str(self.uuid)
        cast(AsyncMock, APPLICATION.domain_client.get_domains_by_contact).return_value = {
            'domains': [{'domain': DomainRef(id=DomainId('42'), fqdn=fqdn), 'domain_history_id': '196'}]}
        cast(AsyncMock, APPLICATION.domain_admin_client.batch_delete_domains).return_value = aiter([])
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = Contact(
            contact_id=self.contact_id, contact_handle='2X4B', contact_history_id=self.contact_history_id)

        with freeze_time(self.close_datetime):
            result = await self.service.approve_futile(request, sentinel.context)

        async with test_session_cls()() as session:
            verification = (await session.execute(select(Verification))).scalars().one()
            event = (await session.execute(select(Event))).scalars().one()
        reply = ApproveFutileReply()
        proto_event = ProtoEvent()
        proto_event.event_datetime.FromDatetime(event.event_datetime)
        proto_event.type = EventType.FAILED
        proto_event.details.update({'domains': ['42'], 'undeleted': {}})
        info = self._get_verification_info(status=ProtoVerificationStatus.failed,
                                           close_datetime=verification.close_datetime, events=[proto_event])
        reply.data.verification.CopyFrom(info)
        self.assertEqual(result, reply)

        # Check verification is closed.
        self.assertIsNotNone(verification.close_datetime)
        self.assertEqual(verification.status, VerificationStatus.FAILED)
        self.assertEqual(event.type, EventType.FAILED)
        self.assertEqual(event.details, {'domains': ['42'], 'undeleted': {}})
        # Check domain call
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.domain_client)),
                         [call.get_domains_by_contact(self.contact_id, {DomainContactRole.HOLDER})])
        # Check domain delete call
        domain = Domain(domain_id=DomainId('42'), domain_history_id=DomainHistoryId('196'), fqdn=fqdn)
        self.assertEqual(cast(AsyncMock, APPLICATION.domain_admin_client).mock_calls,
                         [call.batch_delete_domains([domain])])
        # Check contact call
        flags = {ObjectState.CONTACT_IN_MANUAL_VERIFICATION: False,
                 ObjectState.CONTACT_PASSED_MANUAL_VERIFICATION: False,
                 ObjectState.CONTACT_FAILED_MANUAL_VERIFICATION: True}
        calls = [call.get_contact_info(self.contact_id),
                 call.update_contact_state(self.contact_id, self.contact_history_id, flags)]
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.contact_client)), calls)

    async def test_approve_futile_undeleted(self):
        fqdn = 'kryten.example.org'
        async with atomic() as session:
            await self._make_verification(session, status=VerificationStatus.FUTILE)
        request = ApproveFutileRequest()
        request.verification_id.value = str(self.uuid)
        cast(AsyncMock, APPLICATION.domain_client.get_domains_by_contact).return_value = {
            'domains': [{'domain': DomainRef(id=DomainId('42'), fqdn=fqdn), 'domain_history_id': '196'}]}
        cast(AsyncMock, APPLICATION.domain_admin_client.batch_delete_domains).return_value = aiter(
            [{'domain_id': '42', 'reason': NotDeletedReason.DOMAIN_BLOCKED}])
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = Contact(
            contact_id=self.contact_id, contact_handle='2X4B', contact_history_id=self.contact_history_id)

        with freeze_time(self.close_datetime):
            result = await self.service.approve_futile(request, sentinel.context)

        async with test_session_cls()() as session:
            verification = (await session.execute(select(Verification))).scalars().one()
            event = (await session.execute(select(Event))).scalars().one()
        reply = ApproveFutileReply()
        proto_event = ProtoEvent()
        proto_event.event_datetime.FromDatetime(event.event_datetime)
        proto_event.type = EventType.FAILED
        proto_event.details.update({'domains': ['42'], 'undeleted': {'42': 'domain_blocked'}})
        info = self._get_verification_info(status=ProtoVerificationStatus.failed,
                                           close_datetime=verification.close_datetime, events=[proto_event])
        reply.data.verification.CopyFrom(info)
        self.assertEqual(result, reply)

        # Check verification is closed.
        self.assertIsNotNone(verification.close_datetime)
        self.assertEqual(verification.status, VerificationStatus.FAILED)
        self.assertEqual(event.type, EventType.FAILED)
        self.assertEqual(event.details, {'domains': ['42'], 'undeleted': {'42': 'domain_blocked'}})
        # Check domain call
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.domain_client)),
                         [call.get_domains_by_contact(self.contact_id, {DomainContactRole.HOLDER})])
        # Check domain delete call
        domain = Domain(domain_id=DomainId('42'), domain_history_id=DomainHistoryId('196'), fqdn=fqdn)
        self.assertEqual(cast(AsyncMock, APPLICATION.domain_admin_client).mock_calls,
                         [call.batch_delete_domains([domain])])
        # Check contact call
        flags = {ObjectState.CONTACT_IN_MANUAL_VERIFICATION: False,
                 ObjectState.CONTACT_PASSED_MANUAL_VERIFICATION: False,
                 ObjectState.CONTACT_FAILED_MANUAL_VERIFICATION: True}
        calls = [call.get_contact_info(self.contact_id),
                 call.update_contact_state(self.contact_id, self.contact_history_id, flags)]
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.contact_client)), calls)

    async def test_approve_manual_db_error(self):
        request = ApproveManualRequest()
        request.verification_id.value = str(UUID(int=42))
        await self._test_db_error(self.service.approve_manual, request)

    async def _test_approve_manual_not_exist(self, request: ApproveManualRequest) -> None:
        result = await self.service.approve_manual(request, sentinel.context)

        reply = ApproveManualReply()
        reply.exception.verification_does_not_exist.CopyFrom(VerificationDoesNotExist())
        self.assertEqual(result, reply)

    async def test_approve_manual_empty(self):
        request = ApproveManualRequest()
        await self._test_approve_manual_not_exist(request)

    async def test_approve_manual_invalid(self):
        request = ApproveManualRequest()
        request.verification_id.value = 'INVALID'
        await self._test_approve_manual_not_exist(request)

    async def test_approve_manual_unknown(self):
        request = ApproveManualRequest()
        request.verification_id.value = str(UUID(int=42))
        await self._test_approve_manual_not_exist(request)

    async def _test_approve_manual_incorrect_status(self, status: VerificationStatus) -> None:
        async with atomic() as session:
            await self._make_verification(session, status=status)
        request = ApproveManualRequest()
        request.verification_id.value = str(self.uuid)

        result = await self.service.approve_manual(request, sentinel.context)

        reply = ApproveManualReply()
        reply.exception.incorrect_status.CopyFrom(IncorrectStatus())
        self.assertEqual(result, reply)

    async def test_approve_manual_incorrect_succeeded(self):
        await self._test_approve_manual_incorrect_status(VerificationStatus.SUCCEEDED)

    async def test_approve_manual_incorrect_failed(self):
        await self._test_approve_manual_incorrect_status(VerificationStatus.FAILED)

    async def test_approve_manual_incorrect_canceled(self):
        await self._test_approve_manual_incorrect_status(VerificationStatus.CANCELED)

    async def test_approve_manual_incorrect_manual(self):
        await self._test_approve_manual_incorrect_status(VerificationStatus.MANUAL)

    async def test_approve_manual(self):
        async with atomic() as session:
            await self._make_verification(session)
        request = ApproveManualRequest()
        request.verification_id.value = str(self.uuid)
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = Contact(
            contact_id=self.contact_id, contact_handle='2X4B', contact_history_id=self.contact_history_id)

        result = await self.service.approve_manual(request, sentinel.context)

        async with test_session_cls()() as session:
            verification = (await session.execute(select(Verification))).scalars().one()
            event = (await session.execute(select(Event))).scalars().one()
        reply = ApproveManualReply()
        proto_event = ProtoEvent()
        proto_event.event_datetime.FromDatetime(event.event_datetime)
        proto_event.type = EventType.MANUAL
        info = self._get_verification_info(status=ProtoVerificationStatus.manual,
                                           close_datetime=verification.close_datetime, events=[proto_event])
        reply.data.verification.CopyFrom(info)
        self.assertEqual(result, reply)

        # Check verification is closed.
        self.assertIsNotNone(verification.close_datetime)
        self.assertEqual(verification.status, VerificationStatus.MANUAL)
        self.assertEqual(event.type, EventType.MANUAL)
        self.assertEqual(event.details, {})
        # Check contact call
        flags = {ObjectState.CONTACT_IN_MANUAL_VERIFICATION: False,
                 ObjectState.CONTACT_PASSED_MANUAL_VERIFICATION: True,
                 ObjectState.CONTACT_FAILED_MANUAL_VERIFICATION: False}
        calls = [call.get_contact_info(self.contact_id),
                 call.update_contact_state(self.contact_id, self.contact_history_id, flags)]
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.contact_client)), calls)

    async def test_approve_manual_constrained(self):
        # Test approve_manual on CONSTRAINED verification cancels outzone event.
        async with atomic() as session:
            await self._make_verification(session, status=VerificationStatus.CONSTRAINED)
        request = ApproveManualRequest()
        request.verification_id.value = str(self.uuid)
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = Contact(
            contact_id=self.contact_id, contact_handle='2X4B', contact_history_id=self.contact_history_id)
        cast(AsyncMock, APPLICATION.domain_client.get_domains_by_contact).return_value = {}
        cast(AsyncMock, APPLICATION.domain_admin_client.manage_domain_state_flags).return_value = []

        result = await self.service.approve_manual(request, sentinel.context)

        async with test_session_cls()() as session:
            verification = (await session.execute(select(Verification))).scalars().one()
            event_1, event_2 = (await session.execute(select(Event).order_by(Event.id))).scalars()
        reply = ApproveManualReply()
        proto_event_1 = ProtoEvent()
        proto_event_1.event_datetime.FromDatetime(event_1.event_datetime)
        proto_event_1.type = EventType.OUTZONE
        proto_event_1.details.update({'canceled': True, 'request_ids': {}})
        proto_event_2 = ProtoEvent()
        proto_event_2.event_datetime.FromDatetime(event_2.event_datetime)
        proto_event_2.type = EventType.MANUAL
        info = self._get_verification_info(status=ProtoVerificationStatus.manual,
                                           close_datetime=verification.close_datetime,
                                           events=[proto_event_2, proto_event_1])
        reply.data.verification.CopyFrom(info)
        self.assertEqual(result, reply)

        # Check verification is closed.
        self.assertIsNotNone(verification.close_datetime)
        self.assertEqual(verification.status, VerificationStatus.MANUAL)
        self.assertEqual(event_1.type, EventType.OUTZONE)
        self.assertEqual(event_1.details, {'canceled': True, 'request_ids': {}})
        self.assertEqual(event_2.type, EventType.MANUAL)
        self.assertEqual(event_2.details, {})
        # Check contact call
        flags = {ObjectState.CONTACT_IN_MANUAL_VERIFICATION: False,
                 ObjectState.CONTACT_PASSED_MANUAL_VERIFICATION: True,
                 ObjectState.CONTACT_FAILED_MANUAL_VERIFICATION: False}
        calls = [call.get_contact_info(self.contact_id),
                 call.update_contact_state(self.contact_id, self.contact_history_id,
                                           {ObjectState.SERVER_LINK_PROHIBITED: False}),
                 call.update_contact_state(self.contact_id, self.contact_history_id, flags)]
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.contact_client)), calls)

    async def test_approve_manual_futile(self):
        # Test approve_manual on FUTILE verification cancels futile and outzone events.
        async with atomic() as session:
            await self._make_verification(session, status=VerificationStatus.FUTILE)
        request = ApproveManualRequest()
        request.verification_id.value = str(self.uuid)
        cast(AsyncMock, APPLICATION.contact_client.get_contact_info).return_value = Contact(
            contact_id=self.contact_id, contact_handle='2X4B', contact_history_id=self.contact_history_id)
        cast(AsyncMock, APPLICATION.domain_client.get_domains_by_contact).return_value = {}
        cast(AsyncMock, APPLICATION.domain_admin_client.manage_domain_state_flags).return_value = []

        result = await self.service.approve_manual(request, sentinel.context)

        async with test_session_cls()() as session:
            verification = (await session.execute(select(Verification))).scalars().one()
            event_1, event_2, event_3 = (await session.execute(select(Event).order_by(Event.id))).scalars()
        reply = ApproveManualReply()
        proto_event_1 = ProtoEvent()
        proto_event_1.event_datetime.FromDatetime(event_1.event_datetime)
        proto_event_1.type = EventType.FUTILE
        proto_event_1.details.update({'canceled': True})
        proto_event_2 = ProtoEvent()
        proto_event_2.event_datetime.FromDatetime(event_2.event_datetime)
        proto_event_2.type = EventType.OUTZONE
        proto_event_2.details.update({'canceled': True, 'request_ids': {}})
        proto_event_3 = ProtoEvent()
        proto_event_3.event_datetime.FromDatetime(event_3.event_datetime)
        proto_event_3.type = EventType.MANUAL
        info = self._get_verification_info(status=ProtoVerificationStatus.manual,
                                           close_datetime=verification.close_datetime,
                                           events=[proto_event_3, proto_event_2, proto_event_1])
        reply.data.verification.CopyFrom(info)
        self.assertEqual(result, reply)

        # Check verification is closed.
        self.assertIsNotNone(verification.close_datetime)
        self.assertEqual(verification.status, VerificationStatus.MANUAL)
        self.assertEqual(event_1.type, EventType.FUTILE)
        self.assertEqual(event_1.details, {'canceled': True})
        self.assertEqual(event_2.type, EventType.OUTZONE)
        self.assertEqual(event_2.details, {'canceled': True, 'request_ids': {}})
        self.assertEqual(event_3.type, EventType.MANUAL)
        self.assertEqual(event_3.details, {})
        # Check contact call
        flags = {ObjectState.CONTACT_IN_MANUAL_VERIFICATION: False,
                 ObjectState.CONTACT_PASSED_MANUAL_VERIFICATION: True,
                 ObjectState.CONTACT_FAILED_MANUAL_VERIFICATION: False}
        calls = [call.get_contact_info(self.contact_id),
                 call.update_contact_state(self.contact_id, self.contact_history_id,
                                           {ObjectState.SERVER_LINK_PROHIBITED: False}),
                 call.update_contact_state(self.contact_id, self.contact_history_id, flags)]
        self.assertEqual(mock_awaits(cast(AsyncMock, APPLICATION.contact_client)), calls)

    async def test_set_contact_representative_db_error(self):
        request = SetContactRepresentativeRequest()
        request.verification_id.value = str(UUID(int=42))
        await self._test_db_error(self.service.set_contact_representative, request)

    async def _test_set_contact_representative_not_exist(self, request: SetContactRepresentativeRequest) -> None:
        result = await self.service.set_contact_representative(request, sentinel.context)

        reply = SetContactRepresentativeReply()
        reply.exception.verification_does_not_exist.CopyFrom(VerificationDoesNotExist())
        self.assertEqual(result, reply)

    async def test_set_contact_representative_empty(self):
        request = SetContactRepresentativeRequest()
        await self._test_set_contact_representative_not_exist(request)

    async def test_set_contact_representative_invalid(self):
        request = SetContactRepresentativeRequest()
        request.verification_id.value = 'INVALID'
        await self._test_set_contact_representative_not_exist(request)

    async def test_set_contact_representative_unknown(self):
        request = SetContactRepresentativeRequest()
        request.verification_id.value = str(UUID(int=42))
        await self._test_set_contact_representative_not_exist(request)

    async def _test_set_contact_representative_incorrect_status(self, status: VerificationStatus) -> None:
        async with atomic() as session:
            await self._make_verification(session, status=status)
        request = SetContactRepresentativeRequest()
        request.verification_id.value = str(self.uuid)

        result = await self.service.set_contact_representative(request, sentinel.context)

        reply = SetContactRepresentativeReply()
        reply.exception.incorrect_status.CopyFrom(IncorrectStatus())
        self.assertEqual(result, reply)

    async def test_set_contact_representative_incorrect_succeeded(self):
        await self._test_set_contact_representative_incorrect_status(VerificationStatus.SUCCEEDED)

    async def test_set_contact_representative_incorrect_failed(self):
        await self._test_set_contact_representative_incorrect_status(VerificationStatus.FAILED)

    async def test_set_contact_representative_incorrect_canceled(self):
        await self._test_set_contact_representative_incorrect_status(VerificationStatus.CANCELED)

    async def test_set_contact_representative_incorrect_manual(self):
        await self._test_set_contact_representative_incorrect_status(VerificationStatus.MANUAL)

    async def test_set_contact_representative(self):
        async with atomic() as session:
            await self._make_verification(session)
        request = SetContactRepresentativeRequest()
        request.verification_id.value = str(self.uuid)
        request.name = 'Arnold Rimmer'
        request.organization = 'Jupiter Mining Corporation'
        request.address.street.append('Deck 16')
        request.address.city = 'Red Dwarf'
        request.address.state_or_province = 'Deep Space'
        request.address.postal_code = 'JMC'
        request.address.country_code = 'JU'
        request.telephone = 'Arnie!'
        request.email = 'arnold@example.org'

        result = await self.service.set_contact_representative(request, sentinel.context)

        # Check event created
        async with test_session_cls()() as session:
            event = (await session.execute(select(Event))).scalars().one()
        self.assertEqual(event.type, EventType.SET_CONTACT_REPRESENTATIVE)
        details = {
            'name': 'Arnold Rimmer', 'organization': 'Jupiter Mining Corporation',
            'address': {'street': ['Deck 16'], 'city': 'Red Dwarf', 'state_or_province': 'Deep Space',
                        'postal_code': 'JMC', 'country_code': 'JU'},
            'telephone': 'Arnie!', 'email': 'arnold@example.org'}
        self.assertEqual(event.details, details)

        self.assertEqual(result, SetContactRepresentativeReply())
