from unittest import TestCase
from unittest.mock import Mock, call, patch, sentinel

from testfixtures import Comparison, LogCapture
from typist import HTTPTokenAuth

from verification.handler import VerificationHandler
from verification.settings import Application, Settings


class ApplicationTest(TestCase):
    def setUp(self):
        self.log_handler = LogCapture('verification', propagate=False)
        self.application = Application(Settings)

    def tearDown(self):
        self.log_handler.uninstall()

    def test_handler(self):
        self.assertIsInstance(self.application.handler, VerificationHandler)

    def test_contact_client(self):
        self.application.settings = Mock(registry_netloc=sentinel.netloc, registry_ssl_cert=sentinel.ssl_cert)
        with patch('verification.settings.ContactClient', return_value=sentinel.client) as client_mock:
            with patch('verification.settings.make_credentials', return_value=sentinel.credentials) as cred_mock:
                self.assertEqual(self.application.contact_client, sentinel.client)

        self.assertEqual(client_mock.mock_calls, [call(sentinel.netloc, sentinel.credentials)])
        self.assertEqual(cred_mock.mock_calls, [call(sentinel.ssl_cert)])

    def test_domain_admin_client(self):
        self.application.settings = Mock(registry_netloc=sentinel.netloc, registry_ssl_cert=sentinel.ssl_cert)
        with patch('verification.settings.DomainAdminClient', return_value=sentinel.client) as client_mock:
            with patch('verification.settings.make_credentials', return_value=sentinel.credentials) as cred_mock:
                self.assertEqual(self.application.domain_admin_client, sentinel.client)

        self.assertEqual(client_mock.mock_calls, [call(sentinel.netloc, sentinel.credentials)])
        self.assertEqual(cred_mock.mock_calls, [call(sentinel.ssl_cert)])

    def test_domain_client(self):
        self.application.settings = Mock(registry_netloc=sentinel.netloc, registry_ssl_cert=sentinel.ssl_cert)
        with patch('verification.settings.DomainClient', return_value=sentinel.client) as client_mock:
            with patch('verification.settings.make_credentials', return_value=sentinel.credentials) as cred_mock:
                self.assertEqual(self.application.domain_client, sentinel.client)

        self.assertEqual(client_mock.mock_calls, [call(sentinel.netloc, sentinel.credentials)])
        self.assertEqual(cred_mock.mock_calls, [call(sentinel.ssl_cert)])

    def test_email_client(self):
        self.application.settings = Mock(messenger_netloc=sentinel.netloc, messenger_ssl_cert=sentinel.ssl_cert)
        with patch('verification.settings.EmailMessengerClient', return_value=sentinel.client) as client_mock:
            with patch('verification.settings.make_credentials', return_value=sentinel.credentials) as cred_mock:
                self.assertEqual(self.application.email_client, sentinel.client)

        self.assertEqual(client_mock.mock_calls, [call(sentinel.netloc, sentinel.credentials)])
        self.assertEqual(cred_mock.mock_calls, [call(sentinel.ssl_cert)])

    def test_file_storage(self):
        self.application.settings = Mock(fileman_netloc=sentinel.netloc, fileman_ssl_cert=sentinel.ssl_cert)
        with patch('verification.settings.FileClient', return_value=sentinel.client) as client_mock:
            with patch('verification.settings.make_credentials', return_value=sentinel.credentials) as cred_mock:
                storage = self.application.file_storage

        self.assertEqual(storage.client, sentinel.client)
        self.assertEqual(client_mock.mock_calls, [call(sentinel.netloc, sentinel.credentials)])
        self.assertEqual(cred_mock.mock_calls, [call(sentinel.ssl_cert)])

    def test_letter_client(self):
        self.application.settings = Mock(messenger_netloc=sentinel.netloc, messenger_ssl_cert=sentinel.ssl_cert)
        with patch('verification.settings.LetterMessengerClient', return_value=sentinel.client) as client_mock:
            with patch('verification.settings.make_credentials', return_value=sentinel.credentials) as cred_mock:
                self.assertEqual(self.application.letter_client, sentinel.client)

        self.assertEqual(client_mock.mock_calls, [call(sentinel.netloc, sentinel.credentials)])
        self.assertEqual(cred_mock.mock_calls, [call(sentinel.ssl_cert)])

    def test_representative_client(self):
        self.application.settings = Mock(registry_netloc=sentinel.netloc, registry_ssl_cert=sentinel.ssl_cert)
        with patch('verification.settings.RepresentativeClient', return_value=sentinel.client) as client_mock:
            with patch('verification.settings.make_credentials', return_value=sentinel.credentials) as cred_mock:
                self.assertEqual(self.application.representative_client, sentinel.client)

        self.assertEqual(client_mock.mock_calls, [call(sentinel.netloc, sentinel.credentials)])
        self.assertEqual(cred_mock.mock_calls, [call(sentinel.ssl_cert)])

    def test_secretary_client(self):
        self.application.settings = Mock(secretary_url=sentinel.url, secretary_timeout=sentinel.timeout,
                                    secretary_token=None)
        with patch('verification.settings.SecretaryClient', return_value=sentinel.client) as client_mock:
            self.assertEqual(self.application.secretary_client, sentinel.client)

        self.assertEqual(client_mock.mock_calls, [call(sentinel.url, timeout=sentinel.timeout, auth=None)])

    def test_secretary_client_token(self):
        self.application.settings = Mock(secretary_url=sentinel.url, secretary_timeout=sentinel.timeout,
                                    secretary_token=sentinel.token)
        with patch('verification.settings.SecretaryClient', return_value=sentinel.client) as client_mock:
            self.assertEqual(self.application.secretary_client, sentinel.client)

        self.assertEqual(
            client_mock.mock_calls,
            [call(sentinel.url, timeout=sentinel.timeout, auth=Comparison(HTTPTokenAuth, token=sentinel.token))])

    def test_sms_client(self):
        self.application.settings = Mock(messenger_netloc=sentinel.netloc, messenger_ssl_cert=sentinel.ssl_cert)
        with patch('verification.settings.SmsMessengerClient', return_value=sentinel.client) as client_mock:
            with patch('verification.settings.make_credentials', return_value=sentinel.credentials) as cred_mock:
                self.assertEqual(self.application.sms_client, sentinel.client)

        self.assertEqual(client_mock.mock_calls, [call(sentinel.netloc, sentinel.credentials)])
        self.assertEqual(cred_mock.mock_calls, [call(sentinel.ssl_cert)])

    def test_clear_cache(self):
        self.application.settings = Settings()

        # XXX: Patch Semaphore to prevent test failures in python 3.8 and 3.9.
        with patch('frgal.aio.Semaphore', return_value=sentinel.semaphore):
            handler = self.application.handler
            contact_client = self.application.contact_client
            domain_admin_client = self.application.domain_admin_client
            domain_client = self.application.domain_client
            email_client = self.application.email_client
            file_storage = self.application.file_storage
            letter_client = self.application.letter_client
            representative_client = self.application.representative_client
            secretary_client = self.application.secretary_client
            sms_client = self.application.sms_client

            self.application.clear_cache()

            self.assertIsNot(self.application.handler, handler)
            self.assertIsNot(self.application.contact_client, contact_client)
            self.assertIsNot(self.application.domain_admin_client, domain_admin_client)
            self.assertIsNot(self.application.domain_client, domain_client)
            self.assertIsNot(self.application.email_client, email_client)
            self.assertIsNot(self.application.file_storage, file_storage)
            self.assertIsNot(self.application.letter_client, letter_client)
            self.assertIsNot(self.application.representative_client, representative_client)
            self.assertIsNot(self.application.secretary_client, secretary_client)
            self.assertIsNot(self.application.sms_client, sms_client)
