from unittest import TestCase
from unittest.mock import sentinel

from verification.settings import ENVIRON_NAME
from verification.utils import Result, import_path, load_instance


class ImportPathTest(TestCase):
    def test_class(self):
        cls = import_path('verification.tests.test_utils.ImportPathTest')
        self.assertEqual(cls, ImportPathTest)

    def test_attribute(self):
        attribute = import_path('verification.settings.ENVIRON_NAME')
        self.assertEqual(attribute, ENVIRON_NAME)

    def test_invalid(self):
        data = (
            # path, error_message
            ('', 'is not a valid dotted path'),
            ('.', 'Empty module name'),
            ('.invalid', 'Empty module name'),
            ('invalid', 'is not a valid dotted path'),
            ('invalid.', 'No module named'),
            ('invalid.path', 'No module named'),
            ('invalid..path', 'No module named'),
            ('verification.tests.invalid.TestCase', 'No module named'),
            ('verification.tests.test_utils.', 'does not have a class or attribute'),
            ('verification.tests.test_utils.InvalidTestCase', 'does not have a class or attribute'),
        )
        for path, error in data:
            with self.subTest(path=path):
                with self.assertRaisesRegex(ImportError, error):
                    import_path(path)


class TestClass:
    """Test class for load_instance."""

    def __init__(self, *, keyword=None, another=None):
        self.keyword = keyword
        self.another = another


class LoadInstanceTest(TestCase):
    def test_class_only(self):
        obj = load_instance('verification.tests.test_utils.TestClass')

        self.assertIsInstance(obj, TestClass)
        self.assertIsNone(obj.keyword)
        self.assertIsNone(obj.another)

    def test_class_kwargs(self):
        obj = load_instance('verification.tests.test_utils.TestClass', keyword=sentinel.value, another=sentinel.another)

        self.assertIsInstance(obj, TestClass)
        self.assertEqual(obj.keyword, sentinel.value)
        self.assertEqual(obj.another, sentinel.another)

    def test_invalid_path(self):
        data = (
            # path, error_message
            ('', 'is not a valid dotted path'),
            ('verification.tests.test_utils.InvalidTestCase', 'does not have a class or attribute'),
        )
        for path, error in data:
            with self.subTest(path=path):
                with self.assertRaisesRegex(ImportError, error):
                    load_instance(path)

    def test_invalid_kwargs(self):
        data = (
                {'invalid': sentinel.invalid},
                {'keyword': sentinel.keyword, 'another': sentinel.another, 'invalid': sentinel.invalid},
        )
        for kwargs in data:
            with self.subTest(kwargs=kwargs):
                with self.assertRaises(TypeError):
                    load_instance('verification.tests.test_utils.TestClass', **kwargs)


class ResultTest(TestCase):
    def test_total(self):
        data = (
            # result, total
            (Result(), 0),
            (Result(any=1), 1),
            (Result(any=1, other=2), 3),
        )
        for result, total in data:
            with self.subTest(result=result):
                self.assertEqual(result.total(), total)

    def test_mark_success(self):
        data = (
            # input, output
            (Result(), Result(success=1)),
            (Result(success=42), Result(success=43)),
            (Result(any=42), Result(success=1, any=42)),
        )
        for input, output in data:
            with self.subTest(input=input):
                input.mark_success()
                self.assertEqual(input, output)

    def test_mark_fail(self):
        data = (
            # input, output
            (Result(), Result(fail=1)),
            (Result(fail=42), Result(fail=43)),
            (Result(any=42), Result(fail=1, any=42)),
        )
        for input, output in data:
            with self.subTest(input=input):
                input.mark_fail()
                self.assertEqual(input, output)

    def test_mark_error(self):
        data = (
            # input, output
            (Result(), Result(error=1)),
            (Result(error=42), Result(error=43)),
            (Result(any=42), Result(error=1, any=42)),
        )
        for input, output in data:
            with self.subTest(input=input):
                input.mark_error()
                self.assertEqual(input, output)

    def test_mark_skip(self):
        data = (
            # input, output
            (Result(), Result(skip=1)),
            (Result(skip=42), Result(skip=43)),
            (Result(any=42), Result(skip=1, any=42)),
        )
        for input, output in data:
            with self.subTest(input=input):
                input.mark_skip()
                self.assertEqual(input, output)

    def test_successes(self):
        data = (
            # result, count
            (Result(), 0),
            (Result(success=1), 1),
            (Result(success=42), 42),
            (Result(other=2), 0),
        )
        for result, count in data:
            with self.subTest(result=result):
                self.assertEqual(result.successes, count)

    def test_fails(self):
        data = (
            # result, count
            (Result(), 0),
            (Result(fail=1), 1),
            (Result(fail=42), 42),
            (Result(other=2), 0),
        )
        for result, count in data:
            with self.subTest(result=result):
                self.assertEqual(result.fails, count)

    def test_errors(self):
        data = (
            # result, count
            (Result(), 0),
            (Result(error=1), 1),
            (Result(error=42), 42),
            (Result(other=2), 0),
        )
        for result, count in data:
            with self.subTest(result=result):
                self.assertEqual(result.errors, count)

    def test_skips(self):
        data = (
            # result, count
            (Result(), 0),
            (Result(skip=1), 1),
            (Result(skip=42), 42),
            (Result(other=2), 0),
        )
        for result, count in data:
            with self.subTest(result=result):
                self.assertEqual(result.skips, count)
