import asyncio
import os
import warnings
from copy import copy
from functools import partial
from typing import Any, Dict, List, Optional, Sequence, Tuple, Type, cast
from unittest import IsolatedAsyncioTestCase, TestCase
from unittest.mock import ANY, AsyncMock, Mock, _Call, patch

from click import BaseCommand
from click.testing import CliRunner, Result
from setapp.utils import override_settings as _override_settings
from sqlalchemy import URL
from sqlalchemy.ext.asyncio import AsyncEngine, AsyncSession, async_sessionmaker, create_async_engine
from sqlalchemy.sql import text
from sqlalchemy_utils.functions.database import _sqlite_file_exists
from sqlalchemy_utils.functions.orm import quote

from verification.constants import VerificationStatus
from verification.database import get_engine, get_session_cls
from verification.models import METADATA, ContactData, Domain, Event, Score, ScoreQueue, Token, Verification
from verification.settings import APPLICATION, Application, Settings

override_settings = partial(_override_settings, APPLICATION)


def mock_awaits(mock: AsyncMock) -> List[_Call]:
    """Return a list of awaited call from mock and all its children.

    Awaits are in an order of their coroutine calls.
    """
    # A list of awaited calls.
    awaits = []
    # A queue of mock objects to process.
    queue: List[Tuple[str, Mock]] = []
    queue.append(('', mock))
    # Collect awaits from the structure of mocks.
    # Based on `CallableMixin._increment_mock_call`.
    while queue:
        parent_name, a_mock = queue.pop(0)

        # Collect children to process.
        children = []
        if isinstance(a_mock._mock_return_value, Mock):
            children.append(a_mock._mock_return_value)
        children.extend(a_mock._mock_children.values())

        # Enqueue children.
        for child in children:
            # Create a new parent name for the child.
            if child._mock_new_name == '()':
                dot = ''
            else:
                dot = '.'
            if parent_name:
                new_name = parent_name + dot + child._mock_new_name
            else:
                new_name = child._mock_new_name

            queue.append((new_name, child))

        # Collect calls for the current mock.
        for call in a_mock._mock_await_args_list:
            new_name = parent_name + (call._mock_name or '')
            new_call = _Call((new_name, ) + call._get_call_arguments())
            awaits.append(new_call)

    result = []
    for call in mock.mock_calls:
        if call in awaits:
            awaits.remove(call)
            result.append(call)
    return result


########################################################################################################################
# DATABASE UTILITIES
class TestSession(AsyncSession):
    async def close(self) -> None:
        if self.sync_session._transaction is not None and self.sync_session._transaction.is_active:  # pragma: no cover
            # Flush all the prepared objects.
            await self.flush()
        await super().close()

    def __del__(self):
        if self.sync_session._transaction is not None and self.sync_session._transaction.is_active:  # pragma: no cover
            warnings.warn("The test session hasn't been closed.", ResourceWarning, stacklevel=2)


def test_session_cls() -> async_sessionmaker[AsyncSession]:
    return async_sessionmaker(bind=get_engine(), class_=TestSession)


async def database_exists(url: URL) -> bool:
    """Check if database exists.

    Async versions of sqlalchemy_utils.functions.database_exists.
    Tweaked version to use template1 database to investigate postgresql databases and use environment variables.
    """
    url = copy(url)
    database = url.database
    dialect_name = url.get_dialect().name
    engine = None
    try:
        if dialect_name == 'postgresql':  # pragma: no cover
            opts = url.translate_connect_args(username='user')
            if not opts:
                # No connection parameters, use environment instead.
                database = os.environ.get('PGDATABASE')
            url = url.set(database='template1')
            sql = text("SELECT 1 FROM pg_database WHERE datname= :database").bindparams(database=database)

            engine = create_async_engine(url)
            async with engine.connect() as conn:
                return bool(await conn.scalar(sql))
        else:
            assert dialect_name == 'sqlite'
            return url.database == ":memory:" or _sqlite_file_exists(url.database)
    finally:
        if engine:
            await engine.dispose()


async def create_database(url: URL, encoding: str = 'utf-8', template: Optional[str] = None) -> None:
    """Create database.

    Async versions of sqlalchemy_utils.functions.create_database.
    Tweaked version to use postgresql environment variables, when appropriate.
    """
    url = copy(url)
    dialect_name = url.get_dialect().name
    database = url.database
    if dialect_name == 'postgresql':
        opts = url.translate_connect_args(username='user')
        if not opts:  # pragma: no cover
            # No connection parameters, use environment instead.
            database = os.environ.get('PGDATABASE')
        url = url.set(database='template1')
        engine = create_async_engine(url, isolation_level="AUTOCOMMIT")
        if not template:  # pragma: no cover
            template = "template1"
        async with engine.begin() as conn:
            sql = "CREATE DATABASE {} ENCODING '{}' TEMPLATE {}".format(
                quote(conn, database), encoding, quote(conn, template)
            )
            await conn.execute(text(sql))
        await engine.dispose()
    else:
        assert dialect_name == 'sqlite'
        # SQLite databases are created as you go.
        pass


async def drop_database(url: URL) -> None:
    """Drop database.

    Async versions of sqlalchemy_utils.functions.drop_database.
    Tweaked version to use postgresql environment variables, when appropriate.
    """
    url = copy(url)
    dialect_name = url.get_dialect().name
    database = url.database
    if dialect_name == 'postgresql':  # pragma: no cover
        opts = url.translate_connect_args(username='user')
        if not opts:
            # No connection parameters, use environment instead.
            database = os.environ.get('PGDATABASE')
        url = url.set(database='template1')
        engine = create_async_engine(url, isolation_level="AUTOCOMMIT")
        async with engine.begin() as conn:
            # Disconnect all users from the database we are dropping.
            sql = """
                SELECT pg_terminate_backend(pg_stat_activity.pid)
                FROM pg_stat_activity
                WHERE pg_stat_activity.datname = '{database}'
                AND pid <> pg_backend_pid();
                """.format(database=database)  # noqa: S608
            await conn.execute(text(sql))

            sql = "DROP DATABASE {}".format(quote(conn, database))
            await conn.execute(text(sql))
        await engine.dispose()
    else:
        assert dialect_name == 'sqlite'
        # Skip dropping the sqlite in-memory database. They get dropped when connections are closed.
        if database and database != ':memory:':
            os.remove(database)


class override_database:
    """Decorator and context manager to set up test database."""

    def __init__(self, *, db_connection: Optional[str] = None, db_echo: Optional[bool] = None):
        db_connection = db_connection or APPLICATION.settings.test_db_connection
        kwargs = {}
        if db_echo is not None:  # pragma: no cover
            kwargs['db_echo'] = db_echo
        self._override_settings = override_settings(db_connection=db_connection, **kwargs)

    async def reset_database(self) -> None:
        """Reset database caches."""
        if isinstance(get_engine(), AsyncEngine):
            await get_engine().dispose()
        get_engine.cache_clear()
        get_session_cls.cache_clear()

    async def enable(self) -> None:
        """Set up test database."""
        self._override_settings.enable()
        await self.reset_database()

        # Create and migrate test database.
        if not await database_exists(get_engine().url):
            await create_database(get_engine().url)
        async with get_engine().begin() as conn:
            await conn.run_sync(METADATA.create_all)

    async def disable(self) -> None:
        """Tear down test database."""
        await drop_database(get_engine().url)
        self._override_settings.disable()
        await self.reset_database()

    async def __aenter__(self):
        await self.enable()

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.disable()

    def __call__(self, decorated: Type[IsolatedAsyncioTestCase]) -> Type[IsolatedAsyncioTestCase]:
        """Decorate TestCase."""
        original_setup = decorated.asyncSetUp
        original_teardown = decorated.asyncTearDown

        async def asyncSetUp(inner_self):
            await self.enable()
            try:
                await original_setup(inner_self)
            except Exception:  # pragma: no cover
                await self.disable()
                raise

        async def asyncTearDown(inner_self):
            await original_teardown(inner_self)
            await self.disable()

        decorated.asyncSetUp = asyncSetUp  # type: ignore[method-assign]
        decorated.asyncTearDown = asyncTearDown  # type: ignore[method-assign]
        return decorated


class TransactionTestCase(IsolatedAsyncioTestCase):
    """Test case which creates a transaction before each test.

    The test creates a database at the beginning of the case and drops it at the end.
    """

    _override_database = override_database()

    def __init__(self, methodName='runTest'):
        super().__init__(methodName)
        self.session: AsyncSession = None  # type: ignore[assignment]

    def _callSetUp(self):
        self.session = test_session_cls()()

        super()._callSetUp()  # type: ignore[misc]

    def _callTearDown(self):
        super()._callTearDown()  # type: ignore[misc]

        # Flush all the prepared objects.
        self._callAsync(self.session.flush)  # type: ignore[attr-defined]
        # Rollback and close the session.
        self._callAsync(self.session.rollback)  # type: ignore[attr-defined]
        self._callAsync(self.session.close)  # type: ignore[attr-defined]

    @classmethod
    def setUpClass(cls):
        """Set up the test database."""
        asyncio.run(cls._override_database.enable(), debug=True)

    @classmethod
    def tearDownClass(cls):
        """Tear down the test database."""
        asyncio.run(cls._override_database.disable(), debug=True)


########################################################################################################################
# COMMAND UTILITIES
class TestApplication(Application):
    def load(self, config_file: Optional[str] = None) -> None:
        """Do nothing. Overrides are used instead."""


class CommandMixin:
    """Mixin for command tests."""

    runner = CliRunner(mix_stderr=False)
    command: BaseCommand

    def invoke(self, args: Optional[Sequence[str]] = None, *, settings: Optional[Dict] = None) -> Result:
        """Invoke a command and return a result."""
        # Replace the settings with test settings for the command run:
        # 1. Replace the application for `verification_command` wrapper to bypass settings loading,
        #    but keep parsing of command line arguments.
        # 2. Replace the settings object within the APPLICATION, so it's used everywhere.
        test_app = TestApplication(Settings)
        new_settings = dict(APPLICATION.settings.model_dump(), **(settings or {}))
        test_app.settings = Settings(**new_settings)
        with patch('verification.commands.utils.APPLICATION', new=test_app):
            old_settings = APPLICATION.settings
            try:
                APPLICATION.settings = test_app.settings
                return self.runner.invoke(self.command, args)
            finally:
                APPLICATION.settings = old_settings
                APPLICATION.clear_cache()

    def assertCommandSuccess(self, args: Optional[Sequence[str]] = None, *, stdout: str = ANY, stderr: str = '',
                             settings: Optional[Dict] = None) -> None:
        """Invoke command and check it succeeded."""
        result = self.invoke(args, settings=settings)

        cast(TestCase, self).assertIsNone(result.exception)
        err_msg = "command failed with code {}, stderr: {!r}".format(result.exit_code, result.stderr)
        cast(TestCase, self).assertEqual(result.exit_code, 0, msg=err_msg)
        cast(TestCase, self).assertEqual(result.stdout, stdout)
        cast(TestCase, self).assertEqual(result.stderr, stderr)

    def assertCommandFail(self, args: Optional[Sequence[str]] = None, *, stdout: str = '', stderr: str = '',
                          settings: Optional[Dict] = None) -> None:
        """Invoke command and check it failed."""
        result = self.invoke(args, settings=settings)

        cast(TestCase, self).assertNotEqual(result.exit_code, 0)
        cast(TestCase, self).assertEqual(result.stdout, stdout)
        cast(TestCase, self).assertIn(stderr, result.stderr)


########################################################################################################################
# MODEL UTILITIES
async def make_verification(session: AsyncSession, *, status: VerificationStatus = VerificationStatus.CREATED,
                            **kwargs: Any) -> Verification:
    """Make and save verification."""
    obj = Verification(status=status, **kwargs)
    session.add(obj)
    await session.flush()
    return obj


async def make_domain(session: AsyncSession, **kwargs: Any) -> Domain:
    """Make and save domain."""
    obj = Domain(**kwargs)
    session.add(obj)
    await session.flush()
    return obj


async def make_token(session: AsyncSession, **kwargs: Any) -> Token:
    """Make and save token."""
    obj = Token(**kwargs)
    session.add(obj)
    await session.flush()
    return obj


async def make_event(session: AsyncSession, **kwargs: Any) -> Event:
    """Make and save event."""
    obj = Event(**kwargs)
    session.add(obj)
    await session.flush()
    return obj


async def make_contact_data(session: AsyncSession, **kwargs: Any) -> ContactData:
    """Make and save contact data."""
    obj = ContactData(**kwargs)
    session.add(obj)
    await session.flush()
    return obj


async def make_score_queue(session: AsyncSession, **kwargs: Any) -> ScoreQueue:
    """Make and save contact queued for scoring."""
    obj = ScoreQueue(**kwargs)
    session.add(obj)
    await session.flush()
    return obj


async def make_score(session: AsyncSession, **kwargs: Any) -> Score:
    """Make and save contact score."""
    obj = Score(**kwargs)
    session.add(obj)
    await session.flush()
    return obj
