"""Various utility functions."""
from collections import Counter
from importlib import import_module
from typing import Any


def import_path(path: str) -> Any:
    """Import a dotted path and return the attribute or class defined.

    Raises:
        ImportError: If the import fails.
    """
    try:
        module_name, class_name = path.rsplit('.', 1)
    except ValueError as error:
        raise ImportError("'{}' is not a valid dotted path.".format(path)) from error

    try:
        module = import_module(module_name)
    except ValueError as error:
        raise ImportError(str(error)) from error

    try:
        return getattr(module, class_name)
    except AttributeError as error:
        raise ImportError("'{}' does not have a class or attribute '{}'.".format(module_name, class_name)) from error


def load_instance(_path: str, /, **kwargs: Any) -> Any:
    """Return instance of class specified by a dotted path and keyword arguments."""
    cls = import_path(_path)
    return cls(**kwargs)


class Result(Counter):
    """Collects process statistics.

    Basically a Counter with a defined set of fields.
    """

    # TODO: Remove in favor of Counter.total(), when support for python 3.9 is dropped.
    def total(self) -> int:
        """Compute the sum of the counts."""
        return sum(self.values())

    def mark_success(self) -> None:
        """Mark a success to a result."""
        self.update(success=1)

    def mark_fail(self) -> None:
        """Mark a failure to a result."""
        self.update(fail=1)

    def mark_error(self) -> None:
        """Mark an error to a result."""
        self.update(error=1)

    def mark_skip(self) -> None:
        """Mark a skip to a result."""
        self.update(skip=1)

    @property
    def successes(self) -> int:
        """Return a number of successes."""
        return self['success']

    @property
    def fails(self) -> int:
        """Return a number of failures."""
        return self['fail']

    @property
    def errors(self) -> int:
        """Return a number of errors."""
        return self['error']

    @property
    def skips(self) -> int:
        """Return a number of skips."""
        return self['skip']
